/*	mpi_pi.c

	This program uses C with the MPI library to compute the value
	of pi.  It is designed to be as comparable as possible to the
	other language examples in M. J. Quinn, "Parallel Computing
	Theory And Practice," second edition, McGraw Hill, New York,
	1994.  This code is most similar to the nCUBE C code example
	given on pages 111-112.

	1996 by H. Dietz
	1996 by T. Mattox and R. Fisher
*/

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>

main(int argc,
char **argv)
{
	register double width;
	double sum, lsum;
	register int i, intervals; 
	int nproc, iproc;

	/* Initialize MPI */
	if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1);
        MPI_Comm_size( MPI_COMM_WORLD, &nproc );
	MPI_Comm_rank( MPI_COMM_WORLD, &iproc );

	/* get the number of intervals */
	intervals = atoi(argv[1]);
	width = 1.0 / intervals;

	/* do the local computations */
	lsum = 0;
	for (i=iproc; i<intervals; i+=nproc) {
		register double x = (i + 0.5) * width;
		lsum += 4.0 / (1.0 + x * x);
	}

	/* sum across the local results & scale by width */
	lsum *= width;
	MPI_Reduce(&lsum, &sum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

	/* have only the console PE print the result */
	if (iproc == 0) { 
		printf("Estimation of pi is %14.12lf\n", sum);
	} 

	/* check-out */
	MPI_Finalize();
	exit(0);
}
