/*	pvm_pi.c

	This program uses C with the PVM library to compute the value
	of pi.  It is designed to be as comparable as possible to the
	other language examples in M. J. Quinn, "Parallel Computing
	Theory And Practice," second edition, McGraw Hill, New York,
	1994.  This code is most similar to the nCUBE C code example
	given on pages 111-112.

	1996 by H. Dietz
	1996 by S. Kim, T. Mattox, and R. Fisher
*/

#include <stdlib.h>
#include <stdio.h>
#include <pvm3.h>

#define NPROC	4

main(int argc,
char **argv)
{
	register double lsum, width;
	double sum;
	register int i, intervals; 
	int mytid, iproc, msgtag = 4;
	int tids[NPROC];            /* array of task ids */

	/* enroll in pvm */
	mytid = pvm_mytid();

	/* Join a group and if I am the first instance */
	/* i.e. iproc=0 spawn more copies of myself       */

	iproc = pvm_joingroup("pi");

	if (iproc == 0) {
		tids[0] = pvm_mytid();
		pvm_spawn("pvm_pi", &argv[1], 0, NULL, NPROC-1, &tids[1]);
	}
	pvm_barrier("pi", NPROC);   /* make sure all processes are here */

	/* get the number of intervals */
	intervals = atoi(argv[1]);
	width = 1.0 / intervals;

	lsum = 0.0;
	for (i = iproc; i<intervals; i+=NPROC) {
		register double x = (i + 0.5) * width;
		lsum += 4.0 / (1.0 + x * x);
	}
	
	/* sum across the local results & scale by width */
	sum = lsum * width;
	pvm_reduce(PvmSum, &sum, 1, PVM_DOUBLE, msgtag, "pi", 0);

	/* have only the console PE print the result */
	if (iproc == 0) {
		printf("Estimation of pi is %14.12lf\n", sum);
	}

	/* Program finished leave group and exit pvm */
	pvm_barrier("pi", NPROC);   /* make sure processes are finished*/
	pvm_lvgroup("pi");
	pvm_exit();
	exit(0);
}
