/*	pi.c

	This program uses C with the AFAPI library to compute the
	value of pi.  It is designed to be as comparable as possible
	to the other language examples in M. J. Quinn, "Parallel
	Computing Theory And Practice," second edition, McGraw Hill,
	New York, 1994.  This code is most similar to the nCUBE C code
	example given on pages 111-112, although the AFAPI version is
	significantly shorter and simpler.

	1996 by H. Dietz
*/


#include <stdlib.h>
#include <stdio.h>
#include "afapi.h"


main(int argc,
char **argv)
{
	register double width, sum;
	register int intervals, i;

	/* check-in with AFAPI */
	if (p_init()) exit(1);

	/* get the number of intervals */
	intervals = atoi(argv[1]);
	width = 1.0 / intervals;

	/* do the local computations */
	sum = 0;
	for (i=IPROC; i<intervals; i+=NPROC) {
		register double x = (i + 0.5) * width;
		sum += 4.0 / (1.0 + x * x);
	}

	/* sum across the local results & scale by width */
	sum = p_reduceAdd64f(sum) * width;

	/* have only the console PE print the result */
	if (IPROC == CPROC) {
		printf("Estimation of pi is %14.12lf\n", sum);
	}

	/* check-out */
	p_exit();
	exit(0);
}
