/*	scan.c

	AFAPI scan (parallel prefix) function definitions for
	TTL_PAPERS.

	July 1996 by H. Dietz
*/

#include "afapi.h"
#include "inline.h"

/*	p_scanAnd() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanAnd##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		((p_##suffix##u) datum) &= \
			((p_##suffix##u) p->datum[ which ].u_##suffix); \
	} while (++p <= p_shm_me); \
 \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_scanOr() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanOr##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		((p_##suffix##u) datum) |= \
			((p_##suffix##u) p->datum[ which ].u_##suffix); \
	} while (++p <= p_shm_me); \
 \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_scanXor() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanXor##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	datum = 0; \
	p = p_shm_all; \
	do { \
		((p_##suffix##u) datum) ^= \
			((p_##suffix##u) p->datum[ which ].u_##suffix); \
	} while (++p <= p_shm_me); \
 \
	return(datum); \
}

P_BETA

#undef	P_DECLARE

/*	p_scanMin() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMin##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		register p_##suffix t = p->datum[ which ].u_##suffix; \
		if (t < datum) datum = t; \
	} while (++p <= p_shm_me); \
 \
	return(datum); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_scanMax() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMax##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	do { \
		register p_##suffix t = p->datum[ which ].u_##suffix; \
		if (t > datum) datum = t; \
	} while (++p <= p_shm_me); \
 \
	return(datum); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_scanAdd() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanAdd##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	datum = 0; \
	do { \
		datum += p->datum[ which ].u_##suffix; \
	} while (++p <= p_shm_me); \
 \
	return(P_CAST_##suffix(datum)); \
}

P_GAMMA

#undef P_DECLARE

/*	p_scanMul() functions...
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scanMul##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = P_CAST_##suffix(datum); \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Process the values... */ \
	p = p_shm_all; \
	datum = 1; \
	do { \
		datum *= p->datum[ which ].u_##suffix; \
	} while (++p <= p_shm_me); \
 \
	return(datum); \
}

P_GAMMA

#undef P_DECLARE
