/*	bbpi.c

	Compute Pi using two bb_threads.

	Note kludge to get PE number using pids[] and getpid().
	This only works for the case of two threads.

	1996 by H. Dietz
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "bb_threads.h"


volatile double pi = 0.0;	/* Approximation to pi (shared) */
volatile int intervals;		/* How many intervals for approx? */
volatile int pids[2];		/* Unix PIDs of threads */


void
do_pi(void *data,
size_t len)
{
	register double width, localsum;
	register int i;
	register int iproc = (getpid() != pids[0]);

	/* set width */
	width = 1.0 / intervals;

	/* do the local computations */
	localsum = 0;
	for (i=iproc; i<intervals; i+=2) {
		register double x = (i + 0.5) * width;
		localsum += 4.0 / (1.0 + x * x);
	}
	localsum *= width;

	/* get permission to update pi, update it, and unlock again */
	bb_threads_lock(0);
	pi += localsum;
	bb_threads_unlock(0);
}


int
main(int argc,
char **argv)
{
	/* get the number of intervals */
	intervals = atoi(argv[1]);

	/* set stack size and create lock... */
	bb_threads_stacksize(65536);
	bb_threads_mutexcreate(0);

	/* make two threads... */
	pids[0] = bb_threads_newthread(do_pi, NULL);
	pids[1] = bb_threads_newthread(do_pi, NULL);

	/* cleanup after two threads (really a barrier sync) */
	bb_threads_cleanup(wait(NULL));
	bb_threads_cleanup(wait(NULL));

	/* print the result */
	printf("Estimation of pi is %14.12lf\n", pi);

	/* check-out */
	exit(0);
}
