/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	ieee.c,v $
 * Revision 6.0  96/02/29  13:07:12  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  95/11/20  18:56:50  gdburns
 * Cast IEEE buffers to (void *).
 * 
 * Revision 5.2.1.1  94/09/16  15:45:49  raja
 * Do nothing on SCO.
 * 
 * Revision 5.2  94/08/22  13:34:47  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:31:35  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:08:23  gdburns
 * Ohio Release
 * 
 * Revision 2.2  93/05/04  17:11:15  gdburns
 * pre-release to 2.3
 * 
 *	Function:	- change floating point to/from IEEE big-endian
 *			- handles floats and doubles
 *			- requires XDR support for floating point numbers
 *			- OTB version
 *			- do nothing on SCO
 */

#ifndef SCO
#include <stdio.h>
#include <rpc/types.h>
#include <rpc/xdr.h>

/*
 *	ltoieeef
 *
 *	Function:	- change a float from local to IEEE big-endian
 *	Accepts:	- ptr source float
 *			- ptr destination area
 */
void
ltoieeef(src, dest)

float			*src;
void			*dest;

{
	float		orig;
	XDR		xdrs;

	orig = *src;
	xdrmem_create(&xdrs, (char *) dest, sizeof(float), XDR_ENCODE);

	(void) xdr_float(&xdrs, &orig);
	xdr_destroy(&xdrs);
}

/*
 *	ltoieeed
 *
 *	Function:	- change a double from local to IEEE big-endian
 *	Accepts:	- ptr source double
 *			- ptr destination area
 */
void
ltoieeed(src, dest)

double			*src;
void			*dest;

{
	double		orig;
	XDR		xdrs;

	orig = *src;
	xdrmem_create(&xdrs, (char *) dest, sizeof(double), XDR_ENCODE);

	(void) xdr_double(&xdrs, &orig);
	xdr_destroy(&xdrs);
}

/*
 *	ieeetolf
 *
 *	Function:	- change a float from IEEE big-endian to local
 *	Accepts:	- ptr source area
 *			- ptr destination float
 */
void
ieeetolf(src, dest)

void			*src;
float			*dest;

{
	float		result;
	XDR		xdrs;

	xdrmem_create(&xdrs, (char *) src, sizeof(float), XDR_DECODE);

	(void) xdr_float(&xdrs, &result);
	*dest = result;

	xdr_destroy(&xdrs);
}

/*
 *	ieeetold
 *
 *	Function:	- change a double from IEEE big-endian to local
 *	Accepts:	- ptr source area
 *			- ptr destination double
 */
void
ieeetold(src, dest)

void			*src;
double			*dest;

{
	double		result;
	XDR		xdrs;

	xdrmem_create(&xdrs, (char *) src, sizeof(double), XDR_DECODE);

	(void) xdr_double(&xdrs, &result);
	*dest = result;

	xdr_destroy(&xdrs);
}

/*
 *	mltoieeef
 *
 *	Function:	- change array of floats from local to IEEE big-endian
 *	Accepts:	- ptr array
 *			- # elements in array
 */
void
mltoieeef(array, num)

float			*array;
int			num;

{
	for ( ; num > 0; --num, ++array) {
		ltoieeef(array, (void *) array);
	}
}

/*
 *	mieeetolf
 *
 *	Function:	- change array of floats from IEEE big-endian to local
 *	Accepts:	- ptr array
 *			- # elements in array
 */
void
mieeetolf(array, num)

float			*array;
int			num;

{
	for ( ; num > 0; --num, ++array) {
		ieeetolf((void *) array, array);
	}
}

/*
 *	mltoieeed
 *
 *	Function:	- change array of doubles from local to IEEE big-endian
 *	Accepts:	- ptr array
 *			- # elements in array
 */
void
mltoieeed(array, num)

double			*array;
int			num;

{
	for ( ; num > 0; --num, ++array) {
		ltoieeed(array, (void *) array);
	}
}

/*
 *	mieeetold
 *
 *	Function:	- change array of doubles from IEEE big-endian to local
 *	Accepts:	- ptr array
 *			- # elements in array
 */
void
mieeetold(array, num)

double			*array;
int			num;

{
	for ( ; num > 0; --num, ++array) {
		ieeetold((void *) array, array);
	}
}
#endif /* SCO */
