/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/NJN
 *
 *	$Log:	mpitr_data.c,v $
 * Revision 6.0  96/02/29  13:23:43  gdburns
 * Ohio Release
 * 
 *	Function:	- print MPI message contents
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <mpitrace.h>
#include <portable.h>
#include <t_types.h>

/*
 * constants and macros
 */
#define VCHAR		0
#define VSHORT		1
#define VINT		2
#define VLONG		3
#define VFLOAT		4
#define	VDOUBLE		5
#define	VLONGDOUBLE	6
#define VCOMPLEX        7
#define	V2INT		8
#define V2FLOAT		9
#define V2DOUBLE	10
#define VFLOATINT	11
#define VDOUBLEINT	12
#define VLONGINT	13
#define VSHORTINT	14
#define VLOGIC		15

#define HDISP(x)	(((x) & 0xFFFF) * (((x) >> 16) & 0xFFFF))

/*
 * global functions
 */
void			mpitr_dataprint();

/*
 * local functions
 */
static void		output();
static void		print_bound();
static void		print_char();
static void		print_complex();
static void		print_double();
static void		print_long_double();
static void		print_float();
static void		print_int();
static void		print_logic();
static void		print_short();
static void		print_uchar();
static void		print_uint();
static void		print_ushort();
static void		print_2int();
static void		print_2float();
static void		print_2double();
static void		print_floatint();
static void		print_doubleint();
static void		print_shortint();
static void		unroll();
static void		unroll_contig();
static void		unroll_hvector();
static void		unroll_hindexed();
static void		unroll_struct();

/*
 * local variables
 */
static int		fl_homog;	/* homogeneous LAM flag */
static int		address;	/* line starting address */
static int		eltcount;	/* output line column count */
static int		fl_first;	/* first line flag */
static char		fmtbuf[80];	/* formatting buffer */
/*
 * Maximum number of elements of each type that can be printed on a line.
 */
static int		eltsperline[] = { 8, 8, 4, 4, 4, 4, 1, 2, 2, 2, 2, 2,
					  2, 2, 2, 32 };
/*
 *	mpitr_dataprint
 *
 *	Function:	- print data trace
 *	Accepts:	- trace datatype buffer
 *			- trace data buffer
 *			- data buffer length
 *			- number of leaf elements to print (-1 means all)
 *			- homogeneous LAM flag
 */
void
mpitr_dataprint(dtype, data, datalen, nelts, homog)

char			*dtype;
char			*data;
int			datalen;
int			nelts;
int			homog;

{
	struct trdtype	*ptrace;	/* ptr trace datatype */
	struct trdtype	*p;		/* favourite pointer */
/*
 * Recursively (depth-first) print datatype.
 */
	eltcount = 0;
	address = 0;
	fl_first = 1;
	fl_homog = homog;

	ptrace = (struct trdtype *) (dtype + sizeof(struct trdtype));
	
	while ((datalen > 0) && (nelts != 0)) {
		p = ptrace;

		unroll(&p, &data, &datalen, &nelts);
	}

	printf("\n");
}

/*
 *	unroll
 *
 *	Function:	- unroll datatype & data
 *			- recursive function
 *	Accepts:	- ptr record buffer (value modified)
 *			- ptr data buffer (value modified if non-NULL)
 *			- ptr data buffer length (value modified)
 *			- ptr number elements (value modified)
 */
static void
unroll(pbuffer, pdata, plen, pelts)

struct trdtype		**pbuffer;
char			**pdata;
int			*plen;
int			*pelts;

{
	if ((*plen <= 0) || (*pelts == 0)) return;

	switch((*pbuffer)->trd_dtype) {

	case TRDTUCHAR:
	case TRDTBYTE:
	case TRDTPACKED:
		print_uchar(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTCHAR:
		print_char(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTUINT:
	case TRDTULONG:
		print_uint(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTINT:
	case TRDTLONG:
		print_int(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTLOGIC:
		print_logic(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTSHORT:
		print_short(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTUSHORT:
		print_ushort(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTFLOAT:
		print_float(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTDOUBLE:
		print_double(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTLONGDOUBLE:
		print_long_double(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTCOMPLEX:
		print_complex(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTUB:
	case TRDTLB:
		print_bound(pbuffer, pdata, plen);
		break;

	case TRDT2INT:
	case TRDTLONGINT:
		print_2int(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDT2FLOAT:
		print_2float(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDT2DOUBLE:
		print_2double(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTFLOATINT:
		print_floatint(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTDOUBLEINT:
		print_doubleint(pbuffer, pdata, plen);
		--(*pelts);
		break;

	case TRDTSHORTINT:
		print_shortint(pbuffer, pdata, plen);
		break;

	case TRDTCONTIG:
		unroll_contig(pbuffer, pdata, plen, pelts);
		break;

	case TRDTVECTOR:
		unroll_hvector(pbuffer, pdata, 1, plen, pelts);
		break;

	case TRDTHVECTOR:
		unroll_hvector(pbuffer, pdata, 0, plen, pelts);
		break;

	case TRDTINDEXED:
		unroll_hindexed(pbuffer, pdata, 1, plen, pelts);
		break;

	case TRDTHINDEXED:
		unroll_hindexed(pbuffer, pdata, 0, plen, pelts);
		break;

	case TRDTSTRUCT:
		unroll_struct(pbuffer, pdata, plen, pelts);
		break;

	default:
		break;
	}
}

/*
 *	unroll_contig
 *
 *	Function:	- unroll contiguous datatype & data
 */
static void
unroll_contig(pbuff, pdat, plen, pelts)

struct trdtype		**pbuff;
char			**pdat;
int			*plen;
int			*pelts;

{
	int		count;
	struct trdtype	*savbuff;

	count = (*pbuff)->trd_count;
	savbuff = ++(*pbuff);

	for (; (count > 0) && (*plen > 0) && (*pelts != 0); --count) {
		*pbuff = savbuff;
		unroll(pbuff, pdat, plen, pelts);
	}

	eltcount = 0;
}

/*
 *	unroll_hvector
 *
 *	Function:	- unroll hvector datatype & data
 */
static void
unroll_hvector(pbuff, pdat, fl_vec, plen, pelts)

struct trdtype		**pbuff;
char			**pdat;
int			fl_vec;
int			*plen;
int			*pelts;

{
	int		i, j;
	int		count;
	int		length;
	int		stride;
	int		savaddr;
	struct trdtype	*savbuff;

	count = (*pbuff)->trd_count;
	++(*pbuff);
	length = (*pbuff)->trd_length;
	stride = (*pbuff)->trd_stride;

	if (fl_vec) stride = HDISP(stride);

	savbuff = ++(*pbuff);
	savaddr = address;

	for (i = 0; (i < count) && (*plen > 0) && (*pelts != 0); ++i) {

		address = savaddr;
		savaddr += stride;

		for (j = 0;
		     (j < length) && (*plen > 0) && (*pelts != 0);
		     ++j) {
			*pbuff = savbuff;
			unroll(pbuff, pdat, plen, pelts);
		}

		eltcount = 0;
	}
}

/*
 *	unroll_hindexed
 *
 *	Function:	- unroll hindexed datatype & data
 */
static void
unroll_hindexed(pbuff, pdat, fl_idx, plen, pelts)

struct trdtype		**pbuff;
char			**pdat;
int			fl_idx;
int			*plen;
int			*pelts;

{
	int		i, j;
	int		count;
	int		savaddr;
	int		disp;
	struct trdtype	*savbuff;
	struct trdtype	*idxbuff;

	count = (*pbuff)->trd_count;
	idxbuff = ++(*pbuff);
	*pbuff += count;
	savbuff = *pbuff;
	savaddr = address;

	for (i = 0;
	     (i < count) && (*plen > 0) && (*pelts != 0);
	     ++i, ++idxbuff) {

		disp = idxbuff->trd_disp;
		if (fl_idx) disp = HDISP(disp);

		address = savaddr + idxbuff->trd_disp;

		for (j = 0;
		     (j < idxbuff->trd_length) && (*plen > 0) && (*pelts != 0);
		     ++j) {
			*pbuff = savbuff;
			unroll(pbuff, pdat, plen, pelts);
		}

		eltcount = 0;
	}
}

/*
 *	unroll_struct
 *
 *	Function:	- unroll struct datatype & data
 */
static void
unroll_struct(pbuff, pdat, plen, pelts)

struct trdtype		**pbuff;
char			**pdat;
int			*plen;
int			*pelts;

{
	int		i, j;
	int		count;
	int		length;
	int		savaddr;
	struct trdtype	*savbuff;

	count = (*pbuff)->trd_count;
	++(*pbuff);
	savaddr = address;
 
	for (i = 0; (i < count) && (*plen > 0) && (*pelts != 0); ++i) {

		length = (*pbuff)->trd_length;
		address = savaddr + (*pbuff)->trd_disp;

		savbuff = ++(*pbuff);

		for (j = 0;
		     (j < length) && (*plen > 0) && (*pelts != 0);
		     ++j) {
			*pbuff = savbuff;
			unroll(pbuff, pdat, plen, pelts);
		}

		eltcount = 0;
	}
}

/*
 *	print_uchar
 *
 *	Function:	- print unsigned character
 */
static void
print_uchar(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	unsigned char	c;

	if (*plen >= 1) {

		++(*pbuff);
		c = *((unsigned char *) *pdata);
		++(*pdata);

		sprintf(fmtbuf, isprint(c) ? "%4c" : "%4u", c & 0xFF);
		output(fmtbuf, eltsperline[VCHAR]);

		++address;
	}

	--(*plen);
}

/*
 *	print_char
 *
 *	Function:	- print character
 */
static void
print_char(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	SIGNED char	c;

	if (*plen >= 1) {

		++(*pbuff);
		c = **pdata;
		++(*pdata);

		if (c >= 0) {
			sprintf(fmtbuf, isprint(c) ? "%4c" : "%4u", c & 0xFF);
		} else {
			sprintf(fmtbuf, "%4d", c);
		}

		output(fmtbuf, eltsperline[VCHAR]);

		++address;
	}

	--(*plen);
}

/*
 *	print_int
 *
 *	Function:	- print integer
 */
static void
print_int(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	int4		i;

	if (*plen >= sizeof(int4)) {

		++(*pbuff);
		memcpy((char *) &i, *pdata, sizeof(int4));
		*pdata += sizeof(int4);

		if ( ! fl_homog) ttoli4(&i, &i);
		sprintf(fmtbuf, "%15d", i);
		output(fmtbuf, eltsperline[VINT]);

		address += sizeof(int4);
	}

	*plen -= sizeof(int4);
}

/*
 *	print_uint
 *
 *	Function:	- print unsigned integer
 */
static void
print_uint(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	uint4		i;

	if (*plen >= sizeof(uint4)) {

		++(*pbuff);
		memcpy((char *) &i, *pdata, sizeof(uint4));
		*pdata += sizeof(uint4);

		if ( ! fl_homog) ttoli4(&i, &i);
		sprintf(fmtbuf, "%15u", i);
		output(fmtbuf, eltsperline[VINT]);

		address += sizeof(uint4);
	}

	*plen -= sizeof(uint4);
}

/*
 *	print_logic
 *
 *	Function:	- print boolean
 */
static void
print_logic(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	uint4		i;

	if (*plen >= sizeof(uint4)) {

		++(*pbuff);
		memcpy((char *) &i, *pdata, sizeof(uint4));
		*pdata += sizeof(uint4);

		if ( ! fl_homog) ttoli4(&i, &i);
		sprintf(fmtbuf, "%1s", (i) ? "F" : "T");
		output(fmtbuf, eltsperline[VLOGIC]);

		address += sizeof(uint4);
	}

	*plen -= sizeof(uint4);
}

/*
 *	print_short
 *
 *	Function:	- print short integer
 */
static void
print_short(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	int2		i;

	if (*plen >= sizeof(int2)) {

		++(*pbuff);
		memcpy((char *) &i, *pdata, sizeof(int2));
		*pdata += sizeof(int2);

		if ( ! fl_homog) ttoli2(&i, &i);
		sprintf(fmtbuf, "%7d", (int) i);
		output(fmtbuf, eltsperline[VSHORT]);

		address += sizeof(int2);
	}

	*plen -= sizeof(int2);
}

/*
 *	print_ushort
 *
 *	Function:	- print unsigned short integer
 */
static void
print_ushort(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	uint2		i;

	if (*plen >= sizeof(uint2)) {

		++(*pbuff);
		memcpy((char *) &i, *pdata, sizeof(uint2));
		*pdata += sizeof(uint2);

		if ( ! fl_homog) ttoli2(&i, &i);
		sprintf(fmtbuf, "%7u", (unsigned) i);
		output(fmtbuf, eltsperline[VSHORT]);

		address += sizeof(uint2);
	}

	*plen -= sizeof(uint2);
}

/*
 *	print_float
 *
 *	Function:	- print single-precision floating point
 */
static void
print_float(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float4		f;

	if (*plen >= sizeof(float4)) {

		++(*pbuff);
		memcpy((char *) &f, *pdata, sizeof(float4));
		*pdata += sizeof(float4);

		if ( ! fl_homog) ttolf4(&f, &f);
		sprintf(fmtbuf, "%15g", f);
		output(fmtbuf, eltsperline[VFLOAT]);

		address += sizeof(float4);
	}

	*plen -= sizeof(float4);
}

/*
 *	print_double
 *
 *	Function:	- print double-precision floating point
 */
static void
print_double(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float8		f;

	if (*plen >= sizeof(float8)) {

		++(*pbuff);
		memcpy((char *) &f, *pdata, sizeof(float8));
		*pdata += sizeof(float8);

		if ( ! fl_homog) ttolf8(&f, &f);
		sprintf(fmtbuf, "%#15g", f);
		output(fmtbuf, eltsperline[VDOUBLE]);

		address += sizeof(float8);
	}

	*plen -= sizeof(float8);
}

/*
 *	print_long_double
 *
 *	Function:	- print quad-precision floating point
 */
static void
print_long_double(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float8		d0, d1;
	int		*p0, *p1;
	int		*p2, *p3;
/*
 * Just hex. dump for now.
 */
	if (*plen >= (2 * sizeof(float8))) {

		++(*pbuff);
		memcpy((char *) &d0, *pdata, sizeof(float8));
		*pdata += sizeof(float8);
		memcpy((char *) &d1, *pdata, sizeof(float8));
		*pdata += sizeof(float8);

		if ( ! fl_homog) {
			ttolf8(&d0, &d0);
			ttolf8(&d1, &d1);
		}

		p0 = (int *) &d0;
		p1 = p0 + 1;
		p2 = (int *) &d1;
		p3 = p2 + 1;

		sprintf(fmtbuf, "0x%8X%8X%8X%8X", *p0, *p1, *p2, *p3);
		output(fmtbuf, eltsperline[VLONGDOUBLE]);

		address += 2 * sizeof(float8);
	}

	*plen -= 2 * sizeof(float8);
}

/*
 *	print_complex
 *
 *	Function:	- print complex
 */
static void
print_complex(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float4		fr, fi;

	if (*plen >= (2 * sizeof(float4))) {

		++(*pbuff);
		memcpy((char *) &fr, *pdata, sizeof(float4));
		*pdata += sizeof(float4);
		memcpy((char *) &fi, *pdata, sizeof(float4));
		*pdata += sizeof(float4);

		if ( ! fl_homog) {
			ttolf4(&fr, &fr);
			ttolf4(&fi, &fi);
		}

		sprintf(fmtbuf, "%15g %15g", fr, fi);
		output(fmtbuf, eltsperline[VCOMPLEX]);

		address += 2 * sizeof(float4);
	}

	*plen -= 2 * sizeof(float4);
}

/*
 *	print_bound
 *
 *	Function:	- print upper/lower bound
 */
static void
print_bound(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	if (*plen >= 0) {
		++(*pbuff);
		*pdata += 0;	/* keep it orthogonal & use all vars */
	}

	*plen -= 0;
}

/*
 *	print_2int
 *
 *	Function:	- print 2 integers
 */
static void
print_2int(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	int4		i1, i2;

	if (*plen >= (2 * sizeof(int4))) {

		++(*pbuff);
		memcpy((char *) &i1, *pdata, sizeof(int4));
		*pdata += sizeof(int4);
		memcpy((char *) &i2, *pdata, sizeof(int4));
		*pdata += sizeof(int4);

		if ( ! fl_homog) {
			ttoli4(&i1, &i1);
			ttoli4(&i2, &i2);
		}

		sprintf(fmtbuf, "%15d %15d", i1, i2);
		output(fmtbuf, eltsperline[V2INT]);

		address += 2 * sizeof(int4);
	}

	*plen -= 2 * sizeof(int4);
}

/*
 *	print_2float
 *
 *	Function:	- print 2 floats
 */
static void
print_2float(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float4		f1, f2;

	if (*plen >= (2 * sizeof(float4))) {

		++(*pbuff);
		memcpy((char *) &f1, *pdata, sizeof(float4));
		*pdata += sizeof(float4);
		memcpy((char *) &f2, *pdata, sizeof(float4));
		*pdata += sizeof(float4);

		if ( ! fl_homog) {
			ttolf4(&f1, &f1);
			ttolf4(&f2, &f2);
		}

		sprintf(fmtbuf, "%15g %15g", f1, f2);
		output(fmtbuf, eltsperline[V2FLOAT]);

		address += 2 * sizeof(float4);
	}

	*plen -= 2 * sizeof(float4);
}

/*
 *	print_2double
 *
 *	Function:	- print 2 doubles
 */
static void
print_2double(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float8		f1, f2;

	if (*plen >= (2 * sizeof(float8))) {

		++(*pbuff);
		memcpy((char *) &f1, *pdata, sizeof(float8));
		*pdata += sizeof(float8);
		memcpy((char *) &f2, *pdata, sizeof(float8));
		*pdata += sizeof(float8);

		if ( ! fl_homog) {
			ttolf8(&f1, &f1);
			ttolf8(&f2, &f2);
		}

		sprintf(fmtbuf, "%15g %15g", f1, f2);
		output(fmtbuf, eltsperline[V2DOUBLE]);

		address += 2 * sizeof(float8);
	}

	*plen -= 2 * sizeof(float8);
}

/*
 *	print_floatint
 *
 *	Function:	- print a float and an integer
 */
static void
print_floatint(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float4		f;
	int4		i;

	if (*plen >= (sizeof(float4) + sizeof(int4))) {

		++(*pbuff);
		memcpy((char *) &f, *pdata, sizeof(float4));
		*pdata += sizeof(float4);
		memcpy((char *) &i, *pdata, sizeof(int4));
		*pdata += sizeof(int4);

		if ( ! fl_homog) {
			ttolf4(&f, &f);
			ttoli4(&i, &i);
		}

		sprintf(fmtbuf, "%15g %15d", f, i);
		output(fmtbuf, eltsperline[VFLOATINT]);

		address += sizeof(float4) + sizeof(int4);
	}

	*plen -= sizeof(float4) + sizeof(int4);
}

/*
 *	print_doubleint
 *
 *	Function:	- print a double and an integer
 */
static void
print_doubleint(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	float8		f;
	int4		i;

	if (*plen >= (sizeof(float8) + sizeof(int4))) {

		++(*pbuff);
		memcpy((char *) &f, *pdata, sizeof(float8));
		*pdata += sizeof(float8);
		memcpy((char *) &i, *pdata, sizeof(int4));
		*pdata += sizeof(int4);

		if ( ! fl_homog) {
			ttolf8(&f, &f);
			ttoli4(&i, &i);
		}

		sprintf(fmtbuf, "%15g %15d", f, i);
		output(fmtbuf, eltsperline[VDOUBLEINT]);

		address += sizeof(float8) + sizeof(int4);
	}

	*plen -= sizeof(float8) + sizeof(int4);
}

/*
 *	print_shortint
 *
 *	Function:	- print a short and an integer
 */
static void
print_shortint(pbuff, pdata, plen)

struct trdtype		**pbuff;
char			**pdata;
int			*plen;

{
	int2		s;
	int4		i;

	if (*plen >= (sizeof(int2) + sizeof(int4))) {

		++(*pbuff);
		memcpy((char *) &s, *pdata, sizeof(int2));
		*pdata += sizeof(int2);
		memcpy((char *) &i, *pdata, sizeof(int4));
		*pdata += sizeof(int4);

		if ( ! fl_homog) {
			ttoli2(&s, &s);
			ttoli4(&i, &i);
		}

		sprintf(fmtbuf, "%7d %15d", (int) s, i);
		output(fmtbuf, eltsperline[VSHORTINT]);

		address += sizeof(int2) + sizeof(int4);
	}

	*plen -= sizeof(int2) + sizeof(int4);
}

/*
 *	output
 *
 *	Function:	- print string (representing a basic element) out
 *			- track # elements per line
 *			- print address at start of line
 *			- handle continuation lines
 *	Accepts:	- string representation of element
 *			- max number of elements per line
 */
static void
output(string, maxelts)

char			*string;
int			maxelts;

{
	int		len;
/*
 * Break long lines and print address at the beginning of new lines.
 */
	len = strlen(string);

	if ((eltcount == 0) || (eltcount >= maxelts)) {

		if (fl_first) {
			fl_first = 0;
		} else {
			printf("\n");
		}

		printf("%8x:", address);
		eltcount = 0;
	}
/*
 * Print the string.
 */
	printf(" %s", string);
	eltcount++;
}
