/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	fface.c,v $
 * Revision 6.0  96/02/29  13:30:53  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/26  21:06:42  gdburns
 * Remove setsid().
 * 
 * Revision 5.2  94/08/22  13:47:46  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.2  94/08/19  16:05:56  raja
 * Add SCO support.
 * 
 * Revision 5.1.1.1  94/08/18  10:49:11  vaigl-j
 * Add HP-UX support.
 * 
 * Revision 5.1  94/05/18  10:41:49  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:26:26  gdburns
 * Ohio Release
 * 
 * Revision 2.2.1.6  94/02/24  15:20:54  raja
 * Convert the LAM_O_ flags to UNIX O_ flags.
 * 
 * Revision 2.2.1.5  94/02/17  17:16:00  raja
 * Use getcwd() in Solaris.
 * 
 * Revision 2.2.1.4  94/02/16  14:06:51  gdburns
 * Convert to internal process.
 * 
 * Revision 2.2.1.3  93/05/04  12:12:46  raja
 * Comment out the "extern int" function declarations.
 * Compact the buffer filling in conv_stat().
 * 
 * Revision 2.2.1.2  93/05/03  11:42:45  gdburns
 * Use off_t with lseek().
 * Restrict stat() to POSIX fields.
 * 
 * Revision 2.2.1.1  92/06/05  15:59:40  raja
 * Use uint4 instead of "unsigned int4".
 * 
 * Revision 2.2  92/04/30  14:18:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/02/28  10:45:09  gdburns
 * Give our best effort to stat() and fstat().
 * 
 * Revision 2.1.1.2  91/12/18  16:27:43  gdburns
 * Use NOCTTY on opens.
 * Use setsid() to get away from process groups and
 * controlling terminals.
 * 
 * Revision 2.1.1.1  91/06/05  15:11:33  gdburns
 * Pass a pointer not the whole stat structure in stat() and fstat().
 * 
 * Revision 2.1  91/02/24  20:59:37  gdburns
 * Ohio Release
 * 
 *	Function:	- abstracts native filesystem interface
 *			- Trollius remote file access follows the UNIX I/O
 *			  standard assuming a 32 bit interface (int4).
 *			- In case the standard UNIX calls are not available,
 *			  they may be emulated in this module.
 *			- UNIX version
 */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include "filed.h"
#include <freq.h>
#include <net.h>
#include <typical.h>

/*
 * external functions
 */
/*
 * extern int		open();
 * extern int		close();
 * extern int		read();
 * extern int		write();
 * extern int		chdir();
 * extern int		access();
 * extern int		stat();
 * extern int		fstat();
 * extern int		truncate();
 * extern int		ftruncate();
 * extern int		mkdir();
 * extern int		rmdir();
 * extern int		unlink();
 * extern int		system();
 */
extern off_t		lseek();
extern char		*getwd();
extern pid_t		setsid();

/*
 * local functions
 */
static void		conv_stat();
static int4		conv_flags();

/*
 * UNIX version - boring, so just blurt them out.
 * Here we know that int4 is int.
 * In fmyopen(), convert LAM flags to their UNIX equivalent.
 */

int4
fmyopen(path, flags, mode)

char			*path;
int4			flags;
int4			mode;

{
	int4		fd;

	fd = open(path, conv_flags(flags) | O_NOCTTY, mode);
	return(fd);
}

int4 fmyclose(fd) int4 fd;
{ return(close(fd)); }

int4 fmyread(fd, buf, nbytes) int4 fd; char *buf; int4 nbytes;
{ return(read(fd, buf, nbytes)); }

int4 fmywrite(fd, buf, nbytes) int4 fd; char *buf; int4 nbytes;
{ return(write(fd, buf, nbytes)); }

int4 fmylseek(fd, offset, whence) int4 fd; int4 offset; int4 whence;
{ return(lseek(fd, (off_t) offset, whence)); }

int4 fmychdir(newdir) char *newdir;
{ return(chdir(newdir)); }

char *fmygetwd(dir) char *dir;
#if SUN4_SOL || HPUX || SCO
{ extern char *getcwd(); return(getcwd(dir, MAXNMSGLEN)); }
#else
{ return(getwd(dir)); }
#endif

int4 fmyunlink(name) char *name;
{ return(unlink(name)); }

int4 fmyrmdir(dir) char *dir;
{ return(rmdir(dir)); }

int4 fmysystem(cmd) char *cmd;
{ return(system(cmd)); }

int4 fmymkdir(path, mode) char *path; int4 mode;
{ return(mkdir(path, mode)); }

int4 fmyaccess(path, mode) char *path; int4 mode;
{ return(access(path, mode)); }

#ifndef SCO

int4 fmytrunc(path, len) char *path; int4 len;
{ return(truncate(path, (off_t) len)); }

int4 fmyftrunc(fd, len) int4 fd; int4 len;
{ return(ftruncate(fd, (off_t) len)); }

#else /* SCO */

/*
 * Simulate truncate() by opening the file.
 */
int4 fmytrunc(path, len) char *path; int4 len;
{
	int4		fd;

	fd = open(path, O_WRONLY, 0);
	if (fd < 0) {
		return((int4) ERROR);
	}

	if (chsize(fd, (off_t) len)) {
		return((int4) ERROR);
	}

	return((int4) close(fd));
}

int4 fmyftrunc(fd, len) int4 fd; int4 len;
{ return(chsize(fd, (off_t) len)); }

#endif /* SCO */

/*
 *	fmyinit
 *
 *	Function:	- machine dependent initialization
 *			- in UNIX case, take care of process group/
 *			  controlling terminal headaches
 *	Returns:	- 0 or ERROR
 */
int
fmyinit()

{
/*
 * Create a new session.  This detaches our controlling terminal
 * so we don't get TTIN and TTOU signals when reading/writing the tty.
 * We also get out of the tty's process group so that we don't
 * catch special character signals from it.
 *
 * In the future, we will have worker processes to deal with ttys
 * (and all weird and/or slow devices) so we won't have this hassle.
 */
/*
 * I thought I vaguely understood this subject.  Current testing reveals
 * that neither of the above two problems exists if I fail to call setsid().
 * It is clearly documented that I cannot make this call if I am already
 * a session leader and I want to be able to run lamd from the shell.
 *
 *	if (setsid() < 0) {
 *		return(ERROR);
 *	}
 */
	return(0);
}

/*
 *	fmystat
 */
int4
fmystat(path, buf)

char			*path;
struct stat		*buf;

{
	struct stat	status;
	int4		ret;

	ret = stat(path, &status);
	conv_stat(&status, (unsigned int *) buf);
	return(ret);
}

/*
 *	fmyfstat
 */
int4
fmyfstat(fd, buf)

int4			fd;
struct stat		*buf;

{
	struct stat	status;
	int4		ret;

	ret = fstat(fd, &status);
	conv_stat(&status, (unsigned int *) buf);
	return(ret);
}

/*
 *	conv_stat
 *
 *	Function:	- stores status structure in an int4 array
 *	Accepts:	- ptr to stat structure
 *			- ptr to int4 array
 */
static void
conv_stat(status, buf)

struct stat		*status;
uint4			*buf;

{
	buf[0] = (uint4) status->st_dev;
	buf[1] = (uint4) status->st_ino;
	buf[2] = (uint4) status->st_mode;
	buf[3] = (uint4) status->st_nlink;
	buf[4] = (uint4) status->st_uid;
	buf[5] = (uint4) status->st_gid;
	buf[6] = (uint4) status->st_size;
	buf[7] = (uint4) status->st_atime;
	buf[8] = (uint4) status->st_mtime;
	buf[9] = (uint4) status->st_ctime;
}

/*
 *	conv_flags
 *
 *	Function:	- convert LAM flags to UNIX open() flags
 *	Accepts:	- LAM flags
 *	Returns:	- UNIX flags
 */
static int4
conv_flags(lflags)

int4			lflags;

{
	int4		uflags;

	uflags = 0;

	if (lflags & LAM_O_RDONLY) { uflags |= O_RDONLY; }
	if (lflags & LAM_O_WRONLY) { uflags |= O_WRONLY; }
	if (lflags & LAM_O_RDWR) { uflags |= O_RDWR; }
	if (lflags & LAM_O_APPEND) { uflags |= O_APPEND; }
	if (lflags & LAM_O_CREAT) { uflags |= O_CREAT; }
	if (lflags & LAM_O_EXCL) { uflags |= O_EXCL; }
	if (lflags & LAM_O_TRUNC) { uflags |= O_TRUNC; }

	return(uflags);
}
