/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB
 *
 *	$Log:	pqcreate.c,v $
 * Revision 6.0.1.1  96/06/10  15:55:11  nevin
 * Set MPI working directory
 * 
 * Revision 6.0  96/02/29  13:31:44  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/27  20:11:27  gdburns
 * general modernization
 * 
 * Revision 5.2  94/08/22  13:50:29  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:42:17  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:27:39  gdburns
 * Ohio Release
 *
 *	Function:	- OTB kenyad
 *			- creates a process
 */

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <args.h>
#include <kio.h>
#include <kreq.h>
#include <net.h>
#include <preq.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * global functions
 */
int			pcreate();	/* create process */
char			*argsread();	/* read argv from file */

/*
 * external variables
 */
extern struct pprocess	ptable[];	/* kenya process table */
extern struct preq	pq;		/* kenya request */
extern struct preply	pr;		/* kenya reply */
extern char		**pathv;	/* argv of search paths */

/*
 * external functions
 */
extern void		psendr();
extern int		putenv();
extern int		pfind();
extern char		*_path_env_find();
extern char		*rforget();	/* tag -> filename */

/*
 * local variables
 */
static char		fdbuf[32];		/* space for TROLLIUSFD env */
static char		rtfbuf[32];		/* space for TROLLIUSRTF env */
static char		worldbuf[32];		/* space for LAMWORLD env */
static char		parentbuf[32];		/* space for LAMPARENT env */
static char		dirbuf[PSMAXNAME + 11];	/* space for LAMWORKDIR env */

void
pqcreate()

{
	char		*args;		/* glued argv */
	char		**argv;		/* new process args */
	char		*argvname;	/* argv filename */
	char		*loadname;	/* load module filename */
	int		argc;		/* # args */

	pr.pr_reply = 0;
	loadname = 0;
	argvname = 0;
/*
 * Find the filename of the load module.
 */
	if (pq.pq_rtf & RTF_FLAT) {
		loadname = rforget(LOCAL, pq.pq_loadpt);
	} else {
		loadname = _path_env_find(pq.pq_name, R_OK | X_OK);
	}

	if (loadname == 0) {
		pr.pr_reply = ltot(errno);
	}
/*
 * Convert argv from tag to filename.
 */
	if (pq.pq_argv) {
		argvname = rforget(LOCAL, pq.pq_argv);

		if (argvname == 0) {
			pr.pr_reply = ltot(errno);
		}
/*
 * Read in argv structure.
 */
		else {
			args = argsread(argvname);

			if (args == 0) {
				pr.pr_reply = ltot(errno);
			}
/*
 * Reconstruct the argv structure.
 */
			else {
				argv = argvbreak(args, 0xa);
				if (argv == 0) lampanic("kenyad (argvbreak)");

				free(args);
				argc = argvcount(argv);
			}

			unlink(argvname);
			free(argvname);
		}
	}
/*
 * Construct an argv from the executable name.
 */
	else if (loadname) {
		argc = 1;
		argv = argvbreak(loadname, 0xa);
		if (argv == 0) lampanic("kenyad (argvbreak)");
	}

	else {
		argv = 0;
	}
/*
 * Clean up.
 */
	if (pr.pr_reply) {

		if (loadname) {

			if (pq.pq_rtf & RTF_FLAT) {
				unlink(loadname);
			}

			free(loadname);
		}

		if (argv) {
			free((char *) argv);
		}
	}
/*
 * Create the process.
 */
	else {
		if (pcreate(loadname, argc, argv)) {
			pr.pr_reply = ltot(errno);
		}
	}
/*
 * Reply to client.
 */
	psendr((char *) &pr, sizeof(pr), 0);
}

/*
 *	pcreate
 *
 *	Function:	- creates a new process for a given load module
 *	Accepts:	- load module filename
 *			- command line args
 *	Returns:	- 0 or ERROR
 */
int
pcreate(loadname, argc, argv)

char			*loadname;
int4			argc;
char			**argv;

{
	struct pprocess	*pnew;		/* newborn process pointer */
	int		inew;		/* index of newborn process */
	sigset_t	newset;		/* new signal set */
	sigset_t	oldset;		/* old signal set */
/*
 * Find a slot in the process table.
 */
	inew = pfind((int4) NOTPROCESS);

	if (inew < 0) {
		errno = ENOPDESCRIPTORS;
		return(LAMERROR);
	}

	pnew = ptable + inew;
	pnew->p_loadpt = loadname;
	pnew->p_rtf = pq.pq_rtf | RTF_KENYA;
	pnew->p_argc = argc;
	pnew->p_argv = argv;
	pnew->p_nodeid = pq.pq_src_node;
	pnew->p_event = pq.pq_src_event;
/*
 * Pass Trillium runtime variables through environment.
 */
	sprintf(fdbuf, "TROLLIUSFD=%d:%d:%d", pq.pq_stdin,
			pq.pq_stdout, pq.pq_stderr);

	if (putenv(fdbuf)) lampanic("kenyad (putenv)");

	sprintf(rtfbuf, "TROLLIUSRTF=%d", pnew->p_rtf);

	if (putenv(rtfbuf)) lampanic("kenyad (putenv)");

	sprintf(worldbuf, "LAMWORLD=%d", pq.pq_world);

	if (putenv(worldbuf)) lampanic("kenyad (putenv)");

	sprintf(parentbuf, "LAMPARENT=%d", pq.pq_parent);

	if (putenv(parentbuf)) lampanic("kenyad (putenv)");

	if (pq.pq_rtf & RTF_MPICWD) {

		sprintf(dirbuf, "LAMWORKDIR=%s", loadname);

		if (putenv(dirbuf)) lampanic("kenyad (putenv)");
	}
/*
 * Create the process.
 */
	sigemptyset(&newset);
	sigaddset(&newset, SIGCHLD);
	sigprocmask(SIG_BLOCK, &newset, &oldset);

	pnew->p_pid = kcreate(loadname, argv);

	sigprocmask(SIG_SETMASK, &oldset, (sigset_t *) 0);

	if (pnew->p_pid < 0) {
		return(LAMERROR);
	}

	pr.pr_pid = ltot(pnew->p_pid);
	pr.pr_index = ltot(inew + 1);
	return(0);
}

/*
 *	argsread
 *
 *	Function:	- reads in glued argv structure from a file
 *	Accepts:	- filename
 *	Returns:	- argv ptr or null ptr
 */
char *
argsread(argvname)

char			*argvname;

{
	char		*args;		/* glued argv buffer */
	int		avfd;		/* argv file descriptor */
	struct stat	avstat;		/* argv file status */

	if (stat(argvname, &avstat)) return(0);

	args = malloc((unsigned int) avstat.st_size);
	if (args == 0) lampanic("kenyad (malloc)");

	avfd = open(argvname, O_RDONLY, 0);

	if (avfd < 0) {
		free(args);
		return(0);
	}

	if (read(avfd, args, (int) avstat.st_size) < 0) {
		free(args);
		return(0);
	}

	close(avfd);
	return(args);
}
