/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB
 *
 *	$Log:	pqdoom.c,v $
 * Revision 6.0  96/02/29  13:31:46  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/27  20:11:30  gdburns
 * general modernization
 * 
 * Revision 5.2  94/08/22  13:50:31  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:42:19  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:27:40  gdburns
 * Ohio Release
 * 
 *	Function:	- kenyad
 *			- delivers a signal to local processes
 */

#include <kio.h>
#include <preq.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * public functions
 */
void			pqdoom();
void			pdoom();

/*
 * external variables
 */
extern struct pprocess	ptable[];	/* kenya process table */
extern struct preq	pq;		/* kenya request */
extern struct preply	pr;		/* kenya reply */

/*
 * external functions
 */
extern void		pdetachindex();
extern void		psendr();
extern int		pmatch();

/*
 *	pqdoom
 *
 *	Function:	- signal delivery server
 */
void
pqdoom()

{
	pdoom(&pq, &pr);
/*
 * Send reply of doom operation back to client.
 */
	psendr((char *) &pr, sizeof(pr), 0);
}

/*
 *	pdoom
 *
 *	Function:	- delivers signals to matching processes
 *			- also called internally in lamd
 *			- checks for a match on each used process descriptor
 *			- signals each matched process using kdoom()
 *			- if no process is matched and a pid is supplied,
 *			  attempts to doom process with the kernel only
 */
void
pdoom(request, reply)

struct preq		*request;
struct preply		*reply;

{
	int		i;
	int		found_match;	/* did anything match? */

	found_match = FALSE;
	reply->pr_reply = 0;

	for (i = 0; (i < PMAX) && !pr.pr_reply; ++i) {

		if (!pmatch(ptable + i, request->pq_sel_method,
				request->pq_sel_value)) continue;

		found_match = TRUE;
/*
 * Signal process.
 */
		if (kdoom(ptable[i].p_pid, request->pq_signal)) {

			if (errno != ESERVICE) {

				if (!(ptable[i].p_rtf & RTF_KENYA)) {
					pdetachindex(i);
					found_match = FALSE;
				} else {
					reply->pr_reply = ltot(errno);
				}
			}
		}
	}

	if (!found_match && ((request->pq_sel_method == SELECT_PID) ||
			(request->pq_sel_method == SELECT_INDEX))) {
		reply->pr_reply = ltot(ENOTPROCESS);
	}
}
