/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *      RBD
 *
 *	$Log:	all_shash.c,v $
 * Revision 6.0  96/02/29  13:46:24  gdburns
 * Ohio Release
 * 
 * Revision 5.2  94/08/22  13:56:56  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:55:59  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:34:53  gdburns
 * Ohio Release
 * 
 * Revision 2.2  92/04/30  14:40:10  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/10/20  21:24:05  raja
 * initial revision
 * 
 *	Function:	- generic hash table management code
 *			- fully static version
 *				- ahs_init(): create a hash table
 *			- other functions are macro'ed to the dynamic version
 *			- the record used must have an int4 key as
 *			  its first field
 */


#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <all_hash.h>

/*
 * external variables
 */
extern int		errno;


/*
 *	ahs_init
 *
 *	Function:	- create a static hash table
 *
 *	Accepts:	- size of hash table
 *			- size of hash table element
 *			- value of the null hash key
 *			- operation mode
 *			- hash table array
 *			- LRU counters array
 *			- ptr to static hash table descriptor
 *
 *	Returns:	- ptr to static hash table descriptor or NULL
 */
SHASH * 
ahs_init(size, elemsize, nullkey, mode, htable, lru, ahsd)

int4			size;
int4			elemsize;
int4			nullkey;
int4			mode;
void			*htable;
int4			*lru;
SHASH			*ahsd;

{
	int4		i;		/* favorite counter */
	int4		*p;		/* favorite pointer */
/* 
 * Handle the trivial cases.
 * The element must be big enough to hold the int4 key.
 */
	if ( (size <= INT4_NIL) || (elemsize < sizeof(int4)) ) {
		errno = EINVAL;
		return((SHASH *) 0);
	}
/*
 * Initialize the static hash table descriptor.
 */
	ahsd->ah_maxnelem = size;
	ahsd->ah_elemsize = elemsize;
	ahsd->ah_nelem = INT4_NIL;
	ahsd->ah_nullkey = nullkey;
	ahsd->ah_mode = mode;
	ahsd->ah_table = htable;
	if ((mode & AHLRU) == AHLRU) {
		ahsd->ah_lru = lru;
	} else {
		ahsd->ah_lru = (int4 *) 0;
	}
/*
 * Initialize the hash table if AHNOINIT is not specified.
 * Reset all LRU counters if AHLRU is specified.
 */
	if ((mode & AHNOINIT) != AHNOINIT) {
		for (i = 0, p = (int4 *) ahsd->ah_table; i < size; i++) {
			*p = nullkey;
			p = (int4 *) ((char *) p + elemsize);
		}
	}

	if ((mode & AHLRU) == AHLRU) {
		for (i = 0; i < size; i++) {
			ahsd->ah_lru[i] = INT4_NIL;
		}
	}

	return(ahsd);
}
