/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	NJN
 *
 *	$Log:	mpiformout.c,v $
 * Revision 6.0  96/02/29  13:47:24  gdburns
 * Ohio Release
 * 
 *	Function:	- format/ouput mpi information
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <mpitrace.h>
#include <mpisys.h>

/*
 * global functions
 */
void			format_proc();
void			format_cid();
void			format_tag();
void			format_dtype();
void			print_dtype();
void			print_comm();

/*
 * external variables
 */
extern int		mpitr_comget();
extern int		mpitr_dtypeget();
extern void		mpitr_dtypeprint();
extern void		mpitr_comprint();

/*
 * basic datatypes
 */
static char		*dtbasic[TRDTMAX + 1] = {
				"CONTIG", "VECTOR", "HVECTOR",
				"INDEX", "HINDEX", "STRUCT",
				"BYTE", "CHAR",
				"UCHAR", "SHORT", "USHORT",
				"INT", "UINT", "LOGIC",
				"LONG", "ULONG", "FLOAT",
				"DOUBLE", "COMPLEX", "UB",
				"LB", "PACKED", "2INT",
				"2FLOAT", "2DOUBLE", "FLOAT_INT",
				"DOUBLE_INT", "LONG_INT", 
				"SHORT_INT", "LONG_DOUBLE"
			};

/*
 *	format_proc
 *
 *	Function:	- format a process description for output 
 *	Accepts:	- process node
 *			- process index
 *			- process global rank
 *			- process local rank
 *			- GPS flag (on implies use GPS)
 *			- buffer to format into (returned value)
 */
void
format_proc(node, index, grank, lrank, gps, str)

int			node;
int			index;
int			grank;
int			lrank;
int			gps;
char			*str;

{
	if (node == -1 || grank == -1) {
		strcpy(str, "ANY");
	}
	else if (gps) {
		if (lrank == -1) sprintf(str, "n%d,i%d", node, index);
		else sprintf(str, "n%d,i%d/%d", node, index, lrank);
	}
	else {
		if (lrank == -1) sprintf(str, "%d", grank);
		else sprintf(str, "%d/%d", grank, lrank);
	}
}

/*
 *	format_cid
 *
 *	Function:	- format a cid for output
 *	Accepts:	- context id
 *			- buffer to format into (returned value)
 */
void
format_cid(cid, str)

int			cid;
char			*str;

{
	int 		collective;	/*  is comm collective? */

	collective = (cid < 0);

	cid = __lam_coll2pt_(cid);
	
	if (cid == 0) {
		strcpy(str, "WORLD");
	}
	else if (cid == 1) {
		strcpy(str, "SELF");
	}
	else {
		sprintf(str, "<%d>", cid);
	}

	if (collective) strcat(str, "*");
}

/*
 *	format_tag
 *
 *	Function:	- format a tag for output
 *	Accepts:	- tag
 *			- buffer to format into (returned value)
 */
void
format_tag(tag, str)

int			tag;
char			*str;

{
	if (tag == MPI_ANY_TAG) {
		strcpy(str, "ANY");
	} else {
		sprintf(str, "%d", tag);
	}
}

/*
 *	format_dtype
 *
 *	Function:	- format a datatype for output
 *	Accepts:	- datatype
 *			- buffer to format into (returned value)
 *			- flag, on means print full name
 */
void
format_dtype(dtype, str, full)

int			dtype;
char			*str;
int			full;

{
	*str = 0;
	
	if (dtype < 0) return;

	if (dtype <= TRDTMAX) {
		if (full) strcpy(str, "MPI_");
		strcat(str, dtbasic[dtype]);
	}
	else {
		sprintf(str, "<%d>", dtype);
	}
}

/*
 *	print_dtype
 *
 *	Function:	- gets datatype trace and prints to stdout
 *	Accepts:	- node of process that defined datatype
 *			- pid of process that defined datatype
 *			- datatype label
 *			- prefix for output
 */
void
print_dtype(node, pid, dtype, prefix)

int			node;
int			pid;
int			dtype;
char			*prefix;

{
	char		*dtbuf;		/* ptr datatype trace buffer */
	
	if (mpitr_dtypeget(node, pid, dtype, &dtbuf)) {
		printf("%sData type information unavailable.\n", prefix);
		return;
	}

	mpitr_dtypeprint(dtype, dtbuf, 2, prefix, 0);

	free(dtbuf);
}

/*
 *	print_comm
 *
 *	Function:	- gets communicator trace and prints to stdout
 *	Accepts:	- node of process that defined communicator
 *			- pid of process that defined communicator
 *			- context ID
 *			- GPS flag, (on means show GPS's) 
 */
void
print_comm(node, pid, cid, gps)

int			node;
int			pid;
int			cid;
int			gps;

{
	char		*commbuf;	/* ptr to comm trace buffer */

	if ((__lam_coll2pt_(cid) != 1)
			&& mpitr_comget(node, pid, cid, &commbuf)) {

		printf("Communicator information unavailable.\n");
		return;
	}

	mpitr_comprint(cid, commbuf, gps);

	free(commbuf);
}
