/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/NJN
 *
 *	$Log:	mpitr_dtype.c,v $
 * Revision 6.0  96/02/29  13:47:58  gdburns
 * Ohio Release
 * 
 *	Function:	- MPI datatype trace access functions
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mpitrace.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * constants and macros
 */
#define DISP(f, x)	((f) ? ((x) & 0xFFFF) : (x))

/*
 * global functions
 */
int			mpitr_dtypeget();
void			mpitr_dtypeprint();

/*
 * external functions
 */
extern int		lam_rtrget();

/*
 * local functions
 */
static void		indent();
static void		print_basic();
static void		print_datatype();
static void		print_contig();
static void		print_hvector();
static void		print_hindexed();
static void		print_struct();

/*
 * local variables
 */
static int		colcount;		/* print column count */
static int		nlifconst;		/* newline if contructor */
static char		fmtbuf[80];		/* formatting buffer */
static char		*prefix;		/* prefix for each line */
static char		*obuf;			/* output buffer or NULL */
static int		identchar;		/* # chars per indent level */

/*
 * basic datatypes
 */
static char		*dtbasic[TRDTMAX + 1] = {
				"MPI_CONTIG", "MPI_VECTOR", "MPI_HVECTOR",
				"MPI_INDEX", "MPI_HINDEX", "MPI_STRUCT",
				"MPI_BYTE", "MPI_CHAR",
				"MPI_UCHAR", "MPI_SHORT", "MPI_USHORT",
				"MPI_INT", "MPI_UINT", "MPI_LOGIC",
				"MPI_LONG", "MPI_ULONG", "MPI_FLOAT",
				"MPI_DOUBLE", "MPI_COMPLEX", "MPI_UB",
				"MPI_LB", "MPI_PACKED", "MPI_2INT",
				"MPI_2FLOAT", "MPI_2DOUBLE", "MPI_FLOAT_INT",
				"MPI_DOUBLE_INT", "MPI_LONG_INT", 
				"MPI_SHORT_INT", "MPI_LONG_DOUBLE"
			};

/*
 *	mpitr_dtypeget
 *
 *	Function:	- get datatype trace
 *	Accepts:	- process node
 *			- process pid
 *			- datatype label
 *			- ptr trace datatype buffer (returned value)
 *	Returns:	- 0 or LAMERROR
 */
int
mpitr_dtypeget(nodeid, pid, dtlabel, ptrdtype)

int4			nodeid;
int4			pid;
int4			dtlabel;
char			**ptrdtype;

{
	int		trlen;			/* trace buffer length */
	char		*p;			/* favourite pointer */
	char		*trbuf;			/* datatype trace buffer */
	struct trdtype	*ptrdt;			/* ptr into datatype trace */
	
	*ptrdtype = 0;

	if (dtlabel < 0) {
		errno = EINVAL;
		return(LAMERROR);
	}

	if (dtlabel <= TRDTMAX) {
/*
 * Allocate datatype description.
 */
		trlen = 2 * sizeof(struct trdtype);

		*ptrdtype = malloc((unsigned) trlen);
		if (*ptrdtype == 0) return(LAMERROR);

		ptrdt = (struct trdtype *) *ptrdtype;
		ptrdt->trd_dtype = dtlabel;
		ptrdt->trd_length = trlen;
		
		ptrdt++;
		ptrdt->trd_dtype = dtlabel;
		ptrdt->trd_count = 1;

		return(0);
	}
/*
 * It's a constructed datatype, get its trace representation.
 */
	trlen = lam_rtrget(nodeid, (int4) TRDTYPE, pid, &trbuf);
	if (trlen < 0) return(LAMERROR);
/*
 * Locate the datatype buffer with given number.
 */
	*ptrdtype = 0;
	p = trbuf;

	while (trlen > 0) {

		ptrdt = (struct trdtype *) (p + sizeof(struct trsrc));
		mttoli4((int4 *) ptrdt, sizeof(struct trdtype) / sizeof(int4));
		
		if (ptrdt->trd_dtype == dtlabel) {
			*ptrdtype = (char *) ptrdt;
			break;
		}

		trlen -= (ptrdt->trd_length + sizeof(struct trsrc));
		p += (ptrdt->trd_length + sizeof(struct trsrc));
	}

	if (*ptrdtype == 0) {
		free((char *) trbuf);
		errno = EINVAL;
		return(LAMERROR);
	}
/*
 * Copy datatype record.
 */
	*ptrdtype = malloc((unsigned) (ptrdt->trd_length));
	if (*ptrdtype == 0) {
		free((char *) trbuf);
		return(LAMERROR);
	}

	mttoli4((int4 *) (((char *) ptrdt) + sizeof(struct trdtype)),
		(ptrdt->trd_length - sizeof(struct trdtype)) / sizeof(int4));

	memcpy(*ptrdtype, (char *) ptrdt, ptrdt->trd_length);
	free((char *) trbuf);

	return(0);
}

/*
 *	mpitr_dtypeprint
 *
 *	Function:	- print datatype trace
 *	Accepts:	- datatype index
 *			- trace datatype buffer (may be NULL if basic datatype)
 *			- number of chars to indent by
 *			- string prefix for each output line
 *			- output buffer (NULL means output to stdout)
 */
void
mpitr_dtypeprint(dtype, buf, indent, pfix, outbuf)

int4			dtype;
char			*buf;
int4			indent;			
char			*pfix;
char			*outbuf;

{
	struct trdtype	*p;			/* ptr trace datatype */

	identchar = indent;
	prefix = pfix;
	obuf = outbuf;
	
	if (dtype < 0) {
		if (obuf) {
			sprintf(obuf, "%sinvalid datatype: %d", prefix, dtype);
		} else {
			printf("%sinvalid datatype: %d\n", prefix, dtype);
		}
	}
/*
 * Handle basic datatypes.
 */
	else if (dtype <= TRDTMAX) {
		if (obuf) {
			strcpy(outbuf, prefix);
			strcpy(outbuf, dtbasic[dtype]);
		} else {
			printf("%s%s\n", prefix, dtbasic[dtype]);
		}
	}
	else if (buf == 0) {
		if (obuf) {
			sprintf(obuf, "%sunknown datatype: %d", prefix, dtype);
		} else {
			printf("%sunknown datatype: %d\n", prefix, dtype);
		}
	}
/*
 * Recursively (depth-first) print datatype.
 */
	else {
		p = (struct trdtype *) buf;
		p++;
		colcount = 0;
		nlifconst = 0;
		if (outbuf) *outbuf = '\0';
		
		print_datatype(&p, 0);
	}
}

/*
 *	print_datatype
 *
 *	Function:	- print current datatype
 *			- recursive function
 *	Accepts:	- ptr record buffer (value modified)
 *			- nesting level
 */
static void
print_datatype(pbuffer, level)

struct trdtype		**pbuffer;
int			level;

{
	int		format;			/* datatype format */

	format = (*pbuffer)->trd_dtype;

	switch(format) {

	case TRDTUCHAR:
	case TRDTCHAR:
	case TRDTBYTE:
	case TRDTPACKED:
	case TRDTUINT:
	case TRDTINT:
	case TRDTLOGIC:
	case TRDTLONG:
	case TRDTULONG:
	case TRDTSHORT:
	case TRDTUSHORT:
	case TRDTFLOAT:
	case TRDTDOUBLE:
	case TRDTLONGDOUBLE:
	case TRDTCOMPLEX:
	case TRDTUB:
	case TRDTLB:
	case TRDT2INT:
	case TRDT2FLOAT:
	case TRDT2DOUBLE:
	case TRDTFLOATINT:
	case TRDTDOUBLEINT:
	case TRDTLONGINT:
	case TRDTSHORTINT:

		print_basic(format, pbuffer, level);
		break;

	case TRDTCONTIG:
		print_contig(pbuffer, level);
		break;

	case TRDTVECTOR:
		print_hvector(format, pbuffer, level, 1);
		break;

	case TRDTHVECTOR:
		print_hvector(format, pbuffer, level, 0);
		break;

	case TRDTINDEXED:
		print_hindexed(format, pbuffer, level, 1);
		break;

	case TRDTHINDEXED:
		print_hindexed(format, pbuffer, level, 0);
		break;

	case TRDTSTRUCT:
		print_struct(pbuffer, level);
		break;

	default:
		break;
	}
}

/*
 *	print_basic
 *
 *	Function:	- print basic datatype record
 *	Accepts:	- datatype label
 *			- ptr record buffer (value modified)
 *			- nesting level
 */
static void
print_basic(dtype, pbuff, nlev)

int			dtype;
struct trdtype		**pbuff;
int			nlev;

{
	nlifconst = 0;
	indent(nlev);
	if (obuf) {
		sprintf(fmtbuf, "%s\n", dtbasic[dtype]);
		strcat(obuf, fmtbuf);
	}
	else {
		printf("%s\n", dtbasic[dtype]);
	}
	colcount = 0;
	++(*pbuff);
}

/*
 *	print_contig
 *
 *	Function:	- print contiguous datatype record
 *	Accepts:	- ptr record buffer (value modified)
 *			- nesting level
 */
static void
print_contig(pbuff, nlev)

struct trdtype		**pbuff;
int			nlev;

{
	indent(nlev);

	sprintf(fmtbuf, "%s (%d)", dtbasic[TRDTCONTIG], (*pbuff)->trd_count);

	if (obuf) strcat(obuf, fmtbuf);
	else printf("%s", fmtbuf);

	colcount += strlen(fmtbuf);
	nlifconst = 1;

	++(*pbuff);

	print_datatype(pbuff, nlev + 1);
}

/*
 *	print_hvector
 *
 *	Function:	- print vector or hvector datatype record
 *	Accepts:	- datatype label
 *			- ptr record buffer (value modified)
 *			- nesting level
 *			- vector flag
 */
static void
print_hvector(dtype, pbuff, nlev, fl_vec)

int			dtype;
struct trdtype		**pbuff;
int			nlev;
int			fl_vec;

{
	indent(nlev);

	sprintf(fmtbuf, "%s (%d x %d, %d)", dtbasic[dtype],
			(*pbuff)->trd_count, (*pbuff + 1)->trd_length,
			DISP(fl_vec, (*pbuff + 1)->trd_stride));

	if (obuf) strcat(obuf, fmtbuf);
	else printf("%s", fmtbuf);

	colcount += strlen(fmtbuf);
	nlifconst = 1;

	*pbuff += 2;

	print_datatype(pbuff, nlev + 1);
}

/*
 *	print_hindexed
 *
 *	Function:	- print indexed or hindexed datatype record
 *	Accepts:	- datatype label
 *			- ptr record buffer (value modified)
 *			- nesting level
 *			- indexed flag
 */
static void
print_hindexed(dtype, pbuff, nlev, fl_idx)

int			dtype;
struct trdtype		**pbuff;
int			nlev;
int			fl_idx;

{
	int		i;			/* favourite index */
	int		count;			/* datatype count */

	indent(nlev);

	count = (*pbuff)->trd_count;
	++(*pbuff);

	sprintf(fmtbuf, "%s (%d)", dtbasic[dtype], count);
	
	if (obuf) strcat(obuf, fmtbuf);
	else printf("%s", fmtbuf);

	colcount += strlen(fmtbuf);

	for (i = 0; i < count; ++i) {
		sprintf(fmtbuf, " (%d, %d)", (*pbuff)->trd_length,
				DISP(fl_idx, (*pbuff)->trd_disp));

		if (obuf) strcat(obuf, fmtbuf);
		else printf("%s", fmtbuf);

		colcount += strlen(fmtbuf);
		++(*pbuff);
	}

	nlifconst = 1;

	print_datatype(pbuff, nlev + 1);
}

/*
 *	print_struct
 *
 *	Function:	- print struct datatype record
 *	Accepts:	- ptr record buffer (value modified)
 *			- nesting level
 */
static void
print_struct(pbuff, nlev)

struct trdtype		**pbuff;
int			nlev;

{
	int		i;			/* favourite index */
	int		count;			/* datatype count */
 
	indent(nlev);

	count = (*pbuff)->trd_count;
	++(*pbuff);

	sprintf(fmtbuf, "%s (%d)\n", dtbasic[TRDTSTRUCT], count);

	if (obuf) strcat(obuf, fmtbuf);
	else printf("%s", fmtbuf);

	colcount = 0;
	nlifconst = 0;

	for (i = 0; i < count; ++i) {

		indent(nlev + 1);

		sprintf(fmtbuf, "(%d, %d)",
				(*pbuff)->trd_length, (*pbuff)->trd_disp);

		if (obuf) strcat(obuf, fmtbuf);
		else printf("%s", fmtbuf);

		colcount += strlen(fmtbuf);
		++(*pbuff);

		print_datatype(pbuff, nlev + 2);
	}
}

/*
 *	indent
 *
 *	Function:	- indent line according to level
 *	Accepts:	- indentation level
 */
static void
indent(lev)

int			lev;

{
	lev *= identchar;

	if (nlifconst) {
		if (obuf) strcat(obuf, "\n");
		else printf("\n");

		nlifconst = 0;
		colcount = 0;
	}

	if (obuf) {
		if (colcount == 0) strcat(obuf, prefix);
		if (colcount < lev) {
			for (; colcount < lev; ++colcount) strcat(obuf, " ");
		} else if (colcount > 0) {
			strcat(obuf, " ");
		}
	}
	else {
		if (colcount == 0) printf("%s", prefix);
		if (colcount < lev) {
			for (; colcount < lev; ++colcount) printf(" ");
		} else if (colcount > 0) {
			printf(" ");
		}
	}
}
