/*
 *	Ohio Trollius
 *	Copyright 1993 The Ohio State University
 *	RBD
 *
 *	$Log:	sockface.udp.c,v $
 * Revision 6.0  96/02/29  13:48:50  gdburns
 * Ohio Release
 * 
 * Revision 5.2  94/08/22  13:57:09  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:57:06  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:35:07  gdburns
 * Ohio Release
 * 
 * Revision 2.2  93/10/15  13:59:06  raja
 * pre-release to 2.3
 * 
 *	Function:	- socket interface functions
 *			- UDP version
 *
 *			  fillinetaddr(): fill Internet address structure
 *			  getsrvudpsock(): create server-side socket
 *			  getcltudpsock(): create client-side socket
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <typical.h>

/*
 *	fillinetaddr
 *
 *	Function:	- fill host/port Internet address structure
 *	Accepts:	- host address or NULL (4 bytes, network byte order)
 *			- port number
 *			- ptr Internet address structure
 */
void
fillinetaddr(hostaddr, port, addr)

unsigned char		*hostaddr;
int			port;
struct sockaddr_in	*addr;

{
	memset((char *) addr, 0, sizeof(struct sockaddr_in));
	addr->sin_family = AF_INET;
	addr->sin_port = (port > 0) ? htons((unsigned short) port) : htons(0);
	if (hostaddr) {
		memcpy((char *) &(addr->sin_addr), (char *) hostaddr, 4);
	} else {
		addr->sin_addr.s_addr = htonl(INADDR_ANY);
	}
}

/*
 *	getsrvudpsock
 *
 *	Function:	- create a server-side socket
 *			- Internet domain datagrams
 *			- if pport points to a non-zero positive port
 *			  number, that port number is used otherwise a
 *			  port number is dynamically allocated
 *			- if pport is non-NULL, the server port number
 *			  is returned in it
 *	Accepts:	- ptr to port number (or NULL)
 *	Returns:	- socket descriptor or ERROR
 */
int
getsrvudpsock(pport)

int			*pport;

{
	int		sockd;		/* socket descriptor */
	int		length;		/* length of socket name */
	struct sockaddr_in
			srvaddr;	/* server address info */
/*
 * Create server socket.
 */
	if ((sockd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return(ERROR);
	}
/*
 * Bind the socket.  If a port # is given use it otherwise bind allocates one.
 */
	fillinetaddr((unsigned char *) 0, (pport) ? *pport : 0, &srvaddr);

	if (bind(sockd, (struct sockaddr *) &srvaddr, sizeof(srvaddr))) {
		close(sockd);
		return(ERROR);
	}
/*
 * If bind allocated a port # dynamically and pport is non-NULL,
 * get the port # allocated and return it in pport.
 */
	if (pport && (*pport <= 0)) {
		length = sizeof(srvaddr);
		if (getsockname(sockd,
				(struct sockaddr *) &srvaddr, &length)) {
			close(sockd);
			return(ERROR);
		}

		*pport = (int) ntohs(srvaddr.sin_port);
	}

	return(sockd);
}

/*
 *	getcltudpsock
 *
 *	Function:	- create a client-side socket
 *			- Internet domain datagrams
 *	Returns:	- socket descriptor or ERROR
 */
int
getcltudpsock()

{
	int		sockd;		/* socket descriptor */
	struct sockaddr_in
			myaddr;		/* local address info */
/*
 * Create client socket.
 */
	sockd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockd < 0) {
		return(ERROR);
	}
/*
 * Bind local address.
 */
	fillinetaddr((unsigned char *) 0, 0, &myaddr);

	if (bind(sockd, (struct sockaddr *) &myaddr, sizeof(myaddr))) {
		close(sockd);
		return(ERROR);
	}

	return(sockd);
}
