/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	terror.c,v $
 * Revision 6.0  96/02/29  13:48:51  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/02/06  20:45:00  gdburns
 * Add EBADHOST.
 * 
 * Revision 5.2.1.2  96/01/18  20:41:12  gdburns
 * error code changes
 * 
 * Revision 5.2.1.1  95/04/10  15:44:27  gdburns
 * Extend message formatting buffer.
 * 
 * Revision 5.2  94/08/22  13:57:10  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:57:09  gdburns
 * Ohio Release
 * 
 * Revision 2.3.1.1  94/05/18  10:48:32  raja
 * Add terrorstr(), EMPIINIT, EMPINOINIT.
 * 
 * Revision 2.3  94/04/22  12:35:08  gdburns
 * Ohio Release
 *
 *	Function:	- Trollius error messages
 */

#include <string.h>

#include <portable.h>
#include <terror.h>

/*
 * local variables
 */
static char		*emsg[] = {
		"undiagnosed error",		/* EUNKNOWN */
		"remote node has died",		/* EDEADNODE */
		"out of exit descriptors",	/* ENOEXDESCRIPTORS */
		"out of descriptors",		/* EFULL */
		"not a process",		/* ENOTPROCESS */
		"invalid request",		/* EBADREQUEST */
		"block not allocated",		/* ENOTMALLOCED */
		"GER overflow",			/* EGERFLOW */
		"service unavailable",		/* ESERVICE */
		"bad magic number",		/* EMAGIC */
		"blown stack",			/* EBLOWN */
		"impossible, ergo a bug",	/* EIMPOSSIBLE */
		"limit one i860 process",	/* EI860PROC */
		"unknown host",			/* EBADHOST */
		"no running daemon",		/* ENOKERNEL */
		"not attached to daemon",	/* ENOTATTACHED */
		"unused error code",		/* unused */
		"improper message size",	/* EMSGLEN */
		"end of file",			/* EEOF */
		"cast spanning tree collision",	/* ECOLLIDE */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"can't find shared segment",	/* ESHARE */
		"invalid link index",		/* EBADLINK */
		"bad acknowledgement",		/* EACK */
		"improper command invocation",	/* EUSAGE */
		"bad/unknown option",		/* EOPTION */
		"unused error code",		/* unused */
		"invalid node",			/* EBADNODE */
		"invalid ID specification",	/* EBADIDSPEC */
		"no matching nodes found",	/* ENONODES */
		"out of route descriptors",	/* ENORDESCRIPTORS */
		"out of buffer descriptors",	/* ENOBFDESCRIPTORS */
		"out of boot descriptors",	/* ENOBTDESCRIPTORS */
		"out of flat descriptors",	/* ENOFLDESCRIPTORS */
		"out of kenya descriptors",	/* ENOPDESCRIPTORS */
		"out of process descriptors",	/* ENOKDESCRIPTORS */
		"bad moses request",		/* EBADMOSESREQ */
		"unused error code",		/* unused */
		"invalid address tag",		/* EBADTAG */
		"delete from list failed",	/* EDELETE */
		"no bootstrap present",		/* ENOBOOTSTRAP */
		"invalid record format",	/* EBADREC */
		"bad ascii number",		/* EBADASCIINUM */
		"booted system is too big",	/* EHIGHWATER */
		"process already in group",	/* EINGROUP */
		"limit 32 group procs/node",	/* EGMAXPROC */
		"group has a single process",	/* EGSINGLETON */
		"not a group process",		/* EGNOTPROC */
		"invalid group",		/* EBADGROUP */
		"MPI already initialized",	/* EMPIINIT */
		"MPI not initialized",		/* EMPINOINIT */
			};

static char		*baderror = 
		"?  Get out of town!\n";	/* unknown errno */
static char		buffer[320];		/* complete message */
static char		anum[16];		/* ascii error code */

/*
 * extern variables
 */
extern int		sys_nerr;
extern char		*sys_errlist[];

/*
 * external functions
 */
extern void		itoa();			/* int to ascii */

/*
 *	terrorstr
 *
 *	Function:	- forms Trollius error message
 *			- uses Trollius and UNIX error message tables
 *			- returned message is stored in a static buffer
 *	Accepts:	- user's error message
 *	Returns:	- extended error message
 */
char *
terrorstr(msg)

CONST char		*msg;

{
	buffer[0] = '\0';

	if (msg && *msg) {
		strncpy(buffer, msg, 256);
		buffer[256] = '\0';
		strcat(buffer, ": ");
	}

	if ((errno >= 0) && (errno < sys_nerr) && (errno < ELOW)) {
		strcat(buffer, sys_errlist[errno]);
	} else if ((errno >= ELOW) && (errno <= EHIGH)) {
		strcat(buffer, emsg[errno - ELOW]);
	} else {
		itoa(errno, anum);
		strcat(buffer, anum);
		strcat(buffer, baderror);
	}

	return(buffer);
}

/*
 *	terror
 *
 *	Function:	- prints Trollius error messages on stderr
 *	Accepts:	- user's error message
 */
void
terror(umsg)

CONST char		*umsg;

{
	char		*p;

	p = terrorstr(umsg);
	strcat(p, "\n");
	write(2, p, strlen(p));
}
