/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB
 *
 *	$Log:	rfstate.c,v $
 * Revision 6.0  96/02/29  13:50:11  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  95/11/20  13:10:40  gdburns
 * Remove Cubix.
 * Qualify names for LAM.
 * 
 * Revision 5.2  94/08/22  13:59:18  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/08/18  11:21:10  gdburns
 * Do not rely on the router for LOCAL service.
 * 
 * Revision 5.1  94/05/18  10:58:16  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:38:30  gdburns
 * Ohio Release
 * 
 *	Function:	- gathers state from a remote filed
 *			- receives file descriptors from
 *			  filed until reply code is EEOF
 *
 *	Accepts:	- target node ID
 *			- ptr to file descriptor table
 *			- # of entries in table
 *
 *	Returns:	- number of replied descriptors or ERROR
 */

#include <string.h>

#include <events.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <string.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * local variables
 */
static char		hole[MAXNMSGLEN];

int
lam_rfstate(nodeid, ftbl, maxcount)

int			nodeid;		/* target node ID */
int			maxcount;	/* count of descriptor table */
struct fstate		ftbl[];		/* ptr to descriptor table */

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	struct fstate	*fhole;		/* ptr incoming file descs */
	int		count;		/* # descs in this msg */
	int		totcount;	/* total # replied descs */
	int		i;		/* favourite index */
	int		mask;		/* signal mask */

	totcount = 0;
	fhole = (struct fstate *) hole;
	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQSTATE;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	if (nsend(&nhead)) {
		ksigsetmask(mask);
		return(LAMERROR);
	}

	nhead.nh_event = -getpid();
	nhead.nh_msg = hole;

	do {
		nhead.nh_length = MAXNMSGLEN;

		if (nrecv(&nhead)) {
			ksigsetmask(mask);
			return(LAMERROR);
		}

		count = nhead.nh_length / sizeof(struct fstate);
		totcount += count;

		if (maxcount > 0) {
/*
 * Convert byte ordering (sigh).
 */
			for (i = 0; i < count; ++i) {
				fhole[i].fs_tfd = ttol(fhole[i].fs_tfd);
				fhole[i].fs_tflags = ttol(fhole[i].fs_tflags);
				fhole[i].fs_flow = ttol(fhole[i].fs_flow);
				fhole[i].fs_src_node =
					ttol(fhole[i].fs_src_node);
				fhole[i].fs_src_event =
					ttol(fhole[i].fs_src_event);
			}

			if (count > maxcount) count = maxcount;

			memcpy((char *) ftbl, hole,
				count * sizeof(struct fstate));
			maxcount -= count;
			ftbl += count;
		}

	} while (! reply->fr_errno);

	ksigsetmask(mask);

	if (reply->fr_errno == EEOF) {
		return(totcount);
	} else {
		errno = reply->fr_errno;
		return(LAMERROR);
	}
}

/*
 * backwards compatibility
 */
int rfstate(nodeid, ftbl, maxcount) int nodeid, maxcount; struct fstate ftbl[];
	{ return(lam_rfstate(nodeid, ftbl, maxcount)); }
