/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB
 *
 *	$Log:	rfwrite.c,v $
 * Revision 6.0  96/02/29  13:50:12  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  95/11/20  13:10:42  gdburns
 * Remove Cubix.
 * Qualify names for LAM.
 * 
 * Revision 5.2  94/08/22  13:59:18  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:58:17  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:38:30  gdburns
 * Ohio Release
 *
 *	Function:	- writes to a file using remote filed
 *	Accepts:	- runtime file descriptor
 *			- buffer ptr
 *			- buffer length
 *	Returns:	- 0 or ERROR
 */

#include <string.h>

#include <events.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */

int4
lam_rfwrite(fd, buf, len)

int			fd;
CONST char		*buf;
int4			len;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */
	int4		tot_len = 0;	/* total length written */
	int4		pkt_len;	/* one packet length */
/*
 * Do simple error checking.
 */
	if (_fdvalid(fd)) return((int4) LAMERROR);

	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	nhead.nh_node = _ufd[fd].fu_node;
	nhead.nh_flags = _ufd[fd].fu_tflags & LAM_O_DMSG;

	nhead.nh_msg = (char *) buf;
	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
/*
 * Packetize output buffer.
 */
	do {
		request->fq_src_node = getnodeid();
		request->fq_src_event = -getpid();
		request->fq_req = FQWRITE;
		request->fq_tfd = _ufd[fd].fu_tfd;

		nhead.nh_event = EVFILED;
		nhead.nh_type = 0;
		nhead.nh_length = (len > MAXNMSGLEN) ? MAXNMSGLEN : len;
		pkt_len = nhead.nh_length;
		request->fq_length = pkt_len;

		if (nsend(&nhead)) {
			ksigsetmask(mask);
			return(LAMERROR);
		}

		if (! (_ufd[fd].fu_tflags & F1WAY)) {
			nhead.nh_event = -getpid();
			nhead.nh_type = 0;
			nhead.nh_length = 0;

			if (nrecv(&nhead)) {
				ksigsetmask(mask);
				return(LAMERROR);
			}

			if (reply->fr_errno != 0) {
				ksigsetmask(mask);
				errno = reply->fr_errno;
				return(LAMERROR);
			}

			else if (reply->fr_ret != pkt_len) {
				len = 0;
			}

			else {
				len -= pkt_len;
				nhead.nh_msg += pkt_len;
			}
		} else {
			reply->fr_ret = pkt_len;
			len -= pkt_len;
			nhead.nh_msg += pkt_len;
		}

		tot_len += reply->fr_ret;
	} while (len > 0);

	ksigsetmask(mask);
	return(tot_len);
}

/*
 * backwards compatibility
 */
int4 rfwrite(fd, buf, len) int fd; CONST char *buf; int4 len;
	{ return(lam_rfwrite(fd, buf, len)); }
