/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB
 *
 *	$Log:	kio.h,v $
 * Revision 6.0  96/02/29  13:53:25  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  01:29:46  gdburns
 * Overhaul for new kernel.
 * 
 * Revision 5.2  94/08/22  14:00:35  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/08/18  11:26:13  gdburns
 * overhaul for new kernel
 * 
 * Revision 5.1  94/05/18  12:44:26  gdburns
 * Ohio Release
 * 
 *	Function:	- defines templates and constants for the
 *			  Kernel Input Output block
 */

#ifndef _KIO
#define _KIO

#include <kreq.h>
#include <ksignal.h>
#include <ktrace.h>
#include <portable.h>
#include <rreq.h>

#define MAXRCACHE	29		/* max # of cached routes */
#define	TNEXIT		32		/* maximum number of exit routines */

/*
 * Runtime Flags
 */
#define RTF_KERNEL	0x0001		/* attached to kernel? */
#define RTF_KENYA	0x0002		/* attached to kenya? */
#define RTF_SYSGEN	0x0008		/* system process? */
#define RTF_WAIT	0x0010		/* background process? */
#define RTF_MPIGER	0x0020		/* guaranteed resources protocol */
#define RTF_DEBUG	0x0040		/* for debuggers */
#define RTF_MPIC2C	0x0080		/* direct client to client comm. */
#define RTF_TRACE	0x0100		/* traced process? */
#define RTF_TRSWITCH	0x0200		/* trace on/off switch */
#define RTF_TRON	0x0300		/* comb. of TRACE & TRSWITCH */
#define RTF_MPI		0x0400		/* I am an MPI process. */
#define RTF_FLAT	0x0800		/* loaded by flatd? */
#define RTF_MPIRUN	0x1000		/* started by mpirun? */
#define RTF_HOMOG	0x2000		/* homogeneous cluster */
#define RTF_MPICWD	0x4000		/* change directory in app */

/*
 * Kernel Interface
 */
struct kio_t {
	int4		ki_priority;	/* my kernel process priority */
	int4		ki_rtf;		/* Run Time Flags */
	int4		ki_pid;		/* kernel registered process ID */
	int4		ki_signal;	/* pending signals */
	int4		ki_sigmask;	/* signal masks */
	int4		ki_sigretry;	/* flags to retry after EINTR */
	int4		ki_index;	/* fast identification in kernel */
	int4		ki_nodeid;	/* my node identifier */
	int4		ki_origin;	/* origin node identifier */
	int4		ki_stdin;	/* standard input descriptor */
	int4		ki_stdout;	/* standard output descriptor */
	int4		ki_stderr;	/* standard error descriptor */
/*
 * Signal Handling Table
 */
	void		((*(ki_sig_handlers[TNSIG]))());
	void		((*(ki_sig_default[TNSIG]))());
/*
 * at exit table
 */
	void		((*(ki_exit[TNEXIT]))());
	char		*ki_arg[TNEXIT];
/*
 * route cache
 */
	struct route	ki_route[MAXRCACHE];
/*
 * trace support
 */
	struct trhead	*ki_trhead;	/* trace buffer headers */
	int4		ki_trcurr;	/* current trace buffer index */
	int4		ki_tid;		/* trace process identifier */

	int4		ki_world;	/* expected world population */
	int4		ki_parent;	/* parent group population */
/*
 * debug support
 */
	int4		ki_fyi[KNFYI];	/* debugging information */
	double		ki_blktime;	/* time spent blocking */
};

#endif
