/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	mpisys.h,v $
 * Revision 6.0.1.2  96/04/29  13:39:55  nevin
 * Clash between MPI_CANCEL_SOURCE and MPI_PROC_NULL
 * 
 * Revision 6.0.1.1  96/03/13  15:43:38  nevin
 * Flag for FORTRAN initialization
 * 
 * Revision 6.0  96/03/03  22:36:22  nevin
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/03/03  22:30:53  nevin
 * Fix definition of MPI_NULL_COPY_FN and MPI_NULL_DELETE_FN
 * 
 * Revision 5.2.1.1  96/02/28  21:29:05  nevin
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:00:40  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/08/02  15:08:42  raja
 * MPI_Get_count(), MPI_Get_elements(), and MPI_Test_cancelled()
 * changed to take a pointer to MPI_Status.
 * 
 * Revision 5.1  94/05/18  12:44:32  gdburns
 * Ohio Release
 * 
 *	Function:	- MPI system header file
 *			- the real meat is here
 */

#ifndef _MPISYS
#define _MPISYS

#include <all_list.h>
#include <all_hash.h>
#include <kio.h>
#include <portable.h>

/*
 * misc. constants
 */
#define MPI_SUCCESS		0		/* no errors */
#define MPI_ANY_SOURCE		-1		/* match any source rank */
#define MPI_PROC_NULL		-2		/* rank of null process */
#define MPI_CANCEL_SOURCE	-3		/* successful cancel */
#define MPI_ANY_TAG		-1		/* match any message tag */
#define MPI_GER_TAG		-2		/* used for GER protocol */
#define MPI_MERGE_TAG		-3		/* used to merge inter-comm */
#define MPI_MAX_CID		0xFFE		/* max pt2pt context ID */
#define MPI_GER_CID		0xFFF		/* used for GER protocol */
#define MPI_MAX_PROCESSOR_NAME	256		/* max proc. name length */
#define MPI_MAX_ERROR_STRING	80		/* max error message length */
#define MPI_UNDEFINED		-32766		/* undefined stuff */
#define MPI_CART		1		/* cartesian topology */
#define MPI_GRAPH		2		/* graph topology */
#define MPI_KEYVAL_INVALID	-1		/* invalid key value */
#define MPI_TAG_UB		0		/* predefined attribute key */
#define MPI_HOST		1		/* predefined attribute key */
#define MPI_IO			2		/* predefined attribute key */
#define MPI_WTIME_IS_GLOBAL	3		/* predefined attribute key */
#define MPI_BOTTOM		((void *) 0)
#define MPI_BSEND_OVERHEAD	(sizeof(struct _bsndhdr) + sizeof(void *) - 1)

/*
 * GER (Guaranteed Envelope Resources)
 */
#ifndef MPI_GER
#define MPI_GER			8
#endif

/*
 * error classes
 */
#define MPI_ERR_BUFFER		1	/* invalid buffer pointer */
#define MPI_ERR_COUNT		2	/* invalid count argument */
#define MPI_ERR_TYPE		3	/* invalid datatype argument */
#define MPI_ERR_TAG		4	/* invalid tag argument */
#define MPI_ERR_COMM		5	/* invalid communicator */
#define MPI_ERR_RANK		6	/* invalid rank */
#define MPI_ERR_REQUEST		7	/* invalid request handle */
#define MPI_ERR_ROOT		8	/* invalid root */
#define MPI_ERR_GROUP		9	/* invalid group */
#define MPI_ERR_OP		10	/* invalid operation */
#define MPI_ERR_TOPOLOGY	11	/* invalid topology */
#define MPI_ERR_DIMS		12	/* invalid dimension argument */
#define MPI_ERR_ARG		13	/* invalid argument */
#define MPI_ERR_UNKNOWN		14	/* unknown error */
#define MPI_ERR_TRUNCATE	15	/* message truncated on receive */
#define MPI_ERR_OTHER		16	/* LAM error */
#define MPI_ERR_INTERN		17	/* internal MPI error */
#define MPI_ERR_IN_STATUS	18	/* error code is in status */
#define MPI_ERR_PENDING		19	/* pending request */
#define MPI_ERR_SYSRESOURCE	20	/* out of system resources */
#define MPI_ERR_LASTCODE	21	/* last error code */

/*
 * comparison results
 */
#define MPI_IDENT		1
#define MPI_CONGRUENT		2
#define MPI_SIMILAR		3
#define MPI_UNEQUAL		4

/*
 * typedefs
 */
typedef struct _group		*MPI_Group;
typedef struct _comm		*MPI_Comm;
typedef struct _dtype		*MPI_Datatype;
typedef struct _req		*MPI_Request;
typedef struct _op		*MPI_Op;
typedef struct _status		MPI_Status;
typedef struct _errhdl		*MPI_Errhandler;
typedef void			*MPI_Buffer;
typedef long			MPI_Aint;

/*
 * NULL handles
 */
#define MPI_GROUP_NULL		((MPI_Group) 0)
#define MPI_COMM_NULL		((MPI_Comm) 0)
#define MPI_DATATYPE_NULL	((MPI_Datatype) 0)
#define MPI_REQUEST_NULL	((MPI_Request) 0)
#define MPI_OP_NULL		((MPI_Op) 0)
#define MPI_ERRHANDLER_NULL	((MPI_Errhandler) 0)
#define MPI_NULL_COPY_FN	((MPI_Copy_function *) 0)
#define MPI_NULL_DELETE_FN	((MPI_Delete_function *) 0)

/*
 * LAM constants & macros
 */
#define LAM_COLLMAXLIN		4	/* max # procs for linear coll. op */
#define LAM_COLLMAXDIM		(8 * sizeof(int))	/* max cube dim. */

#define LAM_GPSCMP(p1, p2)	(((p1)->gps_node < (p2)->gps_node) ? -1 : \
				(((p1)->gps_node > (p2)->gps_node) ? 1 : \
				(((p1)->gps_idx < (p2)->gps_idx) ? -1 : \
				(((p1)->gps_idx > (p2)->gps_idx) ? 1 : 0))))

#define RPI_SPLIT(l, c, args) 	((__lam_c2c_) ? c args : l args)

/*
 * structures
 */
struct _dtype {
	int		dt_format;
#define LAM_DTBASIC	0			/* basic datatype */
#define LAM_DTCONTIG	1			/* contiguous */
#define LAM_DTVECTOR	2			/* vector */
#define LAM_DTHVECTOR	3			/* hvector */
#define LAM_DTINDEXED	4			/* indexed */
#define LAM_DTHINDEXED	5			/* hindexed */
#define LAM_DTSTRUCT	6			/* struct */
	int		dt_label;		/* flat-form label */
	int		dt_commit;		/* committed? */
	int		dt_refcount;		/* reference count */
	int		dt_upper;		/* upper extent */
	int		dt_lower;		/* lower extent */
	int		dt_size;		/* basic size */
	int		dt_usrsize;		/* user-visible size */
	int		dt_nelem;		/* # top elements */
	int		dt_count;		/* count */
	int		dt_length;		/* vector length */
	MPI_Aint	dt_stride;		/* vector stride */
	MPI_Datatype	dt_dtype;		/* c/v/i datatype */
	int		*dt_lengths;		/* i/s lengths */
	MPI_Aint	*dt_disps;		/* i/s displacements */
	MPI_Datatype	*dt_dtypes;		/* struct datatypes */
};

struct _gps {
	int		gps_node;		/* node ID */
	int		gps_pid;		/* process ID */
	int		gps_idx;		/* process index */
	int		gps_grank;		/* glob. rank in loc. world */
};

struct _proc {
	struct _gps	p_gps;			/* process GPS */
	void		*p_rpi;			/* RPI specific data */
	int		p_ger_nsnd;		/* #msgs sent there */
	int		p_mode;			/* mode flags */
#define LAM_PFLAG	0x01			/* used for marking */
#define LAM_PDEAD	0x02			/* node died */
};

struct _group {
	int		g_nprocs;		/* # processes */
	int		g_myrank;		/* my local rank */
	int		g_refcount;		/* reference count */
	struct _proc	**g_procs;		/* ptr to processes */
};

struct _comm {
	int		c_inter;		/* intercomm? */
	int		c_contextid;		/* context ID */
	int		c_refcount;		/* reference count */
	int		c_cube_dim;		/* inscribing cube dim. */
	MPI_Group	c_group;		/* local group */
	MPI_Group	c_rgroup;		/* remote group */
	MPI_Errhandler	c_errhdl;		/* error handle */
	HASH		*c_keys;		/* keys cache hash table */
	int		c_topo_type;		/* topology type */
	int		c_topo_nprocs;		/* # topo. processes */
	int		c_topo_ndims;		/* # cart. dimensions */
	int		c_topo_nedges;		/* # graph edges */
	int		*c_topo_dims;		/* cart. dimensions */
	int		*c_topo_coords;		/* cart. coordinates */
	int		*c_topo_index;		/* graph indices */
	int		*c_topo_edges;		/* graph edges */
	int		c_status;		/* status (fault tolerance) */
#define LAM_CLDEAD     0x01			/* local group dead */
#define LAM_CRDEAD     0x02			/* remote group dead */	
};

struct _status { 
	int		MPI_SOURCE;
	int		MPI_TAG;
	int		MPI_ERROR;
	int		st_count;		/* top element count */
	int		st_nelem;		/* low element count */
	int		st_length;		/* message length */
	MPI_Datatype	st_dtype;		/* datatype */
};

struct _req {
	int		rq_type;
#define LAM_RQISEND	0
#define LAM_RQIBSEND	1
#define LAM_RQISSEND	2
#define LAM_RQIRSEND	3
#define LAM_RQIRECV	4
#define LAM_RQIPROBE	5
#define LAM_RQIFAKE	6
	int		rq_state;
#define LAM_RQSINIT	0			/* request initialized */
#define LAM_RQSSTART	1			/* active, nothing done yet */
#define LAM_RQSACTIVE	4			/* active, undone request */
#define LAM_RQSDONE	3			/* request done */
	int		rq_marks;		/* persistent flags */
#define LAM_RQFPERSIST	0x0001			/* persistent request */
#define LAM_RQFDYNBUF	0x0004			/* dynamic buffer */
#define LAM_RQFDYNREQ	0x0008			/* dynamic request */
#define LAM_RQFSOURCE	0x0200			/* source request */
#define LAM_RQFDEST	0x0400			/* destination request */
#define LAM_RQFBLKTYPE	0x0800			/* blocking request type */
	int		rq_flags;		/* active req flags */
#define LAM_RQFCANCEL	0x0002			/* cancelled request */
#define LAM_RQFBLOCK	0x0010			/* blocking request */
#define LAM_RQFTRUNC	0x0040			/* truncated message */
#define LAM_RQFORPHAN	0x0100			/* destroy when done */
#define LAM_RQFCHAR	0x0100000		/* == DRAWMSG */
#define LAM_RQFINT	0x0200000		/* == DINT4MSG */
#define LAM_RQFFLOAT	0x0400000		/* == DFLT4MSG */
#define LAM_RQFDOUBLE	0x0800000		/* == DFLT8MSG */
#define LAM_RQFSHORT	0x01000000		/* == DINT2MSG */
	char		*rq_packbuf;		/* pack buffer */
	int		rq_packsize;		/* pack buffer size */
	int		rq_count;		/* MPI request parameter */
	void		*rq_buf;		/* MPI request parameter */
	MPI_Datatype	rq_dtype;		/* MPI request parameter */
	int		rq_rank;		/* MPI request parameter */
	int		rq_tag;			/* MPI request parameter */
	MPI_Comm	rq_comm;		/* MPI request parameter */
	int		rq_cid;			/* context ID to use */
	int		rq_func;		/* MPI func. which made req. */
	int		rq_seq;			/* seq# of associated msg */
	MPI_Status	rq_status;		/* status info storage */
	struct _bsndhdr	*rq_bsend;		/* ptr bsend header */
	struct _proc	*rq_proc;		/* ptr peer process */
	void		*rq_rpi;		/* RPI specific stuff */
	struct _req	*rq_next;		/* ptr next request */
};

struct _op {
	void		(*op_func)();		/* reduction function */
	int		op_commute;		/* commutative? */
	int		op_f77dtype;		/* f77 datatype */
};

struct _errhdl {
	int		eh_f77hdl;		/* f77 error handle */
	int		eh_refcount;		/* reference count */
	void		(*eh_func)();		/* error function */
};

struct _attrkey {
	int		(*ak_copy)();		/* copy function */
	int		(*ak_del)();		/* delete function */
	void		*ak_extra;		/* extra info */
	int		ak_refcount;		/* reference count */
	int		ak_f77comm;		/* f77 communicator */
};

struct _attr {
	int		a_key;			/* attribute key */
	void		*a_value;		/* attribute value */
};

struct _bsndhdr {
	MPI_Request	bsh_req;
	int		bsh_size;
	struct _bsndhdr	*bsh_prev;
	struct _bsndhdr	*bsh_next;
};

struct _fyimsg {
	int4		fym_src;		/* src global/local rank */
	int4		fym_dest;		/* dest global/local rank */
	int4		fym_count;		/* count */
	int4		fym_dtpid;		/* datatype/pid */
};

struct _fyiproc {
	int4		fyp_me;			/* my global/local rank */
	int4		fyp_peer;		/* peer global/local rank */
	int4		fyp_peergps;		/* peer node/index */
	int4		fyp_root;		/* root global/local rank */
	int4		fyp_rootgps;		/* root node/index */
	int4		fyp_cidtag;		/* context ID & tag */
	int4		fyp_func;		/* top level function */
	int4		fyp_count;		/* count */
	int4		fyp_dtype;		/* datatype */
};


/*
 * external variables
 */
extern struct _comm		lam_mpi_comm_world;
extern struct _comm		lam_mpi_comm_self;
extern MPI_Comm			lam_mpi_comm_parent;
extern struct _group		lam_mpi_group_empty;
extern struct _op		lam_mpi_max, lam_mpi_min;
extern struct _op		lam_mpi_sum, lam_mpi_prod;
extern struct _op		lam_mpi_land, lam_mpi_band;
extern struct _op		lam_mpi_lor, lam_mpi_bor;
extern struct _op		lam_mpi_lxor, lam_mpi_bxor;
extern struct _op		lam_mpi_maxloc, lam_mpi_minloc;
extern struct _dtype		lam_mpi_char, lam_mpi_byte;
extern struct _dtype		lam_mpi_int, lam_mpi_logic;
extern struct _dtype		lam_mpi_short, lam_mpi_long;
extern struct _dtype		lam_mpi_float, lam_mpi_double;
extern struct _dtype		lam_mpi_long_double;
extern struct _dtype		lam_mpi_cplex, lam_mpi_packed;
extern struct _dtype		lam_mpi_unsigned_char;
extern struct _dtype		lam_mpi_unsigned_short;
extern struct _dtype		lam_mpi_unsigned;
extern struct _dtype		lam_mpi_unsigned_long;
extern struct _dtype		lam_mpi_ub, lam_mpi_lb;
extern struct _dtype		lam_mpi_float_int, lam_mpi_double_int;
extern struct _dtype		lam_mpi_long_int, lam_mpi_2int;
extern struct _dtype		lam_mpi_2float, lam_mpi_2double;
extern struct _dtype		lam_mpi_short_int;
extern struct _errhdl		lam_mpi_errors_are_fatal;
extern struct _errhdl		lam_mpi_errors_return;

/*
 * for internal purposes
 */
extern struct _proc		*__lam_myproc_;
extern int                      __lam_f77init_;
extern int			__lam_c2c_;
extern int			__lam_ger_;
extern int			__lam_homog_;
extern int			__lam_poll_delay_;
extern LIST			*__lam_comms_;
extern float8			__lam_clock_skew_;

extern struct kio_t     	_kio;		/* kernel I/O block */

/*
 * useful tracing macro
 */
#define _LAM_TRACE(a)		if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) a

/*
 * MPI predefined handles
 */
#define MPI_COMM_WORLD		((MPI_Comm) &lam_mpi_comm_world)
#define MPI_COMM_SELF		((MPI_Comm) &lam_mpi_comm_self)
#define MPI_GROUP_EMPTY		((MPI_Group) &lam_mpi_group_empty)
#define MPI_MAX			((MPI_Op) &lam_mpi_max)
#define MPI_MIN			((MPI_Op) &lam_mpi_min)
#define MPI_SUM			((MPI_Op) &lam_mpi_sum)
#define MPI_PROD		((MPI_Op) &lam_mpi_prod)
#define MPI_LAND		((MPI_Op) &lam_mpi_land)
#define MPI_BAND		((MPI_Op) &lam_mpi_band)
#define MPI_LOR			((MPI_Op) &lam_mpi_lor)
#define MPI_BOR			((MPI_Op) &lam_mpi_bor)
#define MPI_LXOR		((MPI_Op) &lam_mpi_lxor)
#define MPI_BXOR		((MPI_Op) &lam_mpi_bxor)
#define MPI_MAXLOC		((MPI_Op) &lam_mpi_maxloc)
#define MPI_MINLOC		((MPI_Op) &lam_mpi_minloc)
#define MPI_BYTE		((MPI_Datatype) &lam_mpi_byte)
#define MPI_PACKED		((MPI_Datatype) &lam_mpi_packed)
#define MPI_CHAR		((MPI_Datatype) &lam_mpi_char)
#define MPI_SHORT		((MPI_Datatype) &lam_mpi_short)
#define MPI_INT			((MPI_Datatype) &lam_mpi_int)
#define MPI_LONG		((MPI_Datatype) &lam_mpi_long)
#define MPI_FLOAT		((MPI_Datatype) &lam_mpi_float)
#define MPI_DOUBLE		((MPI_Datatype) &lam_mpi_double)
#define MPI_LONG_DOUBLE		((MPI_Datatype) &lam_mpi_long_double)
#define MPI_CPLEX		((MPI_Datatype) &lam_mpi_cplex)
#define MPI_LOGIC		((MPI_Datatype) &lam_mpi_logic)
#define MPI_UNSIGNED_CHAR	((MPI_Datatype) &lam_mpi_unsigned_char)
#define MPI_UNSIGNED_SHORT	((MPI_Datatype) &lam_mpi_unsigned_short)
#define MPI_UNSIGNED_LONG	((MPI_Datatype) &lam_mpi_unsigned_long)
#define MPI_UNSIGNED		((MPI_Datatype) &lam_mpi_unsigned)
#define MPI_FLOAT_INT		((MPI_Datatype) &lam_mpi_float_int)
#define MPI_DOUBLE_INT		((MPI_Datatype) &lam_mpi_double_int)
#define MPI_LONG_INT		((MPI_Datatype) &lam_mpi_long_int)
#define MPI_SHORT_INT		((MPI_Datatype) &lam_mpi_short_int)
#define MPI_2INT		((MPI_Datatype) &lam_mpi_2int)
#define MPI_2FLOAT		((MPI_Datatype) &lam_mpi_2float)
#define MPI_2DOUBLE		((MPI_Datatype) &lam_mpi_2double)
#define MPI_UB			((MPI_Datatype) &lam_mpi_ub)
#define MPI_LB			((MPI_Datatype) &lam_mpi_lb)
#define MPI_ERRORS_ARE_FATAL	((MPI_Errhandler) &lam_mpi_errors_are_fatal)
#define MPI_ERRORS_RETURN	((MPI_Errhandler) &lam_mpi_errors_return)

/*
 * profiling
 */
#ifdef PROFILELIB

#ifdef PROFILEHDR
#undef PROFILEHDR
#endif

#define MPI_Abort			PMPI_Abort 
#define MPI_Address			PMPI_Address
#define MPI_Allgather			PMPI_Allgather 
#define MPI_Allgatherv			PMPI_Allgatherv 
#define MPI_Allreduce			PMPI_Allreduce 
#define MPI_Alltoall			PMPI_Alltoall 
#define MPI_Alltoallv			PMPI_Alltoallv 
#define MPI_Attr_delete			PMPI_Attr_delete 
#define MPI_Attr_get			PMPI_Attr_get 
#define MPI_Attr_put			PMPI_Attr_put 
#define MPI_Barrier			PMPI_Barrier 
#define MPI_Bcast			PMPI_Bcast 
#define MPI_Bsend			PMPI_Bsend 
#define MPI_Bsend_init			PMPI_Bsend_init 
#define MPI_Buffer_attach		PMPI_Buffer_attach 
#define MPI_Buffer_detach		PMPI_Buffer_detach 
#define MPI_Cancel			PMPI_Cancel 
#define MPI_Cart_coords			PMPI_Cart_coords 
#define MPI_Cart_create			PMPI_Cart_create 
#define MPI_Cart_get			PMPI_Cart_get 
#define MPI_Cart_map			PMPI_Cart_map 
#define MPI_Cart_rank			PMPI_Cart_rank 
#define MPI_Cart_shift			PMPI_Cart_shift 
#define MPI_Cart_sub			PMPI_Cart_sub 
#define MPI_Cartdim_get			PMPI_Cartdim_get 
#define MPI_Comm_compare		PMPI_Comm_compare 
#define MPI_Comm_create			PMPI_Comm_create 
#define MPI_Comm_dup			PMPI_Comm_dup 
#define MPI_Comm_free			PMPI_Comm_free 
#define MPI_Comm_get_contextid		PMPI_Comm_get_contextid 
#define MPI_Comm_group			PMPI_Comm_group 
#define MPI_Comm_rank			PMPI_Comm_rank 
#define MPI_Comm_remote_group		PMPI_Comm_remote_group 
#define MPI_Comm_remote_size		PMPI_Comm_remote_size 
#define MPI_Comm_size			PMPI_Comm_size 
#define MPI_Comm_split			PMPI_Comm_split 
#define MPI_Comm_test_inter		PMPI_Comm_test_inter 
#define MPI_Dims_create			PMPI_Dims_create 
#define MPI_Errhandler_create		PMPI_Errhandler_create 
#define MPI_Errhandler_free		PMPI_Errhandler_free 
#define MPI_Errhandler_get		PMPI_Errhandler_get 
#define MPI_Errhandler_set		PMPI_Errhandler_set 
#define MPI_Error_class			PMPI_Error_class 
#define MPI_Error_string		PMPI_Error_string 
#define MPI_Finalize			PMPI_Finalize 
#define MPI_Gather			PMPI_Gather 
#define MPI_Gatherv			PMPI_Gatherv 
#define MPI_Get_count			PMPI_Get_count 
#define MPI_Get_elements		PMPI_Get_elements 
#define MPI_Get_processor_name		PMPI_Get_processor_name 
#define MPI_Graph_create		PMPI_Graph_create 
#define MPI_Graph_get			PMPI_Graph_get 
#define MPI_Graph_map			PMPI_Graph_map 
#define MPI_Graph_neighbors		PMPI_Graph_neighbors 
#define MPI_Graph_neighbors_count	PMPI_Graph_neighbors_count 
#define MPI_Graphdims_get		PMPI_Graphdims_get 
#define MPI_Group_compare		PMPI_Group_compare 
#define MPI_Group_difference		PMPI_Group_difference 
#define MPI_Group_excl			PMPI_Group_excl 
#define MPI_Group_free			PMPI_Group_free 
#define MPI_Group_incl			PMPI_Group_incl 
#define MPI_Group_intersection		PMPI_Group_intersection 
#define MPI_Group_range_excl		PMPI_Group_range_excl 
#define MPI_Group_range_incl		PMPI_Group_range_incl 
#define MPI_Group_rank			PMPI_Group_rank 
#define MPI_Group_size			PMPI_Group_size 
#define MPI_Group_translate_ranks	PMPI_Group_translate_ranks 
#define MPI_Group_union			PMPI_Group_union 
#define MPI_Ibsend			PMPI_Ibsend 
#define MPI_Init			PMPI_Init 
#define MPI_Initialized			PMPI_Initialized 
#define MPI_Iprobe			PMPI_Iprobe 
#define MPI_Irecv			PMPI_Irecv 
#define MPI_Irsend			PMPI_Irsend 
#define MPI_Isend			PMPI_Isend 
#define MPI_Issend			PMPI_Issend 
#define MPI_Keyval_create		PMPI_Keyval_create 
#define MPI_Keyval_free			PMPI_Keyval_free 
#define MPI_Op_create			PMPI_Op_create 
#define MPI_Op_free			PMPI_Op_free 
#define MPI_Pack			PMPI_Pack 
#define MPI_Pack_size			PMPI_Pack_size 
#define MPI_Pcontrol			PMPI_Pcontrol 
#define MPI_Probe			PMPI_Probe 
#define MPI_Recv			PMPI_Recv 
#define MPI_Recv_init			PMPI_Recv_init 
#define MPI_Reduce			PMPI_Reduce 
#define MPI_Reduce_scatter		PMPI_Reduce_scatter 
#define MPI_Request_free		PMPI_Request_free 
#define MPI_Rsend			PMPI_Rsend 
#define MPI_Rsend_init			PMPI_Rsend_init 
#define MPI_Scan			PMPI_Scan 
#define MPI_Scatter			PMPI_Scatter 
#define MPI_Scatterv			PMPI_Scatterv 
#define MPI_Send			PMPI_Send 
#define MPI_Send_init			PMPI_Send_init 
#define MPI_Sendrecv			PMPI_Sendrecv 
#define MPI_Sendrecv_replace		PMPI_Sendrecv_replace 
#define MPI_Ssend			PMPI_Ssend 
#define MPI_Ssend_init			PMPI_Ssend_init 
#define MPI_Start			PMPI_Start 
#define MPI_Test			PMPI_Test 
#define MPI_Test_cancelled		PMPI_Test_cancelled
#define MPI_Testall			PMPI_Testall 
#define MPI_Testany			PMPI_Testany 
#define MPI_Testsome			PMPI_Testsome 
#define MPI_Topo_test			PMPI_Topo_test 
#define MPI_Type_commit			PMPI_Type_commit 
#define MPI_Type_contiguous		PMPI_Type_contiguous 
#define MPI_Type_extent			PMPI_Type_extent 
#define MPI_Type_free			PMPI_Type_free 
#define MPI_Type_hindexed		PMPI_Type_hindexed 
#define MPI_Type_hvector		PMPI_Type_hvector 
#define MPI_Type_indexed		PMPI_Type_indexed 
#define MPI_Type_lb			PMPI_Type_lb 
#define MPI_Type_size			PMPI_Type_size 
#define MPI_Type_struct			PMPI_Type_struct 
#define MPI_Type_ub			PMPI_Type_ub 
#define MPI_Type_vector			PMPI_Type_vector 
#define MPI_Unpack			PMPI_Unpack 
#define MPI_Wait			PMPI_Wait 
#define MPI_Waitall			PMPI_Waitall 
#define MPI_Waitany			PMPI_Waitany 
#define MPI_Waitsome			PMPI_Waitsome 
#define MPI_Wtick			PMPI_Wtick 
#define MPI_Wtime			PMPI_Wtime 

#endif	/* PROFILELIB */

/*
 * prototypes of LAM MPI library internal functions
 */
#ifndef __ARGS
#if __STDC__ || defined(c_plusplus) || defined(__cplusplus)
#define __ARGS(a)	a
#else
#define __ARGS(a)	()
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern void	__lam_band_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_bkerr_ __ARGS((int, int *, int *, int *));
extern void	__lam_bor_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern int	__lam_bufattach_ __ARGS((void *, int));
extern int	__lam_bufdetach_ __ARGS((void *, int *));
extern int	__lam_buffree_bsend_ __ARGS((MPI_Request));
extern int	__lam_bufinit_ __ARGS((MPI_Request));
extern int	__lam_bufinit_bsend_ __ARGS((MPI_Request));
extern void	__lam_buflength_ __ARGS((MPI_Request));
extern void	__lam_bxor_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern int	__lam_chkarg_ __ARGS((void *, int, MPI_Datatype,
				int, int, MPI_Comm));
extern int	__lam_clocksync_ __ARGS((int, struct _gps *, float8 *));
extern int	__lam_coll2pt_ __ARGS((int));
extern void	__lam_commfault_ __ARGS((int));
extern int	__lam_cubedim_ __ARGS((int));
extern int	__lam_delkey_ __ARGS((MPI_Comm, int));
extern int	__lam_docount_ __ARGS((int, MPI_Datatype, int *, int *));
extern void	__lam_dtcpy_ __ARGS((char *, char *, int, MPI_Datatype));
extern int	__lam_dtflag_ __ARGS((MPI_Datatype));
extern int	__lam_dtsndrcv_ __ARGS((void *, int, MPI_Datatype,
				void *, int, MPI_Datatype, int, MPI_Comm));
extern void	__lam_emptystat_ __ARGS((MPI_Status *));
extern void	__lam_errfatal_ __ARGS((MPI_Comm *, int *));
extern int	__lam_errfunc_ __ARGS((MPI_Comm, int, int));
extern void	__lam_errreturn_ __ARGS((MPI_Comm *, int *));
extern int	__lam_freekey_ __ARGS((int));
extern struct _attrkey *__lam_getattr_ __ARGS((int));
extern int	__lam_getcid_ __ARGS((void));
extern int	__lam_getfunc_ __ARGS((void));
extern struct _attr *__lam_getkey_ __ARGS((MPI_Comm, int));
extern void	__lam_getparam_ __ARGS((int *, int*));
extern int	__lam_hibit_ __ARGS((int, int));
extern int	__lam_inited_ __ARGS((void));
extern void	__lam_initerr_ __ARGS((void));
extern int	__lam_initfault_ __ARGS((void));
extern int	__lam_isend_ __ARGS((void *, int, MPI_Datatype, int,
				int, MPI_Comm, MPI_Request *, int));
extern int	__lam_isend_ger_ __ARGS((MPI_Request *));
extern void	__lam_land_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern int	__lam_linit_ __ARGS((char *, int *, int *, int *, 
				struct _gps **));
extern void	__lam_lor_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_lxor_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_max_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_maxloc_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_min_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_minloc_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern void	__lam_mkcoll_ __ARGS((MPI_Comm));
extern int	__lam_mkerr_ __ARGS((int, int));
extern int	__lam_mkkey_ __ARGS((void));
extern void	__lam_mkpt_ __ARGS((MPI_Comm));
extern struct _proc * __lam_nextproc_ __ARGS((void));
extern int	__lam_nprocs_ __ARGS((void));
extern void	__lam_nukecids_ __ARGS((void));
extern void	__lam_nukefunc_ __ARGS((void));
extern void	__lam_nukekeys_ __ARGS((void));
extern void	__lam_nukeprocs_ __ARGS((void));
extern void	__lam_nuketrace_ __ARGS((void));
extern int	__lam_pack_ __ARGS((char *, int, MPI_Datatype, char *, int));
extern struct _proc * __lam_procadd_ __ARGS((struct _gps *));
extern int	__lam_proccmp_ __ARGS((struct _proc *, struct _proc *));
extern struct _proc *__lam_procfind_ __ARGS((struct _gps *));
extern int	__lam_procrm_ __ARGS((struct _proc *));
extern void	__lam_prod_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern int	__lam_pt2coll_ __ARGS((int));
extern int	__lam_putkey_ __ARGS((MPI_Comm, int));
extern void	__lam_resetfunc_ __ARGS((int));
extern void	__lam_rmcid_ __ARGS((int));
extern int	__lam_send_ __ARGS((void *, int, MPI_Datatype, int, int,
				MPI_Comm, int));
extern void	__lam_setcid_ __ARGS((int));
extern void	__lam_setfunc_ __ARGS((int));
extern void	__lam_setparam_ __ARGS((int, int, int));
extern void	__lam_sum_ __ARGS((void *, void *, int *, MPI_Datatype *));
extern int	__lam_test_ __ARGS((MPI_Request *, int, int *, MPI_Status *));
extern struct _proc *	__lam_topproc_ __ARGS((void));
extern void	__lam_trcffend_ __ARGS((int, int, MPI_Comm,
				MPI_Datatype, int));
extern void	__lam_trcffstart_ __ARGS((int));
extern int	__lam_trcomm_ __ARGS((MPI_Comm));
extern int	__lam_trdtype_ __ARGS((MPI_Datatype));
extern void	__lam_trdtypefree_ __ARGS((MPI_Datatype));
extern int	__lam_trincff_ __ARGS((void));
extern int	__lam_trinit_ __ARGS((char *, float8));
extern void	__lam_trmsg_ __ARGS((int, double, int, int, int, int,
			MPI_Comm, MPI_Datatype, int, int, int, int, int));
extern int	__lam_troff_ __ARGS((void));
extern int	__lam_tron_ __ARGS((void));
extern void	__lam_trstartall_ __ARGS((int, MPI_Request *, double,
				double, double));
extern void	__lam_trsub_ __ARGS((int, double, int));
extern double	__lam_trtimer_ __ARGS((void));
extern void	__lam_trwrapend_ __ARGS((int));
extern int	__lam_trwrapstart_ __ARGS((int));
extern int	__lam_unpack_ __ARGS((char *, int, char *, int, MPI_Datatype));
extern void	__lam_unsyncsql_ __ARGS((int, int, int *, int *,
					int *, int *, int *, int *));
extern void	_m2l_ack __ARGS((struct nmsg *));
extern void	_m2l_fill __ARGS((struct _proc *, struct _proc *, int, int,
				struct nmsg *));
extern struct _proc * _m2l_getsource __ARGS((struct nmsg *));
extern int	_m2l_gettag __ARGS((struct nmsg *));
extern void	_m2l_tail __ARGS((struct nmsg *));
extern void	_mpi_req_add __ARGS((MPI_Request));
extern int	_mpi_req_advance __ARGS((void));
extern void	_mpi_req_blkclr __ARGS((void));
extern void	_mpi_req_blkset __ARGS((MPI_Request));
extern int	_mpi_req_build __ARGS((void *, int, MPI_Datatype, int, int,
				MPI_Comm, int, MPI_Request *));
extern int	_mpi_req_destroy __ARGS((MPI_Request *));
extern int	_mpi_req_end __ARGS((MPI_Request));
extern void	_mpi_req_get __ARGS((int, MPI_Request *));
extern int	_mpi_req_getn __ARGS((void));
extern void	_mpi_req_rem __ARGS((MPI_Request));
extern int	_mpi_req_start __ARGS((MPI_Request));
extern int	_rpi_c2c_addprocs __ARGS((void));
extern int	_rpi_c2c_advance __ARGS((MPI_Request, int));
extern int	_rpi_c2c_build __ARGS((MPI_Request));
extern int	_rpi_c2c_destroy __ARGS((MPI_Request));
extern int	_rpi_c2c_finalize __ARGS((void));
extern int	_rpi_c2c_init __ARGS((void));
extern int	_rpi_c2c_iprobe __ARGS((MPI_Request));
extern int	_rpi_c2c_start __ARGS((MPI_Request, MPI_Request));
extern int	_rpi_lamd_addprocs __ARGS((void));
extern int	_rpi_lamd_advance __ARGS((MPI_Request, int));
extern int	_rpi_lamd_build __ARGS((MPI_Request));
extern int	_rpi_lamd_destroy __ARGS((MPI_Request));
extern int	_rpi_lamd_finalize __ARGS((void));
extern int	_rpi_lamd_init __ARGS((void));
extern int	_rpi_lamd_iprobe __ARGS((MPI_Request));
extern int	_rpi_lamd_start __ARGS((MPI_Request, MPI_Request));

#ifdef __cplusplus
}
#endif

#endif	/* _MPISYS */
