/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/NJN
 *
 *	$Log:	mpitrace.h,v $
 * Revision 6.0  96/02/29  13:53:52  gdburns
 * Ohio Release
 * 
 *	Function:	- MPI tracing constants and templates
 */

#ifndef _MPITRACE
#define _MPITRACE

#include <portable.h>

/*
 * macros
 */
#define	LAM_S2US(time)		((int) ((time) * 1000000.0))

/*
 * constants
 */
#define	LAM_TRMAGIC	1279348022	/* LAM trace file magic number */
#define TRDTINITBUFSIZE	16		/* initial buffer size */
#define TRDPROGMAX	32		/* max length of program name */

#define TRDTCONTIG	((int4) 0)	/* contiguous */
#define TRDTVECTOR	((int4) 1)	/* vector */
#define TRDTHVECTOR	((int4) 2)	/* hvector */
#define TRDTINDEXED	((int4) 3)	/* indexed */
#define TRDTHINDEXED	((int4) 4)	/* hindexed */
#define TRDTSTRUCT	((int4) 5)	/* structure */
#define TRDTBYTE	((int4) 6)	/* byte */
#define TRDTCHAR	((int4) 7)	/* character */
#define TRDTUCHAR	((int4) 8)	/* unsigned character */
#define TRDTSHORT	((int4) 9)	/* short */
#define TRDTUSHORT	((int4) 10)	/* unsigned short */
#define TRDTINT		((int4) 11)	/* integer */
#define TRDTUINT	((int4) 12)	/* unsigned integer */
#define TRDTLOGIC	((int4) 13)	/* logical */
#define TRDTLONG	((int4) 14)	/* long */
#define TRDTULONG	((int4) 15)	/* unsigned long */
#define TRDTFLOAT	((int4) 16)	/* float */
#define TRDTDOUBLE	((int4) 17)	/* double */
#define TRDTCOMPLEX	((int4) 18)	/* complex */
#define TRDTUB		((int4) 19)	/* upper bound */
#define TRDTLB		((int4) 20)	/* lower bound */
#define TRDTPACKED	((int4) 21)	/* packed */
#define TRDT2INT	((int4) 22)	/* 2 ints */
#define TRDT2FLOAT	((int4) 23)	/* 2 floats */
#define TRDT2DOUBLE	((int4) 24)	/* 2 doubles */
#define TRDTFLOATINT	((int4) 25)	/* float-int */
#define TRDTDOUBLEINT	((int4) 26)	/* double-int */
#define TRDTLONGINT	((int4) 27)	/* long-int */
#define TRDTSHORTINT	((int4) 28)	/* short-int */
#define TRDTLONGDOUBLE	((int4) 29)	/* long double */
#define TRDTMAX		((int4) 29)	/* highest flat datatype */

/*
 * trace states
 */
#define TRDTRUN		0			/* running */
#define TRDTBLOCK	1			/* blocked */
#define TRDTSYSTEM	2			/* system overhead */

/*
 * structures
 */
struct trsrc {
	int4		trs_node;	/* node of process that made trace */
	int4		trs_pid;	/* pid of process that made trace */
	int4		trs_listno;	/* list number of the trace */
#define TRRUNTIME	0		/* runtime trace list */	
#define TRCOMM		-2		/* communicator trace list */
#define TRDTYPE		-3		/* datatype trace list */
#define TRONOFF		-4		/* tracing init/on/off trace list */
#define TRWORLD		-5		/* world trace list */
	int4		trs_pad;	/* for alignment */
};

struct trcid {
	int4		trc_cid;	/* cid */
	int4		trc_nlg;	/* size of local group */
	int4		trc_nrg;	/* size of remote group */
	int4		trc_pad;	/* for alignment */
};

struct trdtype {
	int4		trd_param0;
#define trd_dtype	trd_param0	/* datatype */
#define trd_disp	trd_param0	/* displacement (index/struct) */
#define trd_stride	trd_param0	/* stride (vector) */
	int4		trd_param1;
#define trd_count	trd_param1	/* # datatype entries */
#define trd_length	trd_param1	/* length of entry */
};

struct trrthdr {
	float8		trr_time;	/* trace time (sec) */
	int4		trr_type;	/* trace type */
#define TRTINIT		((int4) 0)	/* initialize */
#define TRTSUBCHG	((int4) 1)	/* substrate change */
#define TRTOUTPUT	((int4) 2)	/* output */
#define TRTINPUT	((int4) 3)	/* input */
#define TRTNOIO		((int4) 4)	/* no IO */
#define TRTRON		((int4) 5)	/* tracing on */
#define TRTROFF		((int4) 6)	/* tracing off */
	int4		trr_pad;	/* for alignment */
};

struct tronoff {
	int4		tro_trnum;	/* trace number */
	int4		pad;		/* for alignment */
};

struct trinit {
	float8		tri_skew;		/* clock skew (sec) */
	char		tri_name[TRDPROGMAX];	/* application name */
};

struct trmsg {
	int4		trm_topfunc;	/* top-level function */
	int4		trm_wrapfunc;	/* wrapper function */
	int4            trm_syst;	/* trace time in system (usec) */
	int4		trm_blkt;	/* trace time blocked (usec) */
	int4		trm_peer;	/* peer rank */
	int4		trm_tag;	/* tag */
	int4		trm_cid;	/* context ID */
	int4		trm_dtype;	/* datatype label */
	int4		trm_count;	/* message count */
	int4		trm_mrank;	/* src rank matched in case of wcard */
	int4		trm_mtag;	/* tag matched in case of wcard */
	int4		trm_seqnum;	/* sequence number */
};

struct trxchg {
	int4		trx_topfunc;	/* top-level function */
	int4		trx_sub;	/* new substrate */
};

#endif	/* _MPITRACE */
