/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	MPI.c,v $
 * Revision 6.0.1.2  96/03/13  15:43:02  nevin
 * Flag FORTRAN initialization
 * 
 * Revision 6.0.1.1  96/03/12  12:29:02  nevin
 * Fixed defines of local MPI extension functions
 * 
 * Revision 6.0  96/03/03  22:56:35  nevin
 * Ohio Release
 * 
 * Revision 5.2.1.6  96/03/03  22:55:50  nevin
 * Use MPI_NULL_COPY_FN and MPI_NULL_DELETE_FN
 * 
 * Revision 5.2.1.5  96/03/01  11:12:34  gdburns
 * Common block lamblk_ must be renamed for HP/IBM.
 * 
 * Revision 5.2.1.4  96/02/18  21:17:10  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.3  95/05/28  04:06:14  gdburns
 * Add Fortran to C handle conversion.
 * 
 * Revision 5.2.1.2  95/03/17  14:20:28  raja
 * Remove MPI_TYPE_COUNT.
 * 
 * Revision 5.2.1.1  94/09/26  18:17:58  raja
 * Take care of NULL handles.
 * 
 * Revision 5.2  94/08/22  14:03:03  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.5  94/08/17  16:34:12  raja
 * Change MPI_GET_COUNT(), MPI_GET_ELEMENTS() and
 * MPI_TEST_CANCELLED() to follow the MPI errata sheet.
 * 
 * Revision 5.1.1.4  94/08/12  15:56:12  raja
 * Cast pointer to (char *).
 * 
 * Revision 5.1.1.3  94/08/04  00:31:39  raja
 * Fixed bugs in typealloc() and typefree().
 * 
 * Revision 5.1.1.2  94/06/20  16:23:21  raja
 * Convert MPI_BOTTOM to NULL pointer.
 * 
 * Revision 5.1.1.1  94/05/20  17:59:20  gdburns
 * Make ANSI compilers happy by removing the cast on types[x].
 * 
 * Revision 5.1  94/05/18  12:49:13  gdburns
 * Ohio Release
 * 
 *	Function:	- F77 wrappers for MPI library
 */

#include <stdlib.h>
#include <string.h>

#include <mpi.h>
#include <terror.h>
#include <typical.h>

/*
 * misc. global functions
 */
int			MPI_F2C_comm();
int			MPI_F2C_type();

/*
 * external functions
 */
extern double		MPI_Wtick();
extern double		MPI_Wtime();
extern struct _attrkey	*__lam_getattr_();

/*
 * private variables
 */
static int		type_comm_parent;

/*
 * macros
 */
#define TYPETBLSIZE	64

#define GETHDL(x)	(((x) >= 0) ? types[(x)] : 0)
#define BUF(x)		(((x) == mbot) ? (char *) 0 : (x))

#if RS6K || HPUX
#define mpi_abort_			mpi_abort
#define mpi_address_			mpi_address
#define mpi_allgather_			mpi_allgather
#define mpi_allgatherv_			mpi_allgatherv
#define mpi_allreduce_			mpi_allreduce
#define mpi_alltoall_			mpi_alltoall
#define mpi_alltoallv_			mpi_alltoallv
#define mpi_attr_delete_		mpi_attr_delete
#define mpi_attr_get_			mpi_attr_get
#define mpi_attr_put_			mpi_attr_put
#define mpi_barrier_			mpi_barrier
#define mpi_bcast_			mpi_bcast
#define mpi_bsend_			mpi_bsend
#define mpi_bsend_init_			mpi_bsend_init
#define mpi_buffer_attach_		mpi_buffer_attach
#define mpi_buffer_detach_		mpi_buffer_detach
#define mpi_cancel_			mpi_cancel
#define mpi_cart_coords_		mpi_cart_coords
#define mpi_cart_create_		mpi_cart_create
#define mpi_cart_get_			mpi_cart_get
#define mpi_cart_map_			mpi_cart_map
#define mpi_cart_rank_			mpi_cart_rank
#define mpi_cart_shift_			mpi_cart_shift
#define mpi_cart_sub_			mpi_cart_sub
#define mpi_cartdim_get_		mpi_cartdim_get
#define mpi_comm_compare_		mpi_comm_compare
#define mpi_comm_create_		mpi_comm_create
#define mpi_comm_dup_			mpi_comm_dup
#define mpi_comm_free_			mpi_comm_free
#define mpi_comm_group_			mpi_comm_group
#define mpi_comm_rank_			mpi_comm_rank
#define mpi_comm_remote_group_		mpi_comm_remote_group
#define mpi_comm_remote_size_		mpi_comm_remote_size
#define mpi_comm_size_			mpi_comm_size
#define mpi_comm_split_			mpi_comm_split
#define mpi_comm_test_inter_		mpi_comm_test_inter
#define mpi_dims_create_		mpi_dims_create
#define mpi_errhandler_create_		mpi_errhandler_create
#define mpi_errhandler_free_		mpi_errhandler_free
#define mpi_errhandler_get_		mpi_errhandler_get
#define mpi_errhandler_set_		mpi_errhandler_set
#define mpi_error_class_		mpi_error_class
#define mpi_error_string_		mpi_error_string
#define mpi_finalize_			mpi_finalize
#define mpi_gather_			mpi_gather
#define mpi_gatherv_			mpi_gatherv
#define mpi_get_count_			mpi_get_count
#define mpi_get_elements_		mpi_get_elements
#define mpi_get_processor_name_		mpi_get_processor_name
#define mpi_graph_create_		mpi_graph_create
#define mpi_graph_get_			mpi_graph_get
#define mpi_graph_map_			mpi_graph_map
#define mpi_graph_neighbors_		mpi_graph_neighbors
#define mpi_graph_neighbors_count_	mpi_graph_neighbors_count
#define mpi_graphdims_get_		mpi_graphdims_get
#define mpi_group_compare_		mpi_group_compare
#define mpi_group_difference_		mpi_group_difference
#define mpi_group_excl_			mpi_group_excl
#define mpi_group_free_			mpi_group_free
#define mpi_group_incl_			mpi_group_incl
#define mpi_group_intersection_		mpi_group_intersection
#define mpi_group_range_excl_		mpi_group_range_excl
#define mpi_group_range_incl_		mpi_group_range_incl
#define mpi_group_rank_			mpi_group_rank
#define mpi_group_size_			mpi_group_size
#define mpi_group_translate_ranks_	mpi_group_translate_ranks
#define mpi_group_union_		mpi_group_union
#define mpi_ibsend_			mpi_ibsend
#define mpi_init_			mpi_init
#define mpi_initialized_		mpi_initialized
#define mpi_intercomm_create_		mpi_intercomm_create
#define mpi_intercomm_merge_		mpi_intercomm_merge
#define mpi_iprobe_			mpi_iprobe
#define mpi_irecv_			mpi_irecv
#define mpi_irsend_			mpi_irsend
#define mpi_isend_			mpi_isend
#define mpi_issend_			mpi_issend
#define mpi_keyval_create_		mpi_keyval_create
#define mpi_keyval_free_		mpi_keyval_free
#define mpi_op_create_			mpi_op_create
#define mpi_op_free_			mpi_op_free
#define mpi_pack_			mpi_pack
#define mpi_pack_size_			mpi_pack_size
#define mpi_pcontrol_			mpi_pcontrol
#define mpi_probe_			mpi_probe
#define mpi_recv_			mpi_recv
#define mpi_recv_init_			mpi_recv_init
#define mpi_reduce_			mpi_reduce
#define mpi_reduce_scatter_		mpi_reduce_scatter
#define mpi_request_free_		mpi_request_free
#define mpi_rsend_			mpi_rsend
#define mpi_rsend_init_			mpi_rsend_init
#define mpi_scan_			mpi_scan
#define mpi_scatter_			mpi_scatter
#define mpi_scatterv_			mpi_scatterv
#define mpi_send_			mpi_send
#define mpi_send_init_			mpi_send_init
#define mpi_sendrecv_			mpi_sendrecv
#define mpi_sendrecv_replace_		mpi_sendrecv_replace
#define mpi_ssend_			mpi_ssend
#define mpi_ssend_init_			mpi_ssend_init
#define mpi_start_			mpi_start
#define mpi_startall_			mpi_startall
#define mpi_test_			mpi_test
#define mpi_test_cancelled_		mpi_test_cancelled
#define mpi_testall_			mpi_testall
#define mpi_testany_			mpi_testany
#define mpi_testsome_			mpi_testsome
#define mpi_topo_test_			mpi_topo_test
#define mpi_type_commit_		mpi_type_commit
#define mpi_type_contiguous_		mpi_type_contiguous
#define mpi_type_count_			mpi_type_count
#define mpi_type_extent_		mpi_type_extent
#define mpi_type_free_			mpi_type_free
#define mpi_type_hindexed_		mpi_type_hindexed
#define mpi_type_hvector_		mpi_type_hvector
#define mpi_type_indexed_		mpi_type_indexed
#define mpi_type_lb_			mpi_type_lb
#define mpi_type_size_			mpi_type_size
#define mpi_type_struct_		mpi_type_struct
#define mpi_type_ub_			mpi_type_ub
#define mpi_type_vector_		mpi_type_vector
#define mpi_unpack_			mpi_unpack
#define mpi_wait_			mpi_wait
#define mpi_waitall_			mpi_waitall
#define mpi_waitany_			mpi_waitany
#define mpi_waitsome_			mpi_waitsome
#define mpi_wtick_			mpi_wtick
#define mpi_wtime_			mpi_wtime

#define mpil_comm_id_			mpil_comm_id
#define mpil_comm_gps_			mpil_comm_gps
#define mpil_type_id_			mpil_type_id
#define mpil_signal_			mpil_signal
#define mpil_spawn_			mpil_spawn
#define mpil_comm_parent_		mpil_comm_parent
#define mpil_universe_size_		mpil_universe_size
#define mpil_trace_on_			mpil_trace_on
#define mpil_trace_off_			mpil_trace_off

#define lamblk_				lamblk
#endif

/*
 * external variables
 */
extern int		lamblk_;

/*
 * local functions
 */
static int		typealloc();
static int		typefind();
static void		maketype();
static void		typefree();
static void		c2fstat();
static void		f2cstat();

/*
 * static variables
 */
static int		maxtype = 0;
static int		maxpredef = 0;
static int		ntypes = 0;
static int		typesize = 0;
static void		**types = 0;
static char		*mbot;

/*
 * thin wrappers
 */
void mpi_pcontrol_(l) int *l; { MPI_Pcontrol(*l); }
void mpi_finalize_(ierr) int *ierr; { *ierr = MPI_Finalize(); }
void mpi_initialized_(f, ierr) int *f, *ierr; { *ierr = MPI_Initialized(f); }
void mpi_abort_(comm, err, ierr) int *comm, *err, *ierr;
	{ *ierr = MPI_Abort(GETHDL(*comm), *err); }
void mpi_error_class_(err, c, ierr) int *err, *c, *ierr;
	{ *ierr = MPI_Error_class(*err, c); }
void mpi_error_string_(err, str, l, ierr) int *err; char *str; int *l, *ierr;
	{ *ierr = MPI_Error_string(*err, str, l); }
void mpi_errhandler_free_(hdl, ierr) int *hdl, *ierr;
	{ MPI_Errhandler h; h = GETHDL(*hdl); *ierr = MPI_Errhandler_free(&h);
		typefree(*hdl); *hdl = -1; }
void mpi_errhandler_set_(comm, hdl, ierr) int *comm, *hdl, *ierr;
	{ MPI_Errhandler h; h = GETHDL(*hdl);
		h->eh_f77hdl = (*hdl <= maxpredef) ? -1 : *comm;
		*ierr = MPI_Errhandler_set(GETHDL(*comm), h); }
void mpi_get_processor_name_(name, l, ierr) char *name; int *l, *ierr;
	{ *ierr = MPI_Get_processor_name(name, l); }
void mpi_comm_remote_size_(comm, s, ierr) int *comm, *s, *ierr;
	{ *ierr = MPI_Comm_remote_size(GETHDL(*comm), s); }
void mpi_comm_test_inter_(comm, f, ierr) int *comm, *f, *ierr;
	{ *ierr = MPI_Comm_test_inter(GETHDL(*comm), f); }
void mpi_comm_compare_(c1, c2, r, ierr) int *c1, *c2, *r, *ierr;
	{ *ierr = MPI_Comm_compare(GETHDL(*c1), GETHDL(*c2), r); }
void mpi_comm_rank_(c, rank, ierr) int *c, *rank, *ierr;
	{ *ierr = MPI_Comm_rank(GETHDL(*c), rank); }
void mpi_comm_size_(c, size, ierr) int *c, *size, *ierr;
	{ *ierr = MPI_Comm_size(GETHDL(*c), size); }
void mpi_group_free_(g, ierr) int *g, *ierr;
	{ MPI_Group grp; grp = GETHDL(*g); *ierr = MPI_Group_free(&grp);
		typefree(*g); *g = -1; }
void mpi_group_compare_(g1, g2, r, ierr) int *g1, *g2, *r, *ierr;
	{ *ierr = MPI_Group_compare(GETHDL(*g1), GETHDL(*g2), r); }
void mpi_group_size_(grp, size, ierr) int *grp, *size, *ierr;
	{ *ierr = MPI_Group_size(GETHDL(*grp), size); }
void mpi_group_rank_(grp, rank, ierr) int *grp, *rank, *ierr;
	{ *ierr = MPI_Group_rank(GETHDL(*grp), rank); }
void mpi_group_translate_ranks_(g1, n, r1, g2, r2, ierr)
	int *g1, *n, *r1, *g2, *r2, *ierr;
	{ *ierr = MPI_Group_translate_ranks(GETHDL(*g1), *n, r1,
						GETHDL(*g2), r2); }
void mpi_comm_free_(c, ierr) int *c, *ierr;
	{ MPI_Comm comm; comm = GETHDL(*c); *ierr = MPI_Comm_free(&comm);
		typefree(*c); *c = -1; }
void mpi_dims_create_(nn, nd, d, ierr) int *nn, *nd, *d, *ierr;
	{ *ierr = MPI_Dims_create(*nn, *nd, d); }
void mpi_topo_test_(c, stat, ierr) int *c, *stat, *ierr;
	{ *ierr = MPI_Topo_test(GETHDL(*c), stat); }
void mpi_cartdim_get_(c, n, ierr) int *c, *n, *ierr;
	{ *ierr = MPI_Cartdim_get(GETHDL(*c), n); }
void mpi_cart_get_(c, n, d, p, cd, ierr) int *c, *n, *d, *p, *cd, *ierr;
	{ *ierr = MPI_Cart_get(GETHDL(*c), *n, d, p, cd); }
void mpi_graph_get_(c, mi, me, i, e, ierr) int *c, *mi, *me, *i, *e, *ierr;
	{ *ierr = MPI_Graph_get(GETHDL(*c), *mi, *me, i, e); }
void mpi_graphdims_get_(c, nn, ne, ierr) int *c, *nn, *ne, *ierr;
	{ *ierr = MPI_Graphdims_get(GETHDL(*c), nn, ne); }
void mpi_cart_rank_(c, cd, r, ierr) int *c, *cd, *r, *ierr;
	{ *ierr = MPI_Cart_rank(GETHDL(*c), cd, r); }
void mpi_cart_coords_(c, r, n, cd, ierr) int *c, *r, *n, *cd,*ierr;
	{ *ierr = MPI_Cart_coords(GETHDL(*c), *r, *n, cd); }
void mpi_cart_shift_(c, dr, dp, s, d, ierr) int *c, *dr, *dp, *s, *d, *ierr;
	{ *ierr = MPI_Cart_shift(GETHDL(*c), *dr, *dp, s, d); }
void mpi_cart_map_(comm, n, d, p, r, ierr) int *comm, *n, *d, *p, *r, *ierr;
	{ *ierr = MPI_Cart_map(GETHDL(*comm), *n, d, p, r); }
void mpi_graph_map_(comm, n, i, e, r, ierr) int *comm, *n, *i, *e, *r, *ierr;
	{ *ierr = MPI_Graph_map(GETHDL(*comm), *n, i, e, r); }
void mpi_graph_neighbors_(c, r, m, nbr, ierr) int *c, *r, *m, *nbr, *ierr;
	{ *ierr = MPI_Graph_neighbors(GETHDL(*c), *r, *m, nbr); }
void mpi_graph_neighbors_count_(c, r, n, ierr) int *c, *r, *n, *ierr;
	{ *ierr = MPI_Graph_neighbors_count(GETHDL(*c), *r, n); }
void mpi_pack_(ib, ic, t, ob, oc, p, c, ierr)
	char *ib, *ob; int *ic, *t, *oc, *p, *c, *ierr;
	{ *ierr = MPI_Pack(BUF(ib), *ic, GETHDL(*t), ob, *oc, p, GETHDL(*c)); }
void mpi_unpack_(ib, is, p, ob, oc, t, c, ierr)
	char *ib, *ob; int *is, *p, *oc, *t, *c, *ierr;
	{ *ierr = MPI_Unpack(ib, *is, p, BUF(ob),
			*oc, GETHDL(*t), GETHDL(*c)); }
void mpi_pack_size_(n, t, c, s, ierr) int *n, *t, *c, *s, *ierr;
	{ *ierr = MPI_Pack_size(*n, GETHDL(*t), GETHDL(*c), s); }
void mpi_barrier_(c, ierr) int *c, *ierr;
	{ *ierr = MPI_Barrier(GETHDL(*c)); }
void mpi_bcast_(b, n, t, r, c, ierr) char *b; int *n, *t, *r, *c, *ierr;
	{ *ierr = MPI_Bcast(BUF(b), *n, GETHDL(*t), *r, GETHDL(*c)); }
void mpi_gather_(sb, sc, st, rb, rc, rt, r, c, ierr)
	char *sb, *rb; int *sc, *st, *rc, *rt, *r, *c, *ierr;
	{ *ierr = MPI_Gather(BUF(sb), *sc, GETHDL(*st),
			BUF(rb), *rc, GETHDL(*rt), *r, GETHDL(*c)); }
void mpi_gatherv_(sb, sc, st, rb, rc, rd, rt, r, c, ierr)
	char *sb, *rb; int *sc, *st, *rc, *rd, *rt, *r, *c, *ierr;
	{ *ierr = MPI_Gatherv(BUF(sb), *sc, GETHDL(*st),
			BUF(rb), rc, rd, GETHDL(*rt), *r, GETHDL(*c)); }
void mpi_scatter_(sb, sc, st, rb, rc, rt, r, c, ierr)
	char *sb, *rb; int *sc, *st, *rc, *rt, *r, *c, *ierr;
	{ *ierr = MPI_Scatter(BUF(sb), *sc, GETHDL(*st),
			BUF(rb), *rc, GETHDL(*rt), *r, GETHDL(*c)); }
void mpi_scatterv_(sb, sc, sd, st, rb, rc, rt, r, c, ierr)
	char *sb, *rb; int *sc, *sd, *st, *rc, *rt, *r, *c, *ierr;
	{ *ierr = MPI_Scatterv(BUF(sb), sc, sd, GETHDL(*st),
			BUF(rb), *rc, GETHDL(*rt), *r, GETHDL(*c)); }
void mpi_allgather_(sb, sc, st, rb, rc, rt, c, ierr)
	char *sb, *rb; int *sc, *st, *rc, *rt, *c, *ierr;
	{ *ierr = MPI_Allgather(BUF(sb), *sc, GETHDL(*st),
			BUF(rb), *rc, GETHDL(*rt), GETHDL(*c)); }
void mpi_allgatherv_(sb, sc, st, rb, rc, rd, rt, c, ierr)
	char *sb, *rb; int *sc, *st, *rc, *rd, *rt, *c, *ierr;
	{ *ierr = MPI_Allgatherv(BUF(sb), *sc, GETHDL(*st),
			BUF(rb), rc, rd, GETHDL(*rt), GETHDL(*c)); }
void mpi_alltoall_(sb, sc, st, rb, rc, rt, c, ierr)
	char *sb, *rb; int *sc, *st, *rc, *rt, *c, *ierr;
	{ *ierr = MPI_Alltoall(BUF(sb), *sc, GETHDL(*st),
			BUF(rb), *rc, GETHDL(*rt), GETHDL(*c)); }
void mpi_alltoallv_(sb, sc, sd, st, rb, rc, rd, rt, c, ierr)
	char *sb, *rb; int *sc, *sd, *st, *rc, *rd, *rt, *c, *ierr;
	{ *ierr = MPI_Alltoallv(BUF(sb), sc, sd, GETHDL(*st),
			BUF(rb), rc, rd, GETHDL(*rt), GETHDL(*c)); }
void mpi_reduce_(sb, rb, n, t, op, r, c, ierr)
	char *sb, *rb; int *n, *t, *op, *r, *c, *ierr;
	{ MPI_Op opc; opc = GETHDL(*op);
		opc->op_f77dtype = (*op <= maxpredef) ? -1 : *t;
		*ierr = MPI_Reduce(BUF(sb), BUF(rb), *n,
				GETHDL(*t), opc, *r, GETHDL(*c)); }
void mpi_op_free_(op, ierr) int *op, *ierr;
	{ MPI_Op opc; opc = GETHDL(*op); *ierr = MPI_Op_free(&opc);
		typefree(*op); *op = -1; }
void mpi_allreduce_(sb, rb, n, t, op, c, ierr)
	char *sb, *rb; int *n, *t, *op, *c, *ierr;
	{ MPI_Op opc; opc = GETHDL(*op);
		opc->op_f77dtype = (*op <= maxpredef) ? -1 : *t;
		*ierr = MPI_Allreduce(BUF(sb), BUF(rb), *n,
				GETHDL(*t), opc, GETHDL(*c)); }
void mpi_reduce_scatter_(sb, rb, rn, t, op, comm, ierr)
	char *sb, *rb; int *rn, *t, *op, *comm, *ierr;
	{ MPI_Op opc; opc = GETHDL(*op);
		opc->op_f77dtype = (*op <= maxpredef) ? -1 : *t;
		*ierr = MPI_Reduce_scatter(BUF(sb), BUF(rb), rn,
				GETHDL(*t), opc, GETHDL(*comm)); }
void mpi_scan_(sb, rb, n, t, op, comm, ierr)
	char *sb, *rb; int *n, *t, *op, *comm, *ierr;
	{ MPI_Op opc; opc = GETHDL(*op);
		opc->op_f77dtype = (*op <= maxpredef) ? -1 : *t;
		*ierr = MPI_Scan(BUF(sb), BUF(rb), *n,
				GETHDL(*t), opc, GETHDL(*comm)); }
void mpi_send_(b, n, t, d, tag, c, ierr)
	char *b; int *n, *t, *d, *tag, *c, *ierr;
	{ *ierr = MPI_Send(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c)); }
void mpi_bsend_(b, n, t, d, tag, c, ierr)
	char *b; int *n, *t, *d, *tag, *c, *ierr;
	{ *ierr = MPI_Bsend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c)); }
void mpi_ssend_(b, n, t, d, tag, c, ierr)
	char *b; int *n, *t, *d, *tag, *c, *ierr;
	{ *ierr = MPI_Ssend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c)); }
void mpi_rsend_(b, n, t, d, tag, c, ierr)
	char *b; int *n, *t, *d, *tag, *c, *ierr;
	{ *ierr = MPI_Rsend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c)); }
void mpi_request_free_(rq, ierr) int *rq, *ierr;
	{ MPI_Request req; req = GETHDL(*rq);
		*ierr = MPI_Request_free(&req); typefree(*rq); *rq = -1; }
void mpi_buffer_attach_(b, s, ierr) char *b; int *s, *ierr;
	{ *ierr = MPI_Buffer_attach(BUF(b), *s); }
void mpi_buffer_detach_(b, s, ierr) char *b; int *s, *ierr;
	{ void *junk; *ierr = MPI_Buffer_detach(&junk, s); }
void mpi_start_(r, ierr) int *r, *ierr;
	{ MPI_Request req; req = GETHDL(*r); *ierr = MPI_Start(&req); }
void mpi_cancel_(r, ierr) int *r, *ierr;
	{ MPI_Request req; req = GETHDL(*r); *ierr = MPI_Cancel(&req); }
void mpi_address_(l, a, ierr) char *l; int *a, *ierr;
	{ MPI_Aint ad; *ierr = MPI_Address(BUF(l), &ad); *a = (int) ad; }
void mpi_type_extent_(t, e, ierr) int *t, *e, *ierr;
	{ *ierr = MPI_Type_extent(GETHDL(*t), (MPI_Aint *) e); }
void mpi_type_size_(t, s, ierr) int *t, *s, *ierr;
	{ *ierr = MPI_Type_size(GETHDL(*t), s); }
void mpi_type_lb_(t, d, ierr) int *t, *d, *ierr;
	{ *ierr = MPI_Type_lb(GETHDL(*t), (MPI_Aint *) d); }
void mpi_type_ub_(t, d, ierr) int *t, *d, *ierr;
	{ *ierr = MPI_Type_ub(GETHDL(*t), (MPI_Aint *) d); }
void mpi_type_commit_(t, ierr) int *t, *ierr;
	{ MPI_Datatype dtype; dtype = GETHDL(*t);
		*ierr = MPI_Type_commit(&dtype); }
void mpi_type_free_(t, ierr) int *t, *ierr;
	{ MPI_Datatype dtype; dtype = GETHDL(*t);
		*ierr = MPI_Type_free(&dtype); typefree(*t); *t = -1; }
void mpi_keyval_create_(cf, df, k, es, ierr)
	int (*cf)(); int (*df)(); int *k, *es, *ierr;
	{ if (*cf == 0) cf = MPI_NULL_COPY_FN;
		if (*df == 0) df = MPI_NULL_DELETE_FN;
		*ierr = MPI_Keyval_create(cf, df, k, (void *) *es); }
void mpi_keyval_free_(k, ierr) int *k, *ierr;
	{ *ierr = MPI_Keyval_free(k); }
void mpi_attr_put_(c, k, a, ierr) int *c, *k, *a, *ierr;
	{ struct _attrkey *pk;
		*ierr = MPI_Attr_put(GETHDL(*c), *k, (void *) *a);
		pk = __lam_getattr_(*k); pk->ak_f77comm = *c; }
void mpi_attr_get_(c, k, a, f, ierr) int *c, *k, *a, *f, *ierr;
	{ *ierr = MPI_Attr_get(GETHDL(*c), *k, a, f); }
void mpi_attr_delete_(c, k, ierr) int *c, *k, *ierr;
	{ *ierr = MPI_Attr_delete(GETHDL(*c), *k); }
double mpi_wtime_() { return(MPI_Wtime()); }
double mpi_wtick_() { return(MPI_Wtick()); }

void mpil_comm_id_(c, id, ierr) int *c, *id, *ierr;
	{ *ierr = MPIL_Comm_id(GETHDL(*c), id); }
void mpil_comm_gps_(c, r, n, p, ierr) int *c, *r, *n, *p, *ierr;
	{ *ierr = MPIL_Comm_gps(GETHDL(*c), *r, n, p); }
void mpil_type_id_(t, id, ierr) int *t, *id, *ierr;
	{ *ierr = MPIL_Type_id(GETHDL(*t), id); }
void mpil_signal_(c, r, s, ierr) int *c, *r, *s, *ierr;
	{ *ierr = MPIL_Signal(GETHDL(*c), *r, *s); }
void mpil_trace_on_(ierr) int *ierr; { *ierr = MPIL_Trace_on(); }
void mpil_trace_off_(ierr) int *ierr; { *ierr = MPIL_Trace_off(); }
void mpil_universe_size_(s, ierr) int *s, *ierr;
	{ *ierr = MPIL_Universe_size(s); }

/*
 * thick wrappers
 */
void
mpi_init_(ierror)

int			*ierror;

{
	int		argc;
	char		*argv[1];
	char		**pargv;
/*
 * Do an MPI_Init().
 */
	argc = 0;
	argv[0] = 0;
	pargv = argv;

	__lam_f77init_ = 1;
	*ierror = MPI_Init(&argc, &pargv);
	if (*ierror != MPI_SUCCESS) return;
/*
 * Initialize the type lookup table.
 */
	typesize = TYPETBLSIZE;
	types = (void **) malloc((unsigned) typesize * sizeof(void *));
	if (types == 0) {
		*ierror = __lam_mkerr_(MPI_ERR_OTHER, errno);;
		return;
	}

	memset((char *) types, 0, typesize * sizeof(void *));

	types[ntypes++] = MPI_COMM_WORLD;
	types[ntypes++] = MPI_COMM_SELF;

	types[ntypes++] = MPI_GROUP_EMPTY;

	types[ntypes++] = MPI_ERRORS_ARE_FATAL;
	types[ntypes++] = MPI_ERRORS_RETURN;

	types[ntypes++] = MPI_INT;
	types[ntypes++] = MPI_FLOAT;
	types[ntypes++] = MPI_DOUBLE;
	types[ntypes++] = MPI_CPLEX;
	types[ntypes++] = MPI_LOGIC;
	types[ntypes++] = MPI_CHAR;
	types[ntypes++] = MPI_BYTE;
	types[ntypes++] = MPI_UB;
	types[ntypes++] = MPI_LB;
	types[ntypes++] = MPI_2FLOAT;
	types[ntypes++] = MPI_2DOUBLE;
	types[ntypes++] = MPI_2INT;

	types[ntypes++] = MPI_MAX;
	types[ntypes++] = MPI_MIN;
	types[ntypes++] = MPI_SUM;
	types[ntypes++] = MPI_PROD;
	types[ntypes++] = MPI_LAND;
	types[ntypes++] = MPI_BAND;
	types[ntypes++] = MPI_LOR;
	types[ntypes++] = MPI_BOR;
	types[ntypes++] = MPI_LXOR;
	types[ntypes++] = MPI_BXOR;
	types[ntypes++] = MPI_MAXLOC;
	types[ntypes++] = MPI_MINLOC;
	
	maxpredef = maxtype = ntypes - 1;
/*
 * Make a type for the parent communicator.
 */
	maketype(&type_comm_parent, ierror, lam_mpi_comm_parent);

	mbot = (char *) &lamblk_;
}

void
mpi_errhandler_get_(comm, hdl, ierr)

int			*comm, *hdl, *ierr;

{
	MPI_Errhandler	errhdl;

	*ierr = MPI_Errhandler_get(GETHDL(*comm), &errhdl);
	if (*ierr == MPI_SUCCESS) {
		if ((*hdl = typefind((void *) errhdl)) < 0) {
			if ((*hdl = typealloc()) < 0) {
				*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
				*hdl = -1;
			} else {
				types[*hdl] = errhdl;
			}
		}
	}
}

void
mpi_errhandler_create_(func, hdl, ierr)

void			(*func)();
int			*hdl, *ierr;

{
	MPI_Errhandler	errhdl;

	*ierr = MPI_Errhandler_create((MPI_Handler_function *) func, &errhdl);
	maketype(hdl, ierr, (void *) errhdl);
}

void
mpi_intercomm_merge_(ic, hi, nc, ierr)

int			*ic, *hi, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Intercomm_merge(GETHDL(*ic), *hi, &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_intercomm_create_(lc, ll, pc, pl, tag, nc, ierr)

int			*lc, *ll, *pc, *pl, *tag, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Intercomm_create(GETHDL(*lc), *ll,
					GETHDL(*pc), *pl, *tag, &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_comm_remote_group_(comm, g, ierr)

int			*comm, *g, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Comm_remote_group(GETHDL(*comm), &group);
	maketype(g, ierr, (void *) group);
}

void
mpi_group_union_(g1, g2, ng, ierr)

int			*g1, *g2, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_union(GETHDL(*g1), GETHDL(*g2), &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_group_intersection_(g1, g2, ng, ierr)

int			*g1, *g2, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_intersection(GETHDL(*g1), GETHDL(*g2), &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_group_difference_(g1, g2, ng, ierr)

int			*g1, *g2, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_difference(GETHDL(*g1), GETHDL(*g2), &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_group_incl_(g, n, ranks, ng, ierr)

int			*g, *n, *ranks, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_incl(GETHDL(*g), *n, ranks, &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_group_range_incl_(g, n, ranges, ng, ierr)

int			*g, *n, *ranges, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_range_incl(GETHDL(*g), *n, (void *) ranges, &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_group_excl_(g, n, ranks, ng, ierr)

int			*g, *n, *ranks, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_excl(GETHDL(*g), *n, ranks, &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_group_range_excl_(g, n, ranges, ng, ierr)

int			*g, *n, *ranges, *ng, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Group_range_excl(GETHDL(*g), *n, (void *) ranges, &group);
	maketype(ng, ierr, (void *) group);
}

void
mpi_comm_group_(c, g, ierr)

int			*c, *g, *ierr;

{
	MPI_Group	group;

	*ierr = MPI_Comm_group(GETHDL(*c), &group);
	maketype(g, ierr, (void *) group);
}

void
mpi_op_create_(func, cmt, op, ierr)

void			(*func)();
int			*cmt, *op, *ierr;

{
	MPI_Op		oper;

	*ierr = MPI_Op_create(func, *cmt, &oper);
	maketype(op, ierr, (void *) oper);
}

void
mpi_comm_dup_(c, nc, ierr)

int			*c, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Comm_dup(GETHDL(*c), &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_comm_create_(c, g, nc, ierr)

int			*c, *g, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Comm_create(GETHDL(*c), GETHDL(*g), &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_comm_split_(c, colour, key, nc, ierr)

int			*c, *colour, *key, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Comm_split(GETHDL(*c), *colour, *key, &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_graph_create_(c, n, i, e, re, nc, ierr)

int			*c, *n, *i, *e, *re, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Graph_create(GETHDL(*c), *n, i, e, *re, &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_cart_create_(c, n, d, p, re, nc, ierr)

int			*c, *n, *d, *p, *re, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Cart_create(GETHDL(*c), *n, d, p, *re, &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_cart_sub_(c, rdim, nc, ierr)

int			*c, *rdim, *nc, *ierr;

{
	MPI_Comm	comm;

	*ierr = MPI_Cart_sub(GETHDL(*c), rdim, &comm);
	maketype(nc, ierr, (void *) comm);
}

void
mpi_recv_(b, n, t, s, tag, c, st, ierr)

char			*b;
int			*n, *t, *s, *tag, *c, *st, *ierr;

{
	MPI_Status	stat;

	*ierr = MPI_Recv(BUF(b), *n, GETHDL(*t), *s, *tag, GETHDL(*c), &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, st);
		st[6] = *t;
	}
}

void
mpi_get_count_(st, t, cnt, ierr)

int			*st, *t, *cnt, *ierr;

{
	MPI_Status	stat;

	f2cstat(st, &stat);

	*ierr = MPI_Get_count(&stat, GETHDL(*t), cnt);
	if (*ierr == MPI_SUCCESS) {
		st[3] = stat.st_count;
		st[4] = stat.st_nelem;
	}
}

void
mpi_get_elements_(st, t, cnt, ierr)

int			*st, *t, *cnt, *ierr;

{
	MPI_Status	stat;

	f2cstat(st, &stat);

	*ierr = MPI_Get_elements(&stat, GETHDL(*t), cnt);
	if (*ierr == MPI_SUCCESS) {
		st[3] = stat.st_count;
		st[4] = stat.st_nelem;
	}
}

void
mpi_probe_(s, tag, c, st, ierr)

int			*s, *tag, *c, *st, *ierr;

{
	MPI_Status	stat;

	*ierr = MPI_Probe(*s, *tag, GETHDL(*c), &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, st);
		st[6] = -1;
	}
}

void
mpi_iprobe_(s, tag, c, f, st, ierr)

int			*s, *tag, *c, *f, *st, *ierr;

{
	MPI_Status	stat;

	*ierr = MPI_Iprobe(*s, *tag, GETHDL(*c), f, &stat);
	if ((*ierr == MPI_SUCCESS) && (*f != 0)) {
		c2fstat(&stat, st);
		st[6] = -1;
	}
}

void
mpi_test_cancelled_(st, f, ierr)

int			*st, *f, *ierr;

{
	MPI_Status	stat;

	f2cstat(st, &stat);
	*ierr = MPI_Test_cancelled(&stat, f);
}

void
mpi_send_init_(b, n, t, d, tag, c, r, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *r, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Send_init(BUF(b), *n, GETHDL(*t),
				*d, *tag, GETHDL(*c), &req);
	maketype(r, ierr, (void *) req);
}

void
mpi_bsend_init_(b, n, t, d, tag, c, r, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *r, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Bsend_init(BUF(b), *n, GETHDL(*t),
				*d, *tag, GETHDL(*c), &req);
	maketype(r, ierr, (void *) req);
}

void
mpi_ssend_init_(b, n, t, d, tag, c, r, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *r, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Ssend_init(BUF(b), *n, GETHDL(*t),
				*d, *tag, GETHDL(*c), &req);
	maketype(r, ierr, (void *) req);
}

void
mpi_rsend_init_(b, n, t, d, tag, c, r, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *r, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Rsend_init(BUF(b), *n, GETHDL(*t),
				*d, *tag, GETHDL(*c), &req);
	maketype(r, ierr, (void *) req);
}

void
mpi_recv_init_(b, n, t, s, tag, c, r, ierr)

char			*b;
int			*n, *t, *s, *tag, *c, *r, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Recv_init(BUF(b), *n, GETHDL(*t),
				*s, *tag, GETHDL(*c), &req);
	maketype(r, ierr, (void *) req);
}

void
mpi_startall_(c, r, ierr)

int			*c, *r, *ierr;

{
	MPI_Request	*rq;
	int		i;

	rq = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (rq == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) rq[i] = GETHDL(r[i]);

	*ierr = MPI_Startall(*c, rq);

	free((char *) rq);
}

void
mpi_sendrecv_(sb, sc, st, d, stag, rb, rc, rt, s, rtag, c, sta, ierr)

char			*sb, *rb;
int			*sc, *st, *d, *stag;
int			*rc, *rt, *s, *rtag, *c, *sta, *ierr;

{
	MPI_Status	stat;

	*ierr = MPI_Sendrecv(BUF(sb), *sc, GETHDL(*st), *d, *stag,
		BUF(rb), *rc, GETHDL(*rt), *s, *rtag, GETHDL(*c), &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, sta);
		sta[6] = *rt;
	}
}

void
mpi_wait_(rq, st, ierr)

int			*rq, *st, *ierr;

{
	MPI_Request	req;
	MPI_Status	stat;

	req = GETHDL(*rq);
	*ierr = MPI_Wait(&req, &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, st);
		st[6] = -1;
		if (req == MPI_REQUEST_NULL) {
			typefree(*rq);
			*rq = -1;
		}
	}
}

void
mpi_test_(rq, f, st, ierr)

int			*rq, *f, *st, *ierr;

{
	MPI_Request	req;
	MPI_Status	stat;

	req = GETHDL(*rq);
	*ierr = MPI_Test(&req, f, &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, st);
		st[6] = -1;
		if (req == MPI_REQUEST_NULL) {
			typefree(*rq);
			*rq = -1;
		}
	}
}

void
mpi_waitany_(c, rqs, idx, st, ierr)

int			*c, *rqs, *idx, *st, *ierr;

{
	MPI_Request	*reqs;
	MPI_Status	stat;
	int		i;

	reqs = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (reqs == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) reqs[i] = GETHDL(rqs[i]);

	*ierr = MPI_Waitany(*c, reqs, idx, &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, st);
		st[6] = -1;
		if (reqs[*idx] == MPI_REQUEST_NULL) {
			typefree(rqs[*idx]);
			rqs[*idx] = -1;
		}

		++(*idx);
	}

	free((char *) reqs);
}

void
mpi_testany_(c, rqs, idx, f, st, ierr)

int			*c, *rqs, *idx, *f, *st, *ierr;

{
	MPI_Request	*reqs;
	MPI_Status	stat;
	int		i;

	reqs = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (reqs == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) reqs[i] = GETHDL(rqs[i]);

	*ierr = MPI_Testany(*c, reqs, idx, f, &stat);
	if ((*ierr == MPI_SUCCESS) && (*f != 0)) {
		c2fstat(&stat, st);
		st[6] = -1;
		if (reqs[*idx] == MPI_REQUEST_NULL) {
			typefree(rqs[*idx]);
			rqs[*idx] = -1;
		}

		++(*idx);
	}

	free((char *) reqs);
}

void
mpi_waitall_(c, rqs, sts, ierr)

int			*c, *rqs, *sts, *ierr;

{
	MPI_Request	*reqs;
	MPI_Status	*stats;
	int		i;

	reqs = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (reqs == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	stats = (MPI_Status *) malloc((unsigned) (*c) * sizeof(MPI_Status));
	if (stats == 0) {
		free((char *) reqs);
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) reqs[i] = GETHDL(rqs[i]);

	*ierr = MPI_Waitall(*c, reqs, stats);
	if (*ierr == MPI_SUCCESS) {
		for (i = 0; i < *c; ++i) {
			c2fstat(stats + i, sts);
			sts[6] = -1;
			sts += 7;
			if (reqs[i] == MPI_REQUEST_NULL) {
				typefree(rqs[i]);
				rqs[i] = -1;
			}
		}
	}

	free((char *) reqs);
	free((char *) stats);
}

void
mpi_testall_(c, rqs, f, sts, ierr)

int			*c, *rqs, *f, *sts, *ierr;

{
	MPI_Request	*reqs;
	MPI_Status	*stats;
	int		i;

	reqs = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (reqs == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	stats = (MPI_Status *) malloc((unsigned) (*c) * sizeof(MPI_Status));
	if (stats == 0) {
		free((char *) reqs);
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) reqs[i] = GETHDL(rqs[i]);

	*ierr = MPI_Testall(*c, reqs, f, stats);
	if ((*ierr == MPI_SUCCESS) && (*f != 0)) {
		for (i = 0; i < *c; ++i) {
			c2fstat(stats + i, sts);
			sts[6] = -1;
			sts += 7;
			if (reqs[i] == MPI_REQUEST_NULL) {
				typefree(rqs[i]);
				rqs[i] = -1;
			}
		}
	}

	free((char *) reqs);
	free((char *) stats);
}

void
mpi_waitsome_(c, rqs, oc, idx, st, ierr)

int			*c, *rqs, *oc, *idx, *st, *ierr;

{
	MPI_Request	*reqs;
	MPI_Status	*stats;
	int		i;

	reqs = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (reqs == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	stats = (MPI_Status *) malloc((unsigned) (*c) * sizeof(MPI_Status));
	if (stats == 0) {
		free((char *) reqs);
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) reqs[i] = GETHDL(rqs[i]);

	*ierr = MPI_Waitsome(*c, reqs, oc, idx, stats);
	if (*ierr == MPI_SUCCESS) {
		for (i = 0; i < *c; ++i) {
			if (reqs[i] == MPI_REQUEST_NULL) {
				typefree(rqs[i]);
				rqs[i] = -1;
			}
		}

		for (i = 0; i < *oc; ++i) {
			idx[i] += 1;
			c2fstat(stats + i, st);
			st[6] = -1;
			st += 7;
		}
	}

	free((char *) reqs);
	free((char *) stats);
}

void
mpi_testsome_(c, rqs, oc, idx, st, ierr)

int			*c, *rqs, *oc, *idx, *st, *ierr;

{
	MPI_Request	*reqs;
	MPI_Status	*stats;
	int		i;

	reqs = (MPI_Request *) malloc((unsigned) (*c) * sizeof(MPI_Request));
	if (reqs == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	stats = (MPI_Status *) malloc((unsigned) (*c) * sizeof(MPI_Status));
	if (stats == 0) {
		free((char *) reqs);
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) reqs[i] = GETHDL(rqs[i]);

	*ierr = MPI_Testsome(*c, reqs, oc, idx, stats);
	if ((*ierr == MPI_SUCCESS) && (*oc > 0)) {
		for (i = 0; i < *c; ++i) {
			if (reqs[i] == MPI_REQUEST_NULL) {
				typefree(rqs[i]);
				rqs[i] = -1;
			}
		}

		for (i = 0; i < *oc; ++i) {
			idx[i] += 1;
			c2fstat(stats + i, st);
			st[6] = -1;
			st += 7;
		}
	}

	free((char *) reqs);
	free((char *) stats);
}

void
mpi_isend_(b, n, t, d, tag, c, rq, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *rq, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Isend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c), &req);
	maketype(rq, ierr, (void *) req);
}

void
mpi_ibsend_(b, n, t, d, tag, c, rq, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *rq, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Ibsend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c), &req);
	maketype(rq, ierr, (void *) req);
}

void
mpi_issend_(b, n, t, d, tag, c, rq, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *rq, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Issend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c), &req);
	maketype(rq, ierr, (void *) req);
}

void
mpi_irsend_(b, n, t, d, tag, c, rq, ierr)

char			*b;
int			*n, *t, *d, *tag, *c, *rq, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Irsend(BUF(b), *n, GETHDL(*t), *d, *tag, GETHDL(*c), &req);
	maketype(rq, ierr, (void *) req);
}

void
mpi_irecv_(b, n, t, s, tag, c, rq, ierr)

char			*b;
int			*n, *t, *s, *tag, *c, *rq, *ierr;

{
	MPI_Request	req;

	*ierr = MPI_Irecv(BUF(b), *n, GETHDL(*t), *s, *tag, GETHDL(*c), &req);
	maketype(rq, ierr, (void *) req);
}

void
mpi_sendrecv_replace_(b, n, t, d, stag, s, rtag, c, st, ierr)

char			*b;
int			*n, *t, *d, *stag, *s, *rtag, *c, *st, *ierr;

{
	MPI_Status	stat;

	*ierr = MPI_Sendrecv_replace(BUF(b), *n, GETHDL(*t), *d, *stag,
					*s, *rtag, GETHDL(*c), &stat);
	if (*ierr == MPI_SUCCESS) {
		c2fstat(&stat, st);
		st[6] = *t;
	}
}

void
mpi_type_contiguous_(c, ot, nt, ierr)

int			*c, *ot, *nt, *ierr;

{
	MPI_Datatype	dtype;

	*ierr = MPI_Type_contiguous(*c, GETHDL(*ot), &dtype);
	maketype(nt, ierr, (void *) dtype);
}

void
mpi_type_vector_(c, b, s, ot, nt, ierr)

int			*c, *b, *s, *ot, *nt, *ierr;

{
	MPI_Datatype	dtype;

	*ierr = MPI_Type_vector(*c, *b, *s, GETHDL(*ot), &dtype);
	maketype(nt, ierr, (void *) dtype);
}

void
mpi_type_hvector_(c, b, s, ot, nt, ierr)

int			*c, *b, *s, *ot, *nt, *ierr;

{
	MPI_Datatype	dtype;

	*ierr = MPI_Type_hvector(*c, *b, (MPI_Aint) *s, GETHDL(*ot), &dtype);
	maketype(nt, ierr, (void *) dtype);
}

void
mpi_type_indexed_(c, b, d, ot, nt, ierr)

int			*c, *b, *d, *ot, *nt, *ierr;

{
	MPI_Datatype	dtype;

	*ierr = MPI_Type_indexed(*c, b, d, GETHDL(*ot), &dtype);
	maketype(nt, ierr, (void *) dtype);
}

void
mpi_type_hindexed_(c, b, d, ot, nt, ierr)

int			*c, *b, *d, *ot, *nt, *ierr;

{
	MPI_Datatype	dtype;
	MPI_Aint	*dp;
	int		i;

	dp = (MPI_Aint *) malloc((unsigned) (*c) * sizeof(MPI_Aint));
	if (dp == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) dp[i] = d[i];

	*ierr = MPI_Type_hindexed(*c, b, dp, GETHDL(*ot), &dtype);
	maketype(nt, ierr, (void *) dtype);

	free((char *) dp);
}

void
mpi_type_struct_(c, b, d, t, nt, ierr)

int			*c, *b, *d, *t, *nt, *ierr;

{
	MPI_Datatype	dtype;
	MPI_Datatype	*tp;
	MPI_Aint	*dp;
	int		i;

	tp = (MPI_Datatype *) malloc((unsigned) (*c) * sizeof(MPI_Datatype));
	if (tp == 0) {
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	dp = (MPI_Aint *) malloc((unsigned) (*c) * sizeof(MPI_Aint));
	if (dp == 0) {
		free((char *) tp);
		*ierr = __lam_mkerr_(MPI_ERR_OTHER, errno);
		return;
	}

	for (i = 0; i < *c; ++i) {
		tp[i] = GETHDL(t[i]);
		dp[i] = d[i];
	}

	*ierr = MPI_Type_struct(*c, b, dp, tp, &dtype);
	maketype(nt, ierr, (void *) dtype);

	free((char *) tp);
	free((char *) dp);
}

/*
 * LAM/MPI extensions
 */

void
mpil_spawn_(c, a, r, ic, ierr)

int			*c, *r, *ic, *ierr;
char			*a;

{
	MPI_Comm	comm;

	*ierr = MPIL_Spawn(GETHDL(*c), a, *r, &comm);
	maketype(ic, ierr, (void *) comm);
}

void
mpil_comm_parent_(c, ierr)

int			*c, *ierr;

{
	*c = type_comm_parent;
	*ierr = MPI_SUCCESS;
}

/*
 *	MPI_F2C_comm
 *
 *	Function:	- converts Fortran communicator handle to C handle
 *	Accepts:	- Fortran communicator handle
 *			- C communicator handle (output)
 *	Returns:	- MPI_SUCCESS
 */
int
MPI_F2C_comm(comm_fort, pcomm_c)

int			comm_fort;
MPI_Comm		*pcomm_c;

{
	*pcomm_c = (MPI_Comm) GETHDL(comm_fort);
	return(MPI_SUCCESS);
}

/*
 *	MPI_F2C_type
 *
 *	Accepts:	- Fortran datatype handle
 *			- C datatype handle (output)
 *	Returns:	- MPI_SUCCESS
 */
int
MPI_F2C_type(type_fort, ptype_c)

int			type_fort;
MPI_Datatype		*ptype_c;

{
	*ptype_c = (MPI_Datatype) GETHDL(type_fort);
	return(MPI_SUCCESS);
}

/*
 *	typealloc
 *
 *	Function:	- allocate entry in type lookup table
 *	Returns:	- table index or LAMERROR
 */
static int
typealloc()

{
	int		i;
	unsigned	newsize;
	void		**p;
/*
 * If there is free space at the top of the table, allocate it.
 */
	if (maxtype < (typesize - 1)) {
		++ntypes;
		return(++maxtype);
	}
/*
 * If there is room in the table, search for an empty entry.
 */
	if (ntypes < typesize) {
		for (i = 0, p = types; i < typesize; ++i, ++p) {
			if (*p == 0) {
				++ntypes;
				return(i);
			}
		}

		return(LAMERROR);
	}
/*
 * Double the table size and allocate entry from its top.
 */
	newsize = 2 * typesize;
	p = (void **) realloc((char *) types,
				(unsigned) newsize * sizeof(void *));
	if (p == 0) return(LAMERROR);

	memset((char *) (p + typesize), 0, typesize * sizeof(void *));

	types = p;
	typesize = newsize;

	++ntypes;
	return(++maxtype);
}

/*
 *	typefind
 *
 *	Function:	- locate given entry
 *	Accepts:	- pointer (entry)
 *	Returns:	- table index or LAMERROR
 */
static int
typefind(pointer)

void			*pointer;

{
	int		i;
	int		rem;
	void		**p;

	rem = ntypes;
	for (i = 0, p = types; (i < typesize) && (rem > 0); ++i, ++p) {
		if (*p) {
			--rem;
			if (*p == pointer) return(i);
		}
	}

	return(LAMERROR);
}

/*
 *	typefree
 *
 *	Function:	- free entry in type lookup table
 *	Accepts:	- table index
 */
static void
typefree(index)

int			index;

{
	void		**p;

	if (index < 0) return;

	p = types + index;
	*p = 0;
	--ntypes;

	if (index == maxtype) {
		for (--maxtype, --p; maxtype >= 0; --maxtype, --p) {
			if (*p) break;
		}
	}
}

/*
 *	maketype
 *
 *	Function:	- create a type mapping
 *			- set error code appropriately
 *	Accepts:	- ptr new type
 *			- ptr error code
 *			- new type handle
 */
static void
maketype(newtype, errcode, handle)

int			*newtype;
int			*errcode;
void			*handle;

{
	if ((*errcode == MPI_SUCCESS) && (handle != 0)) {
		if ((*newtype = typealloc()) < 0) {
			*errcode = __lam_mkerr_(MPI_ERR_OTHER, errno);
			*newtype = -1;
		} else {
			types[*newtype] = handle;
		}
	} else {
		*newtype = -1;
	}
}

/*
 *	c2fstat
 *
 *	Function:	- copy MPI_Status from C to F77
 *			- datatype entry not touched
 *	Accepts:	- ptr C status
 *			- ptr F77 status
 */
static void
c2fstat(cstat, f77stat)

MPI_Status		*cstat;
int			*f77stat;

{
	f77stat[0] = cstat->MPI_SOURCE;
	f77stat[1] = cstat->MPI_TAG;
	f77stat[2] = cstat->MPI_ERROR;
	f77stat[3] = cstat->st_count;
	f77stat[4] = cstat->st_nelem;
	f77stat[5] = cstat->st_length;
}

/*
 *	f2cstat
 *
 *	Function:	- copy MPI_Status from F77 to C
 *	Accepts:	- ptr F77 status
 *			- ptr C status
 */
static void
f2cstat(f77stat, cstat)

int			*f77stat;
MPI_Status		*cstat;

{
	cstat->MPI_SOURCE = f77stat[0];
	cstat->MPI_TAG = f77stat[1];
	cstat->MPI_ERROR = f77stat[2];
	cstat->st_count = f77stat[3];
	cstat->st_nelem = f77stat[4];
	cstat->st_length = f77stat[5];
	cstat->st_dtype = (f77stat[6] < 0) ? MPI_DATATYPE_NULL :
				(MPI_Datatype) GETHDL(f77stat[6]);
}
