/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	abort.c,v $
 * Revision 6.0  96/02/29  14:11:50  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:17:30  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  95/06/09  15:36:16  raja
 * Use errno when exiting.
 * 
 * Revision 5.2  94/08/22  14:03:42  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:08  gdburns
 * Ohio Release
 * 
 *	Function:	- attempts to abort all processes in group
 *			- prints error message
 *	Accepts:	- communicator
 *			- error code
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <ksignal.h>
#include <mpi.h>
#include <net.h>
#include <preq.h>

int
MPI_Abort(comm, errcode)

MPI_Comm		comm;
int			errcode;

{
	int		size;			/* group size */
	int		rank;			/* process rank */
	int		err;			/* error code */
	int		i;			/* favourite index */
	struct _proc	**p;			/* favourite pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIABORT);
/*
 * Switch error code to extracted errno value.
 */
	__lam_bkerr_(errcode, &i, &i, &err);
	errcode = err;

	if (comm == MPI_COMM_NULL) comm = MPI_COMM_WORLD;

	err = MPI_Comm_size(comm, &size);
	if (err != MPI_SUCCESS) kexit(errcode);

	err = MPI_Comm_rank(comm, &rank);
	if (err != MPI_SUCCESS) kexit(errcode);
/*
 * Doom all other processes in the group if we know their pids.
 */
	if ((_kio.ki_rtf & RTF_MPIRUN) || (_kio.ki_parent > 0)) {
		p = comm->c_group->g_procs;

		for (i = 0; i < size; ++i, ++p) {

			if (i != rank) {
				rpdoom((*p)->p_gps.gps_node, SELECT_PID,
						(*p)->p_gps.gps_pid, SIGUDIE);
			}
		}
	}
/*
 * Abort the local process if in the group.
 */
	if (rank >= 0) {
/*
 *		MPI_Finalize();
 */
		kexit(errcode);
	}

	__lam_resetfunc_(BLKMPIABORT);
	return(MPI_SUCCESS);
}
