/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	JRV/RBD
 *
 *	$Log:	allgather.c,v $
 * Revision 6.0  96/02/29  14:11:51  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:17:40  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:40:41  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:03:43  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:10  gdburns
 * Ohio Release
 * 
 *	Function:	- gather buffers at all process ranks
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- recv buffer
 *			- recv count
 *			- recv datatype
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Allgather(sbuf, scount, sdtype, rbuf, rcount, rdtype, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
MPI_Comm		comm;

{
	int		err;			/* error code */
	int		size;			/* group size */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIALLGATHER);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHER,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((sdtype == MPI_DATATYPE_NULL) || (rdtype == MPI_DATATYPE_NULL)) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHER,
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if ((scount < 0) || (rcount < 0)) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHER,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIALLGATHER);
/*
 * Gather and broadcast.
 */
	MPI_Comm_size(comm, &size);

	err = MPI_Gather(sbuf, scount, sdtype, rbuf, rcount, rdtype, 0, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHER, err));
	}

	err = MPI_Bcast(rbuf, rcount * size, rdtype, 0, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHER, err));
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIALLGATHER, -1, comm, sdtype, scount);

	__lam_resetfunc_(BLKMPIALLGATHER);
	return(MPI_SUCCESS);
}
