/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	JRV/RBD
 *
 *	$Log:	allgatherv.c,v $
 * Revision 6.0  96/02/29  14:11:52  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:17:46  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:05  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:03:44  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:10  gdburns
 * Ohio Release
 * 
 *	Function:	- gather varying length buffers at all process ranks
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- recv buffer
 *			- recv counts
 *			- displacements
 *			- recv datatype
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Allgatherv(sbuf, scount, sdtype,
			rbuf, rcounts, disps, rdtype, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			*rcounts;
int			*disps;
MPI_Datatype		rdtype;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		size;			/* group size */
	int		err;			/* error code */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIALLGATHERV);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHERV,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((sdtype == MPI_DATATYPE_NULL) || (rdtype == MPI_DATATYPE_NULL)) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHERV,
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (disps == 0) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHERV,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((scount < 0) || (rcounts == 0)) {
		return(__lam_errfunc_(comm, BLKMPIALLGATHERV,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIALLGATHERV);
/*
 * Collect all values at each process, one at a time.
 */
	MPI_Comm_size(comm, &size);

	for (i = 0; i < size; ++i) {

		err = MPI_Gatherv(sbuf, scount, sdtype, rbuf,
					rcounts, disps, rdtype, i, comm);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIALLGATHERV, err));
		}
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIALLGATHERV, -1, comm, sdtype, scount);

	__lam_resetfunc_(BLKMPIALLGATHERV);
	return(MPI_SUCCESS);
}
