/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	JRV/RBD
 *
 *	$Log:	allreduce.c,v $
 * Revision 6.0  96/02/29  14:11:53  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:17:48  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:08  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:03:45  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:11  gdburns
 * Ohio Release
 * 
 *	Function:	- reduce and return data to all processes
 *	Accepts:	- send buffer
 *			- receive buffer
 *			- count of elements
 *			- type of elements
 *			- operation to perform
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Allreduce(sbuf, rbuf, count, dtype, op, comm)

void			*sbuf;
void			*rbuf;
int			count;
MPI_Datatype		dtype;
MPI_Op			op;
MPI_Comm		comm;

{
	int		err;			/* error code */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIALLREDUCE);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return (__lam_errfunc_(comm, BLKMPIALLREDUCE,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(comm, BLKMPIALLREDUCE,
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIALLREDUCE);
/*
 * Reduce to 0 and broadcast.
 */
	err = MPI_Reduce(sbuf, rbuf, count, dtype, op, 0, comm);
	if (err != MPI_SUCCESS) {
		return (__lam_errfunc_(comm, BLKMPIALLREDUCE, err));
	}

	err = MPI_Bcast(rbuf, count, dtype, 0, comm);
	if (err != MPI_SUCCESS) {
		return (__lam_errfunc_(comm, BLKMPIALLREDUCE, err));
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIALLREDUCE, -1, comm, dtype, count);

	__lam_resetfunc_(BLKMPIALLREDUCE);
	return(MPI_SUCCESS);
}
