/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	attrdel.c,v $
 * Revision 6.0  96/03/03  22:53:20  nevin
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/03/03  22:51:43  nevin
 * Correct FORTRAN attribute deletion
 * 
 * Revision 5.2.1.2  96/02/18  21:17:55  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/26  14:19:49  raja
 * Pass values by ref. for F77 delete function.
 * 
 * Revision 5.2  94/08/22  14:03:47  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:14  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:47:59  gdburns
 * new for 5.1
 * 
 *	Function:	- delete attribute
 *			- detach attribute from communicator
 *	Accepts:	- communicator
 *			- attribute key
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern struct _attrkey	*__lam_getattr_();
extern struct _attr	*__lam_getkey_();

int
MPI_Attr_delete(comm, key)

MPI_Comm		comm;
int			key;

{
	struct _attrkey	*p;
	struct _attr	*pk;
	int		err;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIATTRDEL);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIATTRDEL, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	p = __lam_getattr_(key);
	if (p == 0) {
		return(__lam_errfunc_(comm,
			BLKMPIATTRDEL, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Check if the key is attached to the communicator.
 */
	pk = __lam_getkey_(comm, key);
	if (pk == 0) {
		return(__lam_errfunc_(comm, BLKMPIATTRDEL,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Delete the attribute via the callback function.
 */
	if (p->ak_del != MPI_NULL_DELETE_FN) {
		if (p->ak_f77comm < 0) {
			err = (*(p->ak_del))(comm, key,
					pk->a_value, p->ak_extra);
		} else {
			err = (*(p->ak_del))(&p->ak_f77comm, &key,
                                	(int *) &(pk->a_value),
                                	(int *) &(p->ak_extra));
		}

		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm,
				BLKMPIATTRDEL, __lam_mkerr_(err, 0)));
		}
	}
/*
 * Detach the key from the communicator.
 */
	if (__lam_delkey_(comm, key)) {
		return(__lam_errfunc_(comm, BLKMPIATTRDEL,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	__lam_resetfunc_(BLKMPIATTRDEL);
	return(MPI_SUCCESS);
}
