/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	attrput.c,v $
 * Revision 6.0  96/02/29  14:11:57  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:18:00  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:48  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:15  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:00  gdburns
 * new for 5.1
 * 
 *	Function:	- put value in attribute
 *			- assign attribute to communicator
 *	Accepts:	- communicator
 *			- attribute key
 *			- attribute value
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern struct _attrkey	*__lam_getattr_();
extern struct _attr	*__lam_getkey_();

int
MPI_Attr_put(comm, key, value)

MPI_Comm		comm;
int			key;
void			*value;

{
	struct _attrkey	*p;
	struct _attr	*pk;
	int		errcode;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIATTRPUT);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIATTRPUT, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	p = __lam_getattr_(key);
	if (p == 0) {
		return(__lam_errfunc_(comm,
			BLKMPIATTRPUT, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * If the key is not attached to the communicator, attach and fill it.
 */
	pk = (struct _attr *) __lam_getkey_(comm, key);
	if (pk == 0) {
		if (__lam_putkey_(comm, key)) {
			return(__lam_errfunc_(comm, BLKMPIATTRPUT,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}

		pk = (struct _attr *) __lam_getkey_(comm, key);
	}
/*
 * Old attribute, delete previous value and fill new one.
 */
	else if (p->ak_del) {
		if (p->ak_f77comm < 0) {
			errcode = (*(p->ak_del))(&comm, &key,
					pk->a_value, p->ak_extra);
		} else {
			errcode = (*(p->ak_del))(&(p->ak_f77comm),
					&key, pk->a_value, p->ak_extra);
		}

		if (errcode != MPI_SUCCESS) {
			return(__lam_errfunc_(comm,
				BLKMPIATTRPUT, __lam_mkerr_(errcode, 0)));
		}
	}

	pk->a_value = value;

	__lam_resetfunc_(BLKMPIATTRPUT);
	return(MPI_SUCCESS);
}
