/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	bsend.c,v $
 * Revision 6.0.1.1  96/05/10  15:48:59  nevin
 * Handle MPI_PROC_NULL
 * 
 * Revision 6.0  96/02/29  14:12:01  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:18:28  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:50  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:18  gdburns
 * Ohio Release
 * 
 *	Function:	- buffered send
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- destination
 *			- tag
 *			- communicator
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern int		__lam_errfunc_();
extern int		__lam_isend_();
extern int		__lam_trincff_();

int
MPI_Bsend(buf, count, dtype, dest, tag, comm)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			dest;
int			tag;
MPI_Comm		comm;

{
	int		err;			/* error code */
	MPI_Request	req;			/* request */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */

	__lam_initerr_();
/*
 * Handle special case of peer MPI_PROC_NULL.
 */
        if (dest == MPI_PROC_NULL) return(MPI_SUCCESS);

	__lam_setfunc_(BLKMPIBSEND);
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) startt = MPI_Wtime();
/*
 * Request will be orphaned, no need to bother with it here.
 */
	req = MPI_REQUEST_NULL;
	err = __lam_isend_(buf, count, dtype,
			dest, tag, comm, &req, LAM_RQIBSEND);

	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIBSEND, err));
/*
 * Generate a run time trace.
 */
	if (fl_trace) {
		finisht = MPI_Wtime();

		__lam_trmsg_(TRTOUTPUT, startt, LAM_S2US(finisht - startt), 0,
				dest, tag, comm, dtype, count, 0, 0, 
				req->rq_seq, LAM_RQIBSEND);
	}
	
	__lam_resetfunc_(BLKMPIBSEND);
	return(MPI_SUCCESS);
}
