/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	c2cbuf.c,v $
 * Revision 6.0  96/02/29  14:12:04  gdburns
 * Ohio Release
 * 
 *	Function:	- client-to-client message buffering
 */

#include <sys/time.h>
#include <sys/types.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include <all_hash.h>
#include <mpi.h>
#include <portable.h>
#include <rpi.c2c.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * global functions
 */
int			_cbuf_init();
void			_cbuf_end();
struct cbuf_msg *	_cbuf_find();
void			_cbuf_delete();
int			_cbuf_append();

/*
 * external functions
 */
extern int4		next_prime();

/*
 * local functions
 */
static MPI_Comm		cbuf_cid2comm();
static void		cbuf_incgerctr();

/*
 * external variables
 */
extern fd_set		disabledsock;		/* sockets disabled by GER */

/*
 * local variables
 */
static HASH		*buftbl = 0;		/* buffer hash table */

/*
 *	_cbuf_init
 *
 *	Function:	- initialize C2C buffers
 *	Returns:	- 0 or LAMERROR
 */
int
_cbuf_init()

{
	_cbuf_end();

	buftbl = ah_init(INITHASH, sizeof(struct cbuf_cid), NULLCID, 0);
	if (buftbl == 0) return(LAMERROR);

	return(0);
}

/*
 *	_cbuf_end
 *
 *	Function:	- destroy C2C buffers
 */
void
_cbuf_end()

{
	struct cbuf_cid	*p;			/* favourite pointer */
	
	if (buftbl == 0) return;
/*
 * Loop over all context IDs.
 */
	for (p = (struct cbuf_cid *) ah_top(buftbl); p;
			p = (struct cbuf_cid *) ah_next(buftbl, p)) {

		if (p->cbc_envs) {
			al_free(p->cbc_envs);
		}
	}

	ah_free(buftbl);
}

/*
 *	_cbuf_find
 *
 *	Function:	- find a matching message
 *	Accepts:	- ptr request envlope
 *	Returns:	- ptr message or NULL
 */
struct cbuf_msg *
_cbuf_find(rqenv)

struct c2c_envl	*rqenv;

{
	struct cbuf_cid	*pcid;			/* ptr CID entry */
	struct cbuf_msg	*pmsg = 0;		/* ptr message */

	pcid = (struct cbuf_cid *) ah_find(buftbl, rqenv->cce_cid);
	if (pcid && pcid->cbc_envs) {
/*
 * Loop over list of messages finding first match.
 */
		pmsg = al_top(pcid->cbc_envs);

		while (pmsg) {
		
			if (! _c2c_envl_cmp(&(pmsg->cbm_env), rqenv)) {
				return(pmsg);
			}

			pmsg = al_next(pcid->cbc_envs, pmsg);
		}
	}
	
	return(0);
}

/*
 *	_cbuf_delete
 *
 *	Function:	- remove a message
 *			- context ID entries are never deleted
 *	Accepts:	- message
 */
void
_cbuf_delete(msg)

struct cbuf_msg		*msg;

{
	struct cbuf_cid	*pcid;			/* ptr CID entry */

	pcid = (struct cbuf_cid *) ah_find(buftbl, msg->cbm_env.cce_cid);
	if (pcid == 0 || pcid->cbc_envs == 0) return;

	if (__lam_ger_) {
		cbuf_incgerctr(pcid->cbc_comm, msg->cbm_env.cce_rank, -1);
	}

	al_delete(pcid->cbc_envs, msg);
}

/*
 *	_cbuf_append
 *
 *	Function:	- add new message to end of list
 *			- expand hash table if needed
 *	Accepts:	- message
 *	Returns:	- 0 or LAMERROR
 */
int
_cbuf_append(msg)

struct cbuf_msg		*msg;

{
	struct cbuf_cid	*pcid;			/* ptr CID entry */
	struct cbuf_cid	cident;			/* CID entry */
	int		newsize;		/* expanded table size */

	pcid = (struct cbuf_cid *) ah_find(buftbl, msg->cbm_env.cce_cid);
	if (pcid == 0) {
/*
 * Expand hash table if no more free entries.
 */
/* NOTE: if Greg wrote ah_put_and_expand_if_needed(), replace it here. */
/* NOTE: if so, make it call next_prime() and remove its declaration here */

		if (ah_count(buftbl) == (newsize = ah_size(buftbl))) {
			newsize = next_prime(newsize + newsize);
			if (ah_expand(buftbl, newsize)) return(LAMERROR);
		}
/*
 * Create new CID entry.
 */
		cident.cbc_cid = msg->cbm_env.cce_cid;
		cident.cbc_envs = 0;
		if (__lam_ger_) {
			cident.cbc_comm = cbuf_cid2comm(cident.cbc_cid);
if (cident.cbc_comm == MPI_COMM_NULL) kpause();
		}
		if (ah_insert(buftbl, (void *) &cident)) return(LAMERROR);

		pcid = (struct cbuf_cid *)
			ah_find(buftbl, msg->cbm_env.cce_cid);
		
		if (pcid == 0) {
			errno = EIMPOSSIBLE;
			return(LAMERROR);
		}
/* NOTE: if Greg changed ah_insert() to return new pointer, 'dd' here */
/* NOTE: that was a design short-sightedness on my part */
	}
/*
 * Create message list if necessary.
 */
	if (! pcid->cbc_envs) {

		pcid->cbc_envs = al_init(sizeof(struct cbuf_msg), 0);
		if (! pcid->cbc_envs) return(LAMERROR);
	}
/*
 * Link message to end of list.
 */
	if (__lam_ger_) {
		cbuf_incgerctr(pcid->cbc_comm, msg->cbm_env.cce_rank, 1);
	}

	return( (al_append(pcid->cbc_envs, msg)) ? 0 : LAMERROR );
}

/*
 *	cbuf_cid2comm
 *
 *	Function:	- finds communicator from context ID
 *	Accepts:	- context ID
 *	Returns:	- communicator
 */
static MPI_Comm
cbuf_cid2comm(cid)

int			cid;

{
	struct _comm	celem;
	MPI_Comm	comm;
	MPI_Comm	*pcomm;

	celem.c_contextid = cid;
	comm = &celem;
	
	pcomm = al_find(__lam_comms_, &comm);
	
	return((pcomm) ? *pcomm : MPI_COMM_NULL);
}

/*
 *	cbuf_incgerctr
 *
 *	Function:	- increment GER counter
 *			- enable/disable socket according to counter
 *	Accepts:	- context ID
 *			- rank of process in communicator
 *			- increment
 */
static void
cbuf_incgerctr(comm, rank, inc)

MPI_Comm		comm;
int			rank;
int			inc;

{
	struct _proc	*proc;			/* process */
	struct c2c_proc	*c2cp;			/* C2C specific process data */

	if (comm->c_inter) {
		proc = comm->c_rgroup->g_procs[rank];
	} else {
		proc = comm->c_group->g_procs[rank];
	}
	
	c2cp = (struct c2c_proc *) proc->p_rpi;

	c2cp->ccp_nbfde += inc;
	
	if (inc > 0 && c2cp->ccp_nbfde >= MPI_GER) {
		FD_SET(c2cp->ccp_sock, &disabledsock);
	}
	else if (inc < 0 && c2cp->ccp_nbfde < MPI_GER) {
		FD_CLR(c2cp->ccp_sock, &disabledsock);
	}
}
