/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	cancel.c,v $
 * Revision 6.0  96/02/29  14:12:04  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:18:41  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:53  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:21  gdburns
 * Ohio Release
 * 
 *	Function:	- marks immediate request for cancellation
 *	Accepts:	- ptr request
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Cancel(preq)

MPI_Request		*preq;

{
	MPI_Request	req;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICANCEL);
/*
 * Check the arguments.
 */
	if (preq == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICANCEL, __lam_mkerr_(MPI_ERR_REQUEST, 0)));
	}

	if ((req = *preq) == MPI_REQUEST_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICANCEL, __lam_mkerr_(MPI_ERR_REQUEST, 0)));
	}

	if (req->rq_state == LAM_RQSINIT) {
		return(__lam_errfunc_(req->rq_comm,
			BLKMPICANCEL, __lam_mkerr_(MPI_ERR_REQUEST, 0)));
	}
/*
 * In the start state, nothing has happened, so cancel is easy.
 * Move the request to the done state.
 */
	if (req->rq_state == LAM_RQSSTART) {
		req->rq_flags |= LAM_RQFCANCEL;
		req->rq_status.MPI_SOURCE = MPI_CANCEL_SOURCE;
		req->rq_state = LAM_RQSDONE;
	}
/*
 * In the active state, it may still be possible to cancel a send operation.
 * This is an immense pain for many C2C implementations and the same
 * resource cleanup goal can be achieved by having the receiver suck
 * in the message.  Cancelling an active send is no problem for lamd.
 * We just make a remote service call to bufferd and ask that the MPI
 * sequence number from this source be nuked.  We set the cancel flag
 * based on the return code.  I could add this in half an hour.  Screw it.
 */
	else if ((req->rq_state == LAM_RQSACTIVE) &&
			(req->rq_marks & LAM_RQFDEST)) {
	}
/*
 * In the done state, the cancel fails.
 */
	else {
	}

	__lam_resetfunc_(BLKMPICANCEL);
	return(MPI_SUCCESS);
}
