/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartcoords.c,v $
 * Revision 6.0  96/02/29  14:12:05  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:18:50  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:53  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:21  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:03  gdburns
 * new for 5.1
 * 
 *	Function:	- translate rank to coordinates
 *	Accepts:	- communicator
 *			- rank
 *			- max # dimensions to fill
 *			- coordinates array (returned value)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Cart_coords(comm, rank, maxdims, coords)

MPI_Comm		comm;
int			rank;
int			maxdims;
int			*coords;

{
	int		dim;
	int		remprocs;
	int		i;
	int		*d;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTCOORDS);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTCOORDS, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTCOORDS, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_CART) {
		return(__lam_errfunc_(comm, BLKMPICARTCOORDS,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((rank < 0) || (rank >= comm->c_group->g_nprocs)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTCOORDS, __lam_mkerr_(MPI_ERR_RANK, 0)));
	}

	if ((maxdims < 0) || ((maxdims > 0) && (coords == 0))) {
		return(__lam_errfunc_(comm,
			BLKMPICARTCOORDS, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Loop computing the coordinates.
 */
	d = comm->c_topo_dims;
	remprocs = comm->c_topo_nprocs;
	for (i = 0; (i < comm->c_topo_ndims) && (i < maxdims); ++i, ++d) {
		dim = (*d > 0) ? *d : -(*d);
		remprocs /= dim;
		*coords++ = rank / remprocs;
		rank %= remprocs;
	}

	__lam_resetfunc_(BLKMPICARTCOORDS);
	return(MPI_SUCCESS);
}
