/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartget.c,v $
 * Revision 6.0  96/02/29  14:12:08  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:14  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:55  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:24  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:04  gdburns
 * new for 5.1
 * 
 *	Function:	- get info on cartesian topology
 *	Accepts:	- communicator
 *			- size of dimensions array
 *			- dimensions array
 *			- periodicity array
 *			- coordinates array
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Cart_get(comm, maxndims, dims, periods, coords)

MPI_Comm		comm;
int			maxndims;
int			*dims;
int			*periods;
int			*coords;

{
	int		i;
	int		*d;
	int		*c;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTGET);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTGET, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTGET, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_CART) {
		return(__lam_errfunc_(comm, BLKMPICARTGET,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((maxndims < 0) || ((maxndims > 0) &&
			((dims == 0) || (periods == 0) || (coords == 0)))) {
		return(__lam_errfunc_(comm,
			BLKMPICARTGET, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Fill the dimension, period, and coordinate arrays.
 */
	d = comm->c_topo_dims;
	c = comm->c_topo_coords;
	for (i = 0; (i < comm->c_topo_ndims) && (i < maxndims); ++i) {
		if (*d > 0) {
			*dims++ = *d++;
			*periods++ = 0;
		} else {
			*dims++ = -(*d++);
			*periods++ = 1;
		}

		*coords++ = *c++;
	}

	__lam_resetfunc_(BLKMPICARTGET);
	return(MPI_SUCCESS);
}
