/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartmap.c,v $
 * Revision 6.0  96/02/29  14:12:09  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:15  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:56  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:25  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:04  gdburns
 * new for 5.1
 * 
 *	Function:	- map processes to cartesian coordinates
 *	Accepts:	- communicator
 *			- # dimensions
 *			- # procs per dimension
 *			- dimension periodicity
 *			- ptr caller's new rank
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Cart_map(comm, ndims, dims, periods, pnewrank)

MPI_Comm		comm;
int			ndims;
int			*dims;
int			*periods;
int			*pnewrank;

{
	int		nprocs;
	int		rank;
	int		errcode;
	int		i;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTMAP);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTMAP, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTMAP, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (ndims < 1) {
		return(__lam_errfunc_(comm,
			BLKMPICARTMAP, __lam_mkerr_(MPI_ERR_DIMS, 0)));
	}

	if ((dims == 0) || (periods == 0) || (pnewrank == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTMAP, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Compute the # of processes in the grid.
 */
	nprocs = 1;
	for (i = 0, p = dims; i < ndims; ++i, ++p) {
		if (*p <= 0) {
			return(__lam_errfunc_(comm, BLKMPICARTMAP,
					__lam_mkerr_(MPI_ERR_DIMS, 0)));
		}

		nprocs *= *p;
	}
/*
 * Compute my new rank.
 */
	errcode = MPI_Comm_rank(comm, &rank);
	if (errcode != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPICARTMAP, errcode));
	}

	*pnewrank = ((rank < 0) || (rank >= nprocs)) ? MPI_UNDEFINED : rank;

	__lam_resetfunc_(BLKMPICARTMAP);
	return(MPI_SUCCESS);
}
