/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartrank.c,v $
 * Revision 6.0  96/02/29  14:12:10  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:17  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:57  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:25  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:05  gdburns
 * new for 5.1
 * 
 *	Function:	- translate coordinates to rank
 *	Accepts:	- communicator
 *			- coordinates array
 *			- ptr rank (returned value)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Cart_rank(comm, coords, prank)

MPI_Comm		comm;
int			*coords;
int			*prank;

{
	int		rank;
	int		dim;
	int		ord;
	int		factor;
	int		i;
	int		*d;
	int		*c;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTRANK);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTRANK, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTRANK, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_CART) {
		return(__lam_errfunc_(comm, BLKMPICARTRANK,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((coords == 0) || (prank == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTRANK, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Loop over coordinates computing the rank.
 */
	factor = 1;
	rank = 0;
	i = comm->c_topo_ndims - 1;
	d = comm->c_topo_dims + i;
	c = coords + i;

	for (; i >= 0; --i, --c, --d) {
		dim = (*d > 0) ? *d : -(*d);
		ord = *c;
		if ((ord < 0) || (ord >= dim)) {
			if (*d > 0) {
				return(__lam_errfunc_(comm, BLKMPICARTRANK,
					__lam_mkerr_(MPI_ERR_ARG, 0)));
			}

			ord %= dim;
			if (ord < 0) {
				ord += dim;
			}
		}

		rank += factor * ord;
		factor *= dim;
	}

	*prank = rank;

	__lam_resetfunc_(BLKMPICARTRANK);
	return(MPI_SUCCESS);
}
