/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartshift.c,v $
 * Revision 6.0  96/02/29  14:12:11  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:20:18  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/26  14:00:29  raja
 * Return correct source rank.
 * 
 * Revision 5.2  94/08/22  14:03:57  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:26  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:05  gdburns
 * new for 5.1
 * 
 *	Function:	- get src and dest ranks in cartesian shift
 *	Accepts:	- communicator
 *			- dimension of shift
 *			- displacement
 *			- ptr source rank
 *			- ptr destination rank
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Cart_shift(comm, dim, disp, psrc, pdest)

MPI_Comm		comm;
int			dim;
int			disp;
int			*psrc;
int			*pdest;

{
	int		factor;
	int		thisdim;
	int		thisperiod;
	int		ord;
	int		srcord;
	int		destord;
	int		i;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTSHIFT);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTSHIFT, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTSHIFT, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_CART) {
		return(__lam_errfunc_(comm, BLKMPICARTSHIFT,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((dim < 0) || (dim >= comm->c_topo_ndims)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTSHIFT, __lam_mkerr_(MPI_ERR_DIMS, 0)));
	}

	if ((psrc == 0) || (pdest == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTSHIFT, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial case.
 */
	ord = comm->c_group->g_myrank;
	if (disp == 0) {
		*pdest = *psrc = ord;
		__lam_resetfunc_(BLKMPICARTSHIFT);
		return(MPI_SUCCESS);
	}
/*
 * Compute the rank factor and ordinate.
 */
	factor = comm->c_topo_nprocs;
	p = comm->c_topo_dims;
	for (i = 0; (i < comm->c_topo_ndims) && (i <= dim); ++i, ++p) {
		if ((thisdim = *p) > 0) {
			thisperiod = 0;
		} else {
			thisperiod = 1;
			thisdim = -thisdim;
		}

		ord %= factor;
		factor /= thisdim;
	}

	ord /= factor;
/*
 * Check the displacement value and compute the new ranks.
 */
	*psrc = *pdest = MPI_UNDEFINED;

	srcord = ord - disp;
	destord = ord + disp;

	if ( ((destord < 0) || (destord >= thisdim)) && (!thisperiod) ) {
		*pdest = MPI_PROC_NULL;
	} else {
		destord %= thisdim;
		if (destord < 0) destord += thisdim;

		*pdest = comm->c_group->g_myrank + ((destord - ord) * factor);
	}

	if ( ((srcord < 0) || (srcord >= thisdim)) && (!thisperiod) ) {
		*psrc = MPI_PROC_NULL;
	} else {
		srcord %= thisdim;
		if (srcord < 0) srcord += thisdim;

		*psrc = comm->c_group->g_myrank + ((srcord - ord) * factor);
	}

	__lam_resetfunc_(BLKMPICARTSHIFT);
	return(MPI_SUCCESS);
}
