/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartsub.c,v $
 * Revision 6.0  96/02/29  14:12:12  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:20  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:58  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:27  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:06  gdburns
 * new for 5.1
 * 
 *	Function:	- divide cartesian communicator into sub-grids
 *	Accepts:	- old communicator
 *			- array of kept dimensions
 *			- ptr new communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Cart_sub(comm, remdims, pnewcomm)

MPI_Comm		comm;
int			*remdims;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;
	int		errcode;
	int		colour;
	int		key;
	int		colfactor;
	int		keyfactor;
	int		rank;
	int		ndim;
	int		dim;
	int		i;
	int		*d;
	int		*c;
	int		*r;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTSUB);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTSUB, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTSUB, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((remdims == 0) || (pnewcomm == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTSUB, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPICARTSUB);
/*
 * Compute colour and key used in splitting the communicator.
 */
	colour = key = 0;
	colfactor = keyfactor = 1;
	ndim = 0;

	i = comm->c_topo_ndims - 1;
	d = comm->c_topo_dims + i;
	c = comm->c_topo_coords + i;
	r = remdims + i;

	for (; i >= 0; --i, --d, --c, --r) {
		dim = (*d > 0) ? *d : -(*d);

		if (*r == 0) {
			colour += colfactor * (*c);
			colfactor *= dim;
		} else {
			++ndim;
			key += keyfactor * (*c);
			keyfactor *= dim;
		}
	}
/*
 * Split the communicator.
 */
	errcode = MPI_Comm_split(comm, colour, key, pnewcomm);
	if (errcode != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPICARTSUB, errcode));
	}
/*
 * Fill the communicator with topology information. 
 */
	newcomm = *pnewcomm;
	if (newcomm != MPI_COMM_NULL) {
		newcomm->c_topo_type = MPI_CART;
		newcomm->c_topo_nprocs = keyfactor;
		newcomm->c_topo_ndims = ndim;

		newcomm->c_topo_dims = (int *)
				malloc((unsigned) 2 * ndim * sizeof(int));
		if (newcomm->c_topo_dims == 0) {
			return(__lam_errfunc_(comm, BLKMPICARTSUB,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
		newcomm->c_topo_coords = newcomm->c_topo_dims + ndim;

		p = newcomm->c_topo_dims;
		d = comm->c_topo_dims;
		r = remdims;
		for (i = 0; i < comm->c_topo_ndims; ++i, ++d, ++r) {
			if (*r) {
				*p++ = *d;
			}
		}
/*
 * Compute the caller's coordinates.
 */
		errcode = MPI_Comm_rank(newcomm, &rank);
		if (errcode != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPICARTSUB, errcode));
		}

		errcode = MPI_Cart_coords(newcomm, rank,
					ndim, newcomm->c_topo_coords);
		if (errcode != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPICARTSUB, errcode));
		}
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPICARTSUB, -1, comm, 0, 0);

	__lam_resetfunc_(BLKMPICARTSUB);
	return(MPI_SUCCESS);
}
