/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	ccmp.c,v $
 * Revision 6.0  96/02/29  14:12:12  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:22  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:59  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:28  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:06  gdburns
 * new for 5.1
 * 
 *	Function:	- compare two communicators
 *	Accepts:	- first communicator
 *			- second communicator
 *			- ptr comparison result
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Comm_compare(c1, c2, pres)

MPI_Comm		c1;
MPI_Comm		c2;
int			*pres;

{
	int		gcmp;
	int		errcode;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICOMMCMP);
/*
 * Check the arguments.
 */
	if ((c1 == MPI_COMM_NULL) || (c2 == MPI_COMM_NULL)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICOMMCMP, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (pres == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICOMMCMP, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial cases.
 */
	if (c1 == c2) {
		*pres = MPI_IDENT;
		__lam_resetfunc_(BLKMPICOMMCMP);
		return(MPI_SUCCESS);
	}

	if (c1->c_inter != c2->c_inter) {
		*pres = MPI_UNEQUAL;
		__lam_resetfunc_(BLKMPICOMMCMP);
		return(MPI_SUCCESS);
	}
/*
 * Compare the groups.
 */
	errcode = MPI_Group_compare(c1->c_group, c2->c_group, &gcmp);
	if (errcode) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPICOMMCMP, errcode));
	}

	if ((gcmp == MPI_IDENT) && (c1->c_inter)) {
		errcode = MPI_Group_compare(c1->c_rgroup, c2->c_rgroup, &gcmp);
		if (errcode) {
			return(__lam_errfunc_(MPI_COMM_WORLD,
					BLKMPICOMMCMP, errcode));
		}
	}

	*pres = (gcmp == MPI_IDENT) ? MPI_CONGRUENT: gcmp;

	__lam_resetfunc_(BLKMPICOMMCMP);
	return(MPI_SUCCESS);
}
