/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	ccreate.c,v $
 * Revision 6.0  96/02/29  14:12:13  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:24  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:59  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:29  gdburns
 * Ohio Release
 * 
 *	Function:	- create a new communicator
 *	Accepts:	- old communicator
 *			- new group
 *			- ptr new communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern int		__lam_cubedim_();
extern int		__lam_trcomm_();

int
MPI_Comm_create(comm, group, pnewcomm)

MPI_Comm		comm;
MPI_Group		group;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;		/* new communicator */
	int		mycid;			/* local max context ID */
	int		cid;			/* global max context ID */
	int		rank;			/* process rank */
	int		err;			/* error code */

	__lam_initerr_();
	__lam_setfunc_(BLKMPICOMMCREATE);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (group == MPI_GROUP_NULL) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if (pnewcomm == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Create the new context ID using MPI_Allreduce().
 * Processes not in group participate but do not affect the context ID.
 */
	rank = group->g_myrank;

	mycid = (rank == MPI_UNDEFINED) ? 0 : __lam_getcid_();
	if (mycid < 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_INTERN, EFULL)));
	}
/*
 * Set up tracing.
 */
	_LAM_TRACE(__lam_trcffstart_(BLKMPICOMMCREATE));

	err = MPI_Allreduce(&mycid, &cid, 1, MPI_INT, MPI_MAX, comm);
	if (err != MPI_SUCCESS) {
		_LAM_TRACE(__lam_trcffend_(BLKMPICOMMCREATE, -1, comm, 0, 0));
		__lam_resetfunc_(BLKMPICOMMCREATE);
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE, err));
	}

	if (rank == MPI_UNDEFINED) {
		*pnewcomm = MPI_COMM_NULL;
		_LAM_TRACE(__lam_trcffend_(BLKMPICOMMCREATE, -1, comm, 0, 0));
		__lam_resetfunc_(BLKMPICOMMCREATE);
		return(MPI_SUCCESS);
	}
/*
 * Create the new communicator.
 */
	newcomm = (MPI_Comm) malloc(sizeof(struct _comm));
	if (newcomm == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pnewcomm = newcomm;
	newcomm->c_inter = 0;
	newcomm->c_refcount = 1;
	newcomm->c_cube_dim = __lam_cubedim_(group->g_nprocs);
	newcomm->c_group = group;
	++(group->g_refcount);
	newcomm->c_rgroup = MPI_GROUP_NULL;
	newcomm->c_contextid = cid;
	newcomm->c_errhdl = comm->c_errhdl;
	++(comm->c_errhdl->eh_refcount);
	newcomm->c_keys = 0;
	newcomm->c_topo_type = MPI_UNDEFINED;
	newcomm->c_topo_nprocs = 0;
	newcomm->c_topo_ndims = 0;
	newcomm->c_topo_dims = 0;
	newcomm->c_topo_coords = 0;
	newcomm->c_topo_index = 0;
	newcomm->c_topo_nedges = 0;
	newcomm->c_topo_edges = 0;
	newcomm->c_status = 0;

	if (__lam_trcomm_(newcomm)) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	if (! al_insert(__lam_comms_, &newcomm)) {
		return(__lam_errfunc_(comm, BLKMPICOMMCREATE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	__lam_setcid_(cid);
/*
 * Generate run time trace.
 */
	_LAM_TRACE(__lam_trcffend_(BLKMPICOMMCREATE, -1, comm, 0, 0));

	__lam_resetfunc_(BLKMPICOMMCREATE);
	return(MPI_SUCCESS);
}
