/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	cdup.c,v $
 * Revision 6.0.1.1  96/03/20  22:37:54  nevin
 * Typo in graph topology copy fixed
 * 
 * Revision 6.0  96/03/03  22:51:25  nevin
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/03/03  22:49:07  nevin
 * Correct FORTRAN attribute copying
 * 
 * Revision 5.2.1.2  96/02/18  21:20:26  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:00  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:29  gdburns
 * Ohio Release
 * 
 *	Function:	- duplicate a communicator
 *	Accepts:	- old communicator
 *			- ptr new communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>
#include <string.h>

#include <all_hash.h>
#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern int		__lam_cubedim_();
extern int		__lam_trcomm_();
extern struct _attrkey	*__lam_getattr_();

int
MPI_Comm_dup(comm, pnewcomm)

MPI_Comm		comm;
MPI_Comm		*pnewcomm;

{
	MPI_Status	stat;			/* receive status */
	MPI_Comm	newcomm;		/* new communicator */
	int		mycid;			/* my context ID */
	int		cid;			/* global context ID */
	int		err;			/* error code */
	int		fl_copy;		/* attr. copy flag */
	int		size;			/* group size */
	struct _attr	elem;			/* attr. element */
	struct _attr	*p;			/* favourite pointer */
	struct _attrkey	*pk;			/* key pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPICOMMDUP);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
	    return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPICOMMDUP,
		    __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (pnewcomm == 0) {
	    return(__lam_errfunc_(comm, BLKMPICOMMDUP,
		    __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	_LAM_TRACE(__lam_trcffstart_(BLKMPICOMMDUP));
/*
 * Create the new context ID.
 */
	mycid = __lam_getcid_();
	if (mycid < 0) {
	    return(__lam_errfunc_(comm, BLKMPICOMMDUP,
		    __lam_mkerr_(MPI_ERR_INTERN, EFULL)));
	}
/*
 * If intra-comm, do an all-reduce.
 */
	if ( ! comm->c_inter) {
	    err = MPI_Allreduce(&mycid, &cid, 1, MPI_INT, MPI_MAX, comm);
	}
/*
 * If inter-comm, do a local reduce to leader (0) then the leaders
 * exchange the context ID and broadcast to their local group.
 * We cheat by flip-flopping the communicator inter/intra state.
 */
	else {
	    comm->c_inter = 0;
	    err = MPI_Reduce(&mycid, &cid, 1, MPI_INT, MPI_MAX, 0, comm);
	    comm->c_inter = 1;

	    if (err != MPI_SUCCESS) {
		_LAM_TRACE(__lam_trcffend_(BLKMPICOMMDUP, -1, comm, 0, 0));
		__lam_resetfunc_(BLKMPICOMMDUP);
		return(__lam_errfunc_(comm, BLKMPICOMMDUP, err));
	    }

	    if (comm->c_group->g_myrank ==  0) {
		err = MPI_Sendrecv(&cid, 1, MPI_INT, 0, MPI_MERGE_TAG,
			&mycid, 1, MPI_INT, 0, MPI_MERGE_TAG, comm, &stat);
		if (err != MPI_SUCCESS) {
		    _LAM_TRACE(__lam_trcffend_(BLKMPICOMMDUP, -1, comm, 0, 0));
		    __lam_resetfunc_(BLKMPICOMMDUP);
		    return(__lam_errfunc_(comm, BLKMPICOMMDUP, err));
		}

		if (mycid > cid) cid = mycid;
	    }

	    comm->c_inter = 0;
	    err = MPI_Bcast(&cid, 1, MPI_INT, 0, comm);
	    comm->c_inter = 1;
	}

	if (err != MPI_SUCCESS) {
	    _LAM_TRACE(__lam_trcffend_(BLKMPICOMMDUP, -1, comm, 0, 0));
	    __lam_resetfunc_(BLKMPICOMMDUP);
	    return(__lam_errfunc_(comm, BLKMPICOMMDUP, err));
	}
/*
 * Create the new communicator.
 */
	newcomm = (MPI_Comm) malloc(sizeof(struct _comm));
	if (newcomm == 0) {
	    return(__lam_errfunc_(comm, BLKMPICOMMDUP,
		    __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pnewcomm = newcomm;
	newcomm->c_inter = comm->c_inter;
	newcomm->c_group = comm->c_group;
	newcomm->c_rgroup = comm->c_rgroup;
	newcomm->c_contextid = cid;
	newcomm->c_cube_dim = __lam_cubedim_(comm->c_group->g_nprocs);
	newcomm->c_refcount = 1;
	newcomm->c_errhdl = comm->c_errhdl;
	newcomm->c_status = 0;
/*
 * Copy the key attributes using the copy callback functions.
 */
	if (comm->c_keys == 0) {
	    newcomm->c_keys = 0;
	}

	else {
	    newcomm->c_keys = ah_init((int4) ah_size(comm->c_keys),
		    (int4) sizeof(struct _attr),
		    (int4) MPI_KEYVAL_INVALID, INT4_NIL);
	    if (newcomm->c_keys == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMDUP,
			__lam_mkerr_(MPI_ERR_OTHER, errno)));
	    }

	    p = (struct _attr *) ah_top(comm->c_keys);

	    while (p) {
		pk = __lam_getattr_(p->a_key);
		if (pk == 0) {
		    return(__lam_errfunc_(comm, BLKMPICOMMDUP,
			    __lam_mkerr_(MPI_ERR_INTERN, errno)));
		}
/*
 * Call the C or F77 callback function.
 */
		fl_copy = 0;
		if (pk->ak_copy != MPI_NULL_COPY_FN) {

		    if (pk->ak_f77comm < 0) {
			err = (*(pk->ak_copy))(comm, p->a_key, pk->ak_extra,
				p->a_value, &(elem.a_value), &fl_copy);
		    } else {
			err = (*(pk->ak_copy))(&pk->ak_f77comm,
				&p->a_key, (int *) &pk->ak_extra,
				(int *) &p->a_value, (int *) &(elem.a_value),
				&fl_copy);
		    }
		    
		    if (err != MPI_SUCCESS) {
			    return(__lam_errfunc_(comm, BLKMPICOMMDUP,
				    	__lam_mkerr_(err, 0)));
		    }
		}
/*
 * If attribute is copied, insert it and increment key reference count.
 */
		if (fl_copy) {
		    elem.a_key = p->a_key;

		    if (ah_insert(newcomm->c_keys, &elem)) {
			return(__lam_errfunc_(comm, BLKMPICOMMDUP,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
		    }

		    ++(pk->ak_refcount);
		}

		p = (struct _attr *) ah_next(comm->c_keys, p);
	    }
	}
/*
 * Copy the topology.
 */
	newcomm->c_topo_type = comm->c_topo_type;
	newcomm->c_topo_nprocs = comm->c_topo_nprocs;
	newcomm->c_topo_ndims = comm->c_topo_ndims;
	newcomm->c_topo_nedges = comm->c_topo_nedges;
	newcomm->c_topo_dims = 0;
	newcomm->c_topo_coords = 0;
	newcomm->c_topo_index = 0;
	newcomm->c_topo_edges = 0;

	if (comm->c_topo_type == MPI_CART) {
	    size = 2 * comm->c_topo_ndims * sizeof(int);

	    newcomm->c_topo_dims = (int *) malloc((unsigned) size);
	    if (newcomm->c_topo_dims == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMDUP,
			__lam_mkerr_(MPI_ERR_OTHER, errno)));
	    }

	    newcomm->c_topo_coords = newcomm->c_topo_dims + comm->c_topo_ndims;
	    memcpy((char *) newcomm->c_topo_dims,
		    (char *) comm->c_topo_dims, size);
	}

	else if (comm->c_topo_type == MPI_GRAPH) {
	    size = (comm->c_topo_nprocs+comm->c_topo_nedges) * sizeof(int);

	    newcomm->c_topo_index = (int *) malloc((unsigned) size);
	    if (newcomm->c_topo_index == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMDUP,
			__lam_mkerr_(MPI_ERR_OTHER, errno)));
	    }

	    newcomm->c_topo_edges = newcomm->c_topo_index +
		    comm->c_topo_nprocs;
	    memcpy((char *) newcomm->c_topo_index,
		    (char *) comm->c_topo_index, size);
	}

	++(comm->c_errhdl->eh_refcount);
	++(comm->c_group->g_refcount);
	if (comm->c_inter) ++(comm->c_rgroup->g_refcount);

	if (__lam_trcomm_(newcomm)) {
	    return(__lam_errfunc_(comm, BLKMPICOMMDUP,
		__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	if (! al_insert(__lam_comms_, &newcomm)) {
		return(__lam_errfunc_(comm, BLKMPICOMMDUP,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}


	__lam_setcid_(cid);
/*
 * Generate run time trace.
 */
	_LAM_TRACE(__lam_trcffend_(BLKMPICOMMDUP, -1, comm, 0, 0));

	__lam_resetfunc_(BLKMPICOMMDUP);
	return(MPI_SUCCESS);
}
