/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cfree.c,v $
 * Revision 6.0  96/02/29  14:12:15  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:34  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:01  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:30  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:07  gdburns
 * new for 5.1
 * 
 *	Function:	- free a communicator
 *	Accepts:	- ptr communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>

int
MPI_Comm_free(pcomm)

MPI_Comm		*pcomm;

{
	MPI_Comm	comm;
	int		errcode;
	int		key;
	struct _attr	*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICOMMFREE);
/*
 * Check the arguments.
 */
	if (pcomm == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICOMMFREE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	comm = *pcomm;
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICOMMFREE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}
/*
 * Do not delete the communicator if it has pending requests.
 */
	*pcomm = MPI_COMM_NULL;

	if (--(comm->c_refcount) != 0) {
		__lam_resetfunc_(BLKMPICOMMFREE);
		return(MPI_SUCCESS);
	}
/*
 * Free the groups.
 */
	errcode = MPI_Group_free(&(comm->c_group));
	if (errcode != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
				BLKMPICOMMFREE, errcode));
	}

	if (comm->c_inter) {
		errcode = MPI_Group_free(&(comm->c_rgroup));
		if (errcode != MPI_SUCCESS) {
			return(__lam_errfunc_(MPI_COMM_WORLD,
					BLKMPICOMMFREE, errcode));
		}
	}
/*
 * Free the attribute keys.
 */
	if (comm->c_keys) {
		p = (struct _attr *) ah_top(comm->c_keys);
		while (p) {
			key = p->a_key;
			errcode = MPI_Attr_delete(comm, key);
			if (errcode != MPI_SUCCESS) {
				return(__lam_errfunc_(MPI_COMM_WORLD,
						BLKMPICOMMFREE, errcode));
			}

			errcode = MPI_Keyval_free(&key);
			if (errcode != MPI_SUCCESS) {
				return(__lam_errfunc_(MPI_COMM_WORLD,
						BLKMPICOMMFREE, errcode));
			}

			p = (struct _attr *) ah_next(comm->c_keys, (void *) p);
		}
	}
/*
 * Free the error handler.
 */
	errcode = MPI_Errhandler_free(&(comm->c_errhdl));
	if (errcode != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
				BLKMPICOMMFREE, errcode));
	}
/*
 * Free topology information.
 */
	if (comm->c_topo_type == MPI_CART) {
		free((char *) comm->c_topo_dims);
	} else if (comm->c_topo_type == MPI_GRAPH) {
		free((char *) comm->c_topo_index);
	}
/*
 * Free the context ID.
 */
	__lam_rmcid_(comm->c_contextid);
/*
 * Remove the communicator from the communicator list and free it.
 */
	al_delete(__lam_comms_, &comm);
	free((char *) comm);

	__lam_resetfunc_(BLKMPICOMMFREE);
	return(MPI_SUCCESS);
}
