/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	errstring.c,v $
 * Revision 6.0  96/02/29  14:12:25  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/02/18  21:22:19  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.2  95/06/28  20:37:37  raja
 * Add MPI_ERR_SYSRESOURCE.
 * 
 * Revision 5.2.1.1  95/06/13  03:11:05  raja
 * Add MPI_ERR_IN_STATUS, MPI_ERR_PENDING.
 * 
 * Revision 5.2  94/08/22  14:04:13  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:44  gdburns
 * Ohio Release
 * 
 *	Function:	- return the MPI error message
 *	Accepts:	- error code
 *			- message buffer (returned value)
 *			- ptr buffer length (returned value)
 *	Returns:	- MPI_SUCCESS
 */

#include <stdio.h>
#include <string.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * error messages
 */
static char		*mpierrmsg[] = {
	"unused error code",			/* unused */
	"invalid buffer pointer",		/* MPI_ERR_BUFFER */
	"invalid count argument",		/* MPI_ERR_COUNT */
	"invalid datatype argument",		/* MPI_ERR_TYPE */
	"invalid tag argument",			/* MPI_ERR_TAG */
	"invalid communicator",			/* MPI_ERR_COMM */
	"invalid rank",				/* MPI_ERR_RANK */
	"invalid request handle",		/* MPI_ERR_REQUEST */
	"invalid root",				/* MPI_ERR_ROOT */
	"invalid group",			/* MPI_ERR_GROUP */
	"invalid operation",			/* MPI_ERR_OP */
	"invalid topology",			/* MPI_ERR_TOPOLOGY */
	"invalid dimension argument",		/* MPI_ERR_DIMS */
	"invalid argument",			/* MPI_ERR_ARG */
	"unknown error",			/* MPI_ERR_UNKNOWN */
	"message truncated",			/* MPI_ERR_TRUNCATE */
	"unclassified",				/* MPI_ERR_OTHER */
	"internal MPI error",			/* MPI_ERR_INTERN */
	"error code is in status",		/* MPI_ERR_IN_STATUS */
	"pending request",			/* MPI_ERR_PENDING */
	"out of system resources",		/* MPI_ERR_SYSRESOURCE */
	"process in local group is dead",	/* MPI_ERR_LOCALDEAD */
	"process in remote group is dead",	/* MPI_ERR_REMOTEDEAD */
	"last error code (huh?)",		/* MPI_ERR_LASTCODE */
			};

static char		buf[128];

int
MPI_Error_string(errcode, msg, plen)

int			errcode;
char			*msg;
int			*plen;

{
	int		class;			/* error class */
	int		funct;			/* error function */
	int		error;			/* UNIX errno */
	int		err;			/* function error code */
	char		*p;			/* favourite pointer */
	char		*errmsg;		/* error message */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIERRSTRING);
/*
 * Check the arguments.
 */
	if ((msg == 0) || (plen == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIERRSTRING, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Create the error message string.
 */
	err = MPI_SUCCESS;
	errmsg = buf;

	if (errcode == 0) {
		sprintf(buf, "MPI: no errors");
	}

	else {
		__lam_bkerr_(errcode, &class, &funct, &error);
		strcpy(buf, "");
		p = blktype(funct);

		if (p) {
			strcat(buf, p);
			strcat(buf, ": ");
		}

		if ((class > 0) && (class < MPI_ERR_LASTCODE)) {
			strcat(buf, mpierrmsg[class]);
		} else {
			strcat(buf, "unknown error class");
			err = __lam_mkerr_(MPI_ERR_ARG, 0);
		}

		if (error > 0) {
			strcat(buf, ": ");
			errno = error;
			strcat(buf, terrorstr(""));
		}
	}
/*
 * Fill the user's message buffer.
 */
	strncpy(msg, errmsg, MPI_MAX_ERROR_STRING - 1);
	msg[MPI_MAX_ERROR_STRING - 1] = '\0';
	*plen = strlen(msg);

	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIERRSTRING, err));
	}

	__lam_resetfunc_(BLKMPIERRSTRING);
	return(MPI_SUCCESS);
}
