/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/JRV
 *
 *	$Log:	gather.c,v $
 * Revision 6.0  96/02/29  14:12:27  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/28  23:37:16  nevin
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:12  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:04:14  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:46  gdburns
 * Ohio Release
 * 
 *	Function:	- gather buffers at root in process rank order
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- recv buffer
 *			- recv count
 *			- recv datatype
 *			- root
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Gather(sbuf, scount, sdtype, rbuf, rcount, rdtype, root, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
int			root;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		rank;			/* my rank */
	int		size;			/* group size */
	int		err;			/* error code */
	char		*ptmp;			/* temporary buffer */
	MPI_Aint	incr;			/* increment size */
	MPI_Aint	extent;			/* datatype extent */
	MPI_Status	stat;			/* receive status */
	struct _gps	*p;			/* favourite pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGATHER);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPIGATHER,
					__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((sdtype == MPI_DATATYPE_NULL) || (rdtype == MPI_DATATYPE_NULL)) {
		return(__lam_errfunc_(comm, BLKMPIGATHER,
					__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &size);

	if ((root >= size) || (root < 0)) {
		return(__lam_errfunc_(comm, BLKMPIGATHER,
					__lam_mkerr_(MPI_ERR_ROOT, 0)));
	}

	if ((scount < 0) || ((rcount < 0) && (rank == root))) {
		return(__lam_errfunc_(comm, BLKMPIGATHER,
					__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIGATHER);
/*
 * Remember required parameters.
 */
	p = &(comm->c_group->g_procs[root]->p_gps);

	__lam_setparam_(BLKMPIGATHER, root | (p->gps_grank << 16),
				(p->gps_node << 16) | p->gps_idx);
/*
 * Switch to collective communicator.
 */
	__lam_mkcoll_(comm);
/*
 * Everyone but root sends data and returns.
 */
	if (rank != root) {

		err = MPI_Send(sbuf, scount, sdtype, root, BLKMPIGATHER, comm);
		__lam_mkpt_(comm);

		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIGATHER, err));
		}
/*
 * Generate run time trace.
 */
		if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) {
			__lam_trcffend_(BLKMPIGATHER,
						root, comm, sdtype, scount);
		}

		__lam_resetfunc_(BLKMPIGATHER);
		return(MPI_SUCCESS);
	}
/*
 * I am the root, loop receiving the data.
 */
	MPI_Type_extent(rdtype, &extent);
	incr = extent * rcount;

	for (i = 0, ptmp = (char *) rbuf; i < size; ++i, ptmp += incr) {
/*
 * simple optimization
 */
		if (i == rank) {
			err = __lam_dtsndrcv_(sbuf, scount, sdtype, ptmp,
					rcount, rdtype, BLKMPIGATHER, comm);
		} else {
			err = MPI_Recv(ptmp, rcount, rdtype, i,
						BLKMPIGATHER, comm, &stat);
		}

		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIGATHER, err));
		}
	}

	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIGATHER, root, comm, sdtype, scount);

	__lam_resetfunc_(BLKMPIGATHER);
	return(MPI_SUCCESS);
}
