/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/JRV
 *
 *	$Log:	gatherv.c,v $
 * Revision 6.0  96/02/29  14:12:28  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/28  23:37:22  nevin
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:13  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:04:15  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:47  gdburns
 * Ohio Release
 * 
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- recv buffer
 *			- recv counts
 *			- displacements
 *			- recv datatype
 *			- root
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Gatherv(sbuf, scount, sdtype,
			rbuf, rcounts, disps, rdtype, root, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			*rcounts;
int			*disps;
MPI_Datatype		rdtype;
int			root;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		rank;			/* my rank */
	int		size;			/* group size */
	int		err;			/* error code */
	char		*ptmp;			/* temp. buffer */
	MPI_Aint	extent;			/* datatype extent */
	MPI_Status	stat;			/* receive status */
	struct _gps	*p;			/* favourite pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGATHERV);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPIGATHERV,
					__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((sdtype == MPI_DATATYPE_NULL) || (rdtype == MPI_DATATYPE_NULL)) {
		return(__lam_errfunc_(comm, BLKMPIGATHERV,
					__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &size);

	if ((root >= size) || (root < 0)) {
		return(__lam_errfunc_(comm, BLKMPIGATHERV,
					__lam_mkerr_(MPI_ERR_ROOT, 0)));
	}

	if ((rank == root) && (disps == 0)) {
		return(__lam_errfunc_(comm, BLKMPIGATHERV,
					__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((scount < 0) || ((rcounts == 0) && (rank == root))) {
		return(__lam_errfunc_(comm, BLKMPIGATHERV,
					__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIGATHERV);
/*
 * Remember required parameters.
 */
	p = &(comm->c_group->g_procs[root]->p_gps);

	__lam_setparam_(BLKMPIGATHERV, root | (p->gps_grank << 16),
				(p->gps_node << 16) | p->gps_idx);
/*
 * Switch to collective communicator.
 */
	__lam_mkcoll_(comm);
/*
 * Everyone but root sends data and returns.
 */
	if (rank != root) {

		err = MPI_Send(sbuf, scount, sdtype,
					root, BLKMPIGATHERV, comm);
		__lam_mkpt_(comm);

		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIGATHERV, err));
		}
/*
 * Generate run time trace.
 */
		if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) {
			__lam_trcffend_(BLKMPIGATHERV,
						root, comm, sdtype, scount);
		}

		__lam_resetfunc_(BLKMPIGATHERV);
		return(MPI_SUCCESS);
	}
/*
 * I am the root, loop receiving data.
 */
	MPI_Type_extent(rdtype, &extent);

	for (i = 0; i < size; ++i) {

		ptmp = ((char *) rbuf) + (extent * disps[i]);
/*
 * simple optimization
 */
		if (i == rank) {
			err = __lam_dtsndrcv_(sbuf, scount, sdtype,
						ptmp, rcounts[i], rdtype,
						BLKMPIGATHERV, comm);
		} else {
			err = MPI_Recv(ptmp, rcounts[i], rdtype, i,
						BLKMPIGATHERV, comm, &stat);
		}

		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIGATHERV, err));
		}
	}

	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIGATHERV, root, comm, sdtype, scount);

	__lam_resetfunc_(BLKMPIGATHERV);
	return(MPI_SUCCESS);
}
