/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	getcount.c,v $
 * Revision 6.0  96/02/29  14:12:31  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:22:31  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  95/03/29  04:50:50  raja
 * MPI_PACKED matches all datatypes.
 * 
 * Revision 5.2  94/08/22  14:04:17  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/08/18  12:00:57  raja
 * new for 5.2
 * 
 *	Function:	- get # top-level elements
 *	Accepts:	- ptr to status
 *			- datatype
 *			- ptr to count
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Get_count(pstat, dtype, pcount)

MPI_Status		*pstat;
MPI_Datatype		dtype;
int			*pcount;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPIGETCOUNT);
/*
 * Check the arguments.
 */
	if ((pcount == 0) || (pstat == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGETCOUNT, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((dtype == MPI_DATATYPE_NULL) || ((dtype != pstat->st_dtype) &&
				(pstat->st_dtype != MPI_DATATYPE_NULL) &&
				(pstat->st_dtype != MPI_PACKED) &&
				(dtype != MPI_PACKED))) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGETCOUNT, __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}
/*
 * If different datatypes, recompute count.
 */
	if (dtype != pstat->st_dtype) pstat->st_count = -1;

	pstat->st_dtype = dtype;
/*
 * If the count is not available, compute it.
 */
	if (pstat->st_count == -1) {
		__lam_docount_(pstat->st_length, dtype,
				&(pstat->st_count), &(pstat->st_nelem));
	}

	*pcount = pstat->st_count;

	__lam_resetfunc_(BLKMPIGETCOUNT);
	return(MPI_SUCCESS);
}
