/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	gexcl.c,v $
 * Revision 6.0  96/02/29  14:12:35  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:35  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:19  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:50  gdburns
 * Ohio Release
 * 
 *	Function:	- form new group by exclusion
 *	Accepts:	- original group
 *			- # of processes to exclude
 *			- ranks
 *			- ptr new group
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Group_excl(g, n, ranks, png)

MPI_Group		g;
int			n;
int			*ranks;
MPI_Group		*png;

{
	int		i;			/* favourite index */
	int		nproc;			/* # processes */
	struct _proc	**pi, **pj;		/* favourite pointers */
	MPI_Group	ng;			/* new group */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPEXCL);
/*
 * Check the arguments.
 */
	if (g == MPI_GROUP_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPEXCL, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if ((png == 0) || ((n > 0) && (ranks == 0)) || (n > g->g_nprocs)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPEXCL, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial cases.
 */
	if (n == 0) {
		*png = g;
		++(g->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPEXCL);
		return(MPI_SUCCESS);
	}
/*
 * Get the # of processes in the new group.
 * Flag the excluded processes.
 */
	nproc = g->g_nprocs - n;
	pi = g->g_procs;

	for (i = 0; i < n; ++i, ++ranks) {

		if ((*ranks >= 0) && (*ranks < g->g_nprocs)) {
			pi[*ranks]->p_mode |= LAM_PFLAG;
		} else {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPEXCL,
					__lam_mkerr_(MPI_ERR_RANK, 0)));
		}
	}
/*
 * Handle the empty difference case.
 */
	if (nproc == 0) {
		*png = MPI_GROUP_EMPTY;
		++(MPI_GROUP_EMPTY->g_refcount);

		for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		}

		__lam_resetfunc_(BLKMPIGROUPEXCL);
		return(MPI_SUCCESS);
	}
/*
 * Allocate a new group.
 */
	ng = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(nproc * sizeof(struct _proc *)));
	if (ng == 0) {
		for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		}

		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPEXCL,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*png = ng;
	ng->g_nprocs = nproc;
	ng->g_myrank = MPI_UNDEFINED;
	ng->g_refcount = 1;
	ng->g_procs = (struct _proc **) ((char *) ng + sizeof(struct _group));
/*
 * Fill the new group processes.
 */
	pj = ng->g_procs;

	for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {

		if ((*pi)->p_mode & LAM_PFLAG) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		} else {
			*pj = *pi;
			++pj;
		}
	}

	for (i = 0, pi = ng->g_procs; i < ng->g_nprocs; ++i, ++pi) {

		if (*pi == __lam_myproc_) {
			ng->g_myrank = i;
			break;
		}
	}

	__lam_resetfunc_(BLKMPIGROUPEXCL);
	return(MPI_SUCCESS);
}
