/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	gincl.c,v $
 * Revision 6.0  96/02/29  14:12:36  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:38  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:21  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:51  gdburns
 * Ohio Release
 * 
 *	Function:	- form a new group by inclusion
 *	Accepts:	- original group
 *			- # of processes
 *			- ranks 
 *			- ptr new group
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Group_incl(g, n, ranks, png)

MPI_Group		g;
int			n;
int			*ranks;
MPI_Group		*png;

{
	int		i;			/* favourite index */
	struct _proc	**pi, **pj;		/* favourite pointers */
	MPI_Group	ng;			/* new group */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPINCL);
/*
 * Check the arguments.
 */
	if (g == MPI_GROUP_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPINCL, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if ((png == 0) || ((n > 0) && (ranks == 0)) || (n > g->g_nprocs)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPINCL, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial cases.
 */
	if (n == 0) {
		*png = MPI_GROUP_EMPTY;
		++(MPI_GROUP_EMPTY->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPINCL);
		return(MPI_SUCCESS);
	}
/*
 * Allocate a new group.
 */
	ng = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(n * sizeof(struct _proc *)));
	if (ng == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPINCL, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*png = ng;
	ng->g_nprocs = n;
	ng->g_myrank = MPI_UNDEFINED;
	ng->g_refcount = 1;
	ng->g_procs = (struct _proc **) ((char *) ng + sizeof(struct _group));
/*
 * Fill the new group processes.
 */
	pj = g->g_procs;

	for (i = 0, pi = ng->g_procs; i < n; ++i, ++pi, ++ranks) {

		if ((*ranks >= 0) && (*ranks < g->g_nprocs)) {
			*pi = pj[*ranks];
		} else {
			free((char *) ng);
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPINCL,
					__lam_mkerr_(MPI_ERR_RANK, 0)));
		}
	}

	for (i = 0, pi = ng->g_procs; i < n; ++i, ++pi) {

		if (*pi == __lam_myproc_) {
			ng->g_myrank = i;
			break;
		}
	}

	__lam_resetfunc_(BLKMPIGROUPINCL);
	return(MPI_SUCCESS);
}
