/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	ginter.c,v $
 * Revision 6.0  96/02/29  14:12:37  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:40  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:21  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:52  gdburns
 * Ohio Release
 * 
 *	Function:	- form the intersection of two groups
 *	Accepts:	- first group
 *			- second group
 *			- ptr intersection group
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Group_intersection(g1, g2, pgi)

MPI_Group		g1;
MPI_Group		g2;
MPI_Group		*pgi;

{
	int		i, j;			/* favourite indices */
	int		nproc;			/* # procs in inter. */
	struct _proc	**pi, **pj;		/* favourite pointers */
	MPI_Group	gi;			/* intersection group */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPINTER);
/*
 * Check the arguments.
 */
	if ((g1 == MPI_GROUP_NULL) || (g2 == MPI_GROUP_NULL)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPINTER, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if (pgi == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPINTER, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial cases.
 */
	if (g1->g_nprocs == 0) {
		*pgi = g1;
		++(g1->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPINTER);
		return(MPI_SUCCESS);
	}

	if (g2->g_nprocs == 0) {
		*pgi = g2;
		++(g2->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPINTER);
		return(MPI_SUCCESS);
	}
/*
 * Get the # of processes in the intersection.
 * Flag the processes common to both groups.
 */
	nproc = 0;

	for (i = 0, pi = g1->g_procs; i < g1->g_nprocs; ++i, ++pi) {

		for (j = 0, pj = g2->g_procs; j < g2->g_nprocs; ++j, ++pj) {

			if (*pi == *pj) {
				(*pi)->p_mode |= LAM_PFLAG;
				++nproc;
				break;
			}
		}
	}
/*
 * Handle the empty intersection case.
 */
	if (nproc == 0) {
		*pgi = MPI_GROUP_EMPTY;
		++(MPI_GROUP_EMPTY->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPINTER);
		return(MPI_SUCCESS);
	}
/*
 * Allocate a new group.
 */
	gi = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(nproc * sizeof(struct _proc *)));
	if (gi == 0) {
		for (i = 0, pi = g1->g_procs; i < g1->g_nprocs; ++i, ++pi) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		}

		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPINTER,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pgi = gi;
	gi->g_nprocs = nproc;
	gi->g_myrank = MPI_UNDEFINED;
	gi->g_refcount = 1;
	gi->g_procs = (struct _proc **) ((char *) gi + sizeof(struct _group));
/*
 * Fill the new group processes.
 */
	pj = gi->g_procs;

	for (i = 0, pi = g1->g_procs; i < g1->g_nprocs; ++i, ++pi) {

		if ((*pi)->p_mode & LAM_PFLAG) {
			(*pi)->p_mode &= ~LAM_PFLAG;
			*pj = *pi;
			++pj;
		}
	}

	for (i = 0, pi = gi->g_procs; i < gi->g_nprocs; ++i, ++pi) {

		if (*pi == __lam_myproc_) {
			gi->g_myrank = i;
			break;
		}
	}

	__lam_resetfunc_(BLKMPIGROUPINTER);
	return(MPI_SUCCESS);
}
