/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	graphcreate.c,v $
 * Revision 6.0  96/02/29  14:12:38  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:42  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:23  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:54  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:18  gdburns
 * new for 5.1
 * 
 *	Function:	- create a communicator with graph mapping
 *	Accepts:	- old communicator
 *			- # nodes
 *			- node index array
 *			- array of edges
 *			- rank reordering flag
 *			- ptr new communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Graph_create(comm, nnodes, index, edges, reorder, pnewcomm)

MPI_Comm		comm;
int			nnodes;
int			*index;
int			*edges;
int			reorder;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;
	MPI_Group	newgroup;
	int		nedges;
	int		size;
	int		err;
	int		range[1][3];
	int		i;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGRAPHCREATE);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGRAPHCREATE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHCREATE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((nnodes < 1) || (index == 0) || (edges == 0) || (pnewcomm == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHCREATE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIGRAPHCREATE);
/*
 * Create the group for the new communicator.
 */
	err = MPI_Comm_size(comm, &size);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHCREATE, err));
	}

	if (nnodes > size) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHCREATE,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if (nnodes == size) {
		err = MPI_Comm_group(comm, &newgroup);
	} else {
		range[0][0] = 0;
		range[0][1] = nnodes - 1;
		range[0][2] = 1;
		err = MPI_Group_range_incl(comm->c_group, 1, range, &newgroup);
	}

	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHCREATE, err));
	}
/*
 * Create the new communicator.
 */
	err = MPI_Comm_create(comm, newgroup, pnewcomm);
	if (err != MPI_SUCCESS) {
		MPI_Group_free(&newgroup);
		return(__lam_errfunc_(comm, BLKMPIGRAPHCREATE, err));
	}
/*
 * Fill the communicator with topology information. 
 */
	newcomm = *pnewcomm;
	if (newcomm != MPI_COMM_NULL) {
		newcomm->c_topo_type = MPI_GRAPH;
		newcomm->c_topo_nprocs = nnodes;
		newcomm->c_topo_nedges = nedges = index[nnodes - 1];

		newcomm->c_topo_index = (int *)
			malloc((unsigned) (nnodes + nedges) * sizeof(int));
		if (newcomm->c_topo_index == 0) {
			return(__lam_errfunc_(comm, BLKMPIGRAPHCREATE,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
		newcomm->c_topo_edges = newcomm->c_topo_index + nnodes;

		for (i = 0, p = newcomm->c_topo_index; i < nnodes; ++i, ++p) {
			*p = *index++;
		}

		for (i = 0, p = newcomm->c_topo_edges; i < nedges; ++i, ++p) {
			*p = *edges++;
		}
	}

	err = MPI_Group_free(&newgroup);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHCREATE, err));
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIGRAPHCREATE, -1, comm, 0, 0);

	__lam_resetfunc_(BLKMPIGRAPHCREATE);
	return(MPI_SUCCESS);
}
