/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	graphdimsget.c,v $
 * Revision 6.0  96/02/29  14:12:39  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:43  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:23  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:54  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:19  gdburns
 * new for 5.1
 * 
 *	Function:	- get # nodes/edges in graph topology
 *	Accepts:	- communicator
 *			- ptr # nodes (returned value)
 *			- ptr # edges (returned value)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Graphdims_get(comm, pnodes, pedges)

MPI_Comm		comm;
int			*pnodes;
int			*pedges;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPIGRAPHDIMSGET);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGRAPHDIMSGET, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHDIMSGET, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_GRAPH) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHDIMSGET,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((pnodes == 0) || (pedges == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHDIMSGET, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	*pnodes = comm->c_topo_nprocs;
	*pedges = comm->c_topo_nedges;

	__lam_resetfunc_(BLKMPIGRAPHDIMSGET);
	return(MPI_SUCCESS);
}
