/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	graphget.c,v $
 * Revision 6.0  96/02/29  14:12:40  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:44  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:24  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:55  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:19  gdburns
 * new for 5.1
 * 
 *	Function:	- get nodes/edges in graph topology
 *	Accepts:	- communicator
 *			- size of nodes array
 *			- size of edges array
 *			- nodes array
 *			- edges array
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Graph_get(comm, maxnodes, maxedges, nodes, edges)

MPI_Comm		comm;
int			maxnodes;
int			maxedges;
int			*nodes;
int			*edges;

{
	int		i;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGRAPHGET);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGRAPHGET, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHGET, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_GRAPH) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHGET,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((maxnodes < 0) || (maxedges < 0) || (nodes == 0) || (edges == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHGET, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Fill the nodes and edges arrays.
 */
	p = comm->c_topo_index;
	for (i = 0; (i < comm->c_topo_nprocs) && (i < maxnodes); ++i, ++p) {
		*nodes++ = *p;
	}

	p = comm->c_topo_edges;
	for (i = 0; (i < comm->c_topo_nedges) && (i < maxedges); ++i, ++p) {
		*edges++ = *p;
	}

	__lam_resetfunc_(BLKMPIGRAPHGET);
	return(MPI_SUCCESS);
}
