/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	graphmap.c,v $
 * Revision 6.0  96/02/29  14:12:41  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:45  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:24  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:56  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:20  gdburns
 * new for 5.1
 * 
 *	Function:	- map processes to graph coordinates
 *	Accepts:	- old communicator
 *			- # processes
 *			- node index array
 *			- array of edges
 *			- ptr caller's new rank
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Graph_map(comm, nnodes, index, edges, pnewrank)

MPI_Comm		comm;
int			nnodes;
int			*index;
int			*edges;
int			*pnewrank;

{
	int		errcode;
	int		myrank;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGRAPHMAP);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGRAPHMAP, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHMAP, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((nnodes < 1) || (index == 0) || (edges == 0) || (pnewrank == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHMAP, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Assign new rank.
 */
	errcode = MPI_Comm_rank(comm, &myrank);
	if (errcode != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHMAP, errcode));
	}

	*pnewrank =
		((myrank < 0) || (myrank >= nnodes)) ? MPI_UNDEFINED : myrank;

	__lam_resetfunc_(BLKMPIGRAPHMAP);
	return(MPI_SUCCESS);
}
