/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	graphnbr.c,v $
 * Revision 6.0  96/02/29  14:12:41  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:46  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:25  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:57  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:20  gdburns
 * new for 5.1
 * 
 *	Function:	- get neighbours of a process
 *	Accepts:	- communicator
 *			- rank
 *			- size of neighbours array 
 *			- neighbours array (returned value)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Graph_neighbors(comm, rank, maxnbrs, nbrs)

MPI_Comm		comm;
int			rank;
int			maxnbrs;
int			*nbrs;

{
	int		nnbrs;
	int		i;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGRAPHNBR);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGRAPHNBR, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHNBR, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_GRAPH) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHNBR,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((rank < 0) || (rank >= comm->c_group->g_nprocs)) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHNBR, __lam_mkerr_(MPI_ERR_RANK, 0)));
	}

	if ((maxnbrs < 0) || ((maxnbrs > 0) && (nbrs == 0))) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHNBR, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Fill the neighbours.
 */
	nnbrs = comm->c_topo_index[rank];
	p = comm->c_topo_edges;

	if (rank > 0) {
		i = comm->c_topo_index[rank - 1];
		nnbrs -= i;
		p += i; 
	}

	for (i = 0; (i < maxnbrs) && (i < nnbrs); ++i, ++p) {
		*nbrs++ = *p;
	}

	__lam_resetfunc_(BLKMPIGRAPHNBR);
	return(MPI_SUCCESS);
}
