/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	graphnbrcount.c,v $
 * Revision 6.0  96/02/29  14:12:42  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:47  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:26  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:57  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:21  gdburns
 * new for 5.1
 * 
 *	Function:	- get # neighbours for a process
 *	Accepts:	- communicator
 *			- rank
 *			- ptr # neighbours
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Graph_neighbors_count(comm, rank, pnbr)

MPI_Comm		comm;
int			rank;
int			*pnbr;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPIGRAPHNBRCOUNT);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGRAPHNBRCOUNT, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHNBRCOUNT, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_topo_type != MPI_GRAPH) {
		return(__lam_errfunc_(comm, BLKMPIGRAPHNBRCOUNT,
				__lam_mkerr_(MPI_ERR_TOPOLOGY, 0)));
	}

	if ((rank < 0) || (rank >= comm->c_group->g_nprocs)) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHNBRCOUNT, __lam_mkerr_(MPI_ERR_RANK, 0)));
	}

	if (pnbr == 0) {
		return(__lam_errfunc_(comm,
			BLKMPIGRAPHNBRCOUNT, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	*pnbr = comm->c_topo_index[rank];
	if (rank > 0) {
		*pnbr -= comm->c_topo_index[rank - 1];
	}

	__lam_resetfunc_(BLKMPIGRAPHNBRCOUNT);
	return(MPI_SUCCESS);
}
