/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	grexcl.c,v $
 * Revision 6.0  96/02/29  14:12:43  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:48  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:26  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:58  gdburns
 * Ohio Release
 * 
 *	Function:	- form a new group by range exclusion
 *	Accepts:	- original group
 *			- # of ranges
 *			- rank ranges
 *			- ptr new group
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Group_range_excl(g, n, ranges, png)

MPI_Group		g;
int			n;
int			ranges[][3];
MPI_Group		*png;

{
	int		i, j;			/* favourite indices */
	int		nproc;			/* # processes */
	int		low;			/* low range */
	int		high;			/* high range */
	int		stride;			/* range stride */
	struct _proc	**pi, **pj;		/* favourite pointers */
	MPI_Group	ng;			/* new group */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPREXCL);
/*
 * Check the arguments.
 */
	if (g == MPI_GROUP_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPREXCL, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if ((png == 0) || ((n > 0) && (ranges == 0))) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPREXCL, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Flag the processes to exclude.
 */
	nproc = 0;
	pi = g->g_procs;

	for (i = 0; i < n; ++i) {

		low = ranges[i][0];
		high = ranges[i][1];
		stride = ranges[i][2];

		if ((stride == 0) || (low < 0) || (high < 0) ||
				((stride > 0) && (low > high)) ||
				((stride < 0) && (low < high))) {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPREXCL,
					__lam_mkerr_(MPI_ERR_ARG, 0)));
		}

		j = low;
		for (; (stride > 0) ? (j <= high) : (j >= high); j += stride) {

			if ((j >= 0) && (j < g->g_nprocs)) {
				pi[j]->p_mode |= LAM_PFLAG;
				nproc = 1;
			} else {
				return(__lam_errfunc_(MPI_COMM_WORLD,
					BLKMPIGROUPREXCL,
					__lam_mkerr_(MPI_ERR_ARG, 0)));
			}
		}
	}
/*
 * Handle the trivial cases.
 */
	if (nproc == 0) {
		*png = g;
		++(g->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPREXCL);
		return(MPI_SUCCESS);
	}
/*
 * Count the # processes in the group.
 */
	nproc = g->g_nprocs;
	for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {

		if ((*pi)->p_mode & LAM_PFLAG) --nproc;
	}
/*
 * Handle another trivial case.
 */
	if (nproc == 0) {
		*png = MPI_GROUP_EMPTY;
		++(MPI_GROUP_EMPTY->g_refcount);

		for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		}

		__lam_resetfunc_(BLKMPIGROUPREXCL);
		return(MPI_SUCCESS);
	}
/*
 * Allocate a new group.
 */
	ng = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(nproc * sizeof(struct _proc *)));
	if (ng == 0) {
		for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		}

		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPREXCL, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*png = ng;
	ng->g_nprocs = nproc;
	ng->g_myrank = MPI_UNDEFINED;
	ng->g_refcount = 1;
	ng->g_procs = (struct _proc **) ((char *) ng + sizeof(struct _group));
/*
 * Fill the new group ranks.
 */
	j = 0;
	pj = ng->g_procs;

	for (i = 0, pi = g->g_procs; i < g->g_nprocs; ++i, ++pi) {

		if ((*pi)->p_mode & LAM_PFLAG) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		} else {
			*pj = *pi;

			if (*pj == __lam_myproc_) ng->g_myrank = j;

			++j; ++pj;
		}
	}

	__lam_resetfunc_(BLKMPIGROUPREXCL);
	return(MPI_SUCCESS);
}
