/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	grincl.c,v $
 * Revision 6.0  96/02/29  14:12:44  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:50  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:27  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:59  gdburns
 * Ohio Release
 * 
 *	Function:	- form a new group by range inclusion
 *	Accepts:	- original group
 *			- # of ranges
 *			- rank ranges
 *			- ptr new group
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <net.h>
#include <terror.h>

int
MPI_Group_range_incl(g, n, ranges, png)

MPI_Group		g;
int			n;
int			ranges[][3];
MPI_Group		*png;

{
	int		i, j, k;		/* favourite indices */
	int		nproc;			/* # processes */
	int		low;			/* low range */
	int		high;			/* high range */
	int		stride;			/* range stride */
	struct _proc	**p, **q;		/* favourite pointers */
	MPI_Group	ng;			/* new group */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPRINCL);
/*
 * Check the arguments.
 */
	if (g == MPI_GROUP_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPRINCL, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if ((png == 0) || ((n > 0) && (ranges == 0))) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPRINCL, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Count the # of processes to include.
 */
	nproc = 0;
	for (i = 0; i < n; ++i) {

		if ((stride = ranges[i][2]) >= 0) {
			high = ranges[i][1];
			low = ranges[i][0];
		} else {
			high = ranges[i][0];
			low = ranges[i][1];
			stride = -stride;
		}

		if ((stride == 0) || (low > high) || (low < 0) || (high < 0)) {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPRINCL,
					__lam_mkerr_(MPI_ERR_ARG, 0)));
		}

		nproc += 1 + ((high - low) / stride);
	}
/*
 * Handle the trivial cases.
 */
	if (nproc == 0) {
		*png = MPI_GROUP_EMPTY;
		++(MPI_GROUP_EMPTY->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPRINCL);
		return(MPI_SUCCESS);
	}
/*
 * Allocate a new group.
 */
	ng = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(nproc * sizeof(struct _proc *)));
	if (ng == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPRINCL, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*png = ng;
	ng->g_nprocs = nproc;
	ng->g_myrank = MPI_UNDEFINED;
	ng->g_refcount = 1;
	ng->g_procs = (struct _proc **) ((char *) ng + sizeof(struct _group));
/*
 * Fill the new group processes.
 */
	k = 0;
	p = ng->g_procs;
	q = g->g_procs;

	for (i = 0; i < n; ++i) {

		high = ranges[i][1];
		stride = ranges[i][2];
		j = ranges[i][0];

		for (; (stride > 0) ? (j <= high) : (j >= high); j += stride) {

			if ((j >= 0) && (j < g->g_nprocs)) {
				*p = q[j];
				if (*p == __lam_myproc_) ng->g_myrank = k;
			} else {
				*p = 0;
			}

			++p; ++k;
		}
	}

	__lam_resetfunc_(BLKMPIGROUPRINCL);
	return(MPI_SUCCESS);
}
