/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	gtranks.c,v $
 * Revision 6.0  96/02/29  14:12:46  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:52  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:28  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:01  gdburns
 * Ohio Release
 * 
 *	Function:	- get process ranks in a second group
 *	Accepts:	- first group
 *			- # of processes
 *			- array of ranks in first group
 *			- second group
 *			- array of ranks in second group (returned values)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Group_translate_ranks(g1, n, r1, g2, r2)

MPI_Group		g1;
int			n;
int			*r1;
MPI_Group		g2;
int			*r2;

{
	int		i, j;			/* favourite indices */
	int		rank;			/* process rank */
	struct _proc	**p1, **p2;		/* gps pointers */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPTRANKS);
/*
 * Check given arguments.
 */
	if ((g1 == MPI_GROUP_NULL) || (g2 == MPI_GROUP_NULL)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPTRANKS, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if (n == 0) {
		__lam_resetfunc_(BLKMPIGROUPTRANKS);
		return(MPI_SUCCESS);
	}

	if ((n < 0) || (r1 == 0) || (r2 == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPTRANKS, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Loop over the processes, translating their ranks.
 */
	for (i = 0; i < n; ++i) {
		rank = r1[i];

		if ((rank >= 0) && (rank < g1->g_nprocs)) {

			p1 = g1->g_procs + rank;
			p2 = g2->g_procs;

			for (j = 0; j < g2->g_nprocs; ++j, ++p2) {
				if (*p1 == *p2) break;
			}

			r2[i] = (j < g2->g_nprocs) ? j : MPI_UNDEFINED;
		}
		else {
			r2[i] = MPI_UNDEFINED;
		}
	}

	__lam_resetfunc_(BLKMPIGROUPTRANKS);
	return(MPI_SUCCESS);
}
