/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	gunion.c,v $
 * Revision 6.0  96/02/29  14:12:46  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:53  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:29  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:01  gdburns
 * Ohio Release
 * 
 *	Function:	- form the union of two groups
 *	Accepts:	- first group
 *			- second group
 *			- ptr union group
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Group_union(g1, g2, pgu)

MPI_Group		g1;
MPI_Group		g2;
MPI_Group		*pgu;

{
	int		i, j;			/* favourite indices */
	int		nproc;			/* # procs in union */
	struct _proc	**pi, **pj;		/* favourite pointers */
	MPI_Group	gu;			/* union group */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIGROUPUNION);
/*
 * Check the arguments.
 */
	if ((g1 == MPI_GROUP_NULL) || (g2 == MPI_GROUP_NULL)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPUNION, __lam_mkerr_(MPI_ERR_GROUP, 0)));
	}

	if (pgu == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIGROUPUNION, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial cases.
 */
	if (g1->g_nprocs == 0) {
		*pgu = g2;
		++(g2->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPUNION);
		return(MPI_SUCCESS);
	}

	if (g2->g_nprocs == 0) {
		*pgu = g1;
		++(g1->g_refcount);
		__lam_resetfunc_(BLKMPIGROUPUNION);
		return(MPI_SUCCESS);
	}
/*
 * Get the # of processes in the union.
 * Flag the duplicate processes in second group.
 */
	nproc = g1->g_nprocs + g2->g_nprocs;

	for (i = 0, pi = g2->g_procs; i < g2->g_nprocs; ++i, ++pi) {

		for (j = 0, pj = g1->g_procs; j < g1->g_nprocs; ++j, ++pj) {

			if (*pi == *pj) {
				(*pi)->p_mode |= LAM_PFLAG;
				--nproc;
				break;
			}
		}
	}
/*
 * Allocate a new group.
 */
	gu = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(nproc * sizeof(struct _proc *)));
	if (gu == 0) {
		for (i = 0, pi = g2->g_procs; i < g2->g_nprocs; ++i, ++pi) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		}

		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIGROUPUNION,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pgu = gu;
	gu->g_nprocs = nproc;
	gu->g_myrank = g1->g_myrank;
	gu->g_refcount = 1;
	gu->g_procs = (struct _proc **) ((char *) gu + sizeof(struct _group));
/*
 * Fill the new group processes.
 */
	memcpy((char *) gu->g_procs, (char *) g1->g_procs,
				g1->g_nprocs * sizeof(struct _proc *));

	pj = gu->g_procs + g1->g_nprocs;

	for (i = 0, pi = g2->g_procs; i < g2->g_nprocs; ++i, ++pi) {

		if ((*pi)->p_mode & LAM_PFLAG) {
			(*pi)->p_mode &= ~LAM_PFLAG;
		} else {
			*pj = *pi;
			++pj;
		}
	}

	if (gu->g_myrank == MPI_UNDEFINED) {

		for (i = 0, pi = gu->g_procs; i < gu->g_nprocs; ++i, ++pi) {

			if (*pi == __lam_myproc_) {
				gu->g_myrank = i;
				break;
			}
		}
	}

	__lam_resetfunc_(BLKMPIGROUPUNION);
	return(MPI_SUCCESS);
}
