/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	ibsend.c,v $
 * Revision 6.0.1.1  96/04/29  10:35:48  nevin
 * Fake requests status must be set as it is never started
 * 
 * Revision 6.0  96/02/29  14:12:47  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:54  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:30  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:02  gdburns
 * Ohio Release
 * 
 * 
 *	Function:	- immediate buffered send
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- destination
 *			- tag
 *			- communicator
 *			- request (out)
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern int		__lam_errfunc_();
extern int		_mpi_req_build();

int
MPI_Ibsend(buf, count, dtype, dest, tag, comm, preq)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			dest;
int			tag;
MPI_Comm		comm;
MPI_Request		*preq;

{
	int		err;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIIBSEND);
/*
 * Create the fake user request.  We do not add it into the
 * request list.  This is an exception to the design that says
 * that all active (inc. done) requests are to be found in the
 * request list but fake requests, really ibsend requests are
 * very exceptional MPI creatures.
 */
	*preq = MPI_REQUEST_NULL;
	err = _mpi_req_build(buf, count, dtype, dest, tag, comm,
			LAM_RQIFAKE, preq);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIIBSEND, err));
	}

	(*preq)->rq_status.MPI_ERROR = MPI_SUCCESS;
	(*preq)->rq_status.MPI_SOURCE = MPI_PROC_NULL;
	(*preq)->rq_state = LAM_RQSDONE;
/*
 * Bsend() the message.
 */
	err = MPI_Bsend(buf, count, dtype, dest, tag, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIIBSEND, err));
	}

	__lam_resetfunc_(BLKMPIIBSEND);
	return(MPI_SUCCESS);
}
