/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	iccreate.c,v $
 * Revision 6.0  96/02/29  14:12:48  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/28  23:47:46  nevin
 * Send GPS arrays as MPI_INT
 * 
 * Revision 5.2.1.1  1996/02/18 21:22:55  gdburns
 * major overhaul
 *
 * Revision 5.2  94/08/22  14:04:30  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:03  gdburns
 * Ohio Release
 * 
 *	Function:	- create a new inter-communicator
 *	Accepts:	- local intra-communicator
 *			- local leader
 *			- peer intra-communicator
 *			- peer leader
 *			- tag
 *			- ptr new inter-communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern struct _proc	*__lam_procadd_();

int
MPI_Intercomm_create(lcomm, lleader, pcomm, pleader, tag, pnewcomm)

MPI_Comm		lcomm;
int			lleader;
MPI_Comm		pcomm;
int			pleader;
int			tag;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;		/* new communicator */
	MPI_Status	stat;			/* message status */
	MPI_Group	lgroup;			/* local group */
	MPI_Group	rgroup;			/* remote group */
	int		myrank;			/* my rank */
	int		outmsg[2];		/* output message buffer */
	int		inmsg[2];		/* input message buffer */
	int		mycid;			/* my context ID */
	int		cid;			/* global context ID */
	int		err;			/* error code */
	int		i;			/* favourite index */
	int		lgsize;			/* local group size */
	int		rgsize;			/* remote group size */
	int		lgbytes;		/* local group GPS size */
	int		rgbytes;		/* remote group GPS size */
	struct _gps	*lprocs;		/* local process GPS */
	struct _gps	*rprocs;		/* remote process GPS */
	struct _proc	*p;			/* favourite pointer */
	struct _proc	**pp;			/* another pointer */
	struct _gps	*pgp;			/* yet another pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIICOMMCREATE);
/*
 * Check the arguments.
 */
	if (lcomm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIICOMMCREATE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (lcomm->c_inter) {
		return(__lam_errfunc_(lcomm,
			BLKMPIICOMMCREATE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (lcomm->c_group->g_myrank == lleader) {
		if (pcomm == MPI_COMM_NULL) {
			return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE, 
					__lam_mkerr_(MPI_ERR_COMM, 0)));
		}

		if ((pleader < 0) || (pleader >= pcomm->c_group->g_nprocs)) {
			return(__lam_errfunc_(pcomm, BLKMPIICOMMCREATE, 
					__lam_mkerr_(MPI_ERR_RANK, 0)));
		}
	}

	if (tag < 0) {
		return(__lam_errfunc_(lcomm,
			BLKMPIICOMMCREATE, __lam_mkerr_(MPI_ERR_TAG, 0)));
	}

	if ((lleader < 0) || (lleader >= lcomm->c_group->g_nprocs)) {
		return(__lam_errfunc_(lcomm,
			BLKMPIICOMMCREATE, __lam_mkerr_(MPI_ERR_RANK, 0)));
	}

	if (pnewcomm == 0) {
		return(__lam_errfunc_(lcomm,
			BLKMPIICOMMCREATE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	_LAM_TRACE(__lam_trcffstart_(BLKMPIICOMMCREATE));
/*
 * Create the new context ID: reduce-max to leader within the local group,
 * then find the max between the two leaders, then broadcast within group.
 * In the same message, the leaders exchange their local group sizes and
 * broadcast the received group size to their local group.
 */
	lgroup = lcomm->c_group;
	lgsize = lgroup->g_nprocs;
	myrank = lgroup->g_myrank;

	mycid = __lam_getcid_();
	if (mycid < 0) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_INTERN, EFULL)));
	}

	err = MPI_Reduce(&mycid, &cid, 1, MPI_INT, MPI_MAX, lleader, lcomm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE, err));
	}

	if (lleader == myrank) {
		outmsg[0] = cid;
		outmsg[1] = lgsize;

		err = MPI_Sendrecv(outmsg, 2, MPI_INT, pleader, tag,
				inmsg, 2, MPI_INT, pleader, tag, pcomm, &stat);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE, err));
		}

		if (inmsg[0] < cid) inmsg[0] = cid;
	}

	err = MPI_Bcast(inmsg, 2, MPI_INT, lleader, lcomm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE, err));
	}

	cid = inmsg[0];
	rgsize = inmsg[1];
/*
 * Allocate remote group process GPS array.
 */
	rgbytes = rgsize * sizeof(struct _gps);
	rprocs = (struct _gps *) malloc((unsigned) rgbytes);
	if (rprocs == 0) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * Leaders exchange process GPS arrays and broadcast them to their group.
 */
	if (lleader == myrank) {

		lgbytes = lgsize * sizeof(struct _gps);
		lprocs = (struct _gps *) malloc((unsigned) lgbytes);
		if (lprocs == 0) {
			return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
/*
 * Fill local process GPS.
 */
		for (i = 0, pp = lgroup->g_procs; i < lgsize; ++i, ++pp) {
			lprocs[i] = (*pp)->p_gps;
		}

		err = MPI_Sendrecv(lprocs, lgbytes/sizeof(int), MPI_INT,
					pleader, tag, rprocs,
					rgbytes/sizeof(int), MPI_INT,
					pleader, tag, pcomm, &stat);

		free((char *) lprocs);

		if (err != MPI_SUCCESS) {
			free((char *) rprocs);
			return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE, err));
		}
	}

	err = MPI_Bcast(rprocs, rgbytes/sizeof(int), MPI_INT, lleader, lcomm);
	if (err != MPI_SUCCESS) {
		free((char *) rprocs);
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE, err));
	}
/*
 * Create the remote group.
 */
	rgroup = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(rgsize * sizeof(struct _proc **)));
	if (rgroup == 0) {
		free((char *) rprocs);
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	rgroup->g_nprocs = rgsize;
	rgroup->g_myrank = MPI_UNDEFINED;
	rgroup->g_refcount = 1;
	rgroup->g_procs = (struct _proc **)
				((char *) rgroup + sizeof(struct _group));

	for (i = 0, pgp = rprocs; i < rgsize; ++i, ++pgp) {
		p = __lam_procadd_(pgp);
		if (p == 0) {
			return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
		
		rgroup->g_procs[i] = p;
	}

	free((char *) rprocs);
/*
 * Create the new communicator.
 */
	newcomm = (MPI_Comm) malloc(sizeof(struct _comm));
	if (newcomm == 0) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pnewcomm = newcomm;
	newcomm->c_inter = 1;
	newcomm->c_refcount = 1;
	newcomm->c_cube_dim = __lam_cubedim_(lgroup->g_nprocs);
	newcomm->c_group = lgroup;
	++(lgroup->g_refcount);
	newcomm->c_rgroup = rgroup;
	newcomm->c_contextid = cid;
	newcomm->c_errhdl = lcomm->c_errhdl;
	++(lcomm->c_errhdl->eh_refcount);
	newcomm->c_keys = 0;
	newcomm->c_topo_type = MPI_UNDEFINED;
	newcomm->c_topo_nprocs = 0;
	newcomm->c_topo_ndims = 0;
	newcomm->c_topo_dims = 0;
	newcomm->c_topo_index = 0;
	newcomm->c_topo_nedges = 0;
	newcomm->c_topo_edges = 0;
	newcomm->c_status = 0;

	if (! al_insert(__lam_comms_, &newcomm)) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}
	
	if (__lam_trcomm_(newcomm)) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}
	
	__lam_setcid_(cid);
/*
 * setup any new processes
 */
	if (RPI_SPLIT(_rpi_lamd_addprocs, _rpi_c2c_addprocs, ())) {
		return(__lam_errfunc_(lcomm, BLKMPIICOMMCREATE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	_LAM_TRACE(__lam_trcffend_(BLKMPIICOMMCREATE, lleader, lcomm, 0, 0));

	__lam_resetfunc_(BLKMPIICOMMCREATE);
	return(MPI_SUCCESS);
}
