/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/NJN
 *
 *	$Log:	init.c,v $
 * Revision 6.0.1.1  96/03/13  15:41:59  nevin
 * C predefined attribute values are pointers to the real values
 * 
 * Revision 6.0  96/03/03  23:10:34  nevin
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/03/03  22:42:56  nevin
 * Use MPI_NULL_COPY_FN and MPI_NULL_DELETE_FN
 * 
 * Revision 5.2.1.2  96/02/28  23:52:56  nevin
 * Wait at barrier for all to initialize
 * 
 * Revision 5.2.1.1  1996/02/18 21:22:58  gdburns
 * major overhaul
 *
 * Revision 5.2  94/08/22  14:04:32  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/06/09  14:17:46  vaigl-j
 * Predefined ops use C datatypes by default.
 * 
 * Revision 5.1  94/05/18  12:51:05  gdburns
 * Ohio Release
 * 
 *	Function:	- initialize the MPI session
 *			- the kitchen-sink of MPI
 *	Accepts:	- ptr to argc
 *			- ptr to argv
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>
#include <string.h>

#include <all_list.h>
#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * global variables
 */
struct _comm		lam_mpi_comm_world;
struct _comm		lam_mpi_comm_self;
MPI_Comm		lam_mpi_comm_parent;
struct _group		lam_mpi_group_empty;
struct _op		lam_mpi_max, lam_mpi_min;
struct _op		lam_mpi_sum, lam_mpi_prod;
struct _op		lam_mpi_land, lam_mpi_band;
struct _op		lam_mpi_lor, lam_mpi_bor;
struct _op		lam_mpi_lxor, lam_mpi_bxor;
struct _op		lam_mpi_maxloc, lam_mpi_minloc;
struct _dtype		lam_mpi_char, lam_mpi_byte;
struct _dtype		lam_mpi_int, lam_mpi_logic;
struct _dtype		lam_mpi_short, lam_mpi_long;
struct _dtype		lam_mpi_float, lam_mpi_double;
struct _dtype		lam_mpi_long_double;
struct _dtype		lam_mpi_cplex, lam_mpi_packed;
struct _dtype		lam_mpi_unsigned_char;
struct _dtype		lam_mpi_unsigned_short;
struct _dtype		lam_mpi_unsigned;
struct _dtype		lam_mpi_unsigned_long;
struct _dtype		lam_mpi_ub, lam_mpi_lb;
struct _dtype		lam_mpi_float_int, lam_mpi_double_int;
struct _dtype		lam_mpi_long_int, lam_mpi_2int;
struct _dtype		lam_mpi_2float, lam_mpi_2double;
struct _dtype		lam_mpi_short_int;
struct _errhdl		lam_mpi_errors_are_fatal;
struct _errhdl		lam_mpi_errors_return;
struct _proc		*__lam_myproc_;
int			__lam_f77init_ = 0;
int			__lam_c2c_ = 0;
int			__lam_ger_ = 0;
int			__lam_homog_ = 0;
LIST			*__lam_comms_ = 0;
float8			__lam_clock_skew_ = 0.0;

/*
 * static variables
 */
static int		maxtag;			/* max. tag attribute */
static int		host;			/* host rank attribute */
static int		io;			/* I/O rank attribute */
static int		wtimeglob;		/* global time attribute */
static int		mpi_nprocs;		/* # world processes */
static int		mpi_nparent;		/* # parent processes */
static int		mpi_cid;		/* CID of parent intercomm */
static struct _gps	*mpi_procs;		/* array world & parent GPS */

/*
 * external functions
 */
extern struct _proc *	__lam_procfind_();
extern void		__lam_resetfunc_();
extern void		__lam_setcid_();
extern void		__lam_setfunc_();
extern int		__lam_errfunc_();
extern int		__lam_linit_();
extern int		__lam_mkerr_();
extern int		__lam_trcomm_();
extern int		__lam_trinit_();
extern int		__lam_tron_();

/*
 * local functions
 */
static int		init_comm();
static int		init_intercomm();
static int		init_env();
static int		init_rdtype();
static void		init_dtype();
static void		init_errhdl();
static void		init_op();
static void		make_dtype();
static void		make_op();
static int		comm_cmp();

/*
 * local structures for maxloc/minloc operations
 */
struct flt_int {
	float		fi_float;
	int		fi_int;
};

struct dbl_int {
	double		di_double;
	int		di_int;
};

struct long_int {
	long		li_long;
	int		li_int;
};

struct short_int {
	short		li_short;
	int		li_int;
};

/*
 * Create the whole MPI universe.
 */
int
MPI_Init(pargc, pargv)

int			*pargc;
char			***pargv;

{
	int		fl_init;		/* already init'ed? */
	int		err;			/* return error code */
	char		*name;			/* program name */

	__lam_setfunc_(BLKMPIINIT);
/*
 * Check if we have been initialized.
 */
	MPI_Initialized(&fl_init);
	if (fl_init) {
		return(__lam_errfunc_((MPI_Comm) 0,
			BLKMPIINIT, __lam_mkerr_(MPI_ERR_OTHER, EMPIINIT)));
	}
/*
 * Initialize LAM.
 */
	name = (pargc && pargv && (*pargc > 0)) ? **pargv : 0; 

	if (__lam_linit_(name, &mpi_nprocs, &mpi_nparent, &mpi_cid,
			&mpi_procs)) {
		terror("MPI_Init: LAM error");
		exit(errno);
	}
/*
 * Initialize MPI pre-defined "stuff".
 * The order below is important, so don't "hack".
 */
	init_errhdl();
	init_dtype();
	init_op();

	if (init_comm()) {
		free((char *) mpi_procs);
		return(__lam_errfunc_(MPI_COMM_NULL,
			BLKMPIINIT, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	if (init_intercomm()) {
		free((char *) mpi_procs);
		return(__lam_errfunc_(MPI_COMM_NULL,
			BLKMPIINIT, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	if (init_rdtype() || init_env()) {
		return(__lam_errfunc_(MPI_COMM_NULL,
			BLKMPIINIT, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * initialize profiling package
 */
	MPI_Pcontrol(1);
/*
 * cleanup
 */
	free((char *) mpi_procs);
/*
 * Make sure everyone else has also initialized.
 * Rank 0 of a spawned world must inform the root parent when this is so.
 */
	err = MPI_Barrier(MPI_COMM_WORLD);
	if (err != MPI_SUCCESS) return(LAMERROR);
		
	if ((mpi_nparent > 0) && (MPI_COMM_WORLD->c_group->g_myrank == 0)) {
		err = MPI_Send((void *) 0, 0, MPI_BYTE, 0, 0,
				lam_mpi_comm_parent);
		if (err != MPI_SUCCESS) return(LAMERROR);
	}
/*
 * Record an initialization trace.
 * If we are tracing and trace generation is initially to be on, turn it on.
 */
	if (__lam_trinit_(name, __lam_clock_skew_)) return(LAMERROR);

	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) {
		_kio.ki_rtf &= ~RTF_TRSWITCH;
		if (__lam_tron_() < 0) return(LAMERROR);
	}

	__lam_resetfunc_(BLKMPIINIT);
	return(MPI_SUCCESS);
}

/*
 *	init_errhdl
 *
 *	Function:	- initialize pre-defined error handles
 */
static void
init_errhdl()

{
	MPI_ERRORS_ARE_FATAL->eh_func = __lam_errfatal_;
	MPI_ERRORS_ARE_FATAL->eh_f77hdl = -1;
	MPI_ERRORS_ARE_FATAL->eh_refcount = 1;

	MPI_ERRORS_RETURN->eh_func = __lam_errreturn_;
	MPI_ERRORS_RETURN->eh_f77hdl = -1;
	MPI_ERRORS_RETURN->eh_refcount = 1;
}

/*
 *	init_comm
 *
 *	Function:	- initialize pre-defined communicators
 *	Returns:	- 0 or LAMERROR
 */
static int
init_comm()

{
	MPI_Group	group;			/* a process group */
	MPI_Comm	comm;			/* a communicator */
	int		i;			/* favourite index */
	struct _gps	*p;			/* favourite pointer */
	struct _proc	**pp;			/* another pointer */
/*
 * Initialize the list of communicators.
 */
	__lam_comms_ = al_init(sizeof(MPI_Comm), comm_cmp);
	if (__lam_comms_ == 0) return(LAMERROR);
/*
 * Create the "world" communicator.
 */
	group = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(mpi_nprocs * sizeof(struct _proc *)));
	if (group == 0) return(LAMERROR);

	group->g_nprocs = mpi_nprocs;
	group->g_myrank = MPI_UNDEFINED;
	group->g_refcount = 1;
	group->g_procs = (struct _proc **)
				((char *) group + sizeof(struct _group));
/*
 * Fill the group processes and determine my rank.
 */
	p = mpi_procs;
	pp = group->g_procs;

	for (i = 0; i < mpi_nprocs; ++i, ++p, ++pp) {

		*pp = __lam_procfind_(p);
		if (*pp == 0) return(LAMERROR);

		if (*pp == __lam_myproc_) {
			group->g_myrank = i;
		}
	}

	MPI_COMM_WORLD->c_inter = 0;
	MPI_COMM_WORLD->c_refcount = 1;
	MPI_COMM_WORLD->c_cube_dim = __lam_cubedim_(mpi_nprocs);
	MPI_COMM_WORLD->c_group = group;
	MPI_COMM_WORLD->c_rgroup = MPI_GROUP_NULL;
	MPI_COMM_WORLD->c_contextid = 0;
	MPI_COMM_WORLD->c_errhdl = MPI_ERRORS_ARE_FATAL;
	MPI_COMM_WORLD->c_keys = 0;
	MPI_COMM_WORLD->c_topo_type = MPI_UNDEFINED;
	MPI_COMM_WORLD->c_topo_nprocs = 0;
	MPI_COMM_WORLD->c_topo_ndims = 0;
	MPI_COMM_WORLD->c_topo_dims = 0;
	MPI_COMM_WORLD->c_topo_index = 0;
	MPI_COMM_WORLD->c_topo_nedges = 0;
	MPI_COMM_WORLD->c_topo_edges = 0;
	MPI_COMM_WORLD->c_status = 0;

	comm = MPI_COMM_WORLD;
	if (! al_insert(__lam_comms_, &comm)) return(LAMERROR);
	__lam_setcid_(MPI_COMM_WORLD->c_contextid);
/*
 * Set the pid under which run time traces are stored and write the
 * trace for MPI_COMM_WORLD.
 */
	_kio.ki_tid = __lam_myproc_->p_gps.gps_pid;
	if (__lam_trcomm_(MPI_COMM_WORLD)) return(LAMERROR);
/*
 * Create the "self" communicator.
 */
	group = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
						sizeof(struct _proc *));
	if (group == 0) return(LAMERROR);

	group->g_nprocs = 1;
	group->g_myrank = 0;
	group->g_refcount = 1;
	group->g_procs = (struct _proc **)
				((char *) group + sizeof(struct _group));

	group->g_procs[0] = __lam_myproc_;

	MPI_COMM_SELF->c_inter = 0;
	MPI_COMM_SELF->c_refcount = 1;
	MPI_COMM_SELF->c_cube_dim = __lam_cubedim_(group->g_nprocs);
	MPI_COMM_SELF->c_group = group;
	MPI_COMM_SELF->c_rgroup = MPI_GROUP_NULL;
	MPI_COMM_SELF->c_contextid = 1;
	MPI_COMM_SELF->c_errhdl = MPI_ERRORS_ARE_FATAL;
	MPI_COMM_SELF->c_keys = 0;
	MPI_COMM_SELF->c_topo_type = MPI_UNDEFINED;
	MPI_COMM_SELF->c_topo_nprocs = 0;
	MPI_COMM_SELF->c_topo_ndims = 0;
	MPI_COMM_SELF->c_topo_dims = 0;
	MPI_COMM_SELF->c_topo_index = 0;
	MPI_COMM_SELF->c_topo_nedges = 0;
	MPI_COMM_SELF->c_topo_edges = 0;
	MPI_COMM_SELF->c_status = 0;

	comm = MPI_COMM_SELF;
	if (! al_insert(__lam_comms_, &comm)) return(LAMERROR);
	__lam_setcid_(MPI_COMM_SELF->c_contextid);
/*
 * Create the empty group.
 */
	MPI_GROUP_EMPTY->g_nprocs = 0;
	MPI_GROUP_EMPTY->g_myrank = MPI_UNDEFINED;
	MPI_GROUP_EMPTY->g_refcount = 1;
	MPI_GROUP_EMPTY->g_procs = 0;

	return(0);
}

/*
 *	init_intercomm
 *
 *	Function:	- initialize intercomm to parents (if any)
 *	Returns:	- 0 or LAMERROR
 */
static int
init_intercomm()

{
	MPI_Group	group;			/* a process group */
	int		i;			/* favourite index */
	struct _gps	*p;			/* favourite pointer */
	struct _proc	**pp;			/* another pointer */
/*
 * Create the intercomm to the parent world if it exists.
 */
	if (mpi_nparent <= 0) {
		lam_mpi_comm_parent = MPI_COMM_NULL;
		return(0);
	}

	group = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
				(mpi_nparent * sizeof(struct _proc *)));
	if (group == 0) return(LAMERROR);

	lam_mpi_comm_parent = (MPI_Comm) malloc(sizeof(struct _comm));
	if (lam_mpi_comm_parent == 0) {
		free((char *) group);
		return(LAMERROR);
	}
	
	group->g_nprocs = mpi_nparent;
	group->g_myrank = MPI_UNDEFINED;
	group->g_refcount = 1;
	group->g_procs = (struct _proc **)
				((char *) group + sizeof(struct _group));
/*
 * Fill the parent group processes.
 */
	p = mpi_procs + mpi_nprocs;
	pp = group->g_procs;

	for (i = 0; i < mpi_nparent; ++i, ++p, ++pp) {
		*pp = __lam_procfind_(p);
		if (*pp == 0) return(LAMERROR);
	}

	lam_mpi_comm_parent->c_inter = 1;
	lam_mpi_comm_parent->c_refcount = 1;
	lam_mpi_comm_parent->c_cube_dim = __lam_cubedim_(mpi_nprocs);
	lam_mpi_comm_parent->c_group = MPI_COMM_WORLD->c_group;
	++(MPI_COMM_WORLD->c_group->g_refcount);
	lam_mpi_comm_parent->c_rgroup = group;
	lam_mpi_comm_parent->c_contextid = mpi_cid;
	lam_mpi_comm_parent->c_errhdl = MPI_ERRORS_ARE_FATAL;
	lam_mpi_comm_parent->c_keys = 0;
	lam_mpi_comm_parent->c_topo_type = MPI_UNDEFINED;
	lam_mpi_comm_parent->c_topo_nprocs = 0;
	lam_mpi_comm_parent->c_topo_ndims = 0;
	lam_mpi_comm_parent->c_topo_dims = 0;
	lam_mpi_comm_parent->c_topo_index = 0;
	lam_mpi_comm_parent->c_topo_nedges = 0;
	lam_mpi_comm_parent->c_topo_edges = 0;
	lam_mpi_comm_parent->c_status = 0;

	if (! al_insert(__lam_comms_, &lam_mpi_comm_parent)) return(LAMERROR);
	__lam_setcid_(lam_mpi_comm_parent->c_contextid);

	if (__lam_trcomm_(lam_mpi_comm_parent)) {
		--(MPI_COMM_WORLD->c_group->g_refcount);
		free((char *) group);
		free((char *) lam_mpi_comm_parent);
		return(LAMERROR);
	}
	
	return(0);
}

/*
 *	init_dtype
 *
 *	Function:	- initialize basic (intrinsic) datatypes
 */
static void
init_dtype()

{
	make_dtype(MPI_CHAR, sizeof(char), TRDTCHAR);
	make_dtype(MPI_BYTE, sizeof(char), TRDTBYTE);
	make_dtype(MPI_PACKED, sizeof(char), TRDTPACKED);
	make_dtype(MPI_SHORT, sizeof(short), TRDTSHORT);
	make_dtype(MPI_INT, sizeof(int), TRDTINT);
	make_dtype(MPI_LONG, sizeof(long), TRDTLONG);
	make_dtype(MPI_FLOAT, sizeof(float), TRDTFLOAT);
	make_dtype(MPI_DOUBLE, sizeof(double), TRDTDOUBLE);
	make_dtype(MPI_LONG_DOUBLE, 2 * sizeof(double), TRDTLONGDOUBLE);
	make_dtype(MPI_CPLEX, 2 * sizeof(float), TRDTCOMPLEX);
	make_dtype(MPI_LOGIC, sizeof(int), TRDTLOGIC);
	make_dtype(MPI_UNSIGNED_CHAR, sizeof(unsigned char), TRDTUCHAR);
	make_dtype(MPI_UNSIGNED_SHORT, sizeof(unsigned short), TRDTUSHORT);
	make_dtype(MPI_UNSIGNED, sizeof(unsigned int), TRDTUINT);
	make_dtype(MPI_UNSIGNED_LONG, sizeof(unsigned long), TRDTULONG);

	make_dtype(MPI_UB, 0, TRDTUB);
	MPI_UB->dt_nelem = 0;

	make_dtype(MPI_LB, 0, TRDTLB);
	MPI_LB->dt_nelem = 0;
}

/*
 *	init_rdtype
 *
 *	Function:	- initialize the reduction datatypes
 *	Returns:	- 0 or LAMERROR
 */
static int
init_rdtype()

{
	MPI_Datatype	new;			/* new datatype */
	MPI_Datatype	types[4];		/* struct datatypes */
	int		lengths[4];		/* struct lengths */
	MPI_Aint	disps[4];		/* struct displacements */
	struct flt_int	fi[2];			/* float_int data */
	struct dbl_int	di[2];			/* double_int data */
	struct long_int	li[2];			/* long_int data */
	struct short_int
			si[2];			/* short_int data */
	int		err;			/* error code */
/*
 * Create MPI_2INT.
 */
	err = MPI_Type_contiguous(2, MPI_INT, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_2INT, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_2INT->dt_label = TRDT2INT;
	MPI_2INT->dt_commit = 1;
/*
 * Create MPI_2FLOAT.
 */
	err = MPI_Type_contiguous(2, MPI_FLOAT, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_2FLOAT, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_2FLOAT->dt_label = TRDT2FLOAT;
	MPI_2FLOAT->dt_commit = 1;
/*
 * Create MPI_2DOUBLE.
 */
	err = MPI_Type_contiguous(2, MPI_DOUBLE, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_2DOUBLE, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_2DOUBLE->dt_label = TRDT2DOUBLE;
	MPI_2DOUBLE->dt_commit = 1;
/*
 * Create MPI_FLOAT_INT.
 */
	types[0] = MPI_FLOAT; types[1] = MPI_INT; types[2] = MPI_UB;
	lengths[0] = 1; lengths[1] = 1; lengths[2] = 1;

	MPI_Address(&fi[0], &disps[0]);
	MPI_Address(&(fi[0].fi_int), &disps[1]);
	MPI_Address(&fi[1], &disps[2]);

	disps[2] -= disps[0]; disps[1] -= disps[0]; disps[0] = 0;

	err = MPI_Type_struct(3, lengths, disps, types, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_FLOAT_INT, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_FLOAT_INT->dt_label = TRDTFLOATINT;
	MPI_FLOAT_INT->dt_commit = 1;
/*
 * Create MPI_DOUBLE_INT.
 */
	types[0] = MPI_DOUBLE; types[1] = MPI_INT; types[2] = MPI_UB;
	lengths[0] = 1; lengths[1] = 1; lengths[2] = 1;

	MPI_Address(&di[0], &disps[0]);
	MPI_Address(&(di[0].di_int), &disps[1]);
	MPI_Address(&di[1], &disps[2]);

	disps[2] -= disps[0]; disps[1] -= disps[0]; disps[0] = 0;

	err = MPI_Type_struct(3, lengths, disps, types, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_DOUBLE_INT, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_DOUBLE_INT->dt_label = TRDTDOUBLEINT;
	MPI_DOUBLE_INT->dt_commit = 1;
/*
 * Create MPI_LONG_INT.
 */
	types[0] = MPI_LONG; types[1] = MPI_INT; types[2] = MPI_UB;
	lengths[0] = 1; lengths[1] = 1; lengths[2] = 1;

	MPI_Address(&li[0], &disps[0]);
	MPI_Address(&(li[0].li_int), &disps[1]);
	MPI_Address(&li[1], &disps[2]);

	disps[2] -= disps[0]; disps[1] -= disps[0]; disps[0] = 0;

	err = MPI_Type_struct(3, lengths, disps, types, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_LONG_INT, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_LONG_INT->dt_label = TRDTLONGINT;
	MPI_LONG_INT->dt_commit = 1;
/*
 * Create MPI_SHORT_INT.
 */
	types[0] = MPI_SHORT; types[1] = MPI_INT; types[2] = MPI_UB;
	lengths[0] = 1; lengths[1] = 1; lengths[2] = 1;

	MPI_Address(&si[0], &disps[0]);
	MPI_Address(&(si[0].li_int), &disps[1]);
	MPI_Address(&si[1], &disps[2]);

	disps[2] -= disps[0]; disps[1] -= disps[0]; disps[0] = 0;

	err = MPI_Type_struct(3, lengths, disps, types, &new);
	if (err != MPI_SUCCESS) return(LAMERROR);

	memcpy((char *) MPI_SHORT_INT, (char *) new, sizeof(struct _dtype));
	free((char *) new);
	MPI_SHORT_INT->dt_label = TRDTSHORTINT;
	MPI_SHORT_INT->dt_commit = 1;

	return(0);
}

/*
 *	init_op
 *
 *	Function:	- initialize intrinsic reduction operations
 */
static void
init_op()

{
	make_op(MPI_MAX, __lam_max_);
	make_op(MPI_MIN, __lam_min_);
	make_op(MPI_SUM, __lam_sum_);
	make_op(MPI_PROD, __lam_prod_);
	make_op(MPI_LAND, __lam_land_);
	make_op(MPI_BAND, __lam_band_);
	make_op(MPI_LOR, __lam_lor_);
	make_op(MPI_BOR, __lam_bor_);
	make_op(MPI_LXOR, __lam_lxor_);
	make_op(MPI_BXOR, __lam_bxor_);
	make_op(MPI_MAXLOC, __lam_maxloc_);
	make_op(MPI_MINLOC, __lam_minloc_);
}

/*
 *	init_env
 *
 *	Function:	- initialize environment attributes
 *	Returns:	- 0 or LAMERROR
 */
static int
init_env()

{
	int		key;			/* attribute key */
	int		err;			/* error code */
	int		i;			/* favourite index */
	void		*null = 0;		/* junk pointer */
	struct _proc	**p;			/* favourite pointer */
	MPI_Group	world;			/* world group */
/*
 * Create the predefined keys.
 */
	err = MPI_Keyval_create(MPI_NULL_COPY_FN, MPI_NULL_DELETE_FN,
				&key, null);
	if ((err != MPI_SUCCESS) || (key != MPI_TAG_UB)) return(LAMERROR);

	err = MPI_Keyval_create(MPI_NULL_COPY_FN, MPI_NULL_DELETE_FN,
				&key, null);
	if ((err != MPI_SUCCESS) || (key != MPI_HOST)) return(LAMERROR);

	err = MPI_Keyval_create(MPI_NULL_COPY_FN, MPI_NULL_DELETE_FN,
				&key, null);
	if ((err != MPI_SUCCESS) || (key != MPI_IO)) return(LAMERROR);

	err = MPI_Keyval_create(MPI_NULL_COPY_FN, MPI_NULL_DELETE_FN,
				&key, null);
	if ((err != MPI_SUCCESS) || (key != MPI_WTIME_IS_GLOBAL)) {
		return(LAMERROR);
	}
/*
 * Initialize the attribute values.
 */
	maxtag = 32767;
	wtimeglob = 0;
/*
 * We interpret MPI_IO to include the ability to print to a terminal.
 * MPI_IO is set to my rank if I'm on the origin node.
 * Otherwise, it's set to MPI_ANY_SOURCE if all processes are on
 * the origin node, MPI_PROC_NULL if no processes are on the
 * origin node or the lowest ranked process on the origin node.
 */
	world = MPI_COMM_WORLD->c_group;
	p = world->g_procs;

	for (i = 0; i < world->g_nprocs; ++i) {
		if ((*p)->p_gps.gps_node != getorigin()) break;
		p++;
	}

	if (i >= world->g_nprocs) {
		io = MPI_ANY_SOURCE;
	} else if (__lam_myproc_->p_gps.gps_node == getorigin()) {
		io = world->g_myrank;
	} else {
		p = world->g_procs;

		for (i = 0; i < world->g_nprocs; ++i) {
			if ((*p)->p_gps.gps_node == getorigin()) break;
			p++;
		}

		io = (i >= world->g_nprocs) ? MPI_PROC_NULL : i;
	}
/*
 * MPI_HOST is similarly related to the origin node, except if all
 * processes are on the origin, we set this attribute to rank 0.
 */
	p = world->g_procs;

	for (i = 0; i < world->g_nprocs; ++i) {
		if ((*p)->p_gps.gps_node == getorigin()) break;
		p++;
	}

	host = (i >= world->g_nprocs) ? MPI_PROC_NULL : i;
/*
 * Store the attribute values.
 */
	err = MPI_Attr_put(MPI_COMM_WORLD, MPI_TAG_UB,
		((__lam_f77init_) ? (void *) maxtag : (void *) &maxtag));
	if (err != MPI_SUCCESS) return(LAMERROR);

	err = MPI_Attr_put(MPI_COMM_WORLD, MPI_HOST,
		((__lam_f77init_) ? (void *) host : (void *) &host));
	if (err != MPI_SUCCESS) return(LAMERROR);

	err = MPI_Attr_put(MPI_COMM_WORLD, MPI_IO,
		((__lam_f77init_) ? (void *) io : (void *) &io));
	if (err != MPI_SUCCESS) return(LAMERROR);

	err = MPI_Attr_put(MPI_COMM_WORLD, MPI_WTIME_IS_GLOBAL,
		((__lam_f77init_) ? (void *) wtimeglob : (void *) &wtimeglob));
	if (err != MPI_SUCCESS) return(LAMERROR);

	return(0);
}

/*
 *	make_dtype
 *
 *	Function:	- create a basic datatype
 *			- no errors happen here
 *	Accepts:	- MPI datatype
 *			- size of datatype
 *			- datatype label
 */
static void
make_dtype(dtype, size, label)

MPI_Datatype		dtype;
int			size;
int			label;

{
	dtype->dt_format = LAM_DTBASIC;
	dtype->dt_commit = 1;
	dtype->dt_label = label;
	dtype->dt_refcount = 1;
	dtype->dt_upper = size;
	dtype->dt_lower = 0;
	dtype->dt_size = size;
	dtype->dt_usrsize = size;
	dtype->dt_nelem = 1;
	dtype->dt_count = 0;
	dtype->dt_length = 0;
	dtype->dt_stride = 0;
	dtype->dt_dtype = 0;
	dtype->dt_lengths = 0;
	dtype->dt_disps = 0;
	dtype->dt_dtypes = 0;
}

/*
 *	make_op
 *
 *	Function:	- create intrinsic reduction operation
 *			- no errors happen here
 *	Accepts:	- MPI reduction operation
 *			- reduction function
 */
static void
make_op(op, func)

MPI_Op			op;
void			(*func)();

{
	op->op_func = func;
	op->op_commute = 1;
	op->op_f77dtype = -1;
}

/*
 *	comm_cmp
 *
 *	Function:	- compare two communicators
 *	Accepts:	- ptr to two entries
 *	Returns:	- 0 if same communicator, else 1
 */
static int
comm_cmp(c1, c2)

MPI_Comm		*c1, *c2;

{
	return( !( ((*c1)->c_contextid == (*c2)->c_contextid)
			|| ((*c1)->c_contextid == -((*c2)->c_contextid + 1))));
}
