/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	iprobe.c,v $
 * Revision 6.0  96/02/29  14:12:53  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:23:00  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:33  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:06  gdburns
 * Ohio Release
 * 
 *	Function:	- non-blocking check of incoming messages
 *			- message is not received
 *	Accepts:	- source rank
 *			- message tag
 *			- communicator
 *			- flag (returned value)
 *			- status (returned value)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>
#include <terror.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern void		_mpi_req_add();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();
extern void		_mpi_req_rem();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();
extern int		__lam_trincff_();
extern int		_mpi_req_advance();
extern int		_mpi_req_build();
extern int		_mpi_req_destroy();
extern int		_mpi_req_start();
extern int		_mpi_req_end();
extern int		_rpi_c2c_iprobe();
extern int		_rpi_lamd_iprobe();

int
MPI_Iprobe(src, tag, comm, flag, stat)

int			src;
int			tag;
MPI_Comm		comm;
int			*flag;
MPI_Status		*stat;

{
	struct _req	request_storage;	/* request structure */
	MPI_Request	req;			/* request */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	int		r;
	double		startt;			/* start time */

	__lam_initerr_();
/*
 * Check additional arguments.
 */
	if ((flag == 0) || (stat == 0)) {
		return(__lam_errfunc_(comm, BLKMPIIPROBE,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Handle the trivial case.
 */
	if (src == MPI_PROC_NULL) {
		stat->MPI_ERROR = MPI_SUCCESS;
		stat->MPI_SOURCE = MPI_PROC_NULL;
		stat->MPI_TAG = MPI_ANY_TAG;
		stat->st_count = 0;
		stat->st_nelem = 0;
		stat->st_length = 0;
		*flag = 1;
		return(MPI_SUCCESS);
	}

	__lam_setfunc_(BLKMPIIPROBE);
/*
 * tracing
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) startt = MPI_Wtime();
/*
 * Advance the system to give the probe a better chance of success.
 */
	_mpi_req_blkclr();

	err = _mpi_req_advance();
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIIPROBE, err));
	}

	req = &request_storage;
	err = _mpi_req_build((char *) 0, 0, MPI_BYTE, src, tag, comm,
			LAM_RQIPROBE, &req);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIIPROBE, err));
	}

	err = _mpi_req_start(req);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIIPROBE, err));
	}

	r = RPI_SPLIT(_rpi_lamd_iprobe, _rpi_c2c_iprobe, (req));
	if (r < 0) {
		return(__lam_errfunc_(comm, BLKMPIIPROBE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	err = _mpi_req_end(req);
	if (err != MPI_SUCCESS)
			return(__lam_errfunc_(comm, BLKMPIIPROBE, err));
/*
 * Copy the status info if synchronized.
 */
	if (r == 1) {
		*flag = 1;
		*stat = req->rq_status;
	} else {
		*flag = 0;
	}

	err = _mpi_req_destroy(&req);
	if (err != MPI_SUCCESS)
			return(__lam_errfunc_(comm, BLKMPIIPROBE, err));
/*
 * Generate a run time trace.
 */
	if (fl_trace) {
		__lam_trmsg_(TRTNOIO, startt, LAM_S2US(MPI_Wtime() - startt),
				0, src, tag, comm, 0, 0, 0, 0, 0,
				LAM_RQIPROBE);
	}

	__lam_resetfunc_(BLKMPIIPROBE);
	return(MPI_SUCCESS);
}
