/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	irecv.c,v $
 * Revision 6.0  96/02/29  14:12:54  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:23:02  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:34  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:08  gdburns
 * Ohio Release
 * 
 *	Function:	- immediate receive
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- source
 *			- tag
 *			- communicator
 *			- ptr to request (returned value)
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern void		_mpi_req_add();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();
extern int		__lam_errfunc_();
extern int		__lam_trincff_();
extern int		_mpi_req_advance();
extern int		_mpi_req_build();
extern int		_mpi_req_end();

int
MPI_Irecv(buf, count, dtype, src, tag, comm, req)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			src;
int			tag;
MPI_Comm		comm;
MPI_Request		*req;

{
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIIRECV);
/*
 * Setup for tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) startt = MPI_Wtime();

	*req = MPI_REQUEST_NULL;
	err = _mpi_req_build(buf, count, dtype, src, tag, comm,
			LAM_RQIRECV, req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIIRECV, err));

	err = _mpi_req_start(*req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIIRECV, err));

	_mpi_req_add(*req);
	_mpi_req_blkclr();

	err = _mpi_req_advance();
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIIRECV, err));
/*
 * Generate a run time trace.
 */
	if (fl_trace) {
		__lam_trmsg_(TRTNOIO, startt, LAM_S2US(MPI_Wtime() - startt),
				0, src, tag, comm, dtype, count, 0, 0,
				(*req)->rq_seq, LAM_RQIRECV);
	}

	__lam_resetfunc_(BLKMPIIRECV);
	return(MPI_SUCCESS);
}
