/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	isend.c,v $
 * Revision 6.0  96/02/29  14:12:56  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:23:31  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:35  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:11  gdburns
 * Ohio Release
 * 
 *	Function:	- immediate send
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- destination
 *			- tag
 *			- communicator
 *			- request (out)
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern int		__lam_errfunc_();
extern int		__lam_isend_();
extern int		__lam_trincff_();

int
MPI_Isend(buf, count, dtype, dest, tag, comm, req)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			dest;
int			tag;
MPI_Comm		comm;
MPI_Request		*req;

{
	int		err;
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIISEND);
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) startt = MPI_Wtime();
/*
 * Start the send.
 */
	err = __lam_isend_(buf, count, dtype, dest, tag, comm, req,
			LAM_RQISEND);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIISEND, err));
/*
 * Generate a run time trace.
 */
	if (fl_trace) {
		finisht = MPI_Wtime();

		__lam_trmsg_(TRTOUTPUT, startt, LAM_S2US(finisht - startt), 0,
				dest, tag, comm, dtype, count, 0, 0, 
				(*req)->rq_seq, LAM_RQISEND);
	}

	__lam_resetfunc_(BLKMPIISEND);
	return(MPI_SUCCESS);
}
