/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	lamattr.c,v $
 * Revision 6.0  96/02/29  14:12:59  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:23:36  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/26  17:23:57  raja
 * Delete key entry without freeing the key.
 * 
 * Revision 5.2  94/08/22  14:04:38  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:14  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:30  gdburns
 * new for 5.1
 * 
 *	Function:	- manage table of attribute keys
 */

#include <stdlib.h>
#include <string.h>

#include <all_hash.h>
#include <mpi.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>

#define DEFATTRSIZE	8
#define DEFKEYSIZE	11			/* prime number */

/*
 * external functions
 */
extern int4		next_prime();

/*
 * static variables
 */
static int		attrcount = 0;
static int		attrsize = 0;
static struct _attrkey	*attrs = 0;

/*
 *	__lam_mkkey_
 *
 *	Function:	- create a new attribute key
 *	Returns:	- key or ERROR
 */
int
__lam_mkkey_()

{
	struct _attrkey	*p;
	int		i;
/*
 * Allocate initial attribute array if needed.
 * Assign first position.
 */
	if (attrsize == 0) {
		i = DEFATTRSIZE * sizeof(struct _attrkey);
		attrs = (struct _attrkey *) malloc((unsigned) i);
		if (attrs == 0) {
			return(ERROR);
		}

		memset((char *) attrs, 0, i);
		attrsize = DEFATTRSIZE;
		p = attrs;
		i = 0;
	}
/*
 * Expand attribute array if needed.
 * Assign first position in the expanded side.
 */
	else if (attrsize == attrcount) {
		i = attrsize * sizeof(struct _attrkey);
		p = (struct _attrkey *) realloc((char *)attrs, (unsigned) i+i);
		if (p == 0) {
			return(ERROR);
		}

		attrs = p;
		p += attrsize;
		memset((char *) p, 0, i);
		i = attrsize;
		attrsize += attrsize;
	}
/*
 * Find an empty slot and assign it.
 */
	else {
		for (i = 0, p = attrs; i < attrsize; ++i, ++p) {
			if (p->ak_refcount == 0) {
				break;
			}
		}

		if (i == attrsize) {
			errno = EIMPOSSIBLE;
			return(ERROR);
		}
	}
/*
 * Reserve the slot.
 */
	++attrcount;
	p->ak_refcount = 1;
	return(i);
}

/*
 *	__lam_getattr_
 *
 *	Function:	- get ptr to attribute
 *	Accepts:	- attribute key
 *	Returns:	- ptr to attribute or NULL
 */
struct _attrkey *
__lam_getattr_(key)

int			key;

{
	struct _attrkey	*p;

	errno = EINVAL;

	if ((key >= 0) && (key < attrsize)) {
		p = attrs + key;
		if (p->ak_refcount == 0) {
			p = 0;
		}
	} else {
		p = 0;
	}

	return(p);
}

/*
 *	__lam_freekey_
 *
 *	Function:	- free attribute key
 *	Accepts:	- attribute key
 *	Returns:	- 0 or ERROR
 */
int
__lam_freekey_(key)

int			key;

{
	struct _attrkey	*p;

	p = __lam_getattr_(key);
	if (p == 0) {
		errno = EINVAL;
		return(ERROR);
	}
		
	if (p->ak_refcount == 0) {
		errno = EINVAL;
		return(ERROR);
	} else if (p->ak_refcount < 0) {
/*
 * Do nothing, this is a non-erasable attribute.
 */
	} else if (--(p->ak_refcount) == 0) {
		--attrcount;
	}

	return(0);
}

/*
 *	__lam_putkey_
 *
 *	Function:	- attach a key to a communicator
 *	Accepts:	- communicator
 *			- key
 *	Returns:	- 0 or ERROR
 */
int
__lam_putkey_(comm, key)

MPI_Comm		comm;
int			key;

{
	struct _attr	elem;
	struct _attrkey	*p;
/*
 * Check if key is valid.
 */
	p = __lam_getattr_(key);
	if (p == 0) {
		return(ERROR);
	}
/*
 * Allocate the key hash table if needed.
 */
	if (comm->c_keys == 0) {
		comm->c_keys = ah_init(DEFKEYSIZE, (int4) sizeof(struct _attr),
					(int4) MPI_KEYVAL_INVALID, INT4_NIL);
		if (comm->c_keys == 0) {
			return(ERROR);
		}
	}
/*
 * Expand the key hash table if needed.
 */
	else if (ah_count(comm->c_keys) == ah_size(comm->c_keys)) {
		if (ah_expand(comm->c_keys,
				next_prime(1 + ah_size(comm->c_keys)))) {
			return(ERROR);
		}
	}
/*
 * Store new key in hash table.
 */
	elem.a_key = key;
	elem.a_value = 0;

	if (ah_insert(comm->c_keys, &elem)) {
		return(ERROR);
	}

	if (p->ak_refcount > 0) {
		++(p->ak_refcount);
	}

	return(0);
}

/*
 *	__lam_getkey_
 *
 *	Function:	- get key entry in communicator
 *	Accepts:	- communicator
 *			- key
 *	Returns:	- ptr to entry of NULL
 */
struct _attr *
__lam_getkey_(comm, key)

MPI_Comm		comm;
int			key;

{
	struct _attr	*p;

	if (comm->c_keys) {
		p = (struct _attr *) ah_find(comm->c_keys, key);
	} else {
		p = 0;
	}

	return(p);
}

/*
 *	__lam_delkey_
 *
 *	Function:	- delete key from communicator
 *	Accepts:	- communicator
 *			- key
 *	Returns:	- 0 or ERROR
 */
int
__lam_delkey_(comm, key)

MPI_Comm		comm;
int			key;

{
	return(ah_delete(comm->c_keys, key));
}

/*
 *	__lam_nukekeys_
 *
 *	Function:	- deallocate all keys (cleanup)
 */
void
__lam_nukekeys_()

{
	if (attrs) {
		free((char *) attrs);
		attrcount = attrsize = 0;
		attrs = 0;
	}
}
