/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	NJN
 *
 *	$Log:	lamclocks.c,v $
 * Revision 6.0  96/03/02  10:41:48  nevin
 * Ohio Release
 * 
 * 
 *	Function:	- synchronize clocks for tracing
 */

#include <stdlib.h>

#include <typical.h>
#include <net.h>

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * global functions
 */
int			__lam_clocksync_();

/*
 * local functions
 */
static int		sync_leaders();
static int		sync_master();
static int		sync_slave();

/*
 *	__lam_clocksync_
 *
 *	Function:	- world clock synchronizaton
 *	Accepts:	- number of processes in world
 *			- GPS of world
 *			- skew (returned value)
 *	Returns:	- 0 or LAMERROR
 */
int 
__lam_clocksync_(nworld, wgps, skew)

int4			nworld;
struct _gps		*wgps;
float8			*skew;

{
	char	        *msg;			/* ptr message data */
	struct nmsg	nhead;			/* network message header */
	struct _gps	*p;			/* favourite pointer */
	int4		*leaders;		/* array of leader indices */
	int		nleaders;		/* number of leaders */
	int4		mynode;			/* local node */
	int		i;			/* favourite index */
	int		j;			/* another index */
	int		done;			/* flag */
	
	*skew = 0.0;
	leaders = (int4 *) malloc(nworld * sizeof(int4));
	if (leaders == 0) return(LAMERROR);
/*
 * Make the list of leaders on which the world is running.
 */
	leaders[0] = 0;
	nleaders = 1;

	for (i = 1, p = wgps + 1; i < nworld; i++, p++) {

		for (j = 0, done = 0; j < nleaders && !done; j++)
			done = (wgps[leaders[j]].gps_node == p->gps_node);
	
		if ( ! done) leaders[nleaders++] = i;
	}
/*
 * Find the leader process on the local node.
 */
	mynode = getnodeid();

	for (i = 0, p = wgps; i < nworld && p->gps_node != mynode; i++, p++)
		;

	if (i >= nworld) return(LAMERROR);
/*
 * Set up local skew message header.
 */
	msg = (char *) (&nhead.nh_data[0]);
	nhead.nh_node = LOCAL;
	nhead.nh_type = BLKMPIINIT;
	nhead.nh_flags = DFLT8DATA;
	nhead.nh_length = 0;
/*
 * If I am the leader on the local node enter the clock synchronization
 * ping-pong and then send the result to the other processes on the
 * local node.  Otherwise I wait for the skew to be sent from the
 * leader.
 */
	if (p->gps_pid == __lam_myproc_->p_gps.gps_pid) {

		if (sync_leaders(i, nleaders, leaders, wgps, skew))
			return(LAMERROR);

		memcpy(msg, (char *) skew, sizeof(float8));

		for (i++, p++; i < nworld; i++, p++) {
			if (p->gps_node == mynode) {
				nhead.nh_event = - p->gps_pid;
				if (nsend(&nhead)) return(LAMERROR);
			}
		}
	}
	else {
		nhead.nh_event = -(__lam_myproc_->p_gps.gps_pid);
		if (nrecv(&nhead)) return(LAMERROR);

		memcpy((char *) skew, msg, sizeof(float8));
	}

	return(0);
}

/*
 *	sync_leaders
 *
 *	Function:	- synchronize clock on leaders
 *	Accepts:	- my index in leaders array
 *			- number of leaders
 *			- array of leaders
 *			- GPS of world
 *			- clock skew (returned value)
 *	Returns:	- 0 or LAMERROR
 */
static int
sync_leaders(me, n, leaders, wgps, skew)

int			me;
int			n;
int			*leaders;
struct _gps		*wgps;
float8			*skew;

{
	int		i;			/* favourite index */

	if (me == 0) {
/*
 * I am the master leader.
 */
		*skew = 0.0;
		
		for (i = 1; i < n; i++) {
			if (sync_master(wgps, wgps + leaders[i], 10))
				return(LAMERROR);
		}
	}
	else {
/*
 * I am a slave leader.
 */
		if (sync_slave(wgps, wgps + me, skew, 10))
			return(LAMERROR);
	}

	return(0);
}

/*
 *	sync_master
 *
 *	Function:	- perform clock sync from master side
 *	Accepts:	- ptr master GPS
 *			- ptr slave GPS
 *			- number of iterations
 *	Returns:	- 0 or LAMERROR
 */
static int
sync_master(master, slave, iter)

struct _gps	*master;
struct _gps	*slave;
int		iter;

{
	float8 		local1;
	float8 		local2;
	float8		remote;
	float8		skew;
	char	        *imsg;
	char	        *omsg;
	struct nmsg	inmsg;
	struct nmsg	outmsg;
	int		i;
/*
 * Set up outgoing message header.
 */
	outmsg.nh_node = slave->gps_node;
	outmsg.nh_event = - slave->gps_pid;
	outmsg.nh_type = BLKMPIINIT;
	outmsg.nh_flags = DFLT8DATA;
	outmsg.nh_length = 0;
	omsg = (char *) (&outmsg.nh_data[0]);
/*
 * Set up incoming message header.
 */
	inmsg.nh_event = - master->gps_pid;
	inmsg.nh_type = BLKMPIINIT;
	inmsg.nh_flags = DFLT8DATA;
	inmsg.nh_length = 0;
	imsg = (char *) (&inmsg.nh_data[0]);
/*
 * Priming.
 */
	if (nsend(&outmsg)) return(LAMERROR);
	if (nrecv(&inmsg)) return(LAMERROR);
/*
 * Ping-pong clock times.
 */
	skew = 0.0;

	for (i = 0; i < iter; i++) {
		local1 = (float8) ttime();
	
		if (nsend(&outmsg)) return(LAMERROR);
	
		if (nrecv(&inmsg)) return(LAMERROR);

		local2 = (float8) ttime();
		memcpy((char *) &remote, imsg, sizeof(float8));

		skew += (local1 + local2) / 2.0  - remote;
	}
/*
 * Calculate and send delta.
 */
	skew /= (float8) iter;
	memcpy(omsg, (char *) &skew, sizeof(float8));

	if (nsend(&outmsg)) return(LAMERROR);

	return(0);
}

/*
 *	sync_slave
 *
 *	Function:	- perform clock sync from slave side
 *	Accepts:	- ptr master GPS
 *			- ptr slave GPS
 *			- ptr clock skew (returned value)
 *			- number of iterations
 *	Returns:	- 0 or LAMERROR
 */
static int
sync_slave(master, slave, skew, iter)

struct _gps	*master;
struct _gps	*slave;
float8		*skew;
int		iter;

{
	float8		local;
	char	        *imsg;
	char	        *omsg;
	struct nmsg	inmsg;
	struct nmsg	outmsg;
	int		i;
/*
 * Set up outgoing message header.
 */
	outmsg.nh_node = master->gps_node;
	outmsg.nh_event = - master->gps_pid;
	outmsg.nh_type = BLKMPIINIT;
	outmsg.nh_flags = DFLT8DATA;
	outmsg.nh_length = 0;
	omsg = (char *) (&outmsg.nh_data[0]);
/*
 * Set up incoming message header.
 */
	inmsg.nh_event = - slave->gps_pid;
	inmsg.nh_type = BLKMPIINIT;
	inmsg.nh_flags = DFLT8DATA;
	inmsg.nh_length = 0;
	imsg = (char *) (&inmsg.nh_data[0]);
/*
 * Priming.
 */
	if (nrecv(&inmsg)) return(LAMERROR);
	if (nsend(&outmsg)) return(LAMERROR);
/*
 * Ping-pong clock times.
 */
	for (i = 0; i < iter; i++) {

		if (nrecv(&inmsg)) return(LAMERROR);

		local = (float8) ttime();
		memcpy(omsg, (char *) &local, sizeof(float8));

		if (nsend(&outmsg)) return(LAMERROR);

	}

	if (nrecv(&inmsg)) return(LAMERROR);

	memcpy((char *) skew, imsg, sizeof(float8));
	
	return(0);
}
