/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	lamdeferr.c,v $
 * Revision 6.0  96/02/29  14:13:02  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/02/18  21:23:39  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.2  95/09/12  14:58:58  raja
 * Spell out "rank", "comm", "node", "pid".
 * 
 * Revision 5.2.1.1  95/08/25  16:05:40  raja
 * Send errors to tstdout instead of tstderr.
 * 
 * Revision 5.2  94/08/22  14:04:40  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:16  gdburns
 * Ohio Release
 * 
 *	Function:	- default MPI error handlers
 */

#include <unistd.h>

#include <mpi.h>
#include <tstdio.h>

/*
 * global functions
 */
void			__lam_errfatal_();
void			__lam_errreturn_();

/*
 * static variables
 */
static char		mpierrmsg[MPI_MAX_ERROR_STRING];

/*
 *	__lam_errfatal_
 *
 *	Function:	- default MPI error function
 *			- print error message and abort
 *	Accepts:	- ptr communicator
 *			- ptr error code
 */
void
__lam_errfatal_(pcomm, perror)

MPI_Comm		*pcomm;
int			*perror;

{
	int		rank;
	int		cid;
	int		len;
	MPI_Comm	comm;

	MPI_Error_string(*perror, mpierrmsg, &len);

	comm = (pcomm) ? *pcomm : MPI_COMM_WORLD;
	if (comm != MPI_COMM_NULL) {
		MPI_Comm_rank(comm, &rank);
		MPIL_Comm_id(comm, &cid);
		tprintf("%s (rank %d, comm %d)\n", mpierrmsg, rank, cid);
		MPI_Abort(comm, *perror);
	} else {
		tprintf("%s (node %d, pid %d)\n",
				mpierrmsg, getnodeid(), getpid());
		kexit(*perror);
	}
}

/*
 *	__lam_errreturn_
 *
 *	Function:	- predefined error function
 *			- stub that does nothing
 *	Accepts:	- ptr communicator
 *			- ptr error code
 */
void
__lam_errreturn_(pcomm, perror)

MPI_Comm		*pcomm;
int			*perror;

{
}
