/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	NJN/RBD
 *
 *	$Log:	laminit.c,v $
 * Revision 6.0.1.1  96/06/10  16:02:53  nevin
 * Set MPI working directory
 * 
 * Revision 6.0  96/02/29  21:10:11  gdburns
 * Ohio Release
 * 
 *	Function:	- LAM specific initialization for MPI
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>
#include <net.h>
#include <preq.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external functions
 */
extern int		lam_rtrstore();
extern int		lpattach();

/*
 *	__lam_linit_
 *
 *	Function:	- initialize the process
 *	Accepts:	- program name
 *			- # processes in local world (returned)
 *			- # processes in parent world (returned)
 *			- CID to use for parent intercomm (returned)
 *			- array of process GPS (returned)
 *	Returns:	- 0 or LAMERROR
 */
int
__lam_linit_(name, pnworld, pnparent, pcid, pprocgps)

char			*name;
int			*pnworld;
int			*pnparent;
int			*pcid;
struct _gps		**pprocgps;

{
	char		*trworld;		/* world trace */
	int		trwlen;			/* length of world trace */
	char		*str;			/* favourite string */
	int		i;			/* favourite index */
	int		nworld;			/* # processes in world */
	int		nprocs;			/* # processes in total */
	char		*envdir;		/* working dir in env */
	char		workdir[PSMAXNAME];	/* modifiable workdir copy */
	struct _gps	*procs;			/* process GPS array */
	struct _gps	*p;			/* favourite pointer */
	struct nmsg	nhead;			/* network message header */
	struct _proc	*proc;			/* process entry */
/*
 * Attach to the kernel.
 */
	if (kenter(name, 0)) return(LAMERROR);
/*
 * Get the # of processes in the world.
 * Allocate the array of process GPS.
 */
	nworld = (_kio.ki_world == 0) ? getncomp() : _kio.ki_world;

	nprocs = nworld + _kio.ki_parent;
	
	procs = (struct _gps *)
			malloc((unsigned) nprocs * sizeof(struct _gps));

	if (procs == 0) return(LAMERROR);
/*
 * Set various runtime flags.
 */
	__lam_c2c_ = _kio.ki_rtf & RTF_MPIC2C;
	__lam_ger_ = _kio.ki_rtf & RTF_MPIGER;
	__lam_homog_ = _kio.ki_rtf & RTF_HOMOG;
/*
 * Set the CWD from the environment if required (don't report chdir() errors).
 */
	if ((_kio.ki_rtf & RTF_MPICWD) && (envdir = getenv("LAMWORKDIR"))) {

		strncpy(workdir, envdir, PSMAXNAME);
		str = (workdir[0]) ? strrchr(workdir + 1, STRDIR) : 0;
		
		if (str) {
			*str = '\0';
		}
		chdir(workdir);
	}
/*
 * If spawned or started by mpirun, receive the list of GPS.
 * Local world GPS's are first followed by the parents (if any).
 * Otherwise assume one process per node, pids are not needed.
 */
	if ((_kio.ki_parent > 0) || (_kio.ki_rtf & RTF_MPIRUN)) {

		nhead.nh_event = -getpid();
		nhead.nh_type = BLKMPIINIT;
		nhead.nh_flags = 0;
		nhead.nh_length = nprocs * sizeof(struct _gps);
		nhead.nh_msg = (char *) procs;

		if (nrecv(&nhead)) {
			free((char *) procs);
			return(LAMERROR);
		}
		
		if (_kio.ki_parent > 0) *pcid = nhead.nh_data[1];
	}
	else {
		for (i = 0, p = procs; i < nprocs; ++i, ++p) {
			p->gps_node = i;
			p->gps_pid = getpid();
			p->gps_idx = 0;
		}
	}
/*
 * Set the global ranks in the local world and add the local world
 * processes to the process list while finding out my identity.
 */
	for (i = 0, p = procs; i < nworld; ++i, ++p) {
		p->gps_grank = i;
		proc = __lam_procadd_(p);
		if (proc == 0) return(LAMERROR);

		if ((p->gps_node == getnodeid()) && (p->gps_pid == getpid())) {
			__lam_myproc_ = proc;
		}
	}
/*
 * Add the parents (if any) to the process list.
 */
	for (i = 0; i < _kio.ki_parent; ++i, ++p) {
		proc = __lam_procadd_(p);
		if (proc == 0) return(LAMERROR);
	}
/*
 * interface specific initialization
 */
	if (RPI_SPLIT(_rpi_lamd_init, _rpi_c2c_init, ())) {
		return(LAMERROR);
	}
/*
 * Synchronize clocks across the nodes.
 */
	if (_kio.ki_rtf & RTF_TRACE) {
		if (__lam_clocksync_(nprocs, procs, &__lam_clock_skew_))
				return(LAMERROR);
	}
/*
 * Global rank zero records the world trace.
 */
	if ((_kio.ki_rtf & RTF_TRACE)
			&& (__lam_myproc_->p_gps.gps_grank == 0)) {

		trwlen = 2 * sizeof(int4) + nprocs * sizeof(struct _gps);

		trworld = (char *) malloc((unsigned) trwlen);
		if (trworld == 0) return(LAMERROR);

		((int4 *) trworld)[0] = LAM_TRMAGIC;
		((int4 *) trworld)[1] = nprocs;
		memcpy(trworld + 2 * sizeof(int4), (char *) procs,
				nprocs * sizeof(struct _gps));
		
		mltoti4(trworld, trwlen / sizeof(int4));
		
		if (lam_rtrstore(LOCAL, TRWORLD, __lam_myproc_->p_gps.gps_pid,
				trworld, trwlen)) {
			free(trworld);
			return(LAMERROR);
		}

		free(trworld);
	}
/*
 * Turn process into an MPI process. and set the return values.
 */
	_kio.ki_rtf |= RTF_MPI;
	if (lpattach(name)) return(LAMERROR);
/*
 * Set the return values.
 */
	*pnworld = nworld;
	*pnparent = _kio.ki_parent;
	*pprocgps = procs;

	return(0);
}
