/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	lammisc.c,v $
 * Revision 6.0  96/02/29  14:13:06  gdburns
 * Ohio Release
 * 
 *	Function:	- miscellaneous functions
 */

#include <stdlib.h>

#include <mpi.h>
#include <terror.h>

/*
 *	__lam_chkarg_
 *
 *	Function:	- check point-to-point arguments
 *	Accepts:	- buffer/count/datatype/peer rank/tag/comm.
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_chkarg_(buf, count, dtype, rank, tag, comm)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			rank;
int			tag;
MPI_Comm		comm;

{
	if (count < 0) {
		return(__lam_mkerr_(MPI_ERR_COUNT, 0));
	}

	if (dtype == MPI_DATATYPE_NULL || ( ! dtype->dt_commit)) {
		return(__lam_mkerr_(MPI_ERR_TYPE, 0));
	}

	if ((tag < 0) && (tag != MPI_ANY_TAG) &&
			(tag != MPI_GER_TAG) && (tag != MPI_MERGE_TAG)) {
		return(__lam_mkerr_(MPI_ERR_TAG, 0));
	}

	if (comm == MPI_COMM_NULL) {
		return(__lam_mkerr_(MPI_ERR_COMM, 0));
	}

	if ((rank != MPI_PROC_NULL) && (rank != MPI_ANY_SOURCE)) {
		if (comm->c_inter) {
			if ((rank < 0) || (rank >= comm->c_rgroup->g_nprocs)) {
				return(__lam_mkerr_(MPI_ERR_RANK, 0));
			}
		} else {
			if ((rank < 0) || (rank >= comm->c_group->g_nprocs)) {
				return(__lam_mkerr_(MPI_ERR_RANK, 0));
			}
		}
	}

	return(MPI_SUCCESS);
}

/*
 *	__lam_cubedim_
 *
 *	Function:	- get dimension of inscribing hypercube
 *	Accepts:	- # processes
 *	Returns:	- dimension or -1
 */
int
__lam_cubedim_(nprocs)

int			nprocs;

{
	int		dim;
	unsigned	size;

	if (nprocs < 1) return(-1);

	for (dim = 0, size = 1; size < nprocs; ++dim, size <<= 1);

	return(dim);
}

/*
 *	__lam_hibit_
 *
 *	Function:	- get position of the highest bit set
 *	Accepts:	- process rank
 *			- hypercube dimension
 *	Returns:	- position or -1
 */
int
__lam_hibit_(rank, dim)

int			rank;
int			dim;

{
	unsigned	mask;

	if ((rank < 1) || (dim < 1)) return(-1);

	--dim;
	mask = 1 << dim;

	for (; dim >= 0; --dim, mask >>= 1) {
		if (rank & mask) break;
	}

	return(dim);
}

/*
 *	__lam_emtpystat_
 *
 *	Function:	- set status to empty
 *	Accepts:	- ptr status
 */
void
__lam_emptystat_(pstat)

MPI_Status		*pstat;

{
	pstat->MPI_SOURCE = MPI_ANY_SOURCE;
	pstat->MPI_TAG = MPI_ANY_TAG;
	pstat->MPI_ERROR = MPI_ERR_PENDING;
	pstat->st_count = 0;
	pstat->st_nelem = 0;
	pstat->st_length = 0;
	pstat->st_dtype = MPI_DATATYPE_NULL;
}
