/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	lamprocs.c,v $
 * Revision 6.0  96/02/29  14:13:09  gdburns
 * Ohio Release
 * 
 *	Function:	- manage MPI processes information
 */

#include <all_list.h>
#include <mpi.h>
#include <typical.h>

/*
 * static variables
 */
static LIST		*plist = 0;		/* list of processes */
static struct _proc	*curproc = 0;		/* current process iterator */

/*
 *	__lam_proccmp_
 *
 *	Function:	- compare two process entries
 *	Accepts:	- ptr 2 process entries
 *	Returns:	- -1/0/+1 comparison
 */
int
__lam_proccmp_(p1, p2)

struct _proc		*p1, *p2;

{
	return(LAM_GPSCMP(&(p1->p_gps), &(p2->p_gps)));
}

/*
 *	__lam_procadd_
 *
 *	Function:	- add process to the list if not already present
 *	Accepts:	- ptr process GPS
 *	Returns:	- ptr to process or NULL
 */
struct _proc *
__lam_procadd_(pgps)

struct _gps		*pgps;

{
	struct _proc	pelem;			/* process element */
	struct _proc	*p;			/* favourite pointer */
/*
 * Initialize the list if necessary.
 */
	if (plist == 0) {
		plist = al_init(sizeof(struct _proc), __lam_proccmp_);
		if (plist == 0) return(0);
	}
/*
 * Check if the process is in the list.
 */
	pelem.p_gps = *pgps;
	p = al_find(plist, &pelem);

	if (p) {
		return(p);
	}
/*
 * The process is not present so initialize and insert it.
 */
	pelem.p_rpi = 0;
	pelem.p_mode = 0;
	pelem.p_ger_nsnd = 0;

	return(al_insert(plist, &pelem));
}

/*
 *	__lam_procfind_
 *
 *	Function:	- locate process in the list
 *	Accepts:	- ptr process GPS
 *	Returns:	- ptr to process or NULL
 */
struct _proc *
__lam_procfind_(pgps)

struct _gps		*pgps;

{
	struct _proc	pelem;			/* process element */
	struct _proc	*p = 0;			/* favourite pointer */

	if (plist) {
		pelem.p_gps.gps_node = pgps->gps_node;
		pelem.p_gps.gps_idx = pgps->gps_idx;

		p = (struct _proc *) al_find(plist, &pelem);
	}

	return(p);
}

/*
 *	__lam_procrm_
 *
 *	Function:	- remove a process from the list
 *	Accepts:	- ptr to process
 *	Returns:	- 0 or LAMERROR
 */
int
__lam_procrm_(pproc)

struct _proc		*pproc;

{
	return( (plist) ? al_delete(plist, pproc) : LAMERROR);
}

/*
 *	__lam_topproc_
 *
 *	Function:	- reset process list iterator and get top element
 *	Returns:	- top process in list
 */
struct _proc *
__lam_topproc_()

{
	curproc = (plist) ? al_top(plist) : 0;
	return(curproc);
}

/*
 *	__lam_nextproc_
 *
 *	Function:	- get next process in list
 *			- this should NOT be called if processes have been
 *			  removed from the list since
 *			  the last call to this function or __lam_topproc_
 *	Returns:	- next process in list or 0
 */
struct _proc *
__lam_nextproc_()

{
	if (curproc) curproc = al_next(plist, curproc);

	return(curproc);
}

/*
 *	__lam_nprocs_
 *
 *	Function:	- size of process list
 *	Returns:	- number of processes in process list
 */
int
__lam_nprocs_()

{
	return ( (plist) ? al_count(plist) : 0 );
}

/*
 *	__lam_nukeprocs_
 *
 *	Function:	- delete the process list
 */
void
__lam_nukeprocs_()

{
	(void) al_free(plist);
	plist = 0;
	curproc = 0;
}
