/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/JRV
 *
 *	$Log:	lamreduce.c,v $
 * Revision 6.0  96/02/29  14:13:10  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  95/03/17  18:08:30  raja
 * Fix complex product + face-lift.
 * 
 * Revision 5.2  94/08/22  14:04:47  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/05/19  17:09:49  gdburns
 * Add semi-colon to the last member of struct definitions.
 * 
 * Revision 5.1  94/05/18  12:51:23  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:34  gdburns
 * new for 5.1
 * 
 *	Function:	- all predefined reduce operations
 *			- min, max, sum, product
 *			- and, or, xor (bitwise and logical)
 *			- minloc, maxloc
 */

#include <mpi.h>

/*
 * local structures
 */
struct complex { float r; float i; };
struct fltflt { float v; float k; };
struct fltint { float v; int k; };
struct dbldbl { double v; double k; };
struct dblint { double v; int k; };
struct intint { int v; int k; };
struct longint { long v; int k; };

void
__lam_max_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	float		*pfin, *pfout;
	double		*pdin, *pdout;

	if (*datatype == MPI_INT) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			if (*piin > *piout) {
				*piout = *piin;
			}
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			plin = (long *) invec;
			plout = (long *) inoutvec;
			if (*plin > *plout) {
				*plout = *plin;
			}
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			if (*psin > *psout) {
				*psout = *psin;
			}
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			if (*pusin > *pusout) {
				*pusout = *pusin;
			}
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			if (*puin > *puout) {
				*puout = *puin;
			}
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			if (*pulin > *pulout) {
				*pulout = *pulin;
			}
		}
		return;
	}

	if (*datatype == MPI_FLOAT) {
		pfin = (float *) invec;
		pfout = (float *) inoutvec;
		for (i = 0; i < *len; ++i, ++pfin, ++pfout) {
			if (*pfin > *pfout) {
				*pfout = *pfin;
			}
		}
		return;
	}

	if (*datatype == MPI_DOUBLE) {
		pdin = (double *) invec;
		pdout = (double *) inoutvec;
		for (i = 0; i < *len; ++i, ++pdin, ++pdout) {
			if (*pdin > *pdout) {
				*pdout = *pdin;
			}
		}
		return;
	}
}

void
__lam_min_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	float		*pfin, *pfout;
	double		*pdin, *pdout;

	if (*datatype == MPI_INT) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			if (*piin < *piout) {
				*piout = *piin;
			}
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			plin = (long *) invec;
			plout = (long *) inoutvec;
			if (*plin < *plout) {
				*plout = *plin;
			}
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			if (*psin < *psout) {
				*psout = *psin;
			}
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			if (*pusin < *pusout) {
				*pusout = *pusin;
			}
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			if (*puin < *puout) {
				*puout = *puin;
			}
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			if (*pulin < *pulout) {
				*pulout = *pulin;
			}
		}
		return;
	}

	if (*datatype == MPI_FLOAT) {
		pfin = (float *) invec;
		pfout = (float *) inoutvec;
		for (i = 0; i < *len; ++i, ++pfin, ++pfout) {
			if (*pfin < *pfout) {
				*pfout = *pfin;
			}
		}
		return;
	}

	if (*datatype == MPI_DOUBLE) {
		pdin = (double *) invec;
		pdout = (double *) inoutvec;
		for (i = 0; i < *len; ++i, ++pdin, ++pdout) {
			if (*pdin < *pdout) {
				*pdout = *pdin;
			}
		}
		return;
	}
}

void
__lam_sum_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	float		*pfin, *pfout;
	double		*pdin, *pdout;
	struct complex	*pxin, *pxout;

	if (*datatype == MPI_INT) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout += *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout += *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout += *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout += *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout += *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout += *pulin;
		}
		return;
	}

	if (*datatype == MPI_FLOAT) {
		pfin = (float *) invec;
		pfout = (float *) inoutvec;
		for (i = 0; i < *len; ++i, ++pfin, ++pfout) {
			*pfout += *pfin;
		}
		return;
	}

	if (*datatype == MPI_CPLEX) {
		pxin = (struct complex *) invec;
		pxout = (struct complex *) inoutvec;
		for (i = 0; i < *len; ++i, ++pxin, ++pxout) {
			pxout->r += pxin->r;
			pxout->i += pxin->i;
		}
		return;
	}

	if (*datatype == MPI_DOUBLE) {
		pdin = (double *) invec;
		pdout = (double *) inoutvec;
		for (i = 0; i < *len; ++i, ++pdin, ++pdout) {
			*pdout += *pdin;
		}
		return;
	}
}

void
__lam_prod_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	float		*pfin, *pfout;
	float		xr, xi;
	double		*pdin, *pdout;
	struct complex	*pxin, *pxout;


	if (*datatype == MPI_INT) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout *= *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout *= *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout *= *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout *= *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout *= *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout *= *pulin;
		}
		return;
	}

	if (*datatype == MPI_FLOAT) {
		pfin = (float *) invec;
		pfout = (float *) inoutvec;
		for (i = 0; i < *len; ++i, ++pfin, ++pfout) {
			*pfout *= *pfin;
		}
		return;
	}

	if (*datatype == MPI_CPLEX) {
		pxin = (struct complex *) invec;
		pxout = (struct complex *) inoutvec;
		for (i = 0; i < *len; ++i, ++pxin, ++pxout) {
			xr = pxout->r;
			xi = pxout->i;
			pxout->r = (xr * pxin->r) - (xi * pxin->i);
			pxout->i = (xr * pxin->i) + (xi * pxin->r);
		}
		return;
	}

	if (*datatype == MPI_DOUBLE) {
		pdin = (double *) invec;
		pdout = (double *) inoutvec;
		for (i = 0; i < *len; ++i, ++pdin, ++pdout) {
			*pdout *= *pdin;
		}
		return;
	}
}

void
__lam_land_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;

	if ((*datatype == MPI_INT) || (*datatype == MPI_LOGIC)) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout = *piout && *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout = *plout && *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout = *psout && *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout = *pusout && *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout = *puout && *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout = *pulout && *pulin;
		}
		return;
	}
}

void
__lam_lor_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;

	if ((*datatype == MPI_INT) || (*datatype == MPI_LOGIC)) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout = *piout || *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout = *plout || *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout = *psout || *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout = *pusout || *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout = *puout || *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout = *pulout || *pulin;
		}
		return;
	}
}

void
__lam_lxor_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;

	if ((*datatype == MPI_INT) || (*datatype == MPI_LOGIC)) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout = (*piout || *piin) && !(*piout && *piin);
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout = (*plout || *plin) && !(*plout && *plin);
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout = (*psout || *psin) && !(*psout && *psin);
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout = (*pusout || *pusin) && !(*pusout && *pusin);
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout = (*puout || *puin) && !(*puout && *puin);
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout = (*pulout || *pulin) && !(*pulout && *pulin);
		}
		return;
	}
}

void
__lam_band_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	unsigned char	*pucin, *pucout;

	if ((*datatype == MPI_INT) || (*datatype == MPI_LOGIC)) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout &= *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout &= *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout &= *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout &= *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout &= *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout &= *pulin;
		}
		return;
	}

	if (*datatype == MPI_BYTE) {
		pucin = (unsigned char *) invec;
		pucout = (unsigned char *) inoutvec;
		for (i = 0; i < *len; ++i, ++pucin, ++pucout) {
			*pucout &= *pucin;
		}
		return;
	}
}

void
__lam_bor_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	unsigned char	*pucin, *pucout;

	if ((*datatype == MPI_INT) || (*datatype == MPI_LOGIC)) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout |= *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout |= *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout |= *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout |= *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout |= *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout |= *pulin;
		}
		return;
	}

	if (*datatype == MPI_BYTE) {
		pucin = (unsigned char *) invec;
		pucout = (unsigned char *) inoutvec;
		for (i = 0; i < *len; ++i, ++pucin, ++pucout) {
			*pucout |= *pucin;
		}
		return;
	}
}

void
__lam_bxor_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	int		*piin, *piout;
	long		*plin, *plout;
	short		*psin, *psout;
	unsigned short	*pusin, *pusout;
	unsigned	*puin, *puout;
	unsigned long	*pulin, *pulout;
	unsigned char	*pucin, *pucout;

	if ((*datatype == MPI_INT) || (*datatype == MPI_LOGIC)) {
		piin = (int *) invec;
		piout = (int *) inoutvec;
		for (i = 0; i < *len; ++i, ++piin, ++piout) {
			*piout ^= *piin;
		}
		return;
	}

	if (*datatype == MPI_LONG) {
		plin = (long *) invec;
		plout = (long *) inoutvec;
		for (i = 0; i < *len; ++i, ++plin, ++plout) {
			*plout ^= *plin;
		}
		return;
	}

	if (*datatype == MPI_SHORT) {
		psin = (short *) invec;
		psout = (short *) inoutvec;
		for (i = 0; i < *len; ++i, ++psin, ++psout) {
			*psout ^= *psin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_SHORT) {
		pusin = (unsigned short *) invec;
		pusout = (unsigned short *) inoutvec;
		for (i = 0; i < *len; ++i, ++pusin, ++pusout) {
			*pusout ^= *pusin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED) {
		puin = (unsigned *) invec;
		puout = (unsigned *) inoutvec;
		for (i = 0; i < *len; ++i, ++puin, ++puout) {
			*puout ^= *puin;
		}
		return;
	}

	if (*datatype == MPI_UNSIGNED_LONG) {
		pulin = (unsigned long *) invec;
		pulout = (unsigned long *) inoutvec;
		for (i = 0; i < *len; ++i, ++pulin, ++pulout) {
			*pulout ^= *pulin;
		}
		return;
	}

	if (*datatype == MPI_BYTE) {
		pucin = (unsigned char *) invec;
		pucout = (unsigned char *) inoutvec;
		for (i = 0; i < *len; ++i, ++pucin, ++pucout) {
			*pucout ^= *pucin;
		}
		return;
	}
}

void
__lam_maxloc_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	struct fltflt	*pffin, *pffout;
	struct fltint	*pfiin, *pfiout;
	struct dbldbl	*pddin, *pddout;
	struct dblint	*pdiin, *pdiout;
	struct intint	*piiin, *piiout;
	struct longint	*pliin, *pliout;

	if (*datatype == MPI_2FLOAT) {
		pffin = (struct fltflt *) invec;
		pffout = (struct fltflt *) inoutvec;
		for (i = 0; i < *len; ++i, ++pffin, ++pffout) {
			if (pffin->v > pffout->v) {
				pffout->v = pffin->v;
				pffout->k = pffin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_FLOAT_INT) {
		pfiin = (struct fltint *) invec;
		pfiout = (struct fltint *) inoutvec;
		for (i = 0; i < *len; ++i, ++pfiin, ++pfiout) {
			if (pfiin->v > pfiout->v) {
				pfiout->v = pfiin->v;
				pfiout->k = pfiin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_2DOUBLE) {
		pddin = (struct dbldbl *) invec;
		pddout = (struct dbldbl *) inoutvec;
		for (i = 0; i < *len; ++i, ++pddin, ++pddout) {
			if (pddin->v > pddout->v) {
				pddout->v = pddin->v;
				pddout->k = pddin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_DOUBLE_INT) {
		pdiin = (struct dblint *) invec;
		pdiout = (struct dblint *) inoutvec;
		for (i = 0; i < *len; ++i, ++pdiin, ++pdiout) {
			if (pdiin->v > pdiout->v) {
				pdiout->v = pdiin->v;
				pdiout->k = pdiin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_2INT) {
		piiin = (struct intint *) invec;
		piiout = (struct intint *) inoutvec;
		for (i = 0; i < *len; ++i, ++piiin, ++piiout) {
			if (piiin->v > piiout->v) {
				piiout->v = piiin->v;
				piiout->k = piiin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_LONG_INT) {
		pliin = (struct longint *) invec;
		pliout = (struct longint *) inoutvec;
		for (i = 0; i < *len; ++i, ++pliin, ++pliout) {
			if (pliin->v > pliout->v) {
				pliout->v = pliin->v;
				pliout->k = pliin->k;
			}
		}
		return;
	}
}


void
__lam_minloc_(invec, inoutvec, len, datatype)

void			*invec;
void			*inoutvec;
int			*len;
MPI_Datatype		*datatype;

{
	int		i;
	struct fltflt	*pffin, *pffout;
	struct fltint	*pfiin, *pfiout;
	struct dbldbl	*pddin, *pddout;
	struct dblint	*pdiin, *pdiout;
	struct intint	*piiin, *piiout;
	struct longint	*pliin, *pliout;

	if (*datatype == MPI_2FLOAT) {
		pffin = (struct fltflt *) invec;
		pffout = (struct fltflt *) inoutvec;
		for (i = 0; i < *len; ++i, ++pffin, ++pffout) {
			if (pffin->v < pffout->v) {
				pffout->v = pffin->v;
				pffout->k = pffin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_FLOAT_INT) {
		pfiin = (struct fltint *) invec;
		pfiout = (struct fltint *) inoutvec;
		for (i = 0; i < *len; ++i, ++pfiin, ++pfiout) {
			if (pfiin->v < pfiout->v) {
				pfiout->v = pfiin->v;
				pfiout->k = pfiin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_2DOUBLE) {
		pddin = (struct dbldbl *) invec;
		pddout = (struct dbldbl *) inoutvec;
		for (i = 0; i < *len; ++i, ++pddin, ++pddout) {
			if (pddin->v < pddout->v) {
				pddout->v = pddin->v;
				pddout->k = pddin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_DOUBLE_INT) {
		pdiin = (struct dblint *) invec;
		pdiout = (struct dblint *) inoutvec;
		for (i = 0; i < *len; ++i, ++pdiin, ++pdiout) {
			if (pdiin->v < pdiout->v) {
				pdiout->v = pdiin->v;
				pdiout->k = pdiin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_2INT) {
		piiin = (struct intint *) invec;
		piiout = (struct intint *) inoutvec;
		for (i = 0; i < *len; ++i, ++piiin, ++piiout) {
			if (piiin->v < piiout->v) {
				piiout->v = piiin->v;
				piiout->k = piiin->k;
			}
		}
		return;
	}

	if (*datatype == MPI_LONG_INT) {
		pliin = (struct longint *) invec;
		pliout = (struct longint *) inoutvec;
		for (i = 0; i < *len; ++i, ++pliin, ++pliout) {
			if (pliin->v < pliout->v) {
				pliout->v = pliin->v;
				pliout->k = pliin->k;
			}
		}
		return;
	}
}
