/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	lamsend.c,v $
 * Revision 6.0.1.1  96/03/20  23:12:36  nevin
 * Orphan flagging moved to request start
 * 
 * Revision 6.0  96/02/29  14:13:12  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:24:31  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:48  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:24  gdburns
 * Ohio Release
 * 
 *	Function:	- shared code for send functions
 */

#include <mpi.h>
#include <mpitrace.h>
#include <typical.h>

/*
 * public functions
 */
int			__lam_send_();
int			__lam_isend_();

/*
 *	__lam_send_
 *
 *	Function:	- send a blocking MPI message
 *	Accepts:	- MPI send args
 *			- request type
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_send_(buf, count, dtype, dest, tag, comm, reqtype)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			dest;
int			tag;
MPI_Comm		comm;
int			reqtype;

{
	int		err;
	int		fl_trace;		/* do tracing? */
	int		req_err;		/* recoverable error */
	int		seqnum;			/* message sequence number */
	struct _req	req_storage;
	MPI_Request	req;
	double		startt;			/* start time */
	double		finisht;		/* finish time */
/*
 * Handle the trivial cases.
 */
	if (tag == MPI_ANY_TAG) return(__lam_mkerr_(MPI_ERR_TAG, 0));

	if (dest == MPI_ANY_SOURCE) return(__lam_mkerr_(MPI_ERR_RANK, 0));

	if (dest == MPI_PROC_NULL) return(MPI_SUCCESS);
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) {
		startt = MPI_Wtime();
		_kio.ki_blktime = 0.0;
	}

	req = &req_storage;
	err = _mpi_req_build(buf, count, dtype, dest, tag, comm,
			reqtype, &req);
	if (err != MPI_SUCCESS) return(err);

	req->rq_marks |= LAM_RQFBLKTYPE;

	err = _mpi_req_start(req);
	if (err != MPI_SUCCESS) return(err);

	_mpi_req_add(req);
	_mpi_req_blkclr();
	_mpi_req_blkset(req);

	err = _mpi_req_advance();
	if (err != MPI_SUCCESS) return(err);

	req_err = req->rq_status.MPI_ERROR;
	seqnum = req->rq_seq;

	_mpi_req_rem(req);

	err = _mpi_req_destroy(&req);
	if (err != MPI_SUCCESS) return(err);
/*
 * Generate a trace.
 */
	if (fl_trace) {
		finisht = MPI_Wtime();

		__lam_trmsg_(TRTOUTPUT, startt,
				LAM_S2US(finisht - startt - _kio.ki_blktime),
				LAM_S2US(_kio.ki_blktime), dest, tag,
				comm, dtype, count, 0, 0, seqnum, reqtype);
	}

	return(req_err);
}

/*
 *	__lam_isend_
 *
 *	Function:	- send a non-blocking MPI message
 *	Accepts:	- MPI isend args
 *			- request type
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_isend_(buf, count, dtype, dest, tag, comm, preq, reqtype)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			dest;
int			tag;
MPI_Comm		comm;
MPI_Request		*preq;
int			reqtype;

{
	int		err;
/*
 * Handle the trivial cases.
 */
	if (tag == MPI_ANY_TAG) return(__lam_mkerr_(MPI_ERR_TAG, 0));

	if (dest == MPI_ANY_SOURCE) return(__lam_mkerr_(MPI_ERR_RANK, 0));

	*preq = MPI_REQUEST_NULL;
	err = _mpi_req_build(buf, count, dtype, dest, tag, comm,
			reqtype, preq);
	if (err != MPI_SUCCESS) return(err);

	err = _mpi_req_start(*preq);
	if (err != MPI_SUCCESS) return(err);

	_mpi_req_add(*preq);
	_mpi_req_blkclr();

	err = _mpi_req_advance();

	return(err);
}
