/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	lamtest.c,v $
 * Revision 6.0.1.1  96/05/10  15:10:41  nevin
 * Handle MPI_PROC_NULL
 * 
 * Revision 6.0  96/02/29  14:13:13  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/29  11:45:39  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:49  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:25  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:35  gdburns
 * new for 5.1
 * 
 *	Function:	- test if request is done
 *			- destructive and non-destructive modes
 *	Accepts:	- request
 *			- destructive mode flag
 *			- flag (returned)
 *			- status (returned)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <mpi.h>
#include <mpitrace.h>
#include <terror.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_trmsg_();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();
extern void		_mpi_req_rem();
extern int		__lam_mkerr_();
extern int		__lam_trincff_();
extern int		_mpi_req_advance();
extern int		_mpi_req_destroy();
extern int		_mpi_req_end();


int
__lam_test_(preq, destroy, flag, stat)

MPI_Request		*preq;
int			destroy;
int			*flag;
MPI_Status		*stat;

{
	MPI_Request	req;			/* request */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */

	if (preq == 0) return(__lam_mkerr_(MPI_ERR_REQUEST, 0));

	if ((flag == 0) || (stat == 0)) return(__lam_mkerr_(MPI_ERR_ARG, 0));
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) {
	    startt = MPI_Wtime();
	    _kio.ki_blktime = 0.0;
	}
/*
 * Advance the system.
 */
	_mpi_req_blkclr();

	err = _mpi_req_advance();
	if (err != MPI_SUCCESS) return(err);
/*
 * Inactive requests get an empty status.
 */
	req = *preq;

	if ((req == MPI_REQUEST_NULL) || (req->rq_state == LAM_RQSINIT)) {
	    *flag = 1;
	    __lam_emptystat_(stat);
	}
/*
 * If not done reset flag.
 */
	else if (req->rq_state != LAM_RQSDONE) {
	    *flag = 0;
	}
/*
 * Finish the request if in destructive-mode.
 */
	else {
	    *flag = 1;

	    if (destroy) {
		_mpi_req_rem(req);

		err = _mpi_req_end(req);
		if (err != MPI_SUCCESS) return(err);

		*stat = req->rq_status;
/*
 * Generate a run time trace except in the case of cancelled request or
 * a request where the peer is MPI_PROC_NULL.
 */
		if (fl_trace && (req->rq_rank != MPI_PROC_NULL)
				&& !(req->rq_flags & LAM_RQFCANCEL)) {
		    finisht = MPI_Wtime();

		    __lam_trmsg_( (req->rq_type == LAM_RQIRECV)
			    ? TRTINPUT : TRTNOIO, startt,
			    LAM_S2US(finisht - startt - _kio.ki_blktime),
			    LAM_S2US(_kio.ki_blktime), req->rq_rank,
			    req->rq_tag, req->rq_comm, req->rq_dtype,
			    req->rq_count, stat->MPI_SOURCE,
			    stat->MPI_TAG, req->rq_seq, req->rq_type);
		}
/*
 * Reset persistent requests.
 * Destroy requests that are not persistent.
 */
		if (req->rq_marks & LAM_RQFPERSIST) {
		    req->rq_state = LAM_RQSINIT;
		} else {
		    err = _mpi_req_destroy(preq);
		    if (err != MPI_SUCCESS) return(err);
		}

		if (stat->MPI_ERROR != MPI_SUCCESS) {
		    return(stat->MPI_ERROR);
		}
	    }
	}
	
	return(MPI_SUCCESS);
}
