/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	lamupdown.c,v $
 * Revision 6.0  96/02/29  14:13:16  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:24:49  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:51  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:27  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:37  gdburns
 * new for 5.1
 * 
 *	Function:	- take top level info down
 *			- take error info up
 */

#include <mpi.h>

/*
 * static variables
 */
static int		topfunc = 0;
static int		toproot;
static int		toprootgps;

/*
 *	__lam_setfunc_
 *
 *	Function:	- set top level function
 *	Accepts:	- local function type
 */
void
__lam_setfunc_(locfunc)

int			locfunc;

{
	if (topfunc == 0) {
		topfunc = locfunc;
	}
}

/*
 *	__lam_resetfunc_
 *
 *	Function:	- reset function at top level
 *	Accepts:	- local function type
 */
void
__lam_resetfunc_(locfunc)

int			locfunc;

{
	struct _fyiproc	*p;			/* favourite pointer */

	if (topfunc == locfunc) {
		p = (struct _fyiproc *) _kio.ki_fyi;
		topfunc = p->fyp_func = 0;
	}
}

/*
 *	__lam_getfunc_
 *
 *	Function:	- get top level function
 *	Returns:	- top level function
 */
int
__lam_getfunc_()

{
	return(topfunc);
}

/*
 *	__lam_setparam_
 *
 *	Function:	- set top level parameters
 *	Accepts:	- local function type
 *			- root global/local rank (collective comm.)
 *			- root node/index GPS (collective comm.)
 */
void
__lam_setparam_(locfunc, root, rootgps)

int			locfunc;
int			root;
int			rootgps;

{
	if (topfunc == locfunc) {
		toproot = root;
		toprootgps = rootgps;
	}
}

/*
 *	__lam_getparam_
 *
 *	Function:	- get top level parameters
 *	Accepts:	- ptr root ranks
 *			- ptr root GPS
 */
void
__lam_getparam_(proot, prootgps)

int			*proot;
int			*prootgps;

{
	*proot = toproot;
	*prootgps = toprootgps;
}

/*
 *	__lam_mkerr_
 *
 *	Function:	- form an error code
 *	Accepts:	- error class
 *			- error value
 *	Returns:	- error code
 */
int
__lam_mkerr_(class, error)

int			class;
int			error;

{
	int		errcode;

	errcode = ((error & 0xFFFF) << 8) | (topfunc & 0xFF);
	errcode = (errcode << 8) | (class & 0xFF);
	return(errcode);
}

/*
 *	__lam_bkerr_
 *
 *	Function:	- break error code into components
 *	Accepts:	- error code
 *			- ptr class (returned value)
 *			- ptr function (returned value)
 *			- ptr error (returned value)
 */
void
__lam_bkerr_(errcode, class, func, error)

int			errcode;
int			*class;
int			*func;
int			*error;

{
	*class = errcode & 0xFF;
	errcode >>= 8;
	*func = errcode & 0xFF;
	errcode >>= 8;
	*error = errcode & 0xFFFF;
}

/*
 *	__lam_errfunc_
 *
 *	Function:	- handle MPI errors according to error mode
 *			- pass error up to top level
 *			- call error handler at top level
 *	Accepts:	- communicator
 *			- local function type
 *			- error code
 *	Returns:	- error code
 */
int
__lam_errfunc_(comm, locfunc, errcode)

MPI_Comm		comm;
int			locfunc;
int			errcode;

{
	MPI_Errhandler	errhdl;			/* error handler */
	int		class;			/* error class */
	int		func;			/* function type */
	int		error;			/* errno value */
	int		errsave;		/* saved error code */
	int		f77index;		/* f77 error handler */
	struct _fyiproc	*p;			/* favourite pointer */

	errsave = errcode;

	__lam_bkerr_(errcode, &class, &func, &error);

	if (func == locfunc) {

		if (comm) {
			errhdl = comm->c_errhdl;
			f77index = errhdl->eh_f77hdl;

			if (f77index < 0) {
				(errhdl->eh_func)(&comm, &errcode);
			} else {
				(errhdl->eh_func)(&f77index, &errcode);
			}

			p = (struct _fyiproc *) _kio.ki_fyi;
			topfunc = p->fyp_func = 0;
		}
		else {
			comm = MPI_COMM_WORLD;
			__lam_errfatal_(&comm, &errcode);
		}
	}

	return(errsave);
}

/*
 *	__lam_nukefunc_
 *
 *	Function:	- reset top function (cleanup)
 */
void
__lam_nukefunc_()

{
	topfunc = 0;
}
