/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	mpietc.c,v $
 * Revision 6.0  96/02/29  14:13:18  gdburns
 * Ohio Release
 * 
 *	Function:	- misc. shared MPI-related functions
 *			- used by libmpi and MPI tools
 */

#include <mpi.h>

/*
 * macros
 */
#define FLIPCID(x)	(-((x) + 1))

/*
 *	__lam_pt2coll_
 *
 *	Function:	- change context ID from pt2pt to collective
 *	Accepts:	- pt2pt context ID
 *	Returns:	- collective context ID
 */
int
__lam_pt2coll_(cid)

int			cid;

{
	if (cid >= 0) cid = FLIPCID(cid);
	return(cid);
}

/*
 *	__lam_coll2pt_
 *
 *	Function:	- change context ID from collective to pt2pt
 *	Accepts:	- collective context ID
 *	Returns:	- pt2pt context ID
 */
int
__lam_coll2pt_(cid)

int			cid;

{
	if (cid < 0) cid = FLIPCID(cid);
	return(cid);
}

/*
 *	__lam_mkcoll_
 *
 *	Function:	- switch communicator to collective context ID
 *	Accepts:	- communicator
 */
void
__lam_mkcoll_(comm)

MPI_Comm		comm;

{
	if (comm) comm->c_contextid = __lam_pt2coll_(comm->c_contextid);
}

/*
 *	__lam_mkpt_
 *
 *	Function:	- switch communicator to pt2pt context ID
 *	Accepts:	- communicator
 */
void
__lam_mkpt_(comm)

MPI_Comm		comm;

{
	if (comm) comm->c_contextid = __lam_coll2pt_(comm->c_contextid);
}

/*
 *	__lam_unsyncsql_
 *
 *	Function:	- extract SQL sync. information
 *	Accepts:	- event
 *			- type
 *			- ptr tag (returned value)
 *			- ptr context ID (returned value)
 *			- ptr ACK flag (returned value)
 *			- ptr destination index (returned value)
 *			- ptr source node (returned value)
 *			- ptr source index (returned value)
 */
void
__lam_unsyncsql_(event, type, ptag, pcid, pack, pdidx, psnode, psidx)

int			event;
int			type;
int			*ptag;
int			*pcid;
int			*pack;
int			*pdidx;
int			*psnode;
int			*psidx;

{
	*pdidx = event & 0x7FFF;
	*pack = event & 0x10000;
	*pcid = (event >> 18) & 0xFFF;
	if (*pcid & 0x800) *pcid |= 0xFFFFF000;

	*ptag = type & 0xFFFF;
	if (*ptag & 0x8000) *ptag |= 0xFFFF0000;

	type >>= 16;
	*psnode = type & 0x3FF;
	*psidx = (type >> 10) & 0x3F;

	if (*psnode == 0x3FF) *psnode = MPI_ANY_SOURCE;
	if (*psidx == 0x3F) *psidx = MPI_ANY_SOURCE;
}
