/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	mpil_id.c,v $
 * Revision 6.0.1.1  96/07/22  12:58:52  nevin
 * In case of intercommunicator return GPS of rank in remote group.
 * 
 * Revision 6.0  96/02/29  14:13:18  gdburns
 * Ohio Release
 * 
 *	Function:	- identifers for MPI opaque objects
 */

#include <blktype.h>
#include <mpi.h>

/*
 * backward compatibility
 */
int MPI_Comm_get_contextid(comm, cid) MPI_Comm comm; int *cid;
	{ return(MPIL_Comm_id(comm, cid)); }

/*
 *	MPIL_Comm_id
 *
 *	Function:	- returns the context ID
 *	Accepts:	- communicator
 *			- context ID (out)
 */
int
MPIL_Comm_id(comm, pcid)

MPI_Comm		comm;
int			*pcid;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPILCOMMID);

	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (pcid == 0) {
		return(__lam_errfunc_(comm,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	*pcid = comm->c_contextid;

	__lam_resetfunc_(BLKMPILCOMMID);
	return(MPI_SUCCESS);
}

/*
 *	MPIL_Comm_gps
 *
 *	Function:	- translates MPI comm/rank to LAM nid/pid
 *			- if the communicator is an intercommunicator then
 *			  rank is taken to be in the remote group
 *	Accepts:	- communicator
 *			- rank
 *			- node ID (out)
 *			- process ID (out)
 */
int
MPIL_Comm_gps(comm, rank, pnid, ppid)

MPI_Comm		comm;
int			rank;
int			*pnid;
int			*ppid;

{
	MPI_Group	grp;

	__lam_initerr_();
	__lam_setfunc_(BLKMPILCOMMGPS);

	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((pnid == 0) || (ppid == 0)) {
		return(__lam_errfunc_(comm,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	grp = (comm->c_inter) ? comm->c_rgroup : comm->c_group;
	
	if ((rank == MPI_ANY_SOURCE) || (rank < 0) ||
			(rank >= grp->g_nprocs)) {
		return(__lam_errfunc_(comm,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_RANK, 0)));
	}

	*pnid = grp->g_procs[rank]->p_gps.gps_node;
	*ppid = grp->g_procs[rank]->p_gps.gps_pid;

	__lam_resetfunc_(BLKMPILCOMMGPS);
	return(MPI_SUCCESS);
}

/*
 *	MPIL_Type_id
 *
 *	Function:	- returns the type ID
 *	Accepts:	- datatype
 *			- type ID (out)
 */
int
MPIL_Type_id(dtype, ptid)

MPI_Datatype		dtype;
int			*ptid;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPILTYPEID);

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (ptid == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			__lam_getfunc_(), __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	*ptid = dtype->dt_label;

	__lam_resetfunc_(BLKMPILTYPEID);
	return(MPI_SUCCESS);
}
