/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	NJN
 *
 *	$Log:	mpil_trace.c,v $
 * Revision 6.0  96/02/29  14:13:21  gdburns
 * Ohio Release
 * 
 *	Function:	- toggles tracing status
 */

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 *	MPIL_Trace_on
 *
 *	Function:	- enables tracing; begins segment
 */
int
MPIL_Trace_on()

{
	int		err;			/* error code */
	int		trmodesave;		/* save trace mode */
/*
 * This is a NOP if trace mode is off.
 */
	if ( ! (_kio.ki_rtf & RTF_TRACE)) return(MPI_SUCCESS);

	__lam_initerr_();
	__lam_setfunc_(BLKMPILTRACEON);
/*
 * Synchronize. We don't want to trace it though!
 */
	trmodesave = _kio.ki_rtf & RTF_TRSWITCH;
	_kio.ki_rtf &= ~RTF_TRSWITCH;
	
	err = MPI_Barrier(MPI_COMM_WORLD);

	_kio.ki_rtf |= trmodesave;
	
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILTRACEON, err));
	}
/*
 * Turn tracing on.
 */
	if (__lam_tron_() < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILTRACEON,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
	
	__lam_resetfunc_(BLKMPILTRACEON);
	return(MPI_SUCCESS);
}

/*
 *	MPIL_Trace_off
 *
 *	Function:	- disables tracing; ends segment
 */
int
MPIL_Trace_off()

{
	int		err;			/* error code */
	int		trmodesave;		/* save trace mode */
/*
 * This is a NOP if trace mode is off.
 */
	if ( ! (_kio.ki_rtf & RTF_TRACE)) return(MPI_SUCCESS);
	
	__lam_initerr_();
	__lam_setfunc_(BLKMPILTRACEOFF);
/*
 * Synchronize. We don't want to trace it though!
 */
	trmodesave = _kio.ki_rtf & RTF_TRSWITCH;
	_kio.ki_rtf &= ~RTF_TRSWITCH;

	err = MPI_Barrier(MPI_COMM_WORLD);

	_kio.ki_rtf |= trmodesave;

	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILTRACEOFF, err));
	}
/*
 * Turn tracing off.
 */
	if (__lam_troff_() < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILTRACEOFF,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
	
	__lam_resetfunc_(BLKMPILTRACEOFF);
	return(MPI_SUCCESS);
}
