/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	pack.c,v $
 * Revision 6.0  96/02/29  14:13:23  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:25:02  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:15  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:04:53  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:30  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:38  gdburns
 * new for 5.1
 * 
 *	Function:	- pack from data buffer
 *	Accepts:	- data buffer
 *			- data count
 *			- data type
 *			- packing buffer
 *			- packing buffer size
 *			- ptr packing buffer position
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Pack(buf, count, dtype, packbuf, packsize, ppackpos, comm)

void			*buf;
int			count;
MPI_Datatype		dtype;
void			*packbuf;
int			packsize;
int			*ppackpos;
MPI_Comm		comm;

{
	int		size;
	char		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIPACK);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(comm,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (count < 0) {
		return(__lam_errfunc_(comm,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((packsize < 0) || (ppackpos == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((*ppackpos < 0) || (*ppackpos >= packsize)) {
		return(__lam_errfunc_(comm,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((packsize > 0) && (packbuf == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_BUFFER, 0)));
	}
/*
 * Pack the data.
 */
	p = ((char *) packbuf) + *ppackpos;
	size = packsize - *ppackpos;

	size = __lam_pack_((char *) buf, count, dtype, p, size);

	if (size < 0) {
		return(__lam_errfunc_(comm,
			BLKMPIPACK, __lam_mkerr_(MPI_ERR_TRUNCATE, 0)));
	}

	*ppackpos += size;

	__lam_resetfunc_(BLKMPIPACK);
	return(MPI_SUCCESS);
}
