/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	recv.c,v $
 * Revision 6.0  96/02/29  14:13:26  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:25:13  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:55  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:35  gdburns
 * Ohio Release
 * 
 *	Function:	- receive a message
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- source
 *			- tag
 *			- communicator
 *			- ptr to status
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern void		_mpi_req_add();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();
extern void		_mpi_req_rem();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();
extern int		__lam_trincff_();
extern int		_mpi_req_advance();
extern int		_mpi_req_build();
extern int		_mpi_req_destroy();
extern int		_mpi_req_start();
extern int		_mpi_req_end();

int
MPI_Recv(buf, count, dtype, src, tag, comm, stat)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			src;
int			tag;
MPI_Comm		comm;
MPI_Status		*stat;

{
	int		err;
	struct _req	req_storage;
	MPI_Request	req;
	int		seqnum;			/* message sequence number */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */
	
	__lam_initerr_();
/*
 * Check the bozo MPI_PROC_NULL case.  I'm not even tracing this.
 */
	if (src == MPI_PROC_NULL) {
		stat->MPI_ERROR = MPI_SUCCESS;
		stat->MPI_SOURCE = MPI_PROC_NULL;
		stat->MPI_TAG = MPI_ANY_TAG;
		stat->st_count = 0;
		stat->st_nelem = 0;
		stat->st_length = 0;
		return(MPI_SUCCESS);
	}

	__lam_setfunc_(BLKMPIRECV);
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) {
		startt = MPI_Wtime();
		_kio.ki_blktime = 0.0;
	}

	req = &req_storage;
	err = _mpi_req_build(buf, count, dtype, src, tag, comm,
			LAM_RQIRECV, &req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIRECV, err));

	req->rq_marks |= LAM_RQFBLKTYPE;

	err = _mpi_req_start(req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIRECV, err));

	_mpi_req_add(req);
	_mpi_req_blkclr();
	_mpi_req_blkset(req);

	err = _mpi_req_advance();
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIRECV, err));
	
	_mpi_req_rem(req);

	err = _mpi_req_end(req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIRECV, err));

	*stat = req->rq_status;
	seqnum = req->rq_seq;
	
	err = _mpi_req_destroy(&req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIRECV, err));
/*
 * Generate a run time trace.
 */
	if (fl_trace) {
		finisht = MPI_Wtime();
		
		__lam_trmsg_(TRTINPUT, startt,
				LAM_S2US(finisht - startt - _kio.ki_blktime),
				LAM_S2US(_kio.ki_blktime), src, tag, comm,
				dtype, count, stat->MPI_SOURCE, stat->MPI_TAG,
				seqnum, LAM_RQIRECV);
	}
	
	__lam_resetfunc_(BLKMPIRECV);

	if (stat->MPI_ERROR != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIRECV, stat->MPI_ERROR));
	}

	return(MPI_SUCCESS);
}
