/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	reqfree.c,v $
 * Revision 6.0  96/02/29  14:13:29  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:25:21  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/26  14:16:10  raja
 * Free the communicator.
 * 
 * Revision 5.2  94/08/22  14:04:58  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:38  gdburns
 * Ohio Release
 * 
 *	Function:	- free communication request
 *	Accepts:	- ptr to request
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		_mpi_req_rem();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();
extern int		_mpi_req_destroy();
extern int		_mpi_req_end();

int
MPI_Request_free(preq)

MPI_Request		*preq;

{
	MPI_Request	req;			/* request */
	int		err;			/* error code */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIREQFREE);
/*
 * Check arguments.
 */
	if (preq == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIREQFREE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((req = *preq) == MPI_REQUEST_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIREQFREE, __lam_mkerr_(MPI_ERR_REQUEST, 0)));
	}
/*
 * If active, flag it for destruction.
 * Otherwise destroy it.
 */
	if (req->rq_state == LAM_RQSACTIVE || req->rq_state == LAM_RQSSTART) {

		req->rq_flags |= LAM_RQFORPHAN;
	}
	else {
		_mpi_req_rem(req);

		if (req->rq_state == LAM_RQSDONE) {
			err = _mpi_req_end(req);
			if (err != MPI_SUCCESS) {
				return(__lam_errfunc_(MPI_COMM_WORLD,
						BLKMPIREQFREE, err));
			}
		}

		err = _mpi_req_destroy(preq);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(MPI_COMM_WORLD,
					BLKMPIREQFREE, err));
		}
	}

	*preq = MPI_REQUEST_NULL;
	
	__lam_resetfunc_(BLKMPIREQFREE);
	return(MPI_SUCCESS);
}
