/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	NJN
 *
 *	$Log:	rpi.c2c.c,v $
 * Revision 6.0.1.4  96/06/10  16:09:00  nevin
 * Fill rank wildcard in receive request envelope
 * 
 * Revision 6.0.1.3  96/05/10  15:12:50  nevin
 * Handle MPI_PROC_NULL
 * 
 * Revision 6.0.1.2  96/04/18  13:27:15  nevin
 * iov_base may be of type void*
 * 
 * Revision 6.0.1.1  96/03/12  13:31:24  nevin
 * Setting socket buffer sizes not supported by Linux
 * 
 * Revision 6.0  96/02/29  14:13:29  gdburns
 * Ohio Release
 * 
 *	Function:	- client-to-client mode interface functions
 */

#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/uio.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include <blktype.h>
#include <dl_inet.h>
#include <mpi.h>
#include <net.h>
#include <rpi.c2c.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * global functions
 */ 
int			_rpi_c2c_init();
int			_rpi_c2c_addprocs();
int			_rpi_c2c_build();
int			_rpi_c2c_start();
int			_rpi_c2c_destroy();
int			_rpi_c2c_advance();
int			_rpi_c2c_iprobe();
int			_rpi_c2c_finalize();
int			_c2c_envl_cmp();

/*
 * external functions
 */
extern int		ldogetlinks();
extern int		getsrvstrsock();
extern int		getcltstrsock();
extern int		accepttmout();
extern void		__lam_commfault_();

/*
 * local functions
 */
static int		c2c_advmany();
static int		c2c_adv1();
static int		c2c_adv1anysrc();
static int		c2c_wttryadv();
static int		c2c_rdtryadv();
static int		c2c_readadv();
static int		c2c_readbody();
static void		c2c_bodydone();
static void		c2c_sendself();
static int		c2c_recvself();
static void		c2c_sndrcvself();
static int		c2c_anysrcset();
static int		c2c_setsockblk();
static int		c2c_buffer();
static void		c2c_fillwcards();
static void		c2c_fillmpistat();
static int		c2c_readtosink();
static void		c2c_badsocket();
static int		c2c_commdead();
static int		c2c_socketdance();
static struct c2c_proc 	*c2c_proc_init();
static void		c2c_fillsync();
static int		c2c_readenv();
static int		c2c_write();
static int		c2c_sread();
static int		c2c_swrite();
static int		c2c_swritev();
static int		c2c_sselect();

/*
 * local macros and constants
 *
 * convenience macros
 */
#define	is_write(r)	((r)->ccq_state == C2C_WRITE)
#define	is_read(r)	((r)->ccq_state == C2C_READ)
#define is_sendself(r)	((r)->ccq_state == C2C_SENDSELF)
#define is_ack(e)	((e).cce_flags & C2C_ACK)
#define is_2nd(e)	((e).cce_flags & C2C_2ND)
#define is_ssend(e)	((e).cce_flags & C2C_SSEND)
#define is_longproto(e)	((e).cce_flags & C2C_LONG)
#define req_c2c(r)	((struct c2c_req *) (r)->rq_rpi)
#define req_envl(r)	(((struct c2c_req *) (r)->rq_rpi)->ccq_env)
#define req_sock(r)	(((struct c2c_proc *) (r)->rq_proc->p_rpi)->ccp_sock)
#define proc_sock(p)	(((struct c2c_proc *) (p)->p_rpi)->ccp_sock)

#define SINKLEN		512			/* size of data sink (bytes) */

/*
 * global variables
 */
fd_set			disabledsock;		/* sockets disabled by GER */

/*
 * local variables
 */
static int		ger_error;		/* flags possible GER error */
static fd_set		readset;		/* read sockets */
static fd_set		writeset;		/* write sockets */
static fd_set		exceptset;		/* exception sockets */
static fd_set		blksock;		/* blocked mode socket? */
static fd_set		eofiserror;		/* eof on socket is error? */
static struct timeval	zerotime = { 0, 0 };	/* zero timeval */
static char		sink[SINKLEN];		/* data sink */
static struct c2c_proc  *sock2proc[FD_SETSIZE];	/* map sockets to processes */

/*
 *	_rpi_c2c_init
 *
 *	Function:	- primary initialiation of RPI subsystem
 *			- initialize buffering and socket connections
 *	Returns		- 0 or LAMERROR
 */
int
_rpi_c2c_init()

{
    if (_cbuf_init()) return(LAMERROR);

    if (__lam_ger_) {
	FD_ZERO(&disabledsock);
    }

    return(_rpi_c2c_addprocs());
}

/*
 *	_rpi_c2c_addprocs
 *
 *	Function:	- setup for new processes
 *			- makes socket connections
 *	Returns:	- 0 or LAMERROR
 */
int
_rpi_c2c_addprocs()

{
    return(c2c_socketdance());
}

/*
 *	_rpi_c2c_build
 *
 *	Function:	- builds RPI portion of a request from scratch
 *			- one-time cost separated from _rpi_c2c_start()
 *			  to optimize persistent requests
 *	Accepts:	- request
 *	Returns:	- 0 or LAMERROR
 */
int
_rpi_c2c_build(req)

MPI_Request		req;

{
    struct c2c_req	*creq;			/* C2C request */

    creq = (struct c2c_req *) malloc(sizeof(struct c2c_req));
    if (creq == 0) return(LAMERROR);

    req->rq_rpi = (void *) creq;
    return(0);
}

/*
 *	_rpi_c2c_start
 *
 *	Function:	- initializes RPI dependent aspects of a request
 *			- cost per request start - separated from
 *			  _rpi_c2c_build() to optimize persistent requests
 *	Accepts:	- request list
 *			- request
 *	Returns:	- 0 or LAMERROR
 */
int
_rpi_c2c_start(req_top, req)

MPI_Request		req_top;
MPI_Request		req;

{
    struct c2c_req	*creq;			/* C2C request */
    struct cbuf_msg	*bmsg;			/* buffered message */

    if (c2c_commdead(req)) return(0);

    creq = (struct c2c_req *) req->rq_rpi;
/*
 * set common synch/comm parameters
 * a recv peer may be a wildcard but will be set to the actual peer
 * upon matching an incoming mesage
 */
    creq->ccq_peer = req->rq_rank;
    creq->ccq_env.cce_flags = 0;
    creq->ccq_env.cce_tag = req->rq_tag;
    creq->ccq_env.cce_cid = req->rq_cid;
/*
 * receive specific initialization
 */
    if ((req->rq_type == LAM_RQIRECV) || (req->rq_type == LAM_RQIPROBE)) {

	creq->ccq_state = C2C_READ;
	creq->ccq_env.cce_rank = creq->ccq_peer;

	if ((req->rq_proc == __lam_myproc_)
	    || (req->rq_proc == 0 && ! req->rq_comm->c_inter)) {
/*
 * potential receive from self - look for a matching send to self
 * if unmatched a recv on MPI_ANY_SOURCE must still be checked against
 * buffered msgs
 */
	    if (c2c_recvself(req_top, req) || req->rq_proc == __lam_myproc_) {
		return(0);
	    }
	}
/*
 * check for matching buffered envelope/message
 */
	if ((bmsg = _cbuf_find(&creq->ccq_env))) {
	    return(c2c_readadv(bmsg->cbm_proc, req, bmsg));
	}
    }
/*
 * send specific initialization
 */
    else {
	creq->ccq_env.cce_rank = req->rq_comm->c_group->g_myrank;
	creq->ccq_env.cce_seq = req->rq_seq;
	creq->ccq_env.cce_len = req->rq_packsize;
	creq->ccq_msgbuf = req->rq_packbuf;

	if (req->rq_proc == __lam_myproc_) {
/*
 * send to self
 */
	    creq->ccq_state = C2C_SENDSELF;
	    c2c_sendself(req_top, req);
	}

	else {
/*
 * send to another process
 */
	    creq->ccq_state = C2C_WRITE;

	    if (req->rq_packsize > C2C_SHORTMSGLEN) {
/*
 * long message protocol - no message data is sent with first envelope
 */
		creq->ccq_env.cce_flags |= C2C_LONG;
		creq->ccq_nmsgout = 0;
	    } else {
/*
 * short message protocol
 */
		creq->ccq_nmsgout = req->rq_packsize;
		if (req->rq_type == LAM_RQISSEND) {
		    creq->ccq_env.cce_flags |= C2C_SSEND;
		}
	    }
/*
 * prepare for writing of envelope
 */
	    set_out_envelope(creq);
	}
    }
	
    return(0);
}

/*
 *	_rpi_c2c_destroy
 *
 *	Function:	- destroys RPI portion of request
 *	Accepts:	- request
 *	Returns:	- 0 or LAMERROR
 */
int
_rpi_c2c_destroy(req)

MPI_Request		req;

{
    free((char *) req->rq_rpi);
    req->rq_rpi = 0;
    return(0);
}

/*
 *	_rpi_c2c_advance
 *
 *	Function:	- advances requests in C2C mode
 *			- we try to advance all requests as far as possible
 *			  as allowed by RPI
 *	Accepts:	- request list
 *			- block enable flag
 *	Returns:	- 0 or LAMERROR
 */
int
_rpi_c2c_advance(req_top, flag_block)

MPI_Request		req_top;
int			flag_block;

{
    MPI_Request		req;			/* current request */
    MPI_Request		lastreq;		/* last sockIO request */
    int			ret;			/* return value */
    int			nsio = 0;		/* # requests needing sockIO */
    int			sock;			/* socket descriptor */
    int			maxsock = -1;		/* max. socket descriptor */
    struct c2c_req  	*creq;			/* C2C request */
/*
 * Find which requests require socket IO.
 */
    FD_ZERO(&readset);
    FD_ZERO(&writeset);
    FD_ZERO(&exceptset);
    FD_ZERO(&eofiserror);
    ger_error = 0;

    for (req = req_top; req; req = req->rq_next) {
/*
 * If a blocking request is done we may no longer block.
 */
	if (req->rq_state == LAM_RQSDONE) {
	    if (req->rq_flags & LAM_RQFBLOCK) {
		    flag_block = 0;
	    }
	    continue;
	}

	if (c2c_commdead(req)) continue;
	
	creq = (struct c2c_req *) req->rq_rpi;
/*
 * writing request
 */
	if (is_write(creq)) {

	    sock = req_sock(req);
	    FD_SET(sock, &writeset);
	    FD_SET(sock, &exceptset);
	    if (sock > maxsock) {
		maxsock = sock;
	    }
	}
/*
 * reading request
 */
	else if (is_read(creq)) {

	    if (req->rq_proc == 0) {
		ret = c2c_anysrcset(req, &maxsock, flag_block);
		if (ret < 0) return(LAMERROR);
		if (ret == 0) continue;
	    }
	    else {
		sock = req_sock(req);
		if (sock < 0) continue;
		
		if (__lam_ger_ && FD_ISSET(sock, &disabledsock)) {
		    ger_error = 1;
		    continue;
		}
		
		FD_SET(sock, &readset);
		FD_SET(sock, &exceptset);
		if (sock > maxsock) {
		    maxsock = sock;
		}
	    }
	}
	else {
	    continue;
	}

	lastreq = req;
	nsio++;
    }

    if (ger_error && flag_block) {
	errno = EGERFLOW;
	return(LAMERROR);
    }
    
    if (nsio == 1) {
	return(c2c_adv1(lastreq, maxsock, flag_block));
    }
    else if (nsio > 1) {
	return(c2c_advmany(req_top, maxsock, flag_block));
    }

    return(0);
}

/*
 *      _rpi_c2c_iprobe
 *
 *      Function:       - non-blocking probe
 *                      - public interface for peculiar MPI_Iprobe() which
 *                        does not return a request to the user
 *      Accepts:        - request
 *      Returns:        - 0: no msg, 1: msg, -1: error
 */
int
_rpi_c2c_iprobe(req)

MPI_Request             req;

{
    int			err;			/* error code */
/*
 * Link the probe request and advance as far as possible.
 */
    _mpi_req_add(req);
    _mpi_req_blkclr();
    err = _mpi_req_advance();
    if (err != MPI_SUCCESS) return(-1);
/*
 * Unlink the request.
 */
    _mpi_req_rem(req);
/*
 * A message was found if the request is in the done state.
 */
    return((req->rq_state == LAM_RQSDONE) ? 1 : 0);
}

/*
 *	_rpi_c2c_finalize
 *
 *	Function:	- c2c cleanup
 */
int
_rpi_c2c_finalize()

{
    struct _proc	*p;
    
    _cbuf_end();

    for (p = __lam_topproc_(); p; p = __lam_nextproc_()) {
	
	if (proc_sock(p)) {
	    shutdown(proc_sock(p), 2);
	    close(proc_sock(p));
	}

	free((char *) p->p_rpi);
    }

    return(0);
}

/*
 *	c2c_advmany
 *
 *	Function:	- advance multiple requests where possible
 *	Accepts:	- request list
 *			- max. socket descriptor
 *			- blocking flag
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_advmany(req_top, maxsock, fl_block)

MPI_Request		req_top;
int			maxsock;
int			fl_block;

{
    MPI_Request		req;			/* current request */
    struct c2c_req  	*creq;			/* C2C request */
    int			sock;			/* socket descriptor */
    int			nready;			/* # ready sockets */

    if (fl_block) {
	nready = c2c_sselect(maxsock + 1, &readset, &writeset, &exceptset,
				(struct timeval *) 0);
	if (nready <= 0) return(LAMERROR);
    }
    else {
	nready = c2c_sselect(maxsock + 1, &readset, &writeset, &exceptset,
				&zerotime);
	if (nready < 0) return(0);
    }
/*
 * loop and (i) read and try and advance from read ready sockets
 *          (ii) try advance on write ready sockets which have
 *               a request currently writing on them
 */
    for (sock = 0; nready && sock <= maxsock; sock++) {

	if (FD_ISSET(sock, &exceptset)) {
	    --nready;
	    c2c_badsocket(sock);
	    
	    if (FD_ISSET(sock, &readset)) {
		--nready;
		FD_CLR(sock, &readset);
	    }
	    if (FD_ISSET(sock, &writeset)) {
		--nready;
		FD_CLR(sock, &writeset);
	    }

	    continue;
	}

	if (FD_ISSET(sock, &readset)) {
	    --nready;
	    FD_CLR(sock, &readset);
	    if (c2c_setsockblk(sock, FALSE)) return(LAMERROR);

	    if (c2c_rdtryadv(req_top, (MPI_Request) 0, sock) < 0)
		return(LAMERROR);
	    }

	if (FD_ISSET(sock, &writeset) && sock2proc[sock]->ccp_wtreq) {
	    --nready;
	    FD_CLR(sock, &writeset);

	    if (c2c_wttryadv(sock2proc[sock]->ccp_wtreq))
		return(LAMERROR);
	}
    }
/*
 * if there are still some write ready sockets try and progress requests
 * that want to write on them
 */
    for (req = req_top; nready && req; req = req->rq_next) {
		
	if (req->rq_state == LAM_RQSDONE) continue;

	creq = (struct c2c_req *) req->rq_rpi;

	if (is_write(creq)) {
	    sock = req_sock(req);
	    if (FD_ISSET(sock, &writeset)) {
		--nready;
		FD_CLR(sock, &writeset);
		if (c2c_wttryadv(req)) return(LAMERROR);
	    }
	}
    }

    return(0);
}

/*
 *	c2c_adv1
 *
 *	Function:	- advance a single request optimally
 *	Accepts:	- request to be advanced
 *			- max. socket
 *			- blocking flag
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_adv1(req, maxsock, fl_block)

MPI_Request		req;
int			maxsock;
int			fl_block;

{
    if (c2c_commdead(req)) return(0);
    
    if (is_read(req_c2c(req))) {

	if (req->rq_proc == 0) {
	    return(c2c_adv1anysrc(req, maxsock, fl_block));
	}
	else {
	    if (c2c_setsockblk(maxsock, fl_block)) return(LAMERROR);
/*
 * in blocking mode EOF is considered to be an error
 */
	    if (fl_block) {
		FD_SET(maxsock, &eofiserror);
	    }
	    
	    return((c2c_rdtryadv((MPI_Request) 0, req, maxsock) < 0)
				? LAMERROR : 0);
	}
    }
    else {
	return(c2c_wttryadv(req));
    }
}

/*
 *	c2c_adv1anysrc
 *
 *	Function:	- advance a recv. on MPI_ANY_SOURCE request optimally
 *	Accepts:	- request to be advanced
 *			- max. socket
 *			- blocking flag
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_adv1anysrc(req, maxsock, fl_block)

MPI_Request		req;
int			maxsock;
int			fl_block;

{
    int			nready;			/* # ready sockets */
    int			sock;			/* socket descriptor */
    int			ret;			/* return value */

    if (fl_block) {
	nready = c2c_sselect(maxsock + 1, &readset, (fd_set *) 0,
				&exceptset, (struct timeval *) 0);
	if (nready <= 0) return(LAMERROR);

    } else {
	nready = c2c_sselect(maxsock + 1, &readset, (fd_set *) 0,
				&exceptset, &zerotime);
	if (nready < 0) return(0);
    }
/*
 * try reading and advancing from all possible sources exiting
 * on the the first match
 */
    for (sock = 0; nready && sock <= maxsock; sock++) {

	if (FD_ISSET(sock, &exceptset)) {
	    --nready;
	    c2c_badsocket(sock);
	    
	    if (FD_ISSET(sock, &readset)) {
		--nready;
		FD_CLR(sock, &readset);
	    }
	    if (c2c_commdead(req)) return(0);
	    continue;
	}

	if (FD_ISSET(sock, &readset)) {
	    --nready;
	    FD_CLR(sock, &readset);
	    if (c2c_setsockblk(sock, FALSE)) return(LAMERROR);

	    ret = c2c_rdtryadv((MPI_Request) 0, req, sock);
	    if (ret < 0) return(LAMERROR);
	    if (ret == 1) {
		break;
	    }
	}
    }

    return(0);
}

/*
 *	c2c_wttryadv
 *
 *	Function:	- try to advance a writing request
 *			- does not handle requests where the destination is
 *			  the same as the source
 *	Accepts:	- request
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_wttryadv(req)

MPI_Request		req;

{
    struct c2c_proc	*cpcs;			/* destination C2C process */
    struct c2c_req  	*creq;			/* C2C request */
    int			nbytes;			/* # bytes written */
    int			sock;			/* socket to write on */

    if (c2c_commdead(req)) return(0);
    
    creq = (struct c2c_req *) req->rq_rpi;
    sock = req_sock(req);
    cpcs = (struct c2c_proc *) req->rq_proc->p_rpi;
/*
 * associate request and destination process
 */
    if (! cpcs->ccp_wtreq) {
	cpcs->ccp_wtreq = req;
    }
/*
 * write on socket - what is to be written has already been set up
 */
    nbytes = c2c_write(sock, creq);
    if (nbytes <= 0) return(nbytes);
/*
 * switch on stage in the protocol
 */
    if (is_ack(creq->ccq_env)) {
/*
 * sending an ack
 */
	if (creq->ccq_nenvout == 0) {
/*
 * ack has been sent - if long protocol and message len > 0 prepare to
 * receive message tail otherwise done
 */
	    cpcs->ccp_wtreq = 0;
	    if (is_longproto(creq->ccq_env) &&
		creq->ccq_env.cce_len > 0) {
				
		creq->ccq_state = C2C_READ;
		creq->ccq_env.cce_flags &= ~C2C_ACK;
		creq->ccq_env.cce_flags |= C2C_2ND;
		creq->ccq_env.cce_rank = creq->ccq_peer;
	    }
	    else {
		req->rq_state = LAM_RQSDONE;
	    }
	}
    }
/*
 * sending tail of a long protocol message
 */	
    else if (is_2nd(creq->ccq_env)) {

	if (creq->ccq_nmsgout == 0) {
	    cpcs->ccp_wtreq = 0;
	    req->rq_state = LAM_RQSDONE;
	}
    }
/*
 * sending long protocol first envelope
 */
    else if (is_longproto(creq->ccq_env)) {

	if (req->rq_state == LAM_RQSSTART) {
	    req->rq_state = LAM_RQSACTIVE;
	}
		
	if (creq->ccq_nenvout == 0) {
/*
 * prepare to read ack.
 */
	    cpcs->ccp_wtreq = 0;
	    creq->ccq_state = C2C_READ;
	    creq->ccq_env.cce_flags |= C2C_ACK;
	    creq->ccq_env.cce_rank = creq->ccq_peer;
	}
    }
/*
 * short protocol
 */
    else {
	if (req->rq_state == LAM_RQSSTART) {
	    req->rq_state = LAM_RQSACTIVE;
	}

	if (creq->ccq_nenvout == 0 && creq->ccq_nmsgout == 0) {
	    cpcs->ccp_wtreq = 0;
	    if (is_ssend(creq->ccq_env)) {
/*
 * synchronous send - wait for ack.
 */
		creq->ccq_state = C2C_READ;
		creq->ccq_env.cce_flags |= C2C_ACK;
		creq->ccq_env.cce_rank = creq->ccq_peer;
	    } else {
		req->rq_state = LAM_RQSDONE;
	    }
	}
    }

    return(0);
}

/*
 *	c2c_rdtryadv
 *
 *	Function:	- try to advance reading from a socket
 *	Accepts:	- request list (0 if specific request to be matched)
 *			- request to match (0 if whole list)
 *			- socket
 *	Returns:	- 1 (matched new envelope), 0 (no match), or LAMERROR
 */
static int
c2c_rdtryadv(req_top, mreq, sock)

MPI_Request		req_top;
MPI_Request		mreq;
int			sock;

{
    MPI_Request		req;			/* request */
    struct c2c_req	*creq;			/* C2C request */
    struct c2c_proc 	*cpcs;			/* C2C process */
    int			nbytes;			/* # bytes read */

    cpcs = sock2proc[sock];
/*
 * reading the message body
 */
    if (cpcs->ccp_nmsgin || cpcs->ccp_extra) {
	return(c2c_readbody(cpcs));
    }
/*
 * reading envelope
 */
    nbytes = c2c_readenv(cpcs);
    if (nbytes <= 0) return(nbytes);
    if (cpcs->ccp_nenvin > 0) return(0);
/*
 * we have a complete envelope
 * find matching read request if any and advance
 */
    envelope_ttol(&cpcs->ccp_env);
    
    if (mreq && ! _c2c_envl_cmp(&cpcs->ccp_env, &req_envl(mreq))) {
	return(c2c_readadv(cpcs, mreq, (struct cbuf_msg *) 0) ? LAMERROR : 1);
    }
/*
 * there cannot be any matching recvs after a matching probe hence we can
 * return upon the first match (buffering the envelope in the case of a
 * probe) and maintain the invariant of no requests in the list which match
 * buffered envelopes
 */
    for (req = req_top; req; req = req->rq_next) {

	if (req->rq_state == LAM_RQSDONE) continue;
	
	creq = (struct c2c_req *) req->rq_rpi;

	if (is_read(creq) && ! _c2c_envl_cmp(&cpcs->ccp_env, &creq->ccq_env)) {
	    return((c2c_readadv(cpcs, req, (struct cbuf_msg *) 0))
				? LAMERROR : 1);
	}
    }
/*
 * no requests match - buffer the envelope (+ message if short protocol)
 */
    if (! is_longproto(cpcs->ccp_env)) {
	cpcs->ccp_nmsgin = cpcs->ccp_env.cce_len;
    }

    return(c2c_buffer(cpcs));
}

/*
 *	c2c_readadv
 *
 *	Function:	- advance a reading request
 *			- either we have just finished reading an envelope
 *			  from the source process or are processing a buffered
 *			  envelope from ther source process
 *	Accepts:	- source process
 *			- request
 *			- buffered matching env+msg (NULL => read envelope)
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_readadv(cpcs, req, bmsg)

struct c2c_proc		*cpcs;
MPI_Request		req;
struct cbuf_msg		*bmsg;

{
    struct c2c_req  	*creq;			/* C2C request */
    struct c2c_envl 	*env;			/* matching incoming env. */
    int			len;			/* message length */
    int			extra;			/* # bytes to go into sink */
    int			nread;			/* # bytes read */
    
    env = (bmsg) ? &bmsg->cbm_env : &cpcs->ccp_env;
    creq = (struct c2c_req *) req->rq_rpi;
/*
 * switch according to state of the protocol
 * acks and long message tails are never buffered
 */
    if (is_ack(creq->ccq_env)) {
/*
 * received ack - if long protocol send the tail else done
 */
	if (is_longproto(creq->ccq_env)) {
/*
 * set message length to minimum of what sender and receiver specified.
 */
	    if (req->rq_packsize < env->cce_len) {
		creq->ccq_nmsgout = req->rq_packsize;
	    } else {
		creq->ccq_nmsgout = env->cce_len;
	    }

	    if (creq->ccq_nmsgout == 0) {
		req->rq_state = LAM_RQSDONE;
	    } else {
		creq->ccq_state = C2C_WRITE;
		creq->ccq_env.cce_len = creq->ccq_nmsgout;
		creq->ccq_env.cce_flags &= ~C2C_ACK;
		creq->ccq_env.cce_flags |= C2C_2ND;
		creq->ccq_env.cce_rank = req->rq_comm->c_group->g_myrank;
		set_out_envelope(creq);
	    }
	}
	else {
	    req->rq_state = LAM_RQSDONE;
	}
/*
 * prepare to read next envelope
 */
	cpcs->ccp_envbuf = (char *) &cpcs->ccp_env;
	cpcs->ccp_nenvin = sizeof(struct c2c_envl);
    }
	
    else if (is_2nd(creq->ccq_env)) {
/*
 * long protocol 2nd part - suck in the message
 */
	c2c_fillmpistat(req, env->cce_rank, env->cce_tag, env->cce_len);
	cpcs->ccp_rdreq = req;
	cpcs->ccp_nmsgin = env->cce_len;
	cpcs->ccp_msgbuf = req->rq_packbuf;
	cpcs->ccp_extra = 0;

	return(c2c_readbody(cpcs));
    }

    else if (is_longproto(*env)) {
/* 		
 * long protocol first part - if a probe buffer envelope else send ack.
 */
	if (req->rq_state == LAM_RQSSTART) {
	    req->rq_state = LAM_RQSACTIVE;
	}

	c2c_fillwcards(req, env);

	if (req->rq_type == LAM_RQIPROBE) {

	    c2c_fillmpistat(req, env->cce_rank, env->cce_tag, env->cce_len);
	    req->rq_state = LAM_RQSDONE;

	    return((bmsg) ? 0 : c2c_buffer(cpcs));
	}
	else {
	    if (env->cce_len > req->rq_packsize) {
		req->rq_flags |= LAM_RQFTRUNC;
	    }

	    creq->ccq_state = C2C_WRITE;
	    creq->ccq_env.cce_flags |= (C2C_ACK | C2C_LONG);
	    creq->ccq_env.cce_len = req->rq_packsize;
	    creq->ccq_env.cce_rank = req->rq_comm->c_group->g_myrank;
	    creq->ccq_nmsgout = 0;
	    set_out_envelope(creq);
/*
 * prepare to read next envelope
 */
	    cpcs->ccp_envbuf = (char *) &cpcs->ccp_env;
	    cpcs->ccp_nenvin = sizeof(struct c2c_envl);
/*
 * discard envelope if it was buffered - there is no message part
 */
	    if (bmsg) {
		_cbuf_delete(bmsg);
	    }
	}
    }

    else {
/*
 * short protocol
 */
	if (req->rq_state == LAM_RQSSTART) {
	    req->rq_state = LAM_RQSACTIVE;
	}

	c2c_fillwcards(req, env);
		
	if (req->rq_type == LAM_RQIPROBE) {
/*
 * probing - the request is done, set the MPI status and buffer the msg
 */
	    c2c_fillmpistat(req, env->cce_rank, env->cce_tag, env->cce_len);
	    req->rq_state = LAM_RQSDONE;

	    if (bmsg) {
		return(0);
	    } else {
		cpcs->ccp_nmsgin = env->cce_len;
		return(c2c_buffer(cpcs));
	    }
	}
/*
 * check for length mismatch and set what's to be read
 */
	if (env->cce_len < req->rq_packsize) {
	    len = env->cce_len;
	    extra = 0;
	}
	else {
	    len = req->rq_packsize;
	    extra = env->cce_len - req->rq_packsize;
	    if (extra) {
		req->rq_flags |= LAM_RQFTRUNC;
	    }
	}

	c2c_fillmpistat(req, env->cce_rank, env->cce_tag, len);
/*
 * if we have received from a sync send record this so that an
 * ack will be sent upon completion of reading the body
 */
	if (is_ssend(*env)) {
	    creq->ccq_env.cce_flags |= C2C_SSEND;
	}

	if (bmsg) {
/*
 * buffered message - copy what's been read so far and adjust the
 * read counts to complete the transfer
 */
	    if (cpcs->ccp_rdreq == req) {
/*
 * still reading the message body
 */
		nread = env->cce_len - (cpcs->ccp_nmsgin + cpcs->ccp_extra);
	    } else {
		nread = env->cce_len;
	    }
	    
	    memcpy(req->rq_packbuf, bmsg->cbm_buf, min(nread,len));

	    if (nread < len) {
/*
 * still more message body to be read
 */
		cpcs->ccp_nmsgin = len - nread;
		cpcs->ccp_extra = extra;
		cpcs->ccp_msgbuf = req->rq_packbuf + nread;
	    } else {
/*
 * some of the extra was read while buffering
 */
		cpcs->ccp_nmsgin = 0;
		cpcs->ccp_extra = extra - (nread - len);
	    }
/*
 * discard the buffered message
 */
	    if (bmsg->cbm_buf) {
		free(bmsg->cbm_buf);
	    }
	    _cbuf_delete(bmsg);
	}
	else {
/*
 * not a buffered message - set up to read the body (if any)
 */
	    cpcs->ccp_rdreq = req;
	    cpcs->ccp_nmsgin = len;
	    cpcs->ccp_extra = extra;
	    cpcs->ccp_msgbuf = req->rq_packbuf;
	}

	if (cpcs->ccp_rdreq == req) {
	    return(c2c_readbody(cpcs));
	} else {
	    c2c_bodydone(req);
	}
    }
	
    return(0);
}

/*
 *	c2c_readbody
 *
 *	Function:	- read message body from process
 *			- handle retries and error cases
 *	Accepts:	- process
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_readbody(cpcs)

struct c2c_proc		*cpcs;

{
    int			nbytes;			/* # bytes read */
/*
 * still message body to be read
 */
    if (cpcs->ccp_nmsgin > 0) {
	nbytes = c2c_sread(cpcs->ccp_sock, cpcs->ccp_msgbuf, cpcs->ccp_nmsgin);
	if (nbytes <= 0) return(nbytes);

	cpcs->ccp_nmsgin -= nbytes;
	cpcs->ccp_msgbuf += nbytes;
    }

    if (cpcs->ccp_nmsgin == 0) {
/*
 * read any extra (truncation) that's left
 */
	if (cpcs->ccp_extra) {
	    nbytes = c2c_readtosink(cpcs);
	    if (nbytes <= 0) return(nbytes);
	    if (cpcs->ccp_extra > 0) return(0);
	}
/*
 * all the message and any extra has been read
 * if reading for a request change its state
 */
	if (cpcs->ccp_rdreq) {
	    c2c_bodydone(cpcs->ccp_rdreq);
	    cpcs->ccp_rdreq = 0;
	}
/*
 * prepare to read next envelope
 */
	cpcs->ccp_nenvin = sizeof(struct c2c_envl);
	cpcs->ccp_envbuf = (char *) &cpcs->ccp_env;
    }

    return(0);
}

/*
 *	c2c_bodydone
 *
 *	Function:	- change the state of a recv request whose
 *			  body has been completely read
 *	Accepts:	- request
 */
static void
c2c_bodydone(req)

MPI_Request		req;

{
    struct c2c_req	*creq;
    
    creq = (struct c2c_req *) req->rq_rpi;
/*
 * if synchronous send then return an ack
 */
    if (is_ssend(creq->ccq_env)) {
	creq->ccq_state = C2C_WRITE;
	creq->ccq_env.cce_flags |= C2C_ACK;
	creq->ccq_env.cce_rank = req->rq_comm->c_group->g_myrank;
	creq->ccq_nmsgout = 0;
	set_out_envelope(creq);
    }
    else {
	req->rq_state = LAM_RQSDONE;
    }
}

/*
 *	c2c_sendself
 *
 *	Function:	- advance read request matching send to self (if any)
 *	Accepts:	- request list
 *			- send request
 */
static void
c2c_sendself(req_top, send)

MPI_Request		req_top;
MPI_Request		send;

{
    MPI_Request     	recv;			/* receive request */
    struct c2c_envl	*senv;			/* send request envelope */

    senv = &req_envl(send);
/*
 * look for inactive matching receive/probe and advance
 */
    for (recv = req_top; recv; recv = recv->rq_next) {
	if ((recv->rq_state == LAM_RQSSTART) && is_read(req_c2c(recv))
			&& ! _c2c_envl_cmp(senv, &req_envl(recv))) {
	    c2c_sndrcvself(send, recv);
	    return;
	}
    }
}

/*
 *	c2c_recvself
 *
 *	Function:	- advance recv request and matching send
 *                        to self (if any)
 *	Accepts:	- request list
 *			- recv request
 *	Returns:	- 1 (match), 0 (no match)
 */
static int
c2c_recvself(req_top, recv)

MPI_Request		req_top;
MPI_Request		recv;

{
    MPI_Request     	send;			/* send request */
    struct c2c_envl	*renv;			/* recv request envelope */

    renv = &req_envl(recv);
/*
 * look for matching send and advance if found
 */
    for (send = req_top; send; send = send->rq_next) {
	if (send->rq_state == LAM_RQSDONE) continue;
	
	if (is_sendself(req_c2c(send))
		&& ! _c2c_envl_cmp(&req_envl(send), renv)) {

	    c2c_sndrcvself(send, recv);
	    return(1);
	}
    }
    return(0);
}

/*
 *	c2c_sndrcvself
 *
 *	Function:	- advance matching send and receive
 *	Accepts:	- send request
 *			- receive request
 */
static void
c2c_sndrcvself(send, recv)

MPI_Request		send;
MPI_Request		recv;

{
    int			len;			/* # bytes to transfer */
    struct c2c_req  	*c2c_snd;		/* C2C send request */

    c2c_snd = req_c2c(send);
    recv->rq_seq = send->rq_seq;

    if (recv->rq_type == LAM_RQIPROBE) {
/*
 * probe - don't actually receive
 * the message is not buffered but the sender is left in the
 * request list to be matched by a later receive
 */
	c2c_fillmpistat(recv, c2c_snd->ccq_env.cce_rank,
			c2c_snd->ccq_env.cce_tag, c2c_snd->ccq_env.cce_len);
		
	recv->rq_state = LAM_RQSDONE;
    }
    else {
/*
 * not a probe - do the transfer
 *
 * check for mismatched message lengths
 */
	if (send->rq_packsize > recv->rq_packsize) {
	    recv->rq_flags |= LAM_RQFTRUNC;
	    len = recv->rq_packsize;
	} else {
	    len = send->rq_packsize;
	}
/*
 * transfer message
 */
	memcpy(recv->rq_packbuf, send->rq_packbuf, len);

	c2c_fillmpistat(recv, c2c_snd->ccq_env.cce_rank,
			c2c_snd->ccq_env.cce_tag, len);
/*	
 * both sender and receiver are now done
 */
	send->rq_state = recv->rq_state = LAM_RQSDONE;
    }
}

/*
 *	c2c_anysrcset
 *
 *	Function:	- set the read set for all processes
 *			  in the peer group of a receive request
 *			  that has specified MPI_ANY_SOURCE
 *	Accepts:	- request
 *			- maximum socket in set (modified)
 *			- blocking allowed flag
 *	Returns:	- # sockets to be checked, or LAMERROR
 */
static int
c2c_anysrcset(req, maxsock, flag_block)

MPI_Request		req;
int			*maxsock;
int			flag_block;

{
    struct _proc	**p;			/* favourite pointer */
    struct _group	*g;			/* peer group */
    int			i;			/* favourite index */
    int			sock;			/* socket descriptor */
    int			sockio = 0;		/* socket IO to be done? */

    g = (req->rq_comm->c_inter)
	? req->rq_comm->c_rgroup : req->rq_comm->c_group;
	
    for (i = g->g_nprocs, p = g->g_procs; i > 0; i--, p++) {
	if ((sock = proc_sock(*p)) < 0) {
	    continue;
	}

	if (__lam_ger_ && FD_ISSET(sock, &disabledsock)) {
		ger_error = 1;
		continue;
	}

	FD_SET(sock, &readset);
	FD_SET(sock, &exceptset);
	sockio++;
	if (sock > *maxsock) {
	    *maxsock = sock;
	}
    }

    return(sockio);
}

/*
 *	c2c_setsockblk
 *
 *	Function:	- set socket to blk/non-blk mode
 *	Accepts:	- socket descriptor
 *			- blocking mode flag
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_setsockblk(sock, fl_block)

int			sock;
int			fl_block;

{
    if (fl_block) {
/*
 * Switch socket to blocking if needed.
 */
	if ( ! FD_ISSET(sock, &blksock)) {
	    if (fcntl(sock, F_SETFL, 0)) return(LAMERROR);
	    FD_SET(sock, &blksock);
	}
    }
    else {
/*
 * Switch socket to non-blocking if needed.
 */
	if (FD_ISSET(sock, &blksock)) {
	    if (fcntl(sock, F_SETFL, O_NONBLOCK)) return(LAMERROR);
	    FD_CLR(sock, &blksock);
	}
    }

    return(0);
}

/*
 *	c2c_buffer
 *
 *	Function:	- buffer incoming envelopes/messages
 *			- there is never any data to be read to the
 *                        data sink when buffering
 *	Accepts:	- process envelope came in from
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_buffer(cpcs)

struct c2c_proc		*cpcs;

{
    struct cbuf_msg 	msg;			/* buffer list entry */

    msg.cbm_proc = cpcs;
    msg.cbm_env = cpcs->ccp_env;
	
    if (cpcs->ccp_nmsgin > 0) {
/*
 * set up the buffer for the message body and read as much as possible
 */
	msg.cbm_buf = (char *) malloc(cpcs->ccp_nmsgin);
	if (msg.cbm_buf == 0) return(LAMERROR);

	cpcs->ccp_msgbuf = msg.cbm_buf;
	if (c2c_readbody(cpcs) < 0) return(LAMERROR);
    }
    else {
/*
 * no message body - prepare to read next envelope from process
 */
	msg.cbm_buf = 0;
	cpcs->ccp_nenvin = sizeof(struct c2c_envl);
	cpcs->ccp_envbuf = (char *) &cpcs->ccp_env;
    }

    return(_cbuf_append(&msg));
}

/*
 *	c2c_fillwcards
 *
 *	Function:	- replace wildcards in request with matched values
 *			  and fill in the sequence number
 *	Accepts:	- request
 *			- matched envelope
 */
static void
c2c_fillwcards(req, env)

MPI_Request		req;
struct c2c_envl		*env;

{
    struct _group	*g;			/* peer group */
    struct c2c_req  	*creq;			/* C2C request */
		
    req->rq_seq = env->cce_seq;
    creq = (struct c2c_req *) req->rq_rpi;
    
    if (creq->ccq_env.cce_tag == MPI_ANY_TAG) {
	creq->ccq_env.cce_tag = env->cce_tag;
    }

    if (creq->ccq_peer == MPI_ANY_SOURCE) {
	creq->ccq_peer = env->cce_rank;
	creq->ccq_env.cce_rank = env->cce_rank;
	
	g = (req->rq_comm->c_inter)
	    ? req->rq_comm->c_rgroup : req->rq_comm->c_group;

	if (creq->ccq_peer < g->g_nprocs) {
	    req->rq_proc = g->g_procs[creq->ccq_peer];
	}
    }
}

/*
 *	c2c_fillmpistat
 *
 *	Function:	- fill in the MPI status object
 *	Accepts:	- request
 *			- rank
 *			- tag
 *			- message length
 */
static void
c2c_fillmpistat(req, rank, tag, length)

MPI_Request		req;
int			rank;
int			tag;
int			length;

{
    req->rq_status.MPI_SOURCE = rank;
    req->rq_status.MPI_TAG = tag;
    req->rq_status.st_length = length;
}

/*
 *	_c2c_envl_cmp
 *
 *	Function:	- check if envelopes match
 *			- second envelope may contain wildcards and first
 *			  may not
 *	Accepts:	- ptr to envelope
 *			- ptr to request envelope
 *	Returns:	- 0 if match, 1 if not
 */
int
_c2c_envl_cmp(pe, pq)

struct c2c_envl		*pe, *pq;

{
    if ((pe->cce_cid == pq->cce_cid)
	&& ((pe->cce_rank == pq->cce_rank) || (pq->cce_rank == MPI_ANY_SOURCE))
	&& ((pe->cce_tag == pq->cce_tag) || (pq->cce_tag == MPI_ANY_TAG))
	&& ((pe->cce_flags & C2C_ACK) == (pq->cce_flags & C2C_ACK))
	&& ((pe->cce_flags & C2C_2ND) == (pq->cce_flags & C2C_2ND))) {

	return(0);
    }

    return(1);
}

/*
 *	2c_socketdance
 *
 *	Function:	- make C2C socket connections to all processes
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_socketdance()

{
    struct _proc	*p;			/* favourite pointer */
    struct _gps		*mygps;			/* my GPS */
    struct nmsg		inmsg;			/* incoming network msg hdr */
    struct nmsg		outmsg;			/* outgoing network msg hdr */
    struct dolink	*links;			/* links to neighbours */
    int4		nlinks;			/* number of links */
    int			sock;			/* socket descriptor */
    int			servsockd;		/* server socket descriptor */
    int			servport = 0;		/* server port number */
    int			rnode;			/* remote node */
    int			rport;			/* remote port */
    int 		flag;			/* for setting socket opts */
    int			bufsize;		/* C2C socket buffer size */
    int			defsize;		/* kernel default buf. size */
    int			optlen;			/* socket option length */
    unsigned char	*raddr;			/* remote host address */
    struct c2c_proc	*cpcs;			/* C2C specific process data */

    if (__lam_nprocs_() > 1) {
/*
 * Get links to neighbours, initialize server socket, message headers, etc.
 */
	if (ldogetlinks(&links, &nlinks)) return(LAMERROR);

	servsockd = getsrvstrsock(5, &servport);
	if (servsockd < 0) return(LAMERROR);

	mygps = &__lam_myproc_->p_gps;

	inmsg.nh_length = 0;
	outmsg.nh_length = 0;
	outmsg.nh_data[0] = (int4) servport;

	bufsize = C2C_SHORTMSGLEN + sizeof(struct c2c_envl);
    }
/*
 * Loop through all processes, initializing the process data and
 * connecting to those not already connected to.
 */
    p = __lam_topproc_();

    while (p) {

	cpcs = c2c_proc_init(p);
	if (cpcs == 0) return(LAMERROR);
		
	if ((p != __lam_myproc_) && (cpcs->ccp_sock < 0)) {
		
	    if (LAM_GPSCMP(mygps, &p->p_gps) < 0) {
/*
 * Act as a client. First get remote server port number and then connect.
 */
		c2c_fillsync(p, __lam_myproc_, &inmsg);
		if (nrecv(&inmsg)) return(LAMERROR);

		rport = (int) inmsg.nh_data[0];
		rnode = p->p_gps.gps_node;
				
		if (rnode > nlinks) return(LAMERROR);

		raddr = (unsigned char *)
			&links[rnode].dol_addr.sin_addr.s_addr;
				
		sock = getcltstrsock(raddr, rport);
		if (sock < 0) return(LAMERROR);
	    }
	    else {
/*
 * Act as a server. First send server port number to client and then accept.
 */
		c2c_fillsync(__lam_myproc_, p, &outmsg);
		if (nsend(&outmsg)) return(LAMERROR);
				
		sock = accepttmout(servsockd, -1);
		if (sock < 0) return(LAMERROR);
	    }

	    cpcs->ccp_sock = sock;
	    sock2proc[sock] = cpcs;
/*
 * sockets initially in non-blocking mode
 * set send and receive buffer sizes
 */
	    flag = 1;
			
	    if (fcntl(sock, F_SETFL, O_NONBLOCK)) return(LAMERROR);
	    FD_CLR(sock, &blksock);

	    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
			(char *) &flag, sizeof(flag)))
		return(LAMERROR);

#if !defined(I386_LINUX)
	    optlen = sizeof(defsize);
	    if (getsockopt(sock, SOL_SOCKET, SO_SNDBUF,
			(char *) &defsize, &optlen))
		return(LAMERROR);

	    if (defsize < bufsize) {
		if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF,
			    (char *) &bufsize, sizeof(bufsize)))
		return(LAMERROR);
	    }
	    
	    optlen = sizeof(defsize);
	    if (getsockopt(sock, SOL_SOCKET, SO_RCVBUF,
			(char *) &defsize, &optlen))
		return(LAMERROR);

	    if (defsize < bufsize) {
		if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
		    	    (char *) &bufsize, sizeof(bufsize)))
		return(LAMERROR);
	    }
#endif
	}

	p = __lam_nextproc_();
    }

    if (__lam_nprocs_() > 1) {
	close(servsockd);
    }

    return(0);
}

/*
 *	c2c_proc_init
 *
 *	Function:	- initialize C2C specific process data
 *			- if already initialized return ptr to existing data
 *	Accepts:	- ptr process
 *	Returns:	- ptr C2C data or NULL
 */
static struct c2c_proc *
c2c_proc_init(proc)

struct _proc	*proc;

{
    struct c2c_proc	*cpcs;			/* C2C specific data */
	
    if (proc->p_rpi == 0) {
	cpcs = (struct c2c_proc *) malloc(sizeof(struct c2c_proc));
	if (cpcs == 0) return(0);

	cpcs->ccp_sock = -1;
	cpcs->ccp_nbfde = 0;
	cpcs->ccp_rdreq = 0;
	cpcs->ccp_wtreq = 0;
	cpcs->ccp_nmsgin = 0;
	cpcs->ccp_extra = 0;
/*
 * start out ready to read in an envelope
 */
	cpcs->ccp_envbuf = (char *) &cpcs->ccp_env;
	cpcs->ccp_nenvin = sizeof(struct c2c_envl);
	proc->p_rpi = cpcs;
    }
    else {
	cpcs = proc->p_rpi;
    }

    return(cpcs);
}

/*
 *	c2c_fillsync
 *
 *	Function:	- fill in network message sync for C2C socket setup
 *	Accepts:	- source process
 *			- destination process
 *			- network message header (filled)
 */
static void
c2c_fillsync(src, dest, head)

struct _proc            *src;
struct _proc            *dest;
struct nmsg             *head;

{
    head->nh_node = dest->p_gps.gps_node;
    head->nh_event = dest->p_gps.gps_idx & 0x7FFF;
    head->nh_type = (((src->p_gps.gps_idx & 0x1F) << 10) |
                                (src->p_gps.gps_node & 0x1FF)) << 16;
/*
 * Add N2ND flag to fake out mpimsg.
 */
    head->nh_flags = KSYNCSQL | N2ND;
}

/*
 *	c2c_readenv
 *
 *	Function:	- read envelope from a process
 *	Accepts:	- process
 *	Returns:	- # bytes read or LAMERROR
 */
static int
c2c_readenv(cpcs)

struct c2c_proc  	*cpcs;

{
    int			nbytes;			/* # bytes read */

    nbytes = c2c_sread(cpcs->ccp_sock, cpcs->ccp_envbuf, cpcs->ccp_nenvin);

    if (nbytes > 0) {
	cpcs->ccp_nenvin -= nbytes;
	cpcs->ccp_envbuf += nbytes;
    }

    return(nbytes);
}

/*
 *	c2c_write
 *
 *	Function:	- write request envelope/message on socket 
 *	Accepts:	- socket
 *			- C2C request
 *	Returns:	- # bytes written or LAMERROR
 */
static int
c2c_write(sock, creq)

int			sock;
struct c2c_req  	*creq;

{
    struct iovec    	iov[2];                 /* IO vector */
    int             	nbytes;                 /* # bytes written */
        
    if (creq->ccq_nenvout > 0 && creq->ccq_nmsgout > 0) {
 
	iov[0].iov_base = (caddr_t) creq->ccq_envbuf;
	iov[0].iov_len = creq->ccq_nenvout;
 
	iov[1].iov_base = (caddr_t) creq->ccq_msgbuf;
	iov[1].iov_len = creq->ccq_nmsgout;
 
	nbytes = c2c_swritev(sock, iov, 2);
	if (nbytes > 0) {
	    if (nbytes >= creq->ccq_nenvout) {
		creq->ccq_msgbuf += (nbytes - creq->ccq_nenvout);
		creq->ccq_nmsgout -= (nbytes - creq->ccq_nenvout);
		creq->ccq_nenvout = 0;
	    } else {
		creq->ccq_envbuf += nbytes;
		creq->ccq_nenvout -= nbytes;
	    }                       
	}
    }
    else if (creq->ccq_nmsgout > 0) {
 	nbytes = c2c_swrite(sock, creq->ccq_msgbuf, creq->ccq_nmsgout);
	if (nbytes > 0) {
	    creq->ccq_msgbuf += nbytes;
	    creq->ccq_nmsgout -= nbytes;
	}
    }
    else {
	nbytes = c2c_swrite(sock, creq->ccq_envbuf, creq->ccq_nenvout);
	if (nbytes > 0) {
	    creq->ccq_envbuf += nbytes;
	    creq->ccq_nenvout -= nbytes;
	}
    }
 
    return(nbytes);
}

/*
 *	c2c_readtosink
 *
 *	Function:	- read bytes from process socket into data sink
 *	Accepts:	- process
 *	Returns:	- 0 or LAMERROR
 */
static int
c2c_readtosink(cpcs)

struct c2c_proc		*cpcs;

{
    int			nbytes;			/* # bytes read */

    do {
	nbytes = c2c_sread(cpcs->ccp_sock, sink, min(cpcs->ccp_extra,SINKLEN));
	if (nbytes <= 0) return(nbytes);
	cpcs->ccp_extra -= nbytes;
    }
    while (cpcs->ccp_extra > 0);

    return(0);
}

/*
 *	c2c_sread
 *
 *	Function:	- atomic socket read()
 *	Returns:	- # bytes read or LAMERROR
 */
static int
c2c_sread(sock, buf, nbytes)

int			sock;
char			*buf;
int			nbytes;

{
    int			nread = 0;		/* # of bytes read */
    int			r;
    double		starttime;		/* time blocking starts */

    if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) && FD_ISSET(sock, &blksock)) {
	starttime = ttime();
    }

    do {
	r = read(sock, buf, nbytes);

	if (r < 0) {
	    if (errno == EAGAIN) {
		break;
	    }
	    if (errno != EINTR) {
		c2c_badsocket(sock);
		return(0);
	    }
	}
	else if ((r == 0) && (nbytes > 0)) {	/* eof */
	    if (FD_ISSET(sock, &eofiserror)) {
		c2c_badsocket(sock);
	    }
	    return(0);
	}
	else {
	    nread += r;
	    buf += r;
	    nbytes -= r;
	}
    } while (nbytes > 0);

    if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) && FD_ISSET(sock, &blksock)) {
	_kio.ki_blktime += (ttime() - starttime);
    }

    FD_CLR(sock, &eofiserror);
    return(nread);
}

/*
 *	c2c_swrite
 *
 *	Function:	- atomic socket write()
 *	Returns:	- # bytes written or LAMERROR
 */
static int
c2c_swrite(sock, buf, nbytes)

int			sock;
char			*buf;
int			nbytes;

{
    int			nwritten = 0;		/* # of bytes written */
    int			r;

    do {
	r = write(sock, buf, nbytes);

	if (r < 0) {
	    if (errno == EAGAIN) {
		return(nwritten);
	    }
	    if (errno != EINTR) {
		c2c_badsocket(sock);
		return(0);
	    }
	} else if ((r == 0) && (nbytes > 0)) {	/* eof */
            c2c_badsocket(sock);
	    return(0);
	} else {
	    nwritten += r;
	    buf += r;
	    nbytes -= r;
	}
    } while (nbytes > 0);

    return(nwritten);
}

/*
 *	c2c_swritev
 *
 *	Function:	- atomic socket writev()
 *	Returns:	- #bytes written or 0
 */
static int
c2c_swritev(sock, iov, iovcnt)

int			sock;
struct iovec		*iov;
int			iovcnt;

{
    int			nwritten = 0;	/* # of bytes written */
    int			r;
    int			savelen;	/* save full length */
    char		*savebase;	/* save original base ptr */

    if (iovcnt > 0) {
	savelen = iov->iov_len;
	savebase = iov->iov_base;
    }

    do {
	r = writev(sock, iov, iovcnt);

	if (r < 0) {
	    if (errno == EAGAIN) {
		return(nwritten);
	    }
	    if (errno != EINTR) {
		iov->iov_len = savelen;
		iov->iov_base = savebase;
		c2c_badsocket(sock);
		return(0);
	    }
	}
	else if (r == 0) {	/* eof */
	    iov->iov_len = savelen;
	    iov->iov_base = savebase;

	    while ((iovcnt > 0) && (iov->iov_len == 0)) {
		iov++;
		iovcnt--;
	    }

	    if (iovcnt > 0) {
		c2c_badsocket(sock);
		return(0);
	    }

	    return(nwritten);
	}
	else {
	    nwritten += r;

	    while (r > 0) {

		if (r >= iov->iov_len) {
		    r -= iov->iov_len;
		    iov->iov_len = savelen;
		    iov->iov_base = savebase;
		    iov++;
		    iovcnt--;

		    if (iovcnt > 0) {
			savelen = iov->iov_len;
			savebase = iov->iov_base;
		    }
		} else {
		    iov->iov_len -= r;
		    iov->iov_base = (char *) iov->iov_base + r;
		    r = 0;
		}
	    }
	}
    } while (iovcnt > 0);

    return(nwritten);
}

/*
 *	c2c_sselect
 *
 *	Function:	- atomic socket select()
 *	Returns:	- # ready descriptors or LAMERROR
 */
static int
c2c_sselect(width, readfds, writefds, exceptfds, timeout)

int 			width;
fd_set 			*readfds;
fd_set 			*writefds;
fd_set 			*exceptfds;
struct timeval 		*timeout;

{
    int			nready;			/* number ready descriptors */
    double		starttime;		/* time blocking starts */

    if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) && (timeout == 0)) {
	starttime = ttime();
    }
    
    do {
	nready = select(width, readfds, writefds, exceptfds, timeout);

	if (nready < 0 && errno != EINTR) return(LAMERROR);
    }
    while (nready < 0);

    if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) && (timeout == 0)) {
	_kio.ki_blktime += (ttime() - starttime);
    }
    
    return(nready);
}

/*
 *      c2c_commdead
 *
 *      Function:       - checks request for dead communicator
 *      Accepts:        - request
 *      Returns:        - 1 if dead, 0 if OK
 */
static int
c2c_commdead(req)

MPI_Request             req;
 
{
    if (req->rq_comm->c_status & (LAM_CLDEAD | LAM_CRDEAD)) {
	req->rq_state = LAM_RQSDONE;
	req->rq_status.MPI_ERROR = __lam_mkerr_(MPI_ERR_OTHER, EDEADNODE);
	return(1);
    } else {
	return(0);
    }
}

/*
 *	c2c_badsocket
 *
 *	Function:	- invalidate communicators with processes on the node
 *			  a bad socket is connected to
 *	Accepts:	- bad socket descriptor
 */
static void
c2c_badsocket(sock)

int			sock;

{
    struct _proc    *proc;			/* a process */
    struct c2c_proc *cpcs;			/* C2C part of process */

    for (proc = __lam_topproc_(); proc; proc = __lam_nextproc_()) {
	if (proc_sock(proc) == sock) {
	    __lam_commfault_(proc->p_gps.gps_node);

	    cpcs = (struct c2c_proc *) proc->p_rpi;
	    
	    if (cpcs->ccp_rdreq) {
		(void) c2c_commdead(cpcs->ccp_rdreq);
	    }
	    if (cpcs->ccp_wtreq) {
		(void) c2c_commdead(cpcs->ccp_wtreq);
	    }
	}
    }
}
