/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	NJN
 *
 *	$Log:	rpi.c2c.h,v $
 * Revision 6.0  96/02/29  14:13:30  gdburns
 * Ohio Release
 * 
 *	Function:	- client-to-client header file
 */

#ifndef _RPIC2C
#define _RPIC2C

#include <all_hash.h>
#include <all_list.h>
#include <mpisys.h>
#include <portable.h>
#include <t_types.h>

/*
 * constants
 */
#define INITHASH	((int4) 16)		/* initial hash table size */
#define NULLCID		(INT4_MAX)		/* invalid context ID */
#define C2C_SHORTMSGLEN	8192			/* max. length of short msg */

/*
 * structures
 */
struct c2c_envl {
	int4		cce_len;		/* message length */
	int4		cce_tag;		/* message tag (16 bits) */
	int4		cce_flags;		/* flags */
#define C2C_LONG	0x01			/* long protocol */
#define C2C_ACK		0x02			/* ACK message */
#define C2C_2ND		0x04			/* 2nd message chunk */
#define C2C_SSEND	0x08			/* synchronous send */
	int4		cce_rank;		/* peer rank */
	int4		cce_cid;		/* message context ID */
	int4		cce_seq;		/* sequence number */
};

struct c2c_proc {				/* C2C specific process data */
	int		ccp_sock;		/* socket descriptor */
	int		ccp_nbfde;		/* # buffered envelopes */
	struct c2c_envl	ccp_env;		/* incoming envelope */
	char		*ccp_envbuf;		/* envelope buffer pointer */
	int		ccp_nenvin;		/* # env. bytes left to read */
	char		*ccp_msgbuf;		/* message buffer pointer */
	int		ccp_nmsgin;		/* # msg. bytes left to read */
	int		ccp_extra;		/* # bytes to go into sink */
	MPI_Request	ccp_rdreq;		/* req reading from process */
	MPI_Request	ccp_wtreq;		/* req writing to process */
};

struct c2c_req {				/* C2C specific request data */
	int		ccq_state;		/* state of request */
#define C2C_DONE 	1			/* all data transfer done */
#define C2C_WRITE 	2			/* writing to socket */
#define C2C_READ	3			/* reading from socket */
#define C2C_SENDSELF    4			/* sending to self */
	int4		ccq_peer;		/* matched peer */
/*
 * The envelope holds synchronization variables and protocol flags
 * which are set according to the stage of the protocol.
 */
	struct c2c_envl ccq_env;	/* envelope */	
	struct c2c_envl ccq_outenv;	/* outgoing env (Trollius byteorder) */

	int		ccq_nenvout;	/* # bytes of env. left to write */
	char		*ccq_envbuf;	/* ptr into envelope buffer */

	int		ccq_nmsgout;	/* # bytes of msg. left to write */
	char		*ccq_msgbuf;	/* ptr into message buffer */
};

struct cbuf_cid {				/* CID hash table entry */
	int4		cbc_cid;		/* context ID */
	MPI_Comm	cbc_comm;		/* communicator */
	LIST		*cbc_envs;		/* envelope list */
};

struct cbuf_msg {				/* msg./env. list entry */
	struct c2c_proc *cbm_proc;		/* source process */
	struct c2c_envl	cbm_env;		/* envelope */
	char		*cbm_buf;		/* short message data */
};

/*
 * data conversion for C2C envelopes for supporetd machine/CPU types
 */
#if tputer || i386 || i860 || SCO || DOS || CRAY || DECAXP || __alpha

#define set_out_envelope(q) \
	if (__lam_homog_) { \
		(q)->ccq_nenvout = sizeof(struct c2c_envl); \
		(q)->ccq_envbuf = (char *) &((q)->ccq_env); \
	} \
	else { \
		(q)->ccq_nenvout = sizeof(struct c2c_envl); \
		(q)->ccq_outenv = (q)->ccq_env;	\
		mltoti4((int4 *) &((q)->ccq_outenv), \
			sizeof(struct c2c_envl) / sizeof(int4)); \
		(q)->ccq_envbuf = (char *) &((q)->ccq_outenv); \
	}

#define envelope_ttol(e) \
	if (! __lam_homog_) { \
		mttoli4((int4 *) (e), sizeof(struct c2c_envl)/sizeof(int4)); \
	}

#else
#if SUN4_OS || SUN4_SOL || sparc || RS6K || _IBMR2 || SGI || mips || HPUX || hpux

#define set_out_envelope(q) \
	do { \
		(q)->ccq_nenvout = sizeof(struct c2c_envl); \
		(q)->ccq_envbuf = (char *) &((q)->ccq_env); \
	} while(0)

#define envelope_ttol(e)

#else

"Machine architecture not defined."

#endif
#endif

/*
 * prototypes of LAM MPI library C2C internal functions
 */
#ifndef __ARGS
#if __STDC__ || defined(c_plusplus) || defined(__cplusplus)
#define __ARGS(a)	a
#else
#define __ARGS(a)	()
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int	_c2c_envl_cmp __ARGS((struct c2c_envl *, struct c2c_envl *));
extern int	_cbuf_append __ARGS((struct cbuf_msg *));
extern void	_cbuf_delete __ARGS((struct cbuf_msg *));
extern void	_cbuf_end __ARGS((void));
extern struct cbuf_msg * _cbuf_find __ARGS((struct c2c_envl *));
extern int	_cbuf_init __ARGS((void));

#ifdef __cplusplus
}
#endif

#endif	/* _RPIC2C */
