/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	JRV/RBD
 *
 *	$Log:	scan.c,v $
 * Revision 6.0  96/02/29  14:13:33  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:25:48  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:18  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:05:00  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:41  gdburns
 * Ohio Release
 * 
 *	Function:	- partial reduce from 0 to rank i, at each rank i
 *	Accepts:	- send buffer
 *			- receive buffer
 *			- count of elements
 *			- type of elements
 *			- operation to perform
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Scan(sbuf, rbuf, count, dtype, op, comm)

void			*sbuf;
void			*rbuf;
int			count;
MPI_Datatype		dtype;
MPI_Op			op;
MPI_Comm		comm;

{
	int		size;			/* group size */
	int		rank;			/* my rank */
	int		err;			/* error code */
	MPI_Status	stat;			/* recv status */

	__lam_initerr_();
	__lam_setfunc_(BLKMPISCAN);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return (__lam_errfunc_(comm, BLKMPISCAN,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(comm, BLKMPISCAN,
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (count < 0) {
		return(__lam_errfunc_(comm, BLKMPISCAN,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if (op == MPI_OP_NULL) {
		return(__lam_errfunc_(comm, BLKMPISCAN,
				__lam_mkerr_(MPI_ERR_OP, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPISCAN);
/*
 * Initialize.
 */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &size);
/*
 * Switch to collective communicator.
 */
	__lam_mkcoll_(comm);
/*
 * If I'm rank 0, initialize the recv buffer.
 */
	if (rank == 0) {
		err = __lam_dtsndrcv_(sbuf, count, dtype,
				rbuf, count, dtype, BLKMPISCAN, comm);
		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPISCAN, err));
		}
	}
/*
 * Otherwise receive previous buffer and reduce.
 */
	else {
		err = MPI_Recv(rbuf, count, dtype, rank - 1,
					BLKMPISCAN, comm, &stat);
		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPISCAN, err));
		}

		if (op->op_f77dtype >= 0) {
			(op->op_func)(sbuf, rbuf, &count, &(op->op_f77dtype));
		} else {
			(op->op_func)(sbuf, rbuf, &count, &dtype);
		}
	}
/*
 * Send result to next process.
 */
	if (rank < (size - 1)) {
		err = MPI_Send(rbuf, count, dtype, rank + 1, BLKMPISCAN, comm);
		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPISCAN, err));
		}

	}

	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPISCAN, -1, comm, dtype, count);
	
	__lam_resetfunc_(BLKMPISCAN);
	return(MPI_SUCCESS);
}
