/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/JRV
 *
 *	$Log:	scatterv.c,v $
 * Revision 6.0  96/02/29  14:13:34  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/28  23:37:27  nevin
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:20  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:05:01  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:44  gdburns
 * Ohio Release
 * 
 *	Function:	- scatter varying len. buffers from root in rank order
 *	Accepts:	- send buffer
 *			- send counts
 *			- displacements
 *			- send datatype
 *			- recv buffer
 *			- recv count
 *			- recv datatype
 *			- root
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Scatterv(sbuf, scounts, disps, sdtype, rbuf, rcount, rdtype, root, comm)

void			*sbuf;
int			*scounts;
int			*disps;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
int			root;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		rank;			/* my rank */
	int		size;			/* group size */
	int		err;			/* error code */
	char		*ptmp;			/* temp. buffer */
	MPI_Aint	extent;			/* datatype extent */
	MPI_Status	status;			/* recv status */
	struct _gps	*p;			/* favourite pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPISCATTERV);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPISCATTERV,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((sdtype == MPI_DATATYPE_NULL) || (rdtype == MPI_DATATYPE_NULL)) {
		return(__lam_errfunc_(comm, BLKMPISCATTERV,
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &size);

	if ((root >= size) || (root < 0)) {
		return(__lam_errfunc_(comm, BLKMPISCATTERV,
				__lam_mkerr_(MPI_ERR_ROOT, 0)));
	}

	if ((rank == root) && (disps == 0)) {
		return(__lam_errfunc_(comm, BLKMPISCATTERV,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((rcount < 0) || ((scounts == 0) && (rank == root))) {
		return(__lam_errfunc_(comm, BLKMPISCATTERV,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPISCATTERV);
/*
 * Remember required parameters.
 */
	p = &(comm->c_group->g_procs[root]->p_gps);

	__lam_setparam_(BLKMPISCATTERV, root | (p->gps_grank << 16),
				(p->gps_node << 16) | p->gps_idx);
/*
 * Switch to collective communicator.
 */
	__lam_mkcoll_(comm);
/*
 * If not root, receive data.
 */
	if (rank != root) {
		err = MPI_Recv(rbuf, rcount, rdtype,
					root, BLKMPISCATTERV, comm, &status);
		__lam_mkpt_(comm);

		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPISCATTERV, err));
		}
/*
 * Generate run time trace.
 */
		if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) {
			__lam_trcffend_(BLKMPISCATTERV,
						root, comm, rdtype, rcount);
		}
		
		__lam_resetfunc_(BLKMPISCATTERV);
		return(MPI_SUCCESS);
	}
/*
 * I am the root, loop sending data.
 */
	MPI_Type_extent(sdtype, &extent);

	for (i = 0; i < size; ++i) {

		ptmp = ((char *) sbuf) + (extent * disps[i]);
/*
 * simple optimization
 */
		if (i == rank) {
			err = __lam_dtsndrcv_(ptmp, scounts[i], sdtype, rbuf,
					rcount, rdtype, BLKMPISCATTERV,	comm);
		} else {
			err = MPI_Send(ptmp, scounts[i], sdtype,
					i, BLKMPISCATTERV, comm);
		}

		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPISCATTERV, err));
		}
	}

	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPISCATTERV, root, comm, rdtype, rcount);
	
	__lam_resetfunc_(BLKMPISCATTERV);
	return(MPI_SUCCESS);
}
