/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	sendrecv.c,v $
 * Revision 6.0  96/02/29  14:13:37  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:25:55  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:03  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:46  gdburns
 * Ohio Release
 * 
 *	Function:	- send and receive
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- destination
 *			- send tag
 *			- receive buffer
 *			- receive count
 *			- receive datatype
 *			- source
 *			- receive tag
 *			- communicator
 *			- status (out)
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Sendrecv(sbuf, scount, sdtype, dest, stag,
		rbuf, rcount, rdtype, src, rtag, comm, status)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
int			dest;
int			stag;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
int			src;
int			rtag;
MPI_Comm		comm;
MPI_Status		*status;

{
	MPI_Request	req;			/* request */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */

	__lam_initerr_();
	__lam_setfunc_(BLKMPISENDRECV);
/*
 * Check the arguments.
 */
	err = __lam_chkarg_(sbuf, scount, sdtype, dest, stag, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECV, err));
	}

	err = __lam_chkarg_(rbuf, rcount, rdtype, src, rtag, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECV, err));
	}

	if (status == 0) {
		return(__lam_errfunc_(comm,
			BLKMPISENDRECV, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON
			&& __lam_trwrapstart_(BLKMPISENDRECV);
/*
 * Send & receive.
 */
	err = MPI_Irecv(rbuf, rcount, rdtype, src, rtag, comm, &req);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECV, err));
	}

	err = MPI_Send(sbuf, scount, sdtype, dest, stag, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECV, err));
	}

	err = MPI_Wait(&req, status);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECV, err));
	}
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) __lam_trwrapend_(BLKMPISENDRECV);

	__lam_resetfunc_(BLKMPISENDRECV);
	return(MPI_SUCCESS);
}
