/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	sendrecvrep.c,v $
 * Revision 6.0  96/02/29  14:13:37  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:25:56  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:04  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:47  gdburns
 * Ohio Release
 * 
 *	Function:	- send and receive using same buffer
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- destination
 *			- send tag
 *			- source
 *			- receive tag
 *			- communicator
 *			- status (out)
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Sendrecv_replace(buf, count, dtype, dest, stag, src, rtag, comm, status)

void			*buf;
int			count;
MPI_Datatype		dtype;
int			dest;
int			stag;
int			src;
int			rtag;
MPI_Comm		comm;
MPI_Status		*status;

{
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	int		pkpos;			/* packing position */
	int		pksize;			/* packing size */
	char		*pkbuf;			/* packing buffer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPISENDRECVREP);
/*
 * Check the arguments.
 */
	err = __lam_chkarg_(buf, count, dtype, dest, stag, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECVREP, err));
	}

	err = __lam_chkarg_(buf, count, dtype, src, rtag, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECVREP, err));
	}

	if (status == 0) {
		return(__lam_errfunc_(comm,
			BLKMPISENDRECVREP, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON
			&& __lam_trwrapstart_(BLKMPISENDRECVREP);
/*
 * Pack the message to send.
 */
	err = MPI_Pack_size(count, dtype, comm, &pksize);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPISENDRECVREP, err));
	}

	if (pksize > 0) {
		pkbuf = malloc((unsigned) pksize);
		if (pkbuf == 0) {
			return(__lam_errfunc_(comm, BLKMPISENDRECVREP,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
	}
	else {
		pkbuf = 0;
	}

	pkpos = 0;

	err = MPI_Pack(buf, count, dtype, pkbuf, pksize, &pkpos, comm);
	if (err != MPI_SUCCESS) {
		if (pkbuf) free(pkbuf);
		return(__lam_errfunc_(comm, BLKMPISENDRECVREP, err));
	}
/*
 * Send & receive.
 */
	err = MPI_Sendrecv(pkbuf, pksize, MPI_PACKED, dest, stag,
				buf, count, dtype, src, rtag, comm, status);
	if (err != MPI_SUCCESS) {
		if (pkbuf) free(pkbuf);
		return(__lam_errfunc_(comm, BLKMPISENDRECVREP, err));
	}

	if (pkbuf) free(pkbuf);
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) __lam_trwrapend_(BLKMPISENDRECVREP);
	
	__lam_resetfunc_(BLKMPISENDRECVREP);
	return(MPI_SUCCESS);
}
