/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	startall.c,v $
 * Revision 6.0.1.1  96/05/10  15:12:50  nevin
 * Handle MPI_PROC_NULL
 * 
 * Revision 6.0  96/02/29  14:13:41  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/28  23:37:29  nevin
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:06  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:50  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:47  gdburns
 * new for 5.1
 * 
 *	Function:	- start several communication requests
 *	Accepts:	- # requests
 *			- requests array
 *	Returns: 	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <typical.h>

/*
 * external functions
 */
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		_mpi_req_add();
extern void		_mpi_req_blkclr();
extern void		__lam_trstartall_();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();
extern int		_mpi_req_advance();
extern int		_mpi_req_build();
extern int		_mpi_req_start();

/*
 * local functions
 */
static int		hidden_bsend();


int
MPI_Startall(nreq, reqs)

int			nreq;
MPI_Request		*reqs;

{
	int		err;			/* error code */
	int		i;			/* favourite index */
	MPI_Request	*p;			/* favourite pointer */
	MPI_Request	req;			/* request */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		advt;			/* advance time */
	double		finisht;		/* finish time */

	__lam_initerr_();
	__lam_setfunc_(BLKMPISTARTALL);
/*
 * Check the arguments.
 */
	if (nreq < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPISTARTALL,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((nreq > 0) && (reqs == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPISTARTALL,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) startt = MPI_Wtime();
/*
 * Loop linking the requests.
 */
	for (i = 0, p = reqs; i < nreq; ++i, ++p) {

		if ((req = *p) == MPI_REQUEST_NULL) {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPISTARTALL,
					__lam_mkerr_(MPI_ERR_REQUEST, 0)));
		}

		if (req->rq_state != LAM_RQSINIT) {
			return(__lam_errfunc_(req->rq_comm, BLKMPISTARTALL,
					__lam_mkerr_(MPI_ERR_REQUEST, 0)));
		}
/*
 * If a Bsend_init() request is found, declare it done,
 * and create and link its hidden Bsend() counterpart.
 */
		if (req->rq_type == LAM_RQIFAKE) {
			req->rq_state = LAM_RQSDONE;
			err = hidden_bsend(req);
		} else {
			err = _mpi_req_start(req);

			if (err == MPI_SUCCESS) {
				_mpi_req_add(req);
			}
		}

		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(req->rq_comm,
					BLKMPISTARTALL, err));
		}
	}
/*
 * Advance the system.
 */
	if (fl_trace) advt = MPI_Wtime();

	_mpi_req_blkclr();

	err = _mpi_req_advance();
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPISTARTALL, err));
	}
/*
 * Generate a run time trace.
 */
	if (fl_trace) {

		finisht = MPI_Wtime();

		__lam_trstartall_(nreq, reqs, startt, advt, finisht);
	}

	__lam_resetfunc_(BLKMPISTARTALL);
	return(MPI_SUCCESS);
}

/*
 *	hidden_bsend
 *
 *	Function:	- build, start and link in a hidden bsend request
 *	Accepts:	- user bsend request
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
hidden_bsend(req)

MPI_Request		req;

{
	MPI_Request	hreq = 0;		/* hidden request */
	int		cidsave;		/* save context ID */
	int		err;			/* error code */
/*
 * Handle special case of peer MPI_PROC_NULL.
 */
	if (req->rq_rank == MPI_PROC_NULL) {
		return(MPI_SUCCESS);
	}
/*
 * Make sure to use the original request's context ID.  This is to take
 * care of the case where an MPI_Bsend_init is done outside all
 * collective calls and then started inside a collective call or
 * vice-versa.
 */
	cidsave = req->rq_comm->c_contextid;
	req->rq_comm->c_contextid = req->rq_cid;
	
	err = _mpi_req_build(req->rq_buf, req->rq_count, req->rq_dtype,
				req->rq_rank, req->rq_tag, req->rq_comm,
				LAM_RQIBSEND, &hreq);

	req->rq_comm->c_contextid = cidsave;
	if (err != MPI_SUCCESS) return(err);

	err = _mpi_req_start(hreq);
	if (err != MPI_SUCCESS) return(err);

	req->rq_seq = hreq->rq_seq;
	_mpi_req_add(hreq);
	return(MPI_SUCCESS);
}
