/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	tcontig.c,v $
 * Revision 6.0  96/02/29  14:13:44  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:04  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:07  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:52  gdburns
 * Ohio Release
 * 
 *	Function:	- create contiguous derived datatype
 *	Accepts:	- # elements
 *			- element datatype
 *			- ptr new datatype
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Type_contiguous(count, oldtype, newtype)

int			count;
MPI_Datatype		oldtype;
MPI_Datatype		*newtype;

{
	MPI_Datatype	ntype;			/* new datatype */
	int		extent;			/* datatype extent */

	__lam_initerr_();
	__lam_setfunc_(BLKMPITCONTIG);
/*
 * Check the arguments.
 */
	if (oldtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITCONTIG, __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITCONTIG, __lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if (newtype == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITCONTIG, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Create the new datatype.
 */
	ntype = (MPI_Datatype) malloc(sizeof(struct _dtype));
	if (ntype == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITCONTIG, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	extent = oldtype->dt_upper - oldtype->dt_lower;
	++(oldtype->dt_refcount);

	*newtype = ntype;
	ntype->dt_format = LAM_DTCONTIG;
	ntype->dt_commit = 0;
	ntype->dt_refcount = 1;
	ntype->dt_label = -1;

	if ((oldtype == MPI_UB) || (oldtype == MPI_LB)) {
		ntype->dt_upper = 0;
		ntype->dt_lower = 0;
		ntype->dt_size = 0;
		ntype->dt_usrsize = 0;
		ntype->dt_nelem = 0;
	} else {
		ntype->dt_upper = count * extent;
		ntype->dt_lower = 0;
		ntype->dt_size = count * oldtype->dt_size;
		ntype->dt_usrsize = count * oldtype->dt_usrsize;
		ntype->dt_nelem = count;
	}

	ntype->dt_count = count;
	ntype->dt_length = 0;
	ntype->dt_stride = 0;
	ntype->dt_dtype = oldtype;
	ntype->dt_lengths = 0;
	ntype->dt_disps = 0;
	ntype->dt_dtypes = 0;

	__lam_resetfunc_(BLKMPITCONTIG);
	return(MPI_SUCCESS);
}
