/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	testall.c,v $
 * Revision 6.0  96/02/29  14:13:46  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:06  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:09  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:54  gdburns
 * Ohio Release
 * 
 *	Function:	- check if all requests completed
 *	Accepts:	- # of requests
 *			- array of requests
 *			- flag (returned)
 *			- array of status structures (modified)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trwrapend_();
extern int		__lam_trwrapstart_();
extern int		__lam_test_();
extern int		__lam_mkerr_();


int
MPI_Testall(count, reqs, flag, stats)

int			count;
MPI_Request		*reqs;
int			*flag;
MPI_Status		*stats;

{
	int		i;			/* favourite index */
	int		fl_done;		/* request done? */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	MPI_Request	*req;			/* ptr request */
	MPI_Status	*stat;			/* ptr status */

	__lam_initerr_();
	__lam_setfunc_(BLKMPITESTALL);

	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTALL,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && ((reqs == 0) || (stats == 0))) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTALL,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if (flag == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTALL,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON;

	if (fl_trace) {
		__lam_trwrapstart_(BLKMPITESTALL);
	}
/*
 * Initialize all status members.
 */
	for (i = 0, stat = stats; i < count; ++i, ++stat) {
		__lam_emptystat_(stat);
	}
/*
 * Loop testing all requests non-destructively.
 * Stop at the first request not done.
 */
	req = reqs;
	stat = stats;

	for (i = 0; i < count; ++i, ++req, ++stat) {

		err = __lam_test_(req, 0, &fl_done, stat);

		if (err != MPI_SUCCESS) {
			stat->MPI_ERROR = err;

			if (fl_trace) {
				__lam_trwrapend_(BLKMPITESTALL);
			}

			__lam_resetfunc_(BLKMPITESTALL);
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTALL,
					__lam_mkerr_(MPI_ERR_IN_STATUS, 0)));
		}

		if (fl_done == 0) {
			*flag = 0;

			if (fl_trace) {
				__lam_trwrapend_(BLKMPITESTALL);
			}

			__lam_resetfunc_(BLKMPITESTALL);
			return(MPI_SUCCESS);
		}
	}
/*
 * All active requests are done, loop doing the job.
 */
	req = reqs;
	stat = stats;

	for (i = 0; i < count; ++i, ++req, ++stat) {
		err = __lam_test_(req, 1, &fl_done, stat);

		if (err != MPI_SUCCESS) {
			stat->MPI_ERROR = err;

			if (fl_trace) {
				__lam_trwrapend_(BLKMPITESTALL);
			}

			__lam_resetfunc_(BLKMPITESTALL);
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTALL,
					__lam_mkerr_(MPI_ERR_IN_STATUS, 0)));
		}
	}

	*flag = 1;
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) {
		__lam_trwrapend_(BLKMPITESTALL);
	}

	__lam_resetfunc_(BLKMPITESTALL);
	return(MPI_SUCCESS);
}
