/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	testany.c,v $
 * Revision 6.0  96/02/29  14:13:48  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:08  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:10  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:55  gdburns
 * Ohio Release
 * 
 *	Function:	- check if one request completed
 *	Accepts:	- # of requests
 *			- array of requests
 *			- index (returned)
 *			- flag (returned)
 *			- status (returned)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trwrapend_();
extern int		__lam_trwrapstart_();
extern int		__lam_mkerr_();


int
MPI_Testany(count, reqs, index, flag, stat)

int			count;
MPI_Request		*reqs;
int			*index;
int			*flag;
MPI_Status		*stat;

{
	int		i;			/* favourite index */
	int		fl_done;		/* request done flag */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	MPI_Comm	comm;			/* communicator */
	
	__lam_initerr_();
	__lam_setfunc_(BLKMPITESTANY);

	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTANY,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && (reqs == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTANY,
				__lam_mkerr_(MPI_ERR_REQUEST, 0)));
	}

	if ((flag == 0) || (index == 0) || (stat == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTANY,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON;

	if (fl_trace) {
		__lam_trwrapstart_(BLKMPITESTANY);
	}
/*
 * Loop testing all requests.
 * Skip inactive requests.
 */
	__lam_emptystat_(stat);

	*flag = 1;
	*index = MPI_UNDEFINED;
	err = MPI_SUCCESS;

	for (i = 0; i < count; ++i, ++reqs) {

		if ((*reqs == MPI_REQUEST_NULL) ||
			((*reqs)->rq_state == LAM_RQSINIT)) continue;

		*flag = 0;
		comm = (*reqs)->rq_comm;

		err = MPI_Test(reqs, &fl_done, stat);
		if (err != MPI_SUCCESS) break;

		if (fl_done) {
			*flag = 1;
			*index = i;
			break;
		}
	}
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) {
		__lam_trwrapend_(BLKMPITESTANY);
	}

	__lam_resetfunc_(BLKMPITESTANY);

	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPITESTANY, err));
	}

	return(MPI_SUCCESS);
}
