/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	testsome.c,v $
 * Revision 6.0  96/02/29  14:13:49  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:11  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:11  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:55  gdburns
 * Ohio Release
 * 
 *	Function:	- check if at least one request completed
 *			- return info on all completed requests
 *	Accepts:	- # of requests
 *			- array of requests
 *			- output count (returned)
 *			- array of indices
 *			- array of status structures
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trwrapend_();
extern int		__lam_trwrapstart_();
extern int		__lam_mkerr_();


int
MPI_Testsome(count, reqs, outcount, indices, stats)

int			count;
MPI_Request		*reqs;
int			*outcount;
int			*indices;
MPI_Status		*stats;

{
	int		i;			/* favourite index */
	int		fl_done;		/* request done flag */
	int		fl_err;			/* error flag */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	MPI_Status	*p;			/* ptr status */

	__lam_initerr_();
	__lam_setfunc_(BLKMPITESTSOME);
/*
 * Check the arguments.
 */
	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTSOME,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && ((reqs == 0) || (indices == 0) || (stats == 0))) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTSOME,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if (outcount == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTSOME,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON;

	if (fl_trace) {
		__lam_trwrapstart_(BLKMPITESTSOME);
	}
/*
 * Initialize all status members.
 */
	for (i = 0, p = stats; i < count; ++i, ++p) {
		__lam_emptystat_(p);
	}
/*
 * Loop testing all requests.
 * Skip inactive requests.
 */
	*outcount = MPI_UNDEFINED;
	fl_err = 0;

	for (i = 0; i < count; ++i, ++reqs) {

		if ((*reqs == MPI_REQUEST_NULL) ||
			((*reqs)->rq_state == LAM_RQSINIT)) continue;

		err = MPI_Test(reqs, &fl_done, stats);

		if ((err != MPI_SUCCESS) || fl_done) {

			if (*outcount == MPI_UNDEFINED) {
				*outcount = 0;
			}

			++(*outcount);
			*indices++ = i;

			if (err != MPI_SUCCESS) {
				stats->MPI_ERROR = err;
				fl_err = 1;
			}

			++stats;
		}
	}
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) {
		__lam_trwrapend_(BLKMPITESTSOME);
	}

	__lam_resetfunc_(BLKMPITESTSOME);

	if (fl_err) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITESTSOME,
				__lam_mkerr_(MPI_ERR_IN_STATUS, 0)));
	}

	return(MPI_SUCCESS);
}
