/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	tfree.c,v $
 * Revision 6.0  96/02/29  14:13:51  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:15  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:13  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:57  gdburns
 * Ohio Release
 * 
 *	Function:	- mark a datatype to deallocation
 *	Accepts:	- ptr to datatype
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_trdtypefree_();

int
MPI_Type_free(dtype)

MPI_Datatype		*dtype;

{
	MPI_Datatype	dt;			/* datatype */
	int		dtfmt;			/* datatype format */
	int		err;			/* error code */
	int		i;			/* favourite index */

	__lam_initerr_();
	__lam_setfunc_(BLKMPITFREE);
/*
 * Check the arguments.
 */
	if (dtype == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITFREE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	dt = *dtype;
	if (dt == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITFREE, __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}
/*
 * If not last reference, NULL the handle and return.
 */
	if (--(dt->dt_refcount) > 0) {
		*dtype = MPI_DATATYPE_NULL;
		__lam_resetfunc_(BLKMPITFREE);
		return(MPI_SUCCESS);
	}
/*
 * Really free the datatype.
 * Recursively free all datatype children.
 */
	dtfmt = dt->dt_format;

	switch(dtfmt) {

	case LAM_DTCONTIG:
	case LAM_DTVECTOR:
	case LAM_DTHVECTOR:
	case LAM_DTINDEXED:
	case LAM_DTHINDEXED:

	err = MPI_Type_free(&(dt->dt_dtype));
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITFREE, err));
	}

	if ((dtfmt == LAM_DTHINDEXED) || (dtfmt == LAM_DTINDEXED)) {
		if (dt->dt_disps) free((char *) dt->dt_disps);
	}

	break;

	case LAM_DTSTRUCT:

	for (i = 0; i < dt->dt_count; ++i) {

		err = MPI_Type_free(&(dt->dt_dtypes[i]));
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(MPI_COMM_WORLD,
						BLKMPITFREE, err));
		}
	}

	if (dt->dt_dtypes) free((char *) dt->dt_dtypes);
	break;

	case LAM_DTBASIC:
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITFREE, __lam_mkerr_(MPI_ERR_TYPE, 0)));

	default:
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITFREE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Free the top datatype.
 */
	*dtype = MPI_DATATYPE_NULL;
	__lam_trdtypefree_(dt);
	if (dt->dt_label > TRDTMAX) {
		free((char *) dt);
	}

	__lam_resetfunc_(BLKMPITFREE);
	return(MPI_SUCCESS);
}
