/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	tindex.c,v $
 * Revision 6.0  96/02/29  14:13:53  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:19  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:15  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:59  gdburns
 * Ohio Release
 * 
 *	Function:	- create indexed derived datatype
 *	Accepts:	- # blocks
 *			- block length array
 *			- block displacement array
 *			- element datatype
 *			- ptr new datatype
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Type_indexed(count, lengths, disps, oldtype, newtype)

int			count;
int			*lengths;
int			*disps;
MPI_Datatype		oldtype;
MPI_Datatype		*newtype;

{
	MPI_Datatype	ntype;			/* new datatype */
	int		extent;			/* datatype extent */
	int		usrsize;		/* user-vis. size */
	int		lower;			/* lower bound */
	int		upper;			/* upper bound */
	int		usrupper;		/* user-visible upper */
	int		top;			/* temp. top bound */
	int		i;			/* favourite index */

	__lam_initerr_();
	__lam_setfunc_(BLKMPITINDEX);
/*
 * Check the arguments.
 */
	if (oldtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITINDEX, __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITINDEX, __lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((newtype == 0) ||
			((count > 0) && ((lengths == 0) || (disps == 0)))) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITINDEX, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Create the new datatype.
 */
	ntype = (MPI_Datatype) malloc(sizeof(struct _dtype));
	if (ntype == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITINDEX, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	if (count > 0) {
		ntype->dt_disps = (MPI_Aint *) malloc((unsigned) (count *
					(sizeof(MPI_Aint) + sizeof(int))));
		if (ntype->dt_disps == 0) {
			free((char *) ntype);
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITINDEX,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}

		ntype->dt_lengths = (int *) (((char *) ntype->dt_disps) +
						(count * sizeof(MPI_Aint)));
	} else {
		ntype->dt_disps = 0;
		ntype->dt_lengths = 0;
	}

	extent = oldtype->dt_upper - oldtype->dt_lower;
	usrsize = oldtype->dt_usrsize;

	++(oldtype->dt_refcount);

	*newtype = ntype;
	ntype->dt_format = LAM_DTINDEXED;
	ntype->dt_commit = 0;
	ntype->dt_refcount = 1;
	ntype->dt_label = -1;
	ntype->dt_upper = 0;
	ntype->dt_lower = 0;
	ntype->dt_size = 0;
	ntype->dt_nelem = 0;
	ntype->dt_count = count;
	ntype->dt_length = 0;
	ntype->dt_stride = 0;
	ntype->dt_dtype = oldtype;
	ntype->dt_dtypes = 0;

	for (i = 0; i < count; ++i) {

		ntype->dt_disps[i] = disps[i];

		if ((ntype->dt_lengths[i] = lengths[i]) < 0) {
			free((char *) ntype->dt_disps);
			free((char *) ntype);
			return(__lam_errfunc_(MPI_COMM_WORLD,
				BLKMPITINDEX, __lam_mkerr_(MPI_ERR_ARG, 0)));
		}
	}

	if ((oldtype == MPI_UB) || (oldtype == MPI_LB) || (count == 0)) {
		__lam_resetfunc_(BLKMPITINDEX);
		return(MPI_SUCCESS);
	}
/*
 * Compute the extent, # elements, and user-visible size.
 */
	lower = upper = usrupper = -1;

	for (i = 0; i < count; ++i) {

		ntype->dt_nelem += lengths[i];

		if ((lower < 0) || (lower > disps[i])) lower = disps[i];

		top = disps[i] + lengths[i];
		if (upper < top) upper = top;

		top = (disps[i] * extent) + (lengths[i] * usrsize);
		if (usrupper < top) usrupper = top;
	}

	ntype->dt_size = ntype->dt_nelem * oldtype->dt_size;
	ntype->dt_usrsize = usrupper - (lower * extent);
	ntype->dt_upper = upper * extent;
	ntype->dt_lower = lower * extent;

	__lam_resetfunc_(BLKMPITINDEX);
	return(MPI_SUCCESS);
}
