/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	tstruct.c,v $
 * Revision 6.0  96/02/29  14:13:56  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:26  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:19  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:52:02  gdburns
 * Ohio Release
 * 
 *	Function:	- create struct derived datatype
 *	Accepts:	- # blocks
 *			- block length array
 *			- block displacement array (in bytes)
 *			- block element datatype array
 *			- ptr new datatype
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Type_struct(count, lengths, disps, oldtypes, newtype)

int			count;
int			*lengths;
MPI_Aint		*disps;
MPI_Datatype		*oldtypes;
MPI_Datatype		*newtype;

{
	MPI_Datatype	ntype;			/* new datatype */
	MPI_Datatype	oldtype;		/* old datatype */
	int		upper, lower;		/* upper/lower bounds */
	int		ub_upper, lb_lower;	/* UB/LB bounds */
	int		usrupper, usrlower;	/* user-vis. bounds */
	int		fl_ub, fl_lb;		/* UB/LB flags */
	int		top;			/* temp. top bound */
	int		i;			/* favourite index */

	__lam_initerr_();
	__lam_setfunc_(BLKMPITSTRUCT);
/*
 * Check the arguments.
 */
	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITSTRUCT, __lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((newtype == 0) || ((count > 0) &&
		((lengths == 0) || (disps == 0) || (oldtypes == 0)))) {

		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITSTRUCT, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	for (i = 0; i < count; ++i) {

		if (oldtypes[i] == MPI_DATATYPE_NULL) {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITSTRUCT,
					__lam_mkerr_(MPI_ERR_TYPE, 0)));
		}

		if (lengths[i] < 0) {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITSTRUCT,
					__lam_mkerr_(MPI_ERR_ARG, 0)));
		}
	}
/*
 * Create the new datatype.
 */
	ntype = (MPI_Datatype) malloc(sizeof(struct _dtype));
	if (ntype == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPITSTRUCT, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	if (count > 0) {
		ntype->dt_dtypes = (MPI_Datatype *) malloc((unsigned)
					(count * (sizeof(MPI_Datatype) +
					sizeof(MPI_Aint) + sizeof(int))));
		if (ntype->dt_dtypes == 0) {
			free((char *) ntype);
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPITSTRUCT,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}

		ntype->dt_disps = (MPI_Aint *) (((char *) ntype->dt_dtypes) +
					(count * sizeof(MPI_Datatype)));
		ntype->dt_lengths = (int *) (((char *) ntype->dt_disps) +
					(count * sizeof(MPI_Aint)));
	} else {
		ntype->dt_dtypes = 0;
		ntype->dt_disps = 0;
		ntype->dt_lengths = 0;
	}

	*newtype = ntype;
	ntype->dt_format = LAM_DTSTRUCT;
	ntype->dt_commit = 0;
	ntype->dt_refcount = 1;
	ntype->dt_label = -1;
	ntype->dt_upper = 0;
	ntype->dt_lower = 0;
	ntype->dt_size = 0;
	ntype->dt_usrsize = 0;
	ntype->dt_nelem = 0;
	ntype->dt_count = count;
	ntype->dt_length = 0;
	ntype->dt_stride = 0;
	ntype->dt_dtype = 0;
/*
 * Compute the extent, # elements, and user-visible size.
 */
	lower = upper = -1;
	usrlower = usrupper = -1;
	ub_upper = lb_lower = -1;
	fl_lb = fl_ub = 0;

	for (i = 0; i < count; ++i) {

		ntype->dt_disps[i] = disps[i];
		ntype->dt_lengths[i] = lengths[i];
		ntype->dt_dtypes[i] = oldtype = oldtypes[i];
		++(oldtype->dt_refcount);

		if ((usrlower < 0) || (usrlower > disps[i])) {
			usrlower = disps[i];
		}

		if (oldtype == MPI_LB) {
			fl_lb = 1;
			if ((lb_lower < 0) || (lb_lower > disps[i])) {
				lb_lower = disps[i];
			}

			if (usrupper < disps[i]) usrupper = disps[i];
		}
		else if (oldtype == MPI_UB) {
			fl_ub = 1;
			if (ub_upper < disps[i]) ub_upper = disps[i];
			if (usrupper < disps[i]) usrupper = disps[i];
		}
		else {
			ntype->dt_size += lengths[i] * oldtype->dt_size;
			ntype->dt_nelem += lengths[i];

			if ((lower < 0) || (lower > disps[i])) {
				lower = disps[i];
			}

			top = disps[i] + (lengths[i] *
				(oldtype->dt_upper - oldtype->dt_lower));
			if (upper < top) upper = top;

			top = disps[i] + (lengths[i] * oldtype->dt_usrsize);
			if (usrupper < top) usrupper = top;
		}
	}

	ntype->dt_upper = (fl_ub) ? ub_upper : upper;
	ntype->dt_lower = (fl_lb) ? lb_lower : lower;
	ntype->dt_usrsize = usrupper - usrlower;

	__lam_resetfunc_(BLKMPITSTRUCT);
	return(MPI_SUCCESS);
}
