/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	unpack.c,v $
 * Revision 6.0  96/02/29  14:13:58  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:26:34  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:21  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:05:21  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:52:05  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:55  gdburns
 * new for 5.1
 * 
 *	Function:	- unpack to data buffer
 *	Accepts:	- packed buffer
 *			- packed buffer size
 *			- ptr packed buffer position
 *			- data buffer
 *			- data count
 *			- data type
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

int
MPI_Unpack(packbuf, packsize, ppackpos, buf, count, dtype, comm)

void			*packbuf;
int			packsize;
int			*ppackpos;
void			*buf;
int			count;
MPI_Datatype		dtype;
MPI_Comm		comm;

{
	int		size;
	char		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIUNPACK);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(comm,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (count < 0) {
		return(__lam_errfunc_(comm,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((packsize < 0) || (ppackpos == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((*ppackpos < 0) || (*ppackpos >= packsize)) {
		return(__lam_errfunc_(comm,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}


	if ((packsize > 0) && (packbuf == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_BUFFER, 0)));
	}
/*
 * Unpack the data.
 */
	p = ((char *) packbuf) + *ppackpos;
	size = packsize - *ppackpos;

	size = __lam_unpack_(p, size, (char *) buf, count, dtype);

	if (size < 0) {
		return(__lam_errfunc_(comm,
			BLKMPIUNPACK, __lam_mkerr_(MPI_ERR_TRUNCATE, 0)));
	}

	*ppackpos += size;

	__lam_resetfunc_(BLKMPIUNPACK);
	return(MPI_SUCCESS);
}
