/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	wait.c,v $
 * Revision 6.0.1.1  96/05/10  15:12:50  nevin
 * Handle MPI_PROC_NULL
 * 
 * Revision 6.0  96/02/29  14:13:59  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/28  19:18:58  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:22  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/06/14  15:08:56  raja
 * Fix silly typo.
 * 
 * Revision 5.1  94/05/18  12:52:06  gdburns
 * Ohio Release
 * 
 *	Function:	- wait for completion of immediate request
 *	Accepts:	- request
 *			- status
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();
extern void		_mpi_req_rem();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();
extern int		__lam_trincff_();
extern int		_mpi_req_advance();
extern int		_mpi_req_destroy();
extern int		_mpi_req_end();


int
MPI_Wait(preq, stat)

MPI_Request		*preq;
MPI_Status		*stat;

{
	MPI_Request	req;			/* request */
	MPI_Comm	comm;			/* communicator */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */
	
	__lam_initerr_();
	__lam_setfunc_(BLKMPIWAIT);

	if (preq == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAIT,
				__lam_mkerr_(MPI_ERR_REQUEST, 0)));
	}

	if (stat == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAIT,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Inactive requests get an empty status.
 */
	req = *preq;

	if ((req == MPI_REQUEST_NULL) || (req->rq_state == LAM_RQSINIT)) {
		__lam_emptystat_(stat);
		__lam_resetfunc_(BLKMPIWAIT);
		return(MPI_SUCCESS);
	}
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) {
		startt = MPI_Wtime();
		_kio.ki_blktime = 0.0;
	}
/*
 * Block till request is done.
 */
	comm = req->rq_comm;

	if (req->rq_state != LAM_RQSDONE) {
		_mpi_req_blkclr();
		_mpi_req_blkset(req);

		err = _mpi_req_advance();
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIWAIT, err));
		}
	}

	_mpi_req_rem(req);

	err = _mpi_req_end(req);
	if (err != MPI_SUCCESS) return(__lam_errfunc_(comm, BLKMPIWAIT, err));

	*stat = req->rq_status;
/*
 * Generate a run time trace except in the case of a cancelled request or
 * a request where the peer is MPI_PROC_NULL.
 */
	if (fl_trace && (req->rq_rank != MPI_PROC_NULL)
			&& !(req->rq_flags & LAM_RQFCANCEL)) {
		finisht = MPI_Wtime();

		__lam_trmsg_(
			(req->rq_type == LAM_RQIRECV) ? TRTINPUT : TRTNOIO,
			startt, LAM_S2US(finisht - startt - _kio.ki_blktime),
			LAM_S2US(_kio.ki_blktime), req->rq_rank, req->rq_tag,
			req->rq_comm, req->rq_dtype, req->rq_count,
			stat->MPI_SOURCE, stat->MPI_TAG, req->rq_seq,
			req->rq_type);
	}
/*
 * Reset persistent requests.
 * Destroy requests that are not persistent.
 */
	if (req->rq_marks & LAM_RQFPERSIST) {
		req->rq_state = LAM_RQSINIT;
	} else {
		err = _mpi_req_destroy(preq);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIWAIT, err));
		}
	}

	__lam_resetfunc_(BLKMPIWAIT);

	if (stat->MPI_ERROR != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIWAIT, stat->MPI_ERROR));
	}

	return(MPI_SUCCESS);
}
