/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	waitall.c,v $
 * Revision 6.0  96/02/29  14:14:00  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:26:37  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:05:22  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/08/17  14:57:11  raja
 * Change it to a loop of MPI_Waitany().
 * 
 * Revision 5.1  94/05/18  12:52:06  gdburns
 * Ohio Release
 * 
 *	Function:	- wait for all requests to finish
 *	Accepts:	- # of requests
 *			- array of requests
 *			- array of status structures (returned)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trwrapend_();
extern int		__lam_trwrapstart_();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();


int
MPI_Waitall(count, reqs, stats)

int			count;
MPI_Request		*reqs;
MPI_Status		*stats;

{
	int		i;			/* favourite index */
	int		err;			/* error code */
	int		fl_trace;		/* do tracing? */
	int		first;			/* first request ready */
	MPI_Status	firststat;		/* status of first request */
	MPI_Status	*p;			/* favourite pointer */

	__lam_setfunc_(BLKMPIWAITALL);

	if (count < 0) {
	    return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITALL,
		    __lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && ((reqs == 0) || (stats == 0))) {
	    return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITALL,
		    __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = ((_kio.ki_rtf & RTF_TRON) == RTF_TRON);

	if (fl_trace) {
	    __lam_trwrapstart_(BLKMPIWAITALL);
	}
/*
 * Initialize all status members.
 */
	for (i = 0, p = stats; i < count; ++i, ++p) {
	    __lam_emptystat_(p);
	}
/*
 * Loop waiting for all requests.
 */
	for (i = 0; i < count; ++i) {
	    err = MPI_Waitany(count, reqs, &first, &firststat);

	    if (err != MPI_SUCCESS) {

		if (first == MPI_UNDEFINED) {
		    return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITALL, err));
		}

		else {
		    stats[first].MPI_ERROR = err;

		    if (fl_trace) {
			__lam_trwrapend_(BLKMPIWAITALL);
		    }
	
		    __lam_resetfunc_(BLKMPIWAITALL);
		    return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITALL,
			    __lam_mkerr_(MPI_ERR_IN_STATUS, 0)));
		}
	    }

	    if (first == MPI_UNDEFINED) {

		if (fl_trace) {
		    __lam_trwrapend_(BLKMPIWAITALL);
		}

		__lam_resetfunc_(BLKMPIWAITALL);
		return(MPI_SUCCESS);
	    }

	    stats[first] = firststat;
	    stats[first].MPI_ERROR = MPI_SUCCESS;
	}
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) {
	    __lam_trwrapend_(BLKMPIWAITALL);
	}
	
	__lam_resetfunc_(BLKMPIWAITALL);
	return(MPI_SUCCESS);
}
