/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	waitsome.c,v $
 * Revision 6.0  96/02/29  14:14:01  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/22  23:56:08  nevin
 * Skip inactive requests.
 * 
 * Revision 5.2.1.1  1996/02/18 21:26:39  gdburns
 * major overhaul
 *
 * Revision 5.2  94/08/22  14:05:24  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:52:08  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:57  gdburns
 * new for 5.1
 * 
 *	Function:	- wait for at least one request to finish
 *	Accepts:	- # of requests
 *			- array of requests
 *			- # completed requests (returned)
 *			- array of indices (returned)
 *			- array of status (returned)
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <string.h>

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trwrapend_();
extern int		__lam_trwrapstart_();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();


int
MPI_Waitsome(count, reqs, ndone, indices, stats)

int			count;
MPI_Request		*reqs;
int			*ndone;
int			*indices;
MPI_Status		*stats;

{
	int		i;			/* favourite index */
	int		err;			/* error code */
	int		done;			/* test return flag */
	int		first;			/* first request ready */
	int		fl_err;			/* error ocurred? */
	int		fl_trace;		/* do tracing? */
	MPI_Status	firststat;		/* status of first request */
	MPI_Status	*p;			/* favourite pointer */

	__lam_setfunc_(BLKMPIWAITSOME);

	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITSOME,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && ((reqs == 0) || (indices == 0) || (stats == 0))) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITSOME,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if (ndone == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITSOME,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Generate a start wrapper trace.
 */
	fl_trace = ((_kio.ki_rtf & RTF_TRON) == RTF_TRON);

	if (fl_trace) {
		__lam_trwrapstart_(BLKMPIWAITSOME);
	}
/*
 * Loop initializing all status members.
 */
	for (i = 0, p = stats; i < count; ++i, ++p) {
		__lam_emptystat_(p);
	}
/*
 * Wait till the first request is done.
 */
	err = MPI_Waitany(count, reqs, &first, &firststat);

	if (err != MPI_SUCCESS) {

		if (first == MPI_UNDEFINED) {
			return(__lam_errfunc_(MPI_COMM_WORLD,
					BLKMPIWAITSOME, err));
		} else {
			stats[first].MPI_ERROR = err;

			if (fl_trace) {
				__lam_trwrapend_(BLKMPIWAITSOME);
			}

			__lam_resetfunc_(BLKMPIWAITSOME);
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITSOME,
					__lam_mkerr_(MPI_ERR_IN_STATUS, 0)));
		}
	    }

	if (first == MPI_UNDEFINED) {
		*ndone = first;

		if (fl_trace) {
			__lam_trwrapend_(BLKMPIWAITSOME);
		}

		__lam_resetfunc_(BLKMPIWAITSOME);
		return(MPI_SUCCESS);
	}
/*
 * Loop testing if any other request is done and skip inactive requests.
 * I am going to be nice and maintain the sequential order
 * in the result arrays, just as if MPI_Testsome() was used.
 */
	*ndone = 0;
	fl_err = 0;

	for (i = 0; i < count; ++i, ++reqs) {

		if (i == first) {
			++(*ndone);
			*indices++ = i;
			*stats = firststat;
			stats->MPI_ERROR = MPI_SUCCESS;
			++stats;
		}
		else if ((*reqs == MPI_REQUEST_NULL) ||
				((*reqs)->rq_state == LAM_RQSINIT)) {
			continue;
		}
		else {
			err = MPI_Test(reqs, &done, stats);

			if ((err != MPI_SUCCESS) || done) {
				++(*ndone);
				*indices++ = i;

				if (err != MPI_SUCCESS) {
					stats->MPI_ERROR = err;
					fl_err = 1;
				}

				++stats;
			}
		}
	}
/*
 * Generate an end wrapper trace.
 */
	if (fl_trace) {
		__lam_trwrapend_(BLKMPIWAITSOME);
	}

	__lam_resetfunc_(BLKMPIWAITSOME);

	if (fl_err) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITSOME,
				__lam_mkerr_(MPI_ERR_IN_STATUS, 0)));
	}
	
	return(MPI_SUCCESS);
}
