/*
 *      Ohio Trollius
 *      Copyright 1995 The Ohio State University
 *      GDB
 *
 *      $Log:	trbuf.c,v $
 * Revision 6.0  96/02/29  14:03:35  gdburns
 * Ohio Release
 * 
 *	Function:	- functions that manipulate the trace buffer
 *			  data structure
 */

#include <stdlib.h>

#include <all_hash.h>
#include <all_list.h>
#include <debug.h>
#include <portable.h>
#include <terror.h>
#include <trreq.h>

/*
 * external variables
 */
extern char		trqmsg[];

/*
 * local variables
 */
static HASH		*trbhd;
static int4		totalspace;

/*
 * external functions
 */
extern void		trsendr();

/*
 *	trbinit
 *
 *	Function:	- initializes trace buffer hash table
 */
void
trbinit()

{
	totalspace = 0;
	trbhd = ah_init(TRHASHSIZE, sizeof(struct trblist), -1, 0);
	if (trbhd == 0) lampanic("traced (ah_init)");
}

/*
 *	trqstore
 *
 *	Function:	- stores trace data from a client
 *			- no reply message
 *	Accepts:	- client pid
 */
void
trqstore(ptrq)

struct trreq		*ptrq;

{
	struct trblist	newl;
	struct trblist	*ptrl;
	struct trbuffer	newb;
	struct trbuffer	*ptrb;
	LIST		*ld;

	db(("traced: store: length=%d, list=%d, pid=%d\n", ptrq->trq_length,
			ptrq->trq_lnum, ptrq->trq_pid));
	if (ptrq->trq_length <= 0) return;
/*
 * Find the requested trace list.
 */
	ptrl = ah_find(trbhd, ptrq->trq_lnum);

	if (ptrl == 0) {
		newl.trl_lnum = ptrq->trq_lnum;

		if (ah_insert(trbhd, &newl)) {

			if (errno != EFULL) {
				lampanic("traced (ah_insert)");
			}

			if (totalspace < TRMAXSPACE) {

				if (ah_expand(trbhd, ah_size(trbhd) +
						TRHASHSIZE))
						lampanic("traced (ah_expand)");

				if (ah_insert(trbhd, &newl))
						lampanic("traced (ah_insert)");
			} else {
				return;
			}
		}

		ptrl = ah_find(trbhd, ptrq->trq_lnum);
		ptrl->trl_ld = al_init(sizeof(struct trbuffer), (int (*)()) 0);
		if (ptrl->trl_ld == 0) lampanic("traced (al_init)");
	}

	ld = ptrl->trl_ld;
/*
 * Delete old traces.
 */
	if ((totalspace >= TRMAXSPACE) && (al_count(ld) > 0) &&
			(ptrq->trq_lnum >= 0)) {
		ptrb = al_top(ld);
		if (ptrb == 0) lampanic("traced (al_top)");

		db(("traced: store: deleting old traces, length=%d, pid=%d\n",
				ptrb->trb_length, ptrb->trb_pid));

		totalspace -= ptrb->trb_length;
		free(ptrb->trb_buffer);
		al_delete(ld, ptrb);
	}
/*
 * Allocate space for the traces.
 */
	newb.trb_pid = ptrq->trq_pid;
	newb.trb_length = ptrq->trq_length;

	newb.trb_buffer = malloc((unsigned) ptrq->trq_length);
	if (newb.trb_buffer == 0) lampanic("traced (malloc)");

	totalspace += newb.trb_length;
	memcpy(newb.trb_buffer, trqmsg, ptrq->trq_length);

	if (al_append(ld, &newb) == 0) lampanic("traced (al_append)");
}

/*
 *	trqget
 *
 *	Function:	- unloads trace data
 *			- if TRQGET, then maintain trace data
 *			- if TRQFORGET, then free trace data
 *			- sends trace buffers for given list and pid
 *			- list TRANYLIST means all lists
 *			- pid TRANYPID means all processes
 *	Accepts:	- traced request ptr
 */
void
trqget(ptrq)

struct trreq		*ptrq;

{
	struct trreply	reply;
	struct trbuffer	*next;
	struct trbuffer	*ptrb;
	struct trblist	*ptrl;

	reply.trr_errno = 0;
	reply.trr_length = 0;
/*
 * Calculate total length of selected traces.
 */
	ptrl = (ptrq->trq_lnum == TRANYLIST) ? ah_top(trbhd) :
			ah_find(trbhd, ptrq->trq_lnum);

	while (ptrl) {
		ptrb = al_top(ptrl->trl_ld);

		while (ptrb) {

			if ((ptrq->trq_pid == ptrb->trb_pid) ||
					(ptrq->trq_pid == TRANYPID)) {
				reply.trr_length += ptrb->trb_length;
			}

			ptrb = al_next(ptrl->trl_ld, ptrb);
		}

		ptrl = (ptrq->trq_lnum == TRANYLIST) ?
				ah_next(trbhd, ptrl) : 0;
	}

	trsendr(ptrq, &reply, (char *) 0, 0);
/*
 * Send the selected traces.
 */
	ptrl = (ptrq->trq_lnum == TRANYLIST) ? ah_top(trbhd) :
			ah_find(trbhd, ptrq->trq_lnum);

	while (ptrl) {
		ptrb = al_top(ptrl->trl_ld);

		while (ptrb) {
			next = al_next(ptrl->trl_ld, ptrb);

			if ((ptrq->trq_pid == ptrb->trb_pid) ||
					(ptrq->trq_pid == TRANYPID)) {
				trsendr(ptrq, &reply, ptrb->trb_buffer,
						ptrb->trb_length);

				if (ptrq->trq_req == TRQFORGET) {
					totalspace -= ptrb->trb_length;
					free(ptrb->trb_buffer);
					al_delete(ptrl->trl_ld, ptrb);
				}
			}

			ptrb = next;
		}

		ptrl = (ptrq->trq_lnum == TRANYLIST) ?
				ah_next(trbhd, ptrl) : 0;
	}
}

/*
 *	trqsweep
 *
 *	Function:	- sweeps out trace data
 *	Accepts:	- traced request ptr
 */
void
trqsweep(ptrq)

struct trreq		*ptrq;

{
	struct trblist	*ptrl;
	struct trbuffer	*ptrb;
	struct trbuffer	*next;

	ptrl = (ptrq->trq_lnum == TRANYLIST) ? ah_top(trbhd) :
			ah_find(trbhd, ptrq->trq_lnum);

	while (ptrl) {
		ptrb = al_top(ptrl->trl_ld);

		while (ptrb) {
			next = al_next(ptrl->trl_ld, ptrb);

			if ((ptrq->trq_pid == ptrb->trb_pid) ||
					(ptrq->trq_pid == TRANYPID)) {
				free(ptrb->trb_buffer);
				totalspace -= ptrb->trb_length;
				al_delete(ptrl->trl_ld, ptrb);
			}

			ptrb = next;
		}

		ptrl = (ptrq->trq_lnum == TRANYLIST) ?
				ah_next(trbhd, ptrl) : 0;
	}
}

/*
 *	trqudie
 *
 *	Function:	- removes all traces related to a pid
 *	Accepts:	- traced request ptr
 */
void
trqudie(ptrq)

struct trreq		*ptrq;

{
	ptrq->trq_lnum = TRANYLIST;
	trqsweep(ptrq);
}

/*
 *	trqwipe
 *
 *	Function:	- completely resets state of traced
 *	Accepts:	- traced request ptr
 */
void
trqwipe(ptrq)

struct trreq		*ptrq;

{
	struct trblist	*ptrl;
	
	ptrq->trq_lnum = TRANYLIST;
	ptrq->trq_pid = TRANYPID;
	trqsweep(ptrq);
	ptrl = ah_top(trbhd);

	while ((ptrl = ah_top(trbhd))) {
		al_free(ptrl->trl_ld);
		ah_delete(trbhd, ptrl->trl_lnum);
	}
}
