/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	rpcreate.c,v $
 * Revision 6.0  96/02/29  14:05:03  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/11  10:35:01  gdburns
 * Add process ID and index as output parameters.
 * 
 * Revision 5.2  94/08/22  14:13:33  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:58:11  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:46:19  gdburns
 * Ohio Release
 * 
 * 	Function:	- creates a process on a given node
 * 			- executable already loaded with flatd
 * 	Accepts:	- target node ID
 *			- program file name
 * 			- process runtime flags
 * 			- arguments argv structure
 *			- process ID (out)
 *			- process index (out)
 * 	Returns:	- 0 or ERROR
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <args.h>
#include <events.h>
#include <kio.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <preq.h>
#include <rreq.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern struct kio_t	_kio;

int
rpcreate(nodeid, filename, rtf, argv, pid, idx)

int4			nodeid;
CONST char		*filename;
int4			rtf;
char			**argv;
int4			*pid;
int4			*idx;

{
	struct nmsg	nhead;		/* network header */
	struct preq	request;	/* kenya request */
	struct preply	reply;		/* kenya reply */
	int4		mask;
	int4		tag;		/* program/argv flat tag */
	int		packsize;	/* size of packed argv */
	char		*args;		/* glued argv structure */
/*
 * Check trivial cases.
 */
	if (filename == 0) {
		errno = EINVAL;
		return(LAMERROR);
	}

	if (strlen(filename) >= PSMAXNAME) {
		errno = ENAMETOOLONG;
		return(LAMERROR);
	}
/*
 * Pack and tag the arguments.
 */
	if (argv) {
		tag = getpid();
		args = argvglue(argv, 0xa, MAXNMSGLEN);
		if (args == 0) return(LAMERROR);

		packsize = strlen(args) + 1;

		if (rflat(nodeid, args, (int4) packsize,
				(int4) packsize, tag)) {
			rflclean(nodeid, tag);
			return(LAMERROR);
		}

		free(args);
	} else {
		tag = 0;
	}
/*
 * Send the `create' request.
 */
	request.pq_req = ltot(PQCREATE);
	request.pq_rtf = ltot(rtf);
	request.pq_argv = ltot(tag);
	request.pq_stdin = ltot(_kio.ki_stdin);
	request.pq_stdout = ltot(_kio.ki_stdout);
	request.pq_stderr = ltot(_kio.ki_stderr);
	request.pq_world = ltot(_kio.ki_world);
	request.pq_parent = ltot(_kio.ki_parent);
	strcpy(request.pq_name, filename);

	request.pq_src_node = ((nodeid == LOCAL) ||
			(tiscast(getrtype(nodeid)))) ?
			ltot(nodeid) : ltot(getnodeid());
	request.pq_src_event = ltot(-((int4) getpid()));

	nhead.nh_node = nodeid;
	nhead.nh_event = EVKENYAD;
	nhead.nh_flags = 0;
	nhead.nh_type = 0;
	nhead.nh_length = sizeof(request);
	nhead.nh_msg = (char *) &request;
	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	if (nsend(&nhead)) {
		ksigsetmask(mask);
		return(LAMERROR);
	}

	nhead.nh_event = -getpid();
	nhead.nh_length = sizeof(reply);
	nhead.nh_msg = (char *) &reply;

	if (nrecv(&nhead)) {
		ksigsetmask(mask);
		return(LAMERROR);
	}

	ksigsetmask(mask);

	if (reply.pr_reply) {
		errno = ttol(reply.pr_reply);
		return(LAMERROR);
	} else {
		*pid = ttol(reply.pr_pid);
		*idx = ttol(reply.pr_index);
		return(0);
	}
}
