/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB
 *
 *	$Log:	rpgo.c,v $
 * Revision 6.0  96/02/29  14:05:04  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/11  10:34:19  gdburns
 * Add process ID and index as output parameters.
 * 
 * Revision 5.2  94/08/22  14:13:35  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:58:13  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:46:20  gdburns
 * Ohio Release
 * 
 * 	Function:	- creates process on remote node from tagged memory
 * 			- executable already loaded with flatd
 * 			- argv already loaded with flatd
 * 	Accepts:	- target node ID
 * 			- load module tag
 * 			- process runtime flags
 * 			- argv tag
 *			- process ID (out)
 *			- process index (out)
 * 	Returns:	- 0 or ERROR
 */

#include <unistd.h>

#include <events.h>
#include <kio.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <preq.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;

int
rpgo(nodeid, loadtag, rtf, argvtag, pid, idx)

int4			nodeid;
int4			loadtag;
int4			rtf;
int4			argvtag;
int4			*pid;
int4			*idx;

{
	struct nmsg	nhead;		/* network header */
	struct preq	request;	/* kenya request */
	struct preply	reply;		/* kenya reply */
	int4		mask;

	request.pq_loadpt = ltot(loadtag);
	request.pq_rtf = ltot(rtf | RTF_FLAT);
	request.pq_argv = ltot(argvtag);
	request.pq_stdin = ltot(_kio.ki_stdin);
	request.pq_stdout = ltot(_kio.ki_stdout);
	request.pq_stderr = ltot(_kio.ki_stderr);
	request.pq_world = ltot(_kio.ki_world);
	request.pq_parent = ltot(_kio.ki_parent);

	request.pq_src_node = ((nodeid == LOCAL) ||
			tiscast(getrtype(nodeid))) ?
			ltot(nodeid) : ltot(getnodeid());
	request.pq_src_event = ltot(-((int4) getpid()));
	request.pq_req = ltot(PQCREATE);

	nhead.nh_node = nodeid;
	nhead.nh_event = EVKENYAD;
	nhead.nh_flags = 0;
	nhead.nh_type = 0;
	nhead.nh_length = sizeof(request);
	nhead.nh_msg = (char *) &request;
	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	if (nsend(&nhead)) {
		ksigsetmask(mask);
		return(LAMERROR);
	}

	nhead.nh_event = -getpid();
	nhead.nh_length = sizeof(reply);
	nhead.nh_msg = (char *) &reply;

	if (nrecv(&nhead)) {
		ksigsetmask(mask);
		return(LAMERROR);
	}

	ksigsetmask(mask);

	if (reply.pr_reply) {
		errno = ttol(reply.pr_reply);
		return(LAMERROR);
	} else {
		*pid = ttol(reply.pr_pid);
		*idx = ttol(reply.pr_index);
		return(0);
	}
}
