/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	GDB
 *
 *	$Log:	rpstate.c,v $
 * Revision 6.0  96/02/29  14:05:06  gdburns
 * Ohio Release
 * 
 * Revision 5.2  94/08/22  14:13:38  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/08/18  11:47:36  gdburns
 * Use buffers so as to not confuse ksr() in kenyad.
 * 
 * Revision 5.1  94/05/18  12:58:15  gdburns
 * Ohio Release
 * 
 * Revision 2.3.1.1  94/05/18  10:24:45  gdburns
 * Do not use buffers on the local node.
 * 
 * Revision 2.3  94/04/22  12:46:22  gdburns
 * Ohio Release
 * 
 * Revision 2.2.1.5  93/12/05  13:19:55  vaigl-j
 * Now returns parent information.
 * 
 * Revision 2.2.1.4  93/10/13  17:17:37  gdburns
 * Move ksigblock() before nsend() due to new kernel.
 * 
 * Revision 2.2.1.3  93/07/19  20:07:40  gdburns
 * Change to standard naming.
 * Use local node optimization.
 * 
 * Revision 2.2.1.2  93/05/04  16:51:15  gdburns
 * Remove pointer status information.
 * 
 * Revision 2.2.1.1  92/09/07  23:11:46  gdburns
 * facelift due to LAM
 *
 *	Function:	- retrieves status of remote processes
 *	Accepts:	- remote node ID
 *			- process select flags
 *			- process select value
 *			- process state table
 *			- process state table # entries
 *	Returns:	- # retrieved process states or ERROR
 */

#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <preq.h>
#include <t_types.h>
#include <typical.h>

int4
rpstate(nodeid, sflags, svalue, pstable, maxproc)

int4			nodeid;
int4			sflags;
int4			svalue;
struct pstate		pstable[];
int4			maxproc;

{
	struct nmsg	nhead;		/* network nhead */
	struct preq	request;	/* kenya request */
	int4		nproc;		/* # replied states */
	int4		mask;		/* signal mask */
	int4		ret;		/* reply code */

	if (maxproc <= 0) {
		return(0);
	}

	nproc = 0;

	request.pq_sel_method = ltot(sflags);
	request.pq_sel_value = ltot(svalue);
	request.pq_req = ltot(PQSTATE);
	request.pq_maxproc = ltot(maxproc);
	request.pq_src_node = (nodeid == LOCAL) ?
			ltot(nodeid) : ltot(getnodeid());
	request.pq_src_event = ltot((int4) -getpid());

	nhead.nh_node = nodeid;
	nhead.nh_event = EVKENYAD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = sizeof(request);
	nhead.nh_msg = (char *) &request;
	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	if (nsend(&nhead)) {
		ksigsetmask(mask);
		return(ERROR);
	}

	nhead.nh_event = -getpid();
	nhead.nh_length = sizeof(struct pstate);

	do {
		nhead.nh_msg = (char *) pstable;

		if (nrecv(&nhead)) {
			ksigsetmask(mask);
			return(ERROR);
		}
/*
 * Convert byte ordering (sigh).
 */
		pstable->ps_reply = ttol(pstable->ps_reply);
		pstable->ps_index = ttol(pstable->ps_index);
		pstable->ps_pid = ttol(pstable->ps_pid);
		pstable->ps_rtf = ttol(pstable->ps_rtf);
		pstable->ps_event = ttol(pstable->ps_event);
		pstable->ps_nodeid = ttol(pstable->ps_nodeid);
		pstable->ps_kernel.ks_state =
				ttol(pstable->ps_kernel.ks_state);
		pstable->ps_kernel.ks_priority =
				ttol(pstable->ps_kernel.ks_priority);
		pstable->ps_kernel.ks_event =
				ttol(pstable->ps_kernel.ks_event);
		pstable->ps_kernel.ks_type =
				ttol(pstable->ps_kernel.ks_type);
		pstable->ps_kernel.ks_signal =
				ttol(pstable->ps_kernel.ks_signal);

		ret = pstable->ps_reply;
		++nproc;
		++pstable;
	} while ((nproc < maxproc) && !ret);

	ksigsetmask(mask);
	if (ret) nproc--;
	return(nproc);
}
