/*
 *	Ohio Trollius
 *	Copyright 1993 The Ohio State University
 *	SBD/JRV
 *
 *	$Log:	tfprint.c,v $
 * Revision 6.0  96/02/29  14:06:31  gdburns
 * Ohio Release
 * 
 * Revision 5.2  94/08/22  14:17:34  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:58:49  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:46:55  gdburns
 * Ohio Release
 * 
 * Revision 2.2.1.1  93/10/12  11:33:57  vaigl-j
 * Mod for SCO ANSI C compilation.
 * 
 * Revision 2.2  93/05/10  16:38:57  raja
 * pre-release to 2.3
 * 
 *	Function	- tstdio user routines for printing
 *			- all functions call _fmt() in tprint.c
 */

#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <tstdio.h>

extern int		_fmt();


/*
 *	tprintf
 *
 *	Function:	- prints on stdout
 *	Accepts:	- format string
 *			- variable number of items to print
 *	Returns:	- the number of printed characters, or EOF if error
 */
#if __STDC__
int
tprintf(char const *format, ...)
#else
tprintf(format, va_alist)

char			*format;
va_dcl
#endif

{
	char		*storep	= 0;
	TFILE		*fp	= tstdout;
	int		count;
	va_list		arglist;

#if __STDC__
	va_start(arglist, format);
#else
	va_start(arglist);
#endif
	count = _fmt(fp, storep, format, arglist);
	va_end(arglist);
	return(count);
}


/*
 *	tfprintf
 *
 *	Function:	- prints to file pointed to by fp
 *	Accepts:	- file pointer
 *			- format string
 *			- variable number of items to print
 *	Returns:	- the number of printed characters, of EOF if error
 */
#if __STDC__
int
tfprintf(TFILE *fp, char const *format, ...)
#else
int
tfprintf(fp, format, va_alist)

TFILE			*fp;
char			*format;
va_dcl

#endif

{
	char		*storep	= 0;
	int		count;
	va_list		arglist;

#if __STDC__
	va_start(arglist, format);
#else
	va_start(arglist);
#endif
	count = _fmt(fp, storep, format, arglist);
	va_end(arglist);
	return(count);
}


/*
 *	tsprintf
 *
 *	Function:	- prints memory location pointed to by storep
 *	Accepts:	- char pointer to destination
 *			- format string
 *			- variable number of items to format
 *	Returns:	- char pointer to destination
 */	
#if __STDC__
char *
tsprintf(char *storep, char const *format, ...)
#else
char *
tsprintf(storep, format, va_alist)

char			*storep;
char			*format;
va_dcl

#endif

{
	TFILE		*fp = 0;
	int		count;
	va_list		arglist;

#if __STDC__
	va_start(arglist, format);
#else
	va_start(arglist);
#endif
	count = _fmt(fp, storep, format, arglist);
	va_end(arglist);
	
	return(storep + count);
}


/*
 *	tvprintf
 *
 *	Function:	- same as printf(), but this takes a va_list
 *	Accepts:	- format string
 *			- variable number of items to format in a va_list
 *	Returns:	- the number of printed characters, of EOF if error
 */
#if __STDC__
int
tvprintf(char const *format, ...)
#else
int
tvprintf(format, va_alist)

char			*format;
va_dcl
#endif

{
	char		*storep	= 0;
	TFILE		*fp	= tstdout;
	int		count;
	va_list		ap;

#if __STDC__
	va_start(ap, format);
#else
	va_start(ap);
#endif
	count = _fmt(fp, storep, format, ap);
	va_end(ap);
	return(count);
}


/*
 *      tvfprintf
 *
 *	Function:	- same as fprintf(), but this takes a va_list
 *	Accepts:	- file pointer
 *			- format string
 *			- variable number of items to format in a va_list
 *	Returns:	- the number of printed characters, of EOF if error
 */
#if __STDC__
int
tvfprintf(TFILE *fp, char const *format, ...)
#else
int
tvfprintf(fp, format, arglist)

TFILE			*fp;
char			*format;
va_list			arglist;
#endif
{
	char		*storep	= 0;
	int		count;
#if __STDC__
	va_list		arglist;

	va_start(arglist, format);
#endif
	count = _fmt(fp, storep, format, arglist);
#if __STDC__
	va_end(arglist);
#endif
	return(count);
}


/*
 *	tvsprintf
 *
 *	Function:	- same as sprintf, but takes a va_list
 *	Accepts:	- char pointer to destination
 *			- format string
 *			- variable number of items to format in a va_list
 *	Returns:	- char pointer to destination
 */
#if __STDC__
char *
tvsprintf(char *storep, char const *format, ...)
#else
char *
tvsprintf(storep, format, arglist)

char			*storep;
char			*format;
va_list			arglist;
#endif
{
	TFILE		*fp = 0;
	int		count;
#if __STDC__
	va_list		arglist;

	va_start(arglist, format);
#endif
	count = _fmt(fp, storep, format, arglist);
#if __STDC__
	va_end(arglist);
#endif
	return(storep + count);
}
